/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * David Carter - Initial API and implementation
 **********************************************************************/

#ifndef LIBARM4DB_CONF_H
#define LIBARM4DB_CONF_H 1

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <db.h>

#define DEFAULT_LOG_FILE_SIZE   (128 * 1024)
#define DEFAULT_LOG_BUFFER_SIZE (2 * 1024 * 1024)

#if defined (__cplusplus)

#include <string>

typedef enum {
	/* All supported database types */
	Arm4dbBerkeleyDB,
	Arm4dbSqlite,
	Arm4dbMySql
} Arm4dbDatabaseType;

#if defined(HAVE_BDB)
#define DEFAULT_ARM4DB_TYPE	Arm4dbBerkeleyDB
#elif defined(HAVE_SQLITE)
#define DEFAULT_ARM4DB_TYPE	Arm4dbSqlite
#elif defined(HAVE_MYSQL)
#define DEFAULT_ARM4DB_TYPE	Arm4dbMySql
#endif

#ifdef DOCROOT
#define DEFAULT_SCHEMA_LOCATION DOCROOT
#else
#define DEFAULT_SCHEMA_LOCATION "/usr/local/share/" PACKAGE_NAME "-" PACKAGE_VERSION
#endif

class Arm4dbConfig
{
public:
	/* Constructors */

	/* Destructors */
	virtual ~Arm4dbConfig ();

	/* Accessors */
	static const Arm4dbConfig &getConfig (const std::string &filename); /* Singleton factory method */
	static const Arm4dbConfig &getConfig (void);

	const std::string &getConfigFilename (void) const { return m_filename; }
	const std::string &getHostname (void) const { return m_hostname; }
	const std::string &getLogFilename (void) const { return m_logFilename; }
	Arm4dbDatabaseType getDbType (void) const { return m_dbType; }
	int getInstance (void) const { return m_instance; }
	uid_t getUid (void) const { return m_uid; }
	gid_t getGid (void) const { return m_gid; }
	mode_t getUmask (void) const { return m_umask; }
	const std::string &getSchemaLocation (void) const { return m_schemaLocation; }
	const std::string &getDbHome (void) const { return m_dbHome; }
	const char *getDbHomeC (void) const
		{ return const_cast<Arm4dbConfig *> (this)->toC (m_dbHome, &(const_cast<Arm4dbConfig *> (this)->m_dbHomeC)); }
	const std::string &getDbBackup (void) const { return m_dbBackup; }
	const char *getDbBackupC (void) const
		{ return const_cast<Arm4dbConfig *> (this)->toC (m_dbBackup, &(const_cast<Arm4dbConfig *> (this)->m_dbBackupC)); }
	long getDbKey (void) const { return m_dbKey; }
	u_int32_t getDbLogFileSize (void) const { return m_dbLogFileSize; }
	u_int32_t getDbLogBufferSize (void) const { return m_dbLogBufferSize; }
	int getCheckpointInterval (void) const { return m_checkpointInterval; }

	const std::string &getMySqlHost (void) const { return m_mysqlHost; }
	const char *getMySqlHostC (void) const
		{ return const_cast<Arm4dbConfig *> (this)->toC (m_mysqlHost, &(const_cast<Arm4dbConfig *> (this)->m_mysqlHostC)); }
	const std::string &getMySqlUser (void) const { return m_mysqlUser; }
	const char *getMySqlUserC (void) const 
		{ return const_cast<Arm4dbConfig *> (this)->toC (m_mysqlUser, &(const_cast<Arm4dbConfig *> (this)->m_mysqlUserC)); }
	const std::string &getMySqlPassword (void) const { return m_mysqlPassword; }
	const char *getMySqlPasswordC (void) const
		{ return const_cast<Arm4dbConfig *> (this)->toC (m_mysqlPassword, &(const_cast<Arm4dbConfig *> (this)->m_mysqlPasswordC)); }
	const std::string &getMySqlDb (void) const { return m_mysqlDb; }
	const char *getMySqlDbC (void) const
		{ return const_cast<Arm4dbConfig *> (this)->toC (m_mysqlDb, &(const_cast<Arm4dbConfig *> (this)->m_mysqlDbC)); }
	const std::string &getMySqlSocket (void) const { return m_mysqlSocket; }
	const char *getMySqlSocketC (void) const 
		{ return ((m_mysqlSocket.length() > 0) ? 
				  const_cast<Arm4dbConfig *> (this)->toC (m_mysqlSocket, &(const_cast<Arm4dbConfig *> (this)->m_mysqlSocketC))
				  : NULL); }
	int getMySqlPort (void) const { return m_mysqlPort; }

	/* Mutators */
	static void setConfig (const std::string &filename); /* Singleton factory method */
	static void setDBHome (const std::string &home);

	/* Operations */
	void read (const char *config_filename_ptr);

	/* Operators */

private:
	// These are private so that we may implement a singleton
	Arm4dbConfig ();
	Arm4dbConfig (const Arm4dbConfig &config);
	Arm4dbConfig (const std::string &filename);
	Arm4dbConfig &operator= (const Arm4dbConfig &config);

	void setDefaultHostname (void);
	void setFilename (const std::string &filename);
	void setHostname (const std::string &hostname);
	void setLogFilename (const std::string &logFilename);
	void setSchemaLocation (const std::string &schemaLocation);
	void setDbHome (const std::string &dbHome);
	void setDbBackup (const std::string &dbBackup);
	void setMysqlHost (const std::string &host);
	void setMysqlUser (const std::string &user);
	void setMysqlPassword (const std::string &password);
	void setMysqlDb (const std::string &db);
	void setMysqlSocket (const std::string &socket);

	const char *toC (const std::string &cppString, char **cString);
	void clearC (char **cString);

	int parse_integer (const char *value_ptr);
	long parse_long (const char *value_ptr);
	void read_line (FILE *file, char *buffer_ptr, int size);
	char *uncomment (char *buffer_ptr);
	void trim_tail (char *buffer_ptr);
	void get_parameters (char *line_ptr, char **value_pptr);
	void parse_hostname (const char *value_ptr);
	void parse_log_filename (const char *value_ptr);
	void parse_uid (const char *value_ptr);
	void parse_gid (const char *value_ptr);
	void parse_umask (const char *value_ptr);
	void parse_port (const char *value_ptr);
	void parse_schema_location (const char *value_ptr);
	void parse_db (const char *value_ptr);
	void parse_db_home (const char *value_ptr);
	void parse_db_backup (const char *value_ptr);
	void parse_db_key (const char *value_ptr);
	void parse_db_log_buffer_size (const char *value_ptr);
	void parse_db_log_file_size (const char *value_ptr);
	void parse_instance (const char *value_ptr);
	void parse_checkpoint_interval (const char *value_ptr);

	void parse_mysql_host (const char *value_ptr);
	void parse_mysql_user (const char *value_ptr);
	void parse_mysql_password (const char *value_ptr);
	void parse_mysql_db (const char *value_ptr);
	void parse_mysql_socket (const char *value_ptr);
	void parse_mysql_port (const char *value_ptr);

	void adjustDbHome (void);

	std::string			m_filename; // The name of the config file used to generate this config
	std::string			m_hostname;
	std::string			m_logFilename;
	Arm4dbDatabaseType	m_dbType;
	int		    		m_instance;
	uid_t	    		m_uid;
	gid_t	    		m_gid;
	mode_t      		m_umask;
	std::string			m_schemaLocation;
	std::string    		m_dbHome;
	std::string    		m_dbBackup;
	long	    		m_dbKey;
	u_int32_t   		m_dbLogFileSize;
	u_int32_t   		m_dbLogBufferSize;
	int					m_checkpointInterval;

	// MySql parameters
	std::string 		m_mysqlHost;
	std::string 		m_mysqlUser;
	std::string 		m_mysqlPassword;
	std::string 		m_mysqlDb;
	std::string 		m_mysqlSocket;
	int 				m_mysqlPort;

	// Pointers to the C style string equivalents
	char				*m_dbHomeC;
	char				*m_dbBackupC;
	char				*m_mysqlHostC;
	char				*m_mysqlUserC;
	char				*m_mysqlPasswordC;
	char				*m_mysqlDbC;
	char				*m_mysqlSocketC;

	static Arm4dbConfig *m_singletonPtr;   // The agent wide configuration
};

#endif /* __cplusplus */


#endif /* LIBARM4DB_CONF_H */

