/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/

#ifndef ARM4_SHM_H
#define ARM4_SHM_H 1

#include "config.h"

#ifdef HAVE_PTHREAD_SETPSHARED
#define USE_POSIX_PROCESS_SEM 1
#else
#undef USE_POSIX_PROCESS_SEM
#endif

#include <pthread.h>
#ifndef USE_POSIX_PROCESS_SEM
#include <semaphore.h>
#endif /* USE_POSIX_PROCESS_SEM */

#include <sys/time.h>
#include <sys/stat.h>
#include <sys/msg.h>
#include <sys/types.h>

#include <openssl/md5.h>

#include "arm4.h"
#include "appdex.h"

/* IPC file definitions */
#define SHARED_MEMORY_FILE		"arm4.shm"
#define MQUEUE_FILE				"arm4.mq"

/* Forward slash required for Solaris */
#define SEM_SEQUENCE_FILE		"/arm4_sequence.sem"
#define SEM_APPLICATION_FILE	"/arm4_application.sem"
#define SEM_TRANSACTION_FILE	"/arm4_transaction.sem"
#define SEM_METRIC_FILE	        "/arm4_metric.sem"
#define SEM_UTILITY_FILE		"/arm4_utility.sem"

#define SEM_INSTANCE_FILE   "arm4_instance.sem"

#define SEM_INSTANCE_APPLICATION    0
#define SEM_INSTANCE_TRANSACTION    1
#define SEM_INSTANCE_METRIC         2
#define SEM_INSTANCE_COUNT          3

#define ARM4_SHM_MAGIC	0x87654321

#define ARM4_SHM_SEQUENCE_STEP	100		/* This is the number of sequences the daemon can assume	*/
										/* before incrementing the number in the database			*/

/* For now, table sizes are hard coded */
#define ARM4_SHM_APPLICATION_ENTRIES						32
#define ARM4_SHM_TRANSACTION_ENTRIES						64
#define ARM4_SHM_METRIC_ENTRIES						        64

#define ARM4_SHM_MESSAGE_APPLICATION						10
#define ARM4_SHM_MESSAGE_APPLICATION_GROUP					11
#define ARM4_SHM_MESSAGE_APPLICATION_INSTANCE_ID			12
#define ARM4_SHM_MESSAGE_APPLICATION_CONTEXT				13
#define ARM4_SHM_MESSAGE_APPLICATION_ADDRESS				14
#define ARM4_SHM_MESSAGE_TRANSACTION						20
#define ARM4_SHM_MESSAGE_TRANSACTION_CONTEXT				21
#define ARM4_SHM_MESSAGE_TRANSACTION_URI					22
#define ARM4_SHM_MESSAGE_TRANSACTION_METRIC_VALUE			23
#define ARM4_SHM_MESSAGE_TRANSACTION_USER					24
#define ARM4_SHM_MESSAGE_TRANSACTION_DIAG					25
#define ARM4_SHM_MESSAGE_CORRELATOR							30
#define ARM4_SHM_MESSAGE_SEQUENCE							31
#define ARM4_SHM_MESSAGE_REGISTER_APPLICATION				40
#define ARM4_SHM_MESSAGE_REGISTER_APPLICATION_IDENTITY		41
#define ARM4_SHM_MESSAGE_REGISTER_APPLICATION_CONTEXT		42
#define ARM4_SHM_MESSAGE_REGISTER_TRANSACTION				50
#define ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_IDENTITY		51
#define ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_CONTEXT		52
#define ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_URI			53
#define ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_METRIC		54
#define ARM4_SHM_MESSAGE_REGISTER_METRIC    				60

#define ARM4_SHM_MESSAGE_UTILITY		    				1000

#define ARM4_SHM_COMMAND_STOP								1
#define ARM4_SHM_COMMAND_BACKUP								2
#define ARM4_SHM_COMMAND_CHECKPOINT							3

#ifndef HOST_NAME_MAX
#define HOST_NAME_MAX	255
#endif

/* Wrap msgsnd to allow for errors in structure alignment */
#define SAFE_SEND(q,message,size,flags) msgsnd(q, message, size - sizeof(long), flags)

/* Wrappers to add stats to pthreads */
#ifdef HAVE_PTHREAD_STATS
#define PTHREAD_STATS(lock_name) \
	int lock_name ## _pthread_stat_total; \
	int lock_name ## _pthread_stat_wait; \
	int lock_name ## _pthread_stat_active; \
	int lock_name ## _pthread_stat_max_active;
#else
#define PTHREAD_STATS(prefix) ;
#endif

/* Set the library tracing level - each level is cumultative */
typedef enum arm4_shm_tracing {
    arm4_shm_trace_none = 0,
    arm4_shm_trace_arm,
    arm4_shm_trace_arm_internal,
    arm4_shm_trace_arm_db,
    arm4_shm_trace_arm_db_internal,
    arm4_shm_trace_arm_shm,
    arm4_shm_trace_arm_shm_internal
} arm4_shm_tracing_t;

typedef struct timeval arm_time_t;
typedef arm_int64_t arm4_shm_handle_t;

typedef enum {
	arm4_shm_sequence_application = 0,
	arm4_shm_sequence_transaction,
	arm4_shm_sequence_application_instance,
	arm4_shm_sequence_transaction_instance,
	arm4_shm_sequence_transaction_blocking,
	arm4_shm_sequence_max
} arm4_shm_sequence_t;

typedef enum {
	/* These values are bit masks */
	arm4_shm_transaction_trace_all = 1,
	arm4_shm_transaction_trace_n = 2,
	arm4_shm_transaction_trace_interval = 4,
	arm4_shm_transaction_trace_use_default = 8,
	arm4_shm_transaction_trace_honor_request = 16,
	
	/* Used internally */
	arm4_shm_transaction_trace_next = 32
} arm4_shm_transaction_trace_t;

typedef struct arm4_shm_application {
	char md5_digest [MD5_DIGEST_LENGTH];
	arm_id_t appl_id;
	arm_boolean_t is_new;
	arm4_shm_transaction_trace_t default_trace_flags;
	arm_int64_t default_trace_parameter;
	arm_int64_t default_appdex_threshold;
} arm4_shm_application_t;

typedef struct arm4_shm_transaction {
	char md5_digest [MD5_DIGEST_LENGTH];
	arm_id_t tran_id;
	arm_boolean_t is_new;
	arm4_shm_transaction_trace_t trace_flags;
	arm4_appdex_t appdex_flags;
	arm_int64_t trace_parameter;
	arm_int64_t trace_counter;
	arm_int64_t transaction_counter;
	arm_int64_t appdex_threshold;
	arm_int64_t appdex_satisfactory;
	arm_int64_t appdex_tolerable;
	arm_int64_t appdex_frustrating;
} arm4_shm_transaction_t;

typedef struct arm4_shm_metric {
	char md5_digest [MD5_DIGEST_LENGTH];
	arm_id_t metric_id;
} arm4_shm_metric_t;

#if 0
typedef struct msghdr arm4_shm_message_header_t;
#else
typedef struct arm4_shm_message_header {
	long mtype;
	char message [1];
} arm4_shm_message_header_t;
#endif

typedef struct arm4_shm_property_message {
	arm4_shm_handle_t handle; /* This is the index */
	arm_char_t property_name [ARM_PROPERTY_VALUE_MAX_LENGTH];
} arm4_shm_property_message_t;

typedef struct arm4_shm_property {
	long mtype;
	arm4_shm_property_message_t message;
} arm4_shm_property_t;

typedef struct arm4_shm_context_property_message {
	arm4_shm_handle_t handle; /* This is the index */
	int property_index;
	arm_char_t property_name [ARM_PROPERTY_VALUE_MAX_LENGTH];
} arm4_shm_context_property_message_t;

typedef struct arm4_shm_context_property {
	long mtype;
	arm4_shm_context_property_message_t message;
} arm4_shm_context_property_t;

typedef struct arm4_shm_uri_message {
	arm4_shm_handle_t handle; /* This is the index */
	arm_char_t property_name [ARM_PROPERTY_URI_MAX_LENGTH];
} arm4_shm_uri_message_t;

typedef struct arm4_shm_uri {
	long mtype;
	arm4_shm_uri_message_t message;
} arm4_shm_uri_t;

typedef struct arm4_shm_metric_value_message {
	arm4_shm_handle_t handle; /* This is the index */
	arm_metric_t value;
	arm_char_t string32 [ARM_METRIC_STRING32_MAX_LENGTH];
} arm4_shm_metric_value_message_t;

typedef struct arm4_shm_metric_value {
	long mtype;
	arm4_shm_metric_value_message_t message;
} arm4_shm_metric_value_t;

typedef struct arm4_shm_user_message {
	arm4_shm_handle_t handle; /* This is the index */
	char name [ARM_PROPERTY_NAME_MAX_LENGTH];
    arm_boolean_t id_valid;
    arm_id_t id;
} arm4_shm_user_message_t;

typedef struct arm4_shm_user {
	long mtype;
	arm4_shm_user_message_t message;
} arm4_shm_user_t;

typedef struct arm4_shm_diag_message {
	arm4_shm_handle_t handle; /* This is the index */
	arm_char_t detail [ARM_DIAG_DETAIL_MAX_LENGTH];
} arm4_shm_diag_message_t;

typedef struct arm4_shm_diag {
	long mtype;
	arm4_shm_diag_message_t message;
} arm4_shm_diag_t;

typedef struct arm4_shm_application_instance_message {
	arm_app_start_handle_t app_handle; /* This is the index */
	arm_id_t appl_id;
	arm_boolean_t active;
} arm4_shm_application_instance_message_t;

typedef struct arm4_shm_application_instance {
	long mtype;
	arm4_shm_application_instance_message_t message;
} arm4_shm_application_instance_t;

typedef struct arm4_shm_address_message {
	arm_app_start_handle_t app_handle; /* This is the index */
	arm_int16_t address_format;
	arm_int16_t address_length;
	arm_uint8_t address [ARM_PROPERTY_URI_MAX_LENGTH];
	arm_boolean_t id_valid;
	arm_id_t id;
} arm4_shm_address_message_t;

typedef struct arm4_shm_register_application_address {
	long mtype;
	arm4_shm_address_message_t message;
} arm4_shm_address_t;

typedef struct arm4_shm_transaction_instance_message {
	arm_tran_start_handle_t tran_handle; /* This is the index */
	arm_id_t appl_id;
	arm_id_t tran_id;
	arm_app_start_handle_t app_handle;
	arm_time_t arrival_time;
	arm_time_t stop_time;
	arm_time_t response_time;
	arm_time_t blocked_time;
	arm_tran_status_t tran_status;
	arm_boolean_t active;
} arm4_shm_transaction_instance_message_t;

typedef struct arm4_shm_transaction_instance {
	long mtype;
	arm4_shm_transaction_instance_message_t message;
} arm4_shm_transaction_instance_t;

typedef struct arm4_shm_correlator_message {
	arm_tran_start_handle_t tran_handle; /* This is the index */
	arm_correlator_t parent_correlator;
	arm_correlator_t current_correlator;
} arm4_shm_correlator_message_t;

typedef struct arm4_shm_correlator {
	long mtype;
	arm4_shm_correlator_message_t message;
} arm4_shm_correlator_t;

typedef struct arm4_shm_db_sequence_message {
	arm4_shm_sequence_t index;
	arm_int64_t sequence;
} arm4_shm_db_sequence_message_t;

typedef struct arm4_shm_db_sequence {
	long mtype;
	arm4_shm_db_sequence_message_t message;
} arm4_shm_db_sequence_t;

typedef struct arm4_shm_register_application_message {
	int application_index;
	unsigned char digest [MD5_DIGEST_LENGTH];
	arm_char_t app_name [ARM_NAME_MAX_LENGTH];
	arm_charset_t charset;
} arm4_shm_register_application_message_t;

typedef struct arm4_shm_register_application {
	long mtype;
	arm4_shm_register_application_message_t message;
} arm4_shm_register_application_t;

typedef struct arm4_shm_register_application_identity_message {
	arm_id_t appid;
    arm_char_t name [ARM_PROPERTY_NAME_MAX_LENGTH];
    arm_char_t value [ARM_PROPERTY_VALUE_MAX_LENGTH];
} arm4_shm_register_application_identity_message_t;

typedef struct arm4_shm_register_application_identity {
	long mtype;
	arm4_shm_register_application_identity_message_t message;
} arm4_shm_register_application_identity_t;

typedef struct arm4_shm_register_application_context_message {
	arm_id_t appid;
    arm_int32_t index;
    arm_char_t name [ARM_PROPERTY_NAME_MAX_LENGTH];
} arm4_shm_register_application_context_message_t;

typedef struct arm4_shm_register_application_context {
	long mtype;
	arm4_shm_register_application_context_message_t message;
} arm4_shm_register_application_context_t;

typedef struct arm4_shm_register_transaction_message {
	int transaction_index;
	unsigned char digest [MD5_DIGEST_LENGTH];
	arm_char_t tran_name [ARM_NAME_MAX_LENGTH];
	arm_id_t app_id;
} arm4_shm_register_transaction_message_t;

typedef struct arm4_shm_register_transaction {
	long mtype;
	arm4_shm_register_transaction_message_t message;
} arm4_shm_register_transaction_t;

typedef struct arm4_shm_register_transaction_uri_message {
	arm_id_t tran_id;
	arm_char_t property_name [ARM_PROPERTY_URI_MAX_LENGTH];
} arm4_shm_register_transaction_uri_message_t;

typedef struct arm4_shm_register_transaction_uri {
	long mtype;
	arm4_shm_register_transaction_uri_message_t message;
} arm4_shm_register_transaction_uri_t;

typedef struct arm4_shm_register_transaction_metric_message {
	arm_id_t tran_id;
    arm_metric_slot_t slot;
    arm_id_t id;
} arm4_shm_register_transaction_metric_message_t;

typedef struct arm4_shm_register_transaction_metric {
	long mtype;
	arm4_shm_register_transaction_metric_message_t message;
} arm4_shm_register_transaction_metric_t;

typedef struct arm4_shm_register_metric_message {
	int metric_index;
	unsigned char digest [MD5_DIGEST_LENGTH];
	arm_char_t metric_name [ARM_NAME_MAX_LENGTH];
    arm_char_t metric_unit [ARM_NAME_MAX_LENGTH];
	arm_id_t app_id;
    arm_metric_format_t metric_format;
    arm_metric_usage_t metric_usage;
} arm4_shm_register_metric_message_t;

typedef struct arm4_shm_register_metric {
	long mtype;
	arm4_shm_register_metric_message_t message;
} arm4_shm_register_metric_t;

typedef struct arm4_shm_utility_message {
    long command;
} arm4_shm_utility_message_t;

typedef struct arm4_shm_utility {
	long mtype;
	arm4_shm_utility_message_t message;
} arm4_shm_utility_t;

typedef struct arm4_shm_header {
	arm_uint64_t magic;

	/* Sequence numbers */
	arm_int64_t sequence_table [arm4_shm_sequence_max];
#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_t sequence_mutex;
	PTHREAD_STATS(sequence_mutex)
#endif /* USE_POSIX_PROCESS_SEM */

	/* cached application entries */
	int max_applications; /* Number of entries in the table */
	int application_count; /* Number of entries in the table */
	int next_application; /* Next entry to check */
#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_t application_mutex;
	PTHREAD_STATS(application_mutex)
#endif /* USE_POSIX_PROCESS_SEM */
    int application_semaphore; /* Used when waiting for the application ID to be filled in */
	arm4_shm_application_t application_table [ARM4_SHM_APPLICATION_ENTRIES];

	/* cached application aliases */

	/* cached transaction entries */
	int max_transactions; /* Number of entries in the table */
	int transaction_count; /* Number of entries in the table */
	int next_transaction; /* Next entry to check */
#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_t transaction_mutex;
	PTHREAD_STATS(transaction_mutex)
#endif /* USE_POSIX_PROCESS_SEM */
	arm4_shm_transaction_t transaction_table [ARM4_SHM_TRANSACTION_ENTRIES];

	/* cached transaction entries */
	int max_metrics; /* Number of entries in the table */
	int metric_count; /* Number of entries in the table */
	int next_metric; /* Next entry to check */
#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_t metric_mutex;
	PTHREAD_STATS(metric_mutex)
#endif /* USE_POSIX_PROCESS_SEM */
	arm4_shm_metric_t metric_table [ARM4_SHM_METRIC_ENTRIES];

	/* Global configuration entries */
	char hostname [HOST_NAME_MAX];
	arm4_shm_tracing_t trace_level;
    arm_boolean_t null_arm; /* True when no recording is being done */
	pid_t daemon_instance;	/* The process ID of the daemon owning this ARM instance */

#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_t utility_read_mutex; /* Used to signal that an external process is accessing the database */
	PTHREAD_STATS(utility_read_mutex)
#endif /* USE_POSIX_PROCESS_SEM */
} arm4_shm_header_t;

#endif /* ARM4_SHM_H */
