/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"
 
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <db.h>
#include "arm4.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Arm4dbSharedMemory.h"
#include "arm4agent.h"

static void
initialize_configuration (const char *archive_directory, const char *config_filename)
{
	// Set the database configuration
	Arm4dbConfig::setConfig (config_filename);

	if (strlen (archive_directory) > 0)
		Arm4dbConfig::setDBHome (archive_directory);
}

static void
version (void)
{
	printf ("Package %s\n", PACKAGE);
	printf ("Version %s\n", VERSION);
}

static void
usage (void)
{
	printf ("Usage:\n");
#ifdef HAVE_GETOPT_LONG 
	printf ("\tarm4_dump --help|-h\n");
	printf ("\t\tShow this help message\n");
	printf ("\tarm4_dump --version|-V\n");
	printf ("\t\tDisplay implementation version\n");
	printf ("\tarm4_dump [--config|-C config_file] [--archive|-a archive_directory]\n");
    printf ("\t\t--config specifies a path to the configuration file. The default is /etc/arm4.conf\n");
	printf ("\t\t--archive specifies a path to the archive directory\n");
#else
    printf ("\tarm4_dump -h\n");
    printf ("\t\tShow this help message\n");
    printf ("\tarm4_dump -V\n");
    printf ("\t\tDisplay implementation version\n");
    printf ("\tarm4_dump [-C config_file] [-a archive_directory]\n");
    printf ("\t\t-C specifies a path to the configuration file. The default is /etc/arm4.conf\n");
	printf ("\t\t-a specifies a path to the archive directory\n");
#endif
}

int
main (int argc, char *argv[])
{
	int	status;
    int c;
    char config_filename [1024];
    char archive_directory [1024];

    // Initialize the logger
    Logger::getLogger().setName ("arm4_dump");
    Logger::getLogger().setFile (stderr);

    /* Set the default config file */
    strcpy (config_filename, "/etc/arm4.conf");
	archive_directory [0] = '\0';

    while (1)
    {
#ifdef HAVE_GETOPT_LONG 
        static struct option long_options[] =
            {
                    /* These options set a flag. */
                    /* These options don't set a flag. We distinguish them by their indices. */
                {"help",     no_argument,       0, 'h'},
                {"version",  no_argument,       0, 'V'},
                {"config",   required_argument, 0, 'C'},
				{"archive",  required_argument, 0, 'a'},

                {0, 0, 0, 0}
            };

        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long (argc, argv, "a:hVC:",
                        long_options, &option_index);
#else
        c = getopt (argc, argv, "a:hVC:");
#endif 
        /* Detect the end of the options. */
        if (c == -1)
            break;

        switch (c)
        {
        case 0:
            /* If this option set a flag, do nothing else now. */
#ifdef HAVE_GETOPT_LONG 
            if (long_options[option_index].flag != 0)
                break;

            printf ("option %s", long_options[option_index].name);
            if (optarg)
                printf (" with arg %s", optarg);
            printf ("\n");
#endif
            break;

        case '?':
            /* getopt_long already printed an error message. */
            break;

		case 'a':
			strncpy (archive_directory, optarg, sizeof(archive_directory));
			break;

        case 'h':
            usage ();
            exit (0);

        case 'V':
            version ();
            exit (0);

        case 'C':
            strncpy (config_filename, optarg, sizeof(config_filename));
            break;

        default:
            usage ();
            abort ();
        }
    }

    initialize_configuration (archive_directory, config_filename);

	Arm4dbUtilityLock lock;

	Arm4db *arm4db_ptr = Arm4db::getArm4db ();
	status = arm4db_ptr->openDatabasesReadOnly ();
	if (status != ARM_SUCCESS)
	{
		lock.unlock ();
		return 1;
	}

    arm4db_ptr->dump ();
		
	arm4db_ptr->closeDatabases ();

	lock.unlock ();

    return 0;
}

