/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 
#include "config.h"
 
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <arpa/inet.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <db.h>
#include "arm4.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Arm4dbSharedMemory.h"
#include "arm4agent.h"

static int g_detailed = 0;
static int g_summary = 0;

#if 0
// Not currently used
static int
time_less (const arm_time_t *time1_ptr, const arm_time_t *time2_ptr)
{
	if (time1_ptr->tv_sec < time2_ptr->tv_sec)
		return ARM_TRUE;
	else if ((time1_ptr->tv_sec == time2_ptr->tv_sec) &&
				(time1_ptr->tv_usec < time2_ptr->tv_usec))
		return ARM_TRUE;
	
	return ARM_FALSE;
}

static int
time_greater (const arm_time_t *time1_ptr, const arm_time_t *time2_ptr)
{
	if (time1_ptr->tv_sec > time2_ptr->tv_sec)
		return ARM_TRUE;
	else if ((time1_ptr->tv_sec == time2_ptr->tv_sec) &&
				(time1_ptr->tv_usec > time2_ptr->tv_usec))
		return ARM_TRUE;
	
	return ARM_FALSE;
}
#endif

static void
initialize_configuration (const char *archive_directory, const char *config_filename)
{
	// Set the database configuration
	Arm4dbConfig::setConfig (config_filename);

	if (strlen (archive_directory) > 0)
		Arm4dbConfig::setDBHome (archive_directory);
}

static void
version (void)
{
	printf ("Package %s\n", PACKAGE);
	printf ("Version %s\n", VERSION);
}

static void
usage (void)
{
	printf ("Usage:\n");
#ifdef HAVE_GETOPT_LONG 
	printf ("\tarm4_report --version|-V\n");
	printf ("\t\tDisplay implementation version\n");
	printf ("\tarm4_report [--summary|-s] [--detailed|-d] [--help|-h] [--config|-C config_file] [--archive|-a archive_directory]\n");
#else
	printf ("\tarm4_report [-s] [-d] [-h] [-V] [-C config_file] [-a archive_directory]\n");
#endif
}

int
main(int argc, char *argv[])
{
	int rc = 0;
	int c;
    char config_filename [1024];
    char archive_directory [1024];

    // Initialize the logger
    Logger::getLogger().setName ("arm4_report");
    Logger::getLogger().setFile (stderr);

    /* Set the default config file */
    strcpy (config_filename, "/etc/arm4.conf");
	archive_directory [0] = '\0';
 
	while (1)
	{
#ifdef HAVE_GETOPT_LONG 
		static struct option long_options[] =
			{
					/* These options set a flag. */
				{"summary",  no_argument,       &g_summary, 1},
				{"detailed", no_argument,       &g_detailed, 1},
					/* These options don't set a flag. We distinguish them by their indices. */
				{"help",     no_argument,       0, 'h'},
				{"version",  no_argument,       0, 'V'},
                {"config",   required_argument, 0, 'C'},
				{"archive",  required_argument, 0, 'a'},

				{0, 0, 0, 0}
			};

		/* getopt_long stores the option index here. */
		int option_index = 0;
 
		c = getopt_long (argc, argv, "a:dhsVC:",
						long_options, &option_index);
#else
		c = getopt (argc, argv, "a:dhsVC:");
#endif 
		/* Detect the end of the options. */
		if (c == -1)
			break;
 
		switch (c)
		{
		case 0:
			/* If this option set a flag, do nothing else now. */
#ifdef HAVE_GETOPT_LONG 
			if (long_options[option_index].flag != 0)
				break;

			printf ("option %s", long_options[option_index].name);
			if (optarg)
				printf (" with arg %s", optarg);
			printf ("\n");
#endif
			break;

		case 'a':
			strncpy (archive_directory, optarg, sizeof(archive_directory));
			break;
 
		case 'd':
			g_detailed = 1;
			break;

		case 's':
			g_summary = 1;
			break;

		case '?':
			/* getopt_long already printed an error message. */
			break;
			
		case 'h':
			usage ();
			exit (0);

		case 'V':
			version ();
			exit (0);

        case 'C':
            strncpy (config_filename, optarg, sizeof(config_filename));
            break;

		default:
			usage ();
			abort ();
		}
	}

    initialize_configuration (archive_directory, config_filename);

	Arm4dbUtilityLock lock;

	Arm4db *arm4db_ptr = Arm4db::getArm4db ();
	int status = arm4db_ptr->openDatabases ();
	if (status != ARM_SUCCESS)
	{
		lock.unlock ();
		return 1;
	}

	/* At some point we should include options to specify what our report contains */
	if (g_summary || (g_detailed == 0))
		arm4db_ptr->printSummaries ();
	if (g_detailed)
		arm4db_ptr->printTransactions();
		
	arm4db_ptr->closeDatabases ();

	lock.unlock ();
	
	return rc;
}
