/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"
#include <string.h>

#ifdef HAVE_UUID_UUID_H
#include <uuid/uuid.h>
#else
#ifdef HAVE_UUID_H
#include <uuid.h>
#endif
#endif

#include "ArmNative.h"
#include "arm4.h"
#include "arm4agent.h"

// Set this to true for debugging
#if 0
#define ARM4_JNI_STATIC
#else
#define ARM4_JNI_STATIC static
#endif

#ifndef ARM_SUCCESS
#define ARM_SUCCESS		(0)
#define ARM_FAILURE		(-1)
#endif

typedef struct _java_string {
	jstring	string;
	const arm_char_t	*string_ptr;
} java_string_t;

typedef struct _java_properties {
	int property_count;
	int context_count;
	java_string_t	property_names [ARM_PROPERTY_MAX_COUNT];
	java_string_t	property_values [ARM_PROPERTY_MAX_COUNT];
	java_string_t	context_names [ARM_PROPERTY_MAX_COUNT];
	java_string_t	uri;
} java_properties_t;

#if 0 /* Useful routines for debugging */
ARM4_JNI_STATIC void
print_id (const char *prefix_ptr, arm_id_t *id_ptr)
{
	/* For debugging */
	int i;
	
	printf ("%s id:", prefix_ptr);
	for (i = 0; i < 16; i++)
		printf ("%02x", id_ptr->id_u.uint8 [i]);
	printf ("\n");
}

ARM4_JNI_STATIC void
print_handle (const char *prefix_ptr, arm_int64_t handle)
{
	/* For debugging */
	printf ("%s handle: %llu\n", prefix_ptr, handle);
}
#endif

ARM4_JNI_STATIC jbyteArray
get_token_data (JNIEnv *env, jobject token, jsize *length)
{
	jclass token_class;
	jfieldID token_data_id;
	jbyteArray token_data;
	
	if (token == NULL)
	{
		if (length)
			*length = 0;
		return NULL;
	}
	
	token_class = (*env)->GetObjectClass (env, token);
	token_data_id = (*env)->GetFieldID (env, token_class, "m_tokenData", "[B");
	token_data = (const jbyteArray) (*env)->GetObjectField (env, token, token_data_id);
	if (length)
	{
		if (token_data)
			*length = (*env)->GetArrayLength (env, token_data);
		else
			*length = 0;
	}
	
	return token_data;
}

ARM4_JNI_STATIC arm_error_t
get_id (JNIEnv *env, jobject token, arm_id_t *id_ptr)
{
	jbyteArray token_data;
	jbyte *data_ptr;
	jsize length;
	
	token_data = get_token_data (env, token, &length);
	if (token_data && (length == sizeof (arm_id_t)))
	{
		data_ptr = (*env)->GetByteArrayElements (env, token_data, NULL);
		memcpy (id_ptr, data_ptr, sizeof (arm_id_t));
		(*env)->ReleaseByteArrayElements (env, token_data, data_ptr, JNI_ABORT);

		return ARM_SUCCESS;
	}

	return ARM_FAILURE;
}

ARM4_JNI_STATIC void
get_correlator (JNIEnv *env, jobject token, arm_correlator_t *correlator_ptr)
{
	jbyteArray token_data;
	jbyte *data_ptr;
	jsize length;
	
	token_data = get_token_data (env, token, &length);
	if (token_data && (length <= ARM_CORR_MAX_LENGTH))
	{
		data_ptr = (*env)->GetByteArrayElements (env, token_data, NULL);
		memcpy (correlator_ptr, data_ptr, length);
		(*env)->ReleaseByteArrayElements (env, token_data, data_ptr, JNI_ABORT);
	}
}

ARM4_JNI_STATIC jobject
create_correlator (JNIEnv *env, const arm_correlator_t *correlator_ptr)
{
	jclass correlator_class;
	jmethodID constructor_id;
	jbyteArray correlator_array;
	jobject correlator_object;
	arm_correlator_length_t length;
	
	correlator_class = (*env)->FindClass (env, "org/arm4/arm40/transaction/ArmCorrelator");
	if (correlator_class)
	{
		/* Get the constructor ID */
		constructor_id = (*env)->GetMethodID (env, correlator_class, "<init>", "([B)V");
		if (constructor_id == NULL)
			return NULL;
		
		/* Get the required array size */
		arm_get_correlator_length (correlator_ptr, &length);
		if (length <= 0)
			return NULL;

		/* Copy the id to a byte array */
		correlator_array = (*env)->NewByteArray(env, length);
		(*env)->SetByteArrayRegion (env, correlator_array, 0, length, (jbyte *) correlator_ptr);
		
		/* Create the object */
		correlator_object = (*env)->NewObject (env, correlator_class, constructor_id, correlator_array);
		return correlator_object;
	}
	
	return NULL;
}

ARM4_JNI_STATIC void
set_correlator (JNIEnv *env, jobject tran, const arm_correlator_t *correlator_ptr)
{
	jclass tran_class;
	jfieldID tran_correlator_id;
	jobject correlator_object;
	
	tran_class = (*env)->GetObjectClass (env, tran);
	tran_correlator_id = (*env)->GetFieldID (env, tran_class, "m_currentCorrelator", "Lorg/arm4/arm40/transaction/ArmCorrelator;");
	correlator_object = create_correlator (env, correlator_ptr);
	(*env)->SetObjectField (env, tran, tran_correlator_id, correlator_object);
}

ARM4_JNI_STATIC jobject
create_id (JNIEnv *env, const arm_id_t *id_ptr)
{
	jclass id_class;
	jmethodID constructor_id;
	jbyteArray id_array;
	jobject id_object;
	
	id_class = (*env)->FindClass (env, "org/arm4/arm40/transaction/ArmID");
	if (id_class)
	{
		/* Get the constructor ID */
		constructor_id = (*env)->GetMethodID (env, id_class, "<init>", "([B)V");
		if (constructor_id == NULL)
			return NULL;

		/* Copy the id to a byte array */
		id_array = (*env)->NewByteArray(env, sizeof (arm_id_t));
		(*env)->SetByteArrayRegion (env, id_array, 0, sizeof (arm_id_t), (jbyte *) id_ptr);
		
		/* Create the object */
		id_object = (*env)->NewObject (env, id_class, constructor_id, id_array);
		return id_object;
	}
	
	return NULL;
}

ARM4_JNI_STATIC arm_tran_start_handle_t
get_tran_handle (JNIEnv *env, jobject tran)
{
	jclass tran_class;
	jfieldID tran_handle_id;
	arm_tran_start_handle_t tran_handle;
	
	tran_class = (*env)->GetObjectClass (env, tran);
	tran_handle_id = (*env)->GetFieldID (env, tran_class, "m_transactionHandle", "J");
	tran_handle = (arm_tran_start_handle_t) (*env)->GetLongField (env, tran, tran_handle_id);
	
	return tran_handle;
}

ARM4_JNI_STATIC void
set_tran_handle (JNIEnv *env, jobject tran, arm_tran_start_handle_t tran_handle)
{
	jclass tran_class;
	jfieldID tran_handle_id;
	
	tran_class = (*env)->GetObjectClass (env, tran);
	tran_handle_id = (*env)->GetFieldID (env, tran_class, "m_transactionHandle", "J");
	(*env)->SetLongField (env, tran, tran_handle_id, tran_handle);
}

ARM4_JNI_STATIC arm_app_start_handle_t
get_app_handle (JNIEnv *env, jobject application_object)
{
	arm_app_start_handle_t app_handle;
	jclass arm_application_class;
	jfieldID application_handle_id;

	arm_application_class = (*env)->GetObjectClass (env, application_object);
	application_handle_id = (*env)->GetFieldID (env, arm_application_class, "m_appHandle", "J");
	app_handle = (*env)->GetLongField (env, application_object, application_handle_id);
	
	return app_handle;
}

ARM4_JNI_STATIC void
set_app_handle (JNIEnv *env, jobject application_object, arm_app_start_handle_t app_handle)
{
	jclass arm_application_class;
	jfieldID application_handle_id;

	arm_application_class = (*env)->GetObjectClass (env, application_object);
	application_handle_id = (*env)->GetFieldID (env, arm_application_class, "m_appHandle", "J");
	(*env)->SetLongField (env, application_object, application_handle_id, app_handle);
}

ARM4_JNI_STATIC void
get_tran_id (JNIEnv *env, jobject transaction_definition_object, arm_id_t *tran_id_ptr)
{
	jclass transaction_definition_class;
	jfieldID tran_id_id;
	jobject tran_id_object;

	transaction_definition_class = (*env)->GetObjectClass (env, transaction_definition_object);
	tran_id_id = (*env)->GetFieldID (env, transaction_definition_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");
	tran_id_object = (*env)->GetObjectField (env, transaction_definition_object, tran_id_id);
	get_id (env, tran_id_object, tran_id_ptr);
}

ARM4_JNI_STATIC void
get_tran_alias (JNIEnv *env, jobject transaction_definition_object, arm_id_t **tran_alias_pptr)
{
	jclass transaction_definition_class;
	jfieldID tran_id_id;
	jobject tran_id_object;

	transaction_definition_class = (*env)->GetObjectClass (env, transaction_definition_object);
	if (transaction_definition_class == NULL)
	{
		*tran_alias_pptr = NULL;
		return;
	}
	tran_id_id = (*env)->GetFieldID (env, transaction_definition_class, "m_alias", "Lorg/arm4/arm40/transaction/ArmID;");
	if (tran_id_id == NULL)
	{
		*tran_alias_pptr = NULL;
		return;
	}
	tran_id_object = (*env)->GetObjectField (env, transaction_definition_object, tran_id_id);
	if (tran_id_object == NULL)
	{
		*tran_alias_pptr = NULL;
		return;
	}
	get_id (env, tran_id_object, *tran_alias_pptr);
}

ARM4_JNI_STATIC void
get_app_id (JNIEnv *env, jobject application_definition_object, arm_id_t *app_id_ptr)
{
	jclass application_definition_class;
	jfieldID application_id_id;
	jobject application_id_object;

	application_definition_class = (*env)->GetObjectClass (env, application_definition_object);
	application_id_id = (*env)->GetFieldID (env, application_definition_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");
	application_id_object = (*env)->GetObjectField (env, application_definition_object, application_id_id);
	get_id (env, application_id_object, app_id_ptr);
}

ARM4_JNI_STATIC arm_buffer4_t *
init_buffer (arm_buffer4_t *buffer_ptr, arm_subbuffer_t **sub_array)
{
	/* Initialize */
	buffer_ptr->count = 0;
	buffer_ptr->subbuffer_array = sub_array;
	
	return buffer_ptr;
}

ARM4_JNI_STATIC void
add_subbuffer (arm_buffer4_t *buffer_ptr, arm_subbuffer_t *subbuffer_ptr, int array_size)
{
	/* Check that the new size is within bounds, and that there's a subbuffer to add */
	if (subbuffer_ptr && (buffer_ptr->count < array_size))
		buffer_ptr->subbuffer_array [buffer_ptr->count++] = subbuffer_ptr;
}

ARM4_JNI_STATIC void
init_subbuffer (arm_subbuffer_t *buffer_ptr, arm_subbuffer_format_t format)
{
	buffer_ptr->format = format;
}

ARM4_JNI_STATIC arm_error_t
init_array_list_methods (JNIEnv *env, jobject list_object, jclass *class_ptr, jmethodID *get_id_ptr, jmethodID *size_id_ptr)
{
	static jclass array_list_class = NULL;
	static jmethodID get_id = NULL;
	static jmethodID size_id = NULL;

	if (array_list_class == NULL)
	{
		array_list_class = (*env)->GetObjectClass (env, list_object);
		if (array_list_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;

		get_id = (*env)->GetMethodID (env, array_list_class, "get", "(I)Ljava/lang/Object;");
		if (get_id == NULL)
		{
			array_list_class = NULL;
			return ARM_FAILURE_INTERNAL_ERROR;
		}
		
		size_id = (*env)->GetMethodID (env, array_list_class, "size", "()I");
		if (size_id == NULL)
		{
			array_list_class = NULL;
			return ARM_FAILURE_INTERNAL_ERROR;
		}
	}
	*class_ptr = array_list_class;
	*get_id_ptr = get_id;
	*size_id_ptr = size_id;
	
	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_property_fields (JNIEnv *env, jobject property_object, jclass *class_ptr, jfieldID *name_id_ptr, jfieldID *value_id_ptr)
{
	static jclass property_class = NULL;
	static jfieldID property_name_id = NULL;
	static jfieldID property_value_id = NULL;
	
	if (property_class == NULL)
	{
		property_class = (*env)->GetObjectClass (env, property_object);
		if (property_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;
		
		property_name_id = (*env)->GetFieldID (env, property_class, "m_name", "Ljava/lang/String;");
		if (property_name_id == NULL)
		{
			property_class = NULL;
			return ARM_FAILURE_INTERNAL_ERROR;
		}
		
		property_value_id = (*env)->GetFieldID (env, property_class, "m_value", "Ljava/lang/String;");
		if (property_value_id == NULL)
		{
			property_class = NULL;
			return ARM_FAILURE_INTERNAL_ERROR;
		}
	}
	
	*class_ptr = property_class;
	*name_id_ptr = property_name_id;
	*value_id_ptr = property_value_id;
	
	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_identity_properties (JNIEnv *env, jobject app_identity_object, arm_subbuffer_app_identity_t *app_identity_ptr,
		arm_property_t *property_array_ptr, arm_char_t **context_name_pptr, java_properties_t *property_ptr)
{
	jint i, size;
	
	jclass app_identity_class;
	jfieldID identity_properties_id;
	jobject identity_properties_object;
	jfieldID context_properties_id;
	jobject context_properties_object;
	
	jclass array_list_class;
	jmethodID get_id;
	jmethodID size_id;
	
	jobject property_object;
	jclass property_class;
	jfieldID property_name_id;
	jfieldID property_value_id;
	jobject name_object;
	jobject value_object;
	
	arm_error_t status;
	
	/* Build the subbuffer */
	init_subbuffer ((arm_subbuffer_t *) app_identity_ptr, ARM_SUBBUFFER_APP_IDENTITY);
	app_identity_ptr->identity_property_count = 0;
	app_identity_ptr->identity_property_array = property_array_ptr;
	app_identity_ptr->context_name_count = 0;
	app_identity_ptr->context_name_array = (const arm_char_t **) context_name_pptr;
	
	property_ptr->property_count = 0;
	property_ptr->context_count = 0;
	property_ptr->uri.string_ptr = NULL;

	app_identity_class = (*env)->GetObjectClass (env, app_identity_object);
	if (app_identity_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;

	identity_properties_id = (*env)->GetFieldID (env, app_identity_class, "m_identityProperties", "Ljava/util/ArrayList;");
	if (identity_properties_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	identity_properties_object = (*env)->GetObjectField (env, app_identity_object, identity_properties_id);
	if (identity_properties_object)
	{
		status = init_array_list_methods (env, identity_properties_object, &array_list_class, &get_id, &size_id);
		if (status != ARM_SUCCESS)
			return status;
		
		size = (*env)->CallIntMethod (env, identity_properties_object, size_id);
		if (size > ARM_PROPERTY_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		app_identity_ptr->identity_property_count = size;
		property_ptr->property_count = size;
		for (i = 0; i < size; i++)
		{
			property_object = (*env)->CallObjectMethod (env, identity_properties_object, get_id, i);
			if (property_object == NULL)
			{
				property_array_ptr [i].name = NULL;
				property_array_ptr [i].value = NULL;
				property_ptr->property_names [i].string_ptr = NULL;
				property_ptr->property_values [i].string_ptr = NULL;
			} else {
				status = init_property_fields (env, property_object, &property_class, &property_name_id, &property_value_id);
				if (status != ARM_SUCCESS)
					return status;

				name_object = (*env)->GetObjectField (env, property_object, property_name_id);
				if (name_object == NULL)
					property_array_ptr [i].name = NULL;
				else
					property_array_ptr [i].name = (*env)->GetStringUTFChars(env, name_object, NULL); // This needs to be released at some point
				property_ptr->property_names [i].string = name_object;
				property_ptr->property_names [i].string_ptr = property_array_ptr [i].name;

				value_object = (*env)->GetObjectField (env, property_object, property_value_id);
				if (value_object == NULL)
					property_array_ptr [i].value = NULL;
				else
					property_array_ptr [i].value = (*env)->GetStringUTFChars(env, value_object, NULL); // This needs to be released at some point

				property_ptr->property_values [i].string = value_object;
				property_ptr->property_values [i].string_ptr = property_array_ptr [i].value;
			}
		}
	}
	
	context_properties_id = (*env)->GetFieldID (env, app_identity_class, "m_contextProperties", "Ljava/util/ArrayList;");
	if (context_properties_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	context_properties_object = (*env)->GetObjectField (env, app_identity_object, context_properties_id);
	if (context_properties_object)
	{
		status = init_array_list_methods (env, context_properties_object, &array_list_class, &get_id, &size_id);
		if (status != ARM_SUCCESS)
			return status;
		
		size = (*env)->CallIntMethod (env, context_properties_object, size_id);
		if (size > ARM_PROPERTY_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		app_identity_ptr->context_name_count = size;
		property_ptr->context_count = size;
		for (i = 0; i < size; i++)
		{
			property_object = (*env)->CallObjectMethod (env, context_properties_object, get_id, i);
			if (property_object == NULL)
			{
				context_name_pptr [i] = NULL;
				property_ptr->context_names [i].string_ptr = NULL;
			} else {
				status = init_property_fields (env, property_object, &property_class, &property_name_id, &property_value_id);
				if (status != ARM_SUCCESS)
					return status;

				name_object = (*env)->GetObjectField (env, property_object, property_name_id);
				if (name_object == NULL)
					context_name_pptr [i] = NULL;
				else
					context_name_pptr [i] = (arm_char_t *) (*env)->GetStringUTFChars(env, name_object, NULL); // This needs to be released at some point

				property_ptr->context_names [i].string = name_object;
				property_ptr->context_names [i].string_ptr = context_name_pptr [i];
			}
		}
	}
	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_app_context (JNIEnv *env, jobject app_identity_object, arm_subbuffer_app_context_t *app_context_ptr,
		arm_char_t **context_value_pptr)
{
	jint i, size;
	
	jclass app_identity_class;
	jfieldID context_properties_id;
	jobject context_properties_object;
	
	jclass array_list_class;
	jmethodID get_id;
	jmethodID size_id;
	
	jobject property_object;
	jclass property_class;
	jfieldID property_name_id;
	jfieldID property_value_id;
	jobject value_object;
	
	arm_error_t status;
	
	/* Build the subbuffer */
	init_subbuffer ((arm_subbuffer_t *) app_context_ptr, ARM_SUBBUFFER_APP_CONTEXT);
	app_context_ptr->context_value_count = 0;
	app_context_ptr->context_value_array = (const arm_char_t **) context_value_pptr;

	app_identity_class = (*env)->GetObjectClass (env, app_identity_object);
	if (app_identity_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;

	context_properties_id = (*env)->GetFieldID (env, app_identity_class, "m_contextProperties", "Ljava/util/ArrayList;");
	if (context_properties_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	context_properties_object = (*env)->GetObjectField (env, app_identity_object, context_properties_id);
	if (context_properties_object)
	{
		status = init_array_list_methods (env, context_properties_object, &array_list_class, &get_id, &size_id);
		if (status != ARM_SUCCESS)
			return status;
		
		size = (*env)->CallIntMethod (env, context_properties_object, size_id);
		if (size > ARM_PROPERTY_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		app_context_ptr->context_value_count = size;
		for (i = 0; i < size; i++)
		{
			property_object = (*env)->CallObjectMethod (env, context_properties_object, get_id, i);
			if (property_object == NULL)
			{
				context_value_pptr [i] = NULL;
			} else {
				status = init_property_fields (env, property_object, &property_class, &property_name_id, &property_value_id);
				if (status != ARM_SUCCESS)
					return status;

				value_object = (*env)->GetObjectField (env, property_object, property_value_id);
				if (value_object == NULL)
					context_value_pptr [i] = NULL;
				else
					context_value_pptr [i] = (arm_char_t *) (*env)->GetStringUTFChars(env, value_object, NULL); // This needs to be released at some point
			}
		}
	}
	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_system_address (JNIEnv *env, jobject system_address_object,
					 arm_subbuffer_system_address_t *system_address_ptr,
					 arm_uint8_t *system_address_buffer_ptr)
{
	jclass system_address_class;
	jfieldID address_format_id;

	jbyteArray token_data;
	jbyte *data_ptr;
	jsize length;

	jfieldID address_id_id;
	jobject address_id_object;
	
	arm_error_t status;
	

	// Build the subbuffer
	init_subbuffer ((arm_subbuffer_t *) system_address_ptr, ARM_SUBBUFFER_SYSTEM_ADDRESS);
	system_address_ptr->address = system_address_buffer_ptr;

	// Set default values that allow us to recover gracefully from errors
	system_address_ptr->address_length = 0;
	system_address_ptr->id_valid = ARM_FALSE;

	system_address_class = (*env)->GetObjectClass (env, system_address_object);
	if (system_address_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;

	address_format_id = (*env)->GetFieldID (env, system_address_class, "m_addressFormat", "S");
	if (address_format_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	system_address_ptr->address_format = (*env)->GetShortField (env, system_address_object, address_format_id);

	token_data = get_token_data (env, system_address_object, &length);
	if (token_data)
	{
		if (length < 0)
			return ARM_FAILURE_INVALID_ARGUMENT;

		// The standard has no upper bound on size, but we trim it to something reasonable
		if (length > ARM_PROPERTY_URI_MAX_LENGTH)
			length = ARM_PROPERTY_URI_MAX_LENGTH;

		if (length > 0)
		{
			data_ptr = (*env)->GetByteArrayElements (env, token_data, NULL);
			memcpy (system_address_buffer_ptr, data_ptr, length);
			(*env)->ReleaseByteArrayElements (env, token_data, data_ptr, JNI_ABORT);
		}
	}
	system_address_ptr->address_length = length;

	address_id_id = (*env)->GetFieldID (env, system_address_class, "m_addressId", "Lorg/arm4/arm40/transaction/ArmID;");
	address_id_object = (*env)->GetObjectField (env, system_address_class, address_id_id);
	status = get_id (env, address_id_object, &(system_address_ptr->id));
	system_address_ptr->id_valid = (status == ARM_SUCCESS) ? ARM_TRUE : ARM_FALSE;

	return ARM_SUCCESS;
}

ARM4_JNI_STATIC void
free_identity_properties (JNIEnv *env, java_properties_t *property_ptr)
{
	int i;
	
	for (i = 0; i < property_ptr->property_count; i++)
	{
		if (property_ptr->property_names [i].string_ptr)
		{
			(*env)->ReleaseStringUTFChars (env, property_ptr->property_names [i].string, property_ptr->property_names [i].string_ptr);
		}
		if (property_ptr->property_values [i].string_ptr)
		{
			(*env)->ReleaseStringUTFChars (env, property_ptr->property_values [i].string, property_ptr->property_values [i].string_ptr);
		}
	}

	for (i = 0; i < property_ptr->context_count; i++)
	{
		if (property_ptr->context_names [i].string_ptr)
		{
			(*env)->ReleaseStringUTFChars (env, property_ptr->context_names [i].string, property_ptr->context_names [i].string_ptr);
		}
	}

	if (property_ptr->uri.string_ptr)
	{
		(*env)->ReleaseStringUTFChars (env, property_ptr->uri.string, property_ptr->uri.string_ptr);
	}
}

ARM4_JNI_STATIC arm_error_t
init_tran_identity_properties (JNIEnv *env, jobject tran_identity_object, arm_subbuffer_tran_identity_t *tran_identity_ptr,
		arm_property_t *property_array_ptr, arm_char_t **context_name_pptr, java_properties_t *property_ptr)
{
	jint i, size;
	
	jclass tran_identity_class;
	jfieldID identity_properties_id;
	jobject identity_properties_object;
	jfieldID context_properties_id;
	jobject context_properties_object;
	
	jclass array_list_class;
	jmethodID get_id;
	jmethodID size_id;
	
	jobject property_object;
	jclass property_class;
	jfieldID property_name_id;
	jfieldID property_value_id;
	jfieldID uri_id;
	jobject name_object;
	jobject value_object;
	jobject uri_object;
	
	arm_error_t status;
	
	/* Build the subbuffer */
	init_subbuffer ((arm_subbuffer_t *) tran_identity_ptr, ARM_SUBBUFFER_TRAN_IDENTITY);
	tran_identity_ptr->identity_property_count = 0;
	tran_identity_ptr->identity_property_array = property_array_ptr;
	tran_identity_ptr->context_name_count = 0;
	tran_identity_ptr->context_name_array = (const arm_char_t **) context_name_pptr;

	property_ptr->property_count = 0;
	property_ptr->context_count = 0;
	property_ptr->uri.string_ptr = NULL;

	tran_identity_class = (*env)->GetObjectClass (env, tran_identity_object);
	if (tran_identity_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;

	identity_properties_id = (*env)->GetFieldID (env, tran_identity_class, "m_identityProperties", "Ljava/util/ArrayList;");
	if (identity_properties_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	identity_properties_object = (*env)->GetObjectField (env, tran_identity_object, identity_properties_id);
	if (identity_properties_object)
	{
		status = init_array_list_methods (env, identity_properties_object, &array_list_class, &get_id, &size_id);
		if (status != ARM_SUCCESS)
			return status;
		
		size = (*env)->CallIntMethod (env, identity_properties_object, size_id);
		if (size > ARM_PROPERTY_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		tran_identity_ptr->identity_property_count = size;
		property_ptr->property_count = size;
		for (i = 0; i < size; i++)
		{
			property_object = (*env)->CallObjectMethod (env, identity_properties_object, get_id, i);
			if (property_object == NULL)
			{
				property_array_ptr [i].name = NULL;
				property_array_ptr [i].value = NULL;
				property_ptr->property_names [i].string_ptr = NULL;
				property_ptr->property_values [i].string_ptr = NULL;
			} else {
				status = init_property_fields (env, property_object, &property_class, &property_name_id, &property_value_id);
				if (status != ARM_SUCCESS)
					return status;

				name_object = (*env)->GetObjectField (env, property_object, property_name_id);
				if (name_object == NULL)
					property_array_ptr [i].name = NULL;
				else
					property_array_ptr [i].name = (*env)->GetStringUTFChars(env, name_object, NULL); // This needs to be released at some point
				property_ptr->property_names [i].string = name_object;
				property_ptr->property_names [i].string_ptr = property_array_ptr [i].name;

				value_object = (*env)->GetObjectField (env, property_object, property_value_id);
				if (value_object == NULL)
					property_array_ptr [i].value = NULL;
				else
					property_array_ptr [i].value = (*env)->GetStringUTFChars(env, value_object, NULL); // This needs to be released at some point
				property_ptr->property_values [i].string = value_object;
				property_ptr->property_values [i].string_ptr = property_array_ptr [i].value;
			}
		}
	}
	
	context_properties_id = (*env)->GetFieldID (env, tran_identity_class, "m_contextProperties", "Ljava/util/ArrayList;");
	if (context_properties_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	context_properties_object = (*env)->GetObjectField (env, tran_identity_object, context_properties_id);
	if (context_properties_object)
	{
		status = init_array_list_methods (env, context_properties_object, &array_list_class, &get_id, &size_id);
		if (status != ARM_SUCCESS)
			return status;
		
		size = (*env)->CallIntMethod (env, context_properties_object, size_id);
		if (size > ARM_PROPERTY_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		tran_identity_ptr->context_name_count = size;
		property_ptr->context_count = size;
		for (i = 0; i < size; i++)
		{
			property_object = (*env)->CallObjectMethod (env, context_properties_object, get_id, i);
			if (property_object == NULL)
			{
				context_name_pptr [i] = NULL;
				property_ptr->context_names [i].string_ptr = NULL;
			} else {
				status = init_property_fields (env, property_object, &property_class, &property_name_id, &property_value_id);
				if (status != ARM_SUCCESS)
					return status;

				name_object = (*env)->GetObjectField (env, property_object, property_name_id);
				if (name_object == NULL)
					context_name_pptr [i] = NULL;
				else
					context_name_pptr [i] = (arm_char_t *) (*env)->GetStringUTFChars(env, name_object, NULL); // This needs to be released at some point

				property_ptr->context_names [i].string = name_object;
				property_ptr->context_names [i].string_ptr = context_name_pptr [i];
			}
		}
	}
	
	uri_id = (*env)->GetFieldID (env, tran_identity_class, "m_uri", "Ljava/lang/String;");
	if (uri_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	uri_object = (*env)->GetObjectField (env, tran_identity_object, uri_id);
	if (uri_object)
		tran_identity_ptr->uri = (arm_char_t *) (*env)->GetStringUTFChars(env, uri_object, NULL); // This needs to be released at some point
	else
		tran_identity_ptr->uri = NULL;

	property_ptr->uri.string = uri_object;
	property_ptr->uri.string_ptr = tran_identity_ptr->uri;

	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_tran_context (JNIEnv *env, jobject tran_context_object, jobject tran_context_uri_object, arm_subbuffer_tran_context_t *tran_context_ptr,
		arm_char_t **context_value_pptr)
{
	jint i, size;
	
	jclass array_list_class;
	jmethodID get_id;
	jmethodID size_id;
	
	jobject value_object;
	
	arm_error_t status;
	
	/* Build the subbuffer */
	init_subbuffer ((arm_subbuffer_t *) tran_context_ptr, ARM_SUBBUFFER_TRAN_CONTEXT);
	tran_context_ptr->context_value_count = 0;
	tran_context_ptr->context_value_array = (const arm_char_t **) context_value_pptr;

	if (tran_context_object)
	{
		status = init_array_list_methods (env, tran_context_object, &array_list_class, &get_id, &size_id);
		if (status != ARM_SUCCESS)
			return status;
		
		size = (*env)->CallIntMethod (env, tran_context_object, size_id);
		if (size > ARM_PROPERTY_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		tran_context_ptr->context_value_count = size;
		for (i = 0; i < size; i++)
		{
			value_object = (*env)->CallObjectMethod (env, tran_context_object, get_id, i);
			if (value_object == NULL)
			{
				context_value_pptr [i] = NULL;
			} else {
				context_value_pptr [i] = (arm_char_t *) (*env)->GetStringUTFChars(env, value_object, NULL); // This needs to be released at some point
			}
		}
	}
	
	if (tran_context_uri_object)
		tran_context_ptr->uri = (arm_char_t *) (*env)->GetStringUTFChars(env, tran_context_uri_object, NULL); // This needs to be released at some point
	else
		tran_context_ptr->uri = NULL;

	return ARM_SUCCESS;
}

ARM4_JNI_STATIC jint
get_int32_metric_value (JNIEnv *env, jclass metric_class, jobject metric_object)
{
	jmethodID get_id;

	get_id = (*env)->GetMethodID (env, metric_class, "get", "()I");
	if (get_id == NULL)
		return 0;

	return (*env)->CallIntMethod (env, metric_object, get_id);
}

ARM4_JNI_STATIC jlong
get_int64_metric_value (JNIEnv *env, jclass metric_class, jobject metric_object)
{
	jmethodID get_id;

	get_id = (*env)->GetMethodID (env, metric_class, "get", "()J");
	if (get_id == NULL)
		return 0;

	return (*env)->CallLongMethod (env, metric_object, get_id);
}

ARM4_JNI_STATIC void
get_float_counter_metric_value (JNIEnv *env, jclass metric_class, jobject metric_object, arm_metric_cntrdivr32_t *value_ptr)
{
	jmethodID get_id;

	get_id = (*env)->GetMethodID (env, metric_class, "get", "()F");
	if (get_id == NULL)
	{
		value_ptr->counter = 0;
		value_ptr->divisor = 1;
		return;
	}

	*((float *) value_ptr) = (*env)->CallFloatMethod (env, metric_object, get_id);
}

ARM4_JNI_STATIC void
get_float_gauge_metric_value (JNIEnv *env, jclass metric_class, jobject metric_object, arm_metric_gaugedivr32_t *value_ptr)
{
	jmethodID get_id;

	get_id = (*env)->GetMethodID (env, metric_class, "get", "()F");
	if (get_id == NULL)
	{
		value_ptr->gauge = 0;
		value_ptr->divisor = 1;
		return;
	}

	*((float *) value_ptr) = (*env)->CallFloatMethod (env, metric_object, get_id);
}

ARM4_JNI_STATIC const arm_char_t *
get_string_metric_value (JNIEnv *env, jclass metric_class, jobject metric_object)
{
	jmethodID get_id;
	jobject metric_value_object;

	get_id = (*env)->GetMethodID (env, metric_class, "get", "()Ljava/lang/String;");
	if (get_id == NULL)
	{
		return NULL;
	}

	metric_value_object = (*env)->CallObjectMethod (env, metric_object, get_id);
	if (metric_value_object == NULL)
		return NULL;

	/* This needs to be released later */
	return (*env)->GetStringUTFChars(env, metric_value_object, NULL);
}

ARM4_JNI_STATIC arm_error_t
init_metric_value (JNIEnv *env, arm_metric_t *metric_value_ptr,
				   jobject metric_object, jboolean metric_valid, int slot)
{
	jclass metric_class;

	jfieldID metric_definition_id;
	jobject metric_definition_object;
	jclass metric_definition_class;

	jfieldID metric_usage_id;
	jfieldID metric_format_id;

	/* This has to be done every time, since each metric can have a different class */
	metric_class = (*env)->GetObjectClass (env, metric_object);
	if (metric_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* Get the metric definition */
	metric_definition_id = (*env)->GetFieldID (env, metric_class, "m_definition", "Lorg/arm4/arm40/metric/ArmMetricDefinition;");
	metric_definition_object = (*env)->GetObjectField (env, metric_object, metric_definition_id);
	metric_definition_class = (*env)->GetObjectClass (env, metric_definition_object);
	if (metric_definition_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	metric_value_ptr->valid = metric_valid;
	metric_value_ptr->slot = slot;

	metric_usage_id = (*env)->GetFieldID (env, metric_definition_class, "m_usage", "S");
	metric_value_ptr->usage = (*env)->GetShortField (env, metric_definition_object, metric_usage_id);

	metric_format_id = (*env)->GetFieldID (env, metric_definition_class, "m_format", "S");
	metric_value_ptr->format = (*env)->GetShortField (env, metric_definition_object, metric_format_id);

	switch (metric_value_ptr->format)
	{
	case ARM_METRIC_FORMAT_COUNTER32:
		metric_value_ptr->metric_u.counter32 = get_int32_metric_value (env, metric_class, metric_object);
		break;

	case ARM_METRIC_FORMAT_COUNTER64:
		metric_value_ptr->metric_u.counter64 = get_int64_metric_value (env, metric_class, metric_object);
		break;

	case ARM_METRIC_FORMAT_CNTRDIVR32:
		/* Indicate to the agent that this actually a floating point object! */
		metric_value_ptr->format = ARM_METRIC_FORMAT_COUNTERFLOAT32;
		get_float_counter_metric_value (env, metric_class, metric_object,
										&(metric_value_ptr->metric_u.counterdivisor32));
		break;

	case ARM_METRIC_FORMAT_GAUGE32:
		metric_value_ptr->metric_u.gauge32 = get_int32_metric_value (env, metric_class, metric_object);
		break;

	case ARM_METRIC_FORMAT_GAUGE64:
		metric_value_ptr->metric_u.gauge64 = get_int64_metric_value (env, metric_class, metric_object);
		break;

	case ARM_METRIC_FORMAT_GAUGEDIVR32:
		/* Indicate to the agent that this actually a floating point object! */
		metric_value_ptr->format = ARM_METRIC_FORMAT_GAUGEFLOAT32;
		get_float_gauge_metric_value (env, metric_class, metric_object,
									  &(metric_value_ptr->metric_u.gaugedivisor32));
		break;

	case ARM_METRIC_FORMAT_NUMERICID32:
		metric_value_ptr->metric_u.numericid32 = get_int32_metric_value (env, metric_class, metric_object);
		break;

	case ARM_METRIC_FORMAT_NUMERICID64:
		metric_value_ptr->metric_u.numericid64 = get_int64_metric_value (env, metric_class, metric_object);
		break;

	case ARM_METRIC_FORMAT_STRING32:
		metric_value_ptr->metric_u.string32 = get_string_metric_value (env, metric_class, metric_object);
		break;

	default:
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_metric_values (JNIEnv *env, jobject metric_group_object, arm_subbuffer_metric_values_t *metric_values_ptr,
	arm_metric_t *metric_value_ptr)
{
	jint i, size;
	
	jclass metric_group_class;
	jmethodID get_metric_id;
	jmethodID valid_id;

	jfieldID metric_id;
	jobject metric_object;
	
	jobject value_object;
	jboolean value_valid;
	
	arm_error_t status;
	
	/* Build the subbuffer */
	init_subbuffer ((arm_subbuffer_t *) metric_values_ptr, ARM_SUBBUFFER_METRIC_VALUES);
	metric_values_ptr->count = 0;
	metric_values_ptr->metric_value_array = (const arm_metric_t *) metric_value_ptr;

	if (metric_group_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	metric_group_class = (*env)->GetObjectClass (env, metric_group_object);
	if (metric_group_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	get_metric_id = (*env)->GetMethodID (env, metric_group_class, "getMetric", "(I)Lorg/opengroup/arm40/metric/ArmMetric;");
	if (get_metric_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	valid_id = (*env)->GetMethodID (env, metric_group_class, "isMetricValid", "(I)Z");
	if (valid_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	metric_id = (*env)->GetFieldID (env, metric_group_class, "m_metrics", "[Lorg/opengroup/arm40/metric/ArmMetric;");
	if (metric_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	metric_object = (*env)->GetObjectField (env, metric_group_object, metric_id);
	size = (*env)->GetArrayLength (env, metric_object);
	if (size > ARM_METRIC_MAX_COUNT)
		return ARM_FAILURE_INVALID_ARGUMENT;
	
	metric_values_ptr->count = size;
	for (i = 0; i < size; i++)
	{
		value_object = (*env)->CallObjectMethod (env, metric_group_object, get_metric_id, i);
		value_valid = (*env)->CallBooleanMethod (env, metric_group_object, valid_id, i);
		if (value_object == NULL)
		{
			return ARM_FAILURE_INVALID_ARGUMENT;
		} else {
			status = init_metric_value (env, &(metric_value_ptr [i]), value_object, value_valid, i);
			if (status != ARM_SUCCESS)
				return status;
		}
	}

	return ARM_SUCCESS;
}

/* This is required to release any metric strings */
#if 0
ARM4_JNI_STATIC arm_error_t
release_metric_values (JNIEnv *env, jobject metric_group_object, arm_subbuffer_metric_values_t *metric_values_ptr,
	arm_metric_t *metric_value_ptr)
#endif

ARM4_JNI_STATIC arm_error_t
init_metric_definition (JNIEnv *env, const arm_id_t *app_id_ptr, arm_metric_binding_t *metric_binding_ptr, 
						jobject definition_object, int slot)
{
	jclass metric_definition_class;

	jfieldID metric_name_id;
	jobject metric_name_object;
	const arm_char_t *metric_name_str;

	jfieldID metric_units_id;
	jobject metric_units_object;
	const arm_char_t *metric_units_str;

	jfieldID metric_usage_id;
	jshort metric_usage;

	jfieldID metric_format_id;
	jshort metric_format;

	jfieldID metric_id_id;
	jobject metric_id_object;
	arm_id_t input_metric_id, *input_metric_id_ptr;

	arm_error_t status;

	/* This needs to be checked each time, as each definition is a derived class */
	metric_definition_class = (*env)->GetObjectClass (env, definition_object);
	if (metric_definition_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;


	metric_binding_ptr->slot = slot;

	/* Get the parameters */
	metric_name_id = (*env)->GetFieldID (env, metric_definition_class, "m_name",
												"Ljava/lang/String;");
	if (metric_name_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	metric_name_object = (*env)->GetObjectField (env, definition_object, metric_name_id);
	if (metric_name_object)
		metric_name_str = (*env)->GetStringUTFChars(env, metric_name_object, NULL);
	else
		metric_name_str = NULL;

	metric_units_id = (*env)->GetFieldID (env, metric_definition_class, "m_units",
												"Ljava/lang/String;");
	if (metric_units_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	metric_units_object = (*env)->GetObjectField (env, definition_object, metric_units_id);
	if (metric_units_object)
		metric_units_str = (*env)->GetStringUTFChars(env, metric_units_object, NULL);
	else
		metric_units_str = NULL;

	metric_usage_id = (*env)->GetFieldID (env, metric_definition_class, "m_usage", "S");
	if (metric_usage_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	metric_usage = (*env)->GetShortField (env, definition_object, metric_usage_id);

	metric_format_id = (*env)->GetFieldID (env, metric_definition_class, "m_format", "S");
	if (metric_format_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	metric_format = (*env)->GetShortField (env, definition_object, metric_format_id);

	/* Transform floating point metrics to something the agent understands */
	if (metric_format == ARM_METRIC_FORMAT_CNTRDIVR32)
		metric_format = ARM_METRIC_FORMAT_COUNTERFLOAT32;
	else if (metric_format == ARM_METRIC_FORMAT_GAUGEDIVR32)
		metric_format = ARM_METRIC_FORMAT_GAUGEFLOAT32;

	input_metric_id_ptr = NULL;
	metric_id_id = (*env)->GetFieldID (env, metric_definition_class, "m_metricId",
												"Lorg/arm4/arm40/transaction/ArmID;");
	if (metric_id_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	metric_id_object = (*env)->GetObjectField (env, definition_object, metric_id_id);
	if (metric_id_object)
	{
		get_id (env, metric_id_object, &input_metric_id);
		input_metric_id_ptr = &input_metric_id;
	}

	status = arm_register_metric(
		app_id_ptr,
		metric_name_str,
		metric_format,
		metric_usage,
		metric_units_str,
		input_metric_id_ptr,
		0,
		NULL,
		&(metric_binding_ptr->id));

	/* Free any strings */
	if (metric_name_str)
		(*env)->ReleaseStringUTFChars (env, metric_name_id, metric_name_str);
	if (metric_units_str)
		(*env)->ReleaseStringUTFChars (env, metric_units_id, metric_units_str);

	return ARM_SUCCESS;
}

ARM4_JNI_STATIC arm_error_t
init_metrics (JNIEnv *env, jobject metric_group_definition_object, const arm_id_t *app_id_ptr,
	arm_subbuffer_metric_bindings_t *metric_buffer_ptr,
	arm_metric_binding_t *metric_binding_ptr)
{
	jint i, size;
	
	jclass metric_group_definition_class;
	jmethodID get_metric_definition_id;
	jfieldID metric_definition_id;

	jobject metric_definition_object;
	jobject definition_object;
	
	arm_error_t status;
	
	/* Build the subbuffer */
	init_subbuffer ((arm_subbuffer_t *) metric_buffer_ptr, ARM_SUBBUFFER_METRIC_BINDINGS);
	metric_buffer_ptr->count = 0;
	metric_buffer_ptr->metric_binding_array = (const arm_metric_binding_t *) metric_binding_ptr;

	metric_group_definition_class = (*env)->GetObjectClass (env, metric_group_definition_object);
	if (metric_group_definition_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	get_metric_definition_id = (*env)->GetMethodID (env, metric_group_definition_class, "getMetricDefinition",
										 "(I)Lorg/opengroup/arm40/metric/ArmMetricDefinition;");
	if (get_metric_definition_id == NULL)
	{
		metric_group_definition_class = NULL;
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	metric_definition_id = (*env)->GetFieldID (env, metric_group_definition_class, "m_definitions",
											    "[Lorg/opengroup/arm40/metric/ArmMetricDefinition;");
	if (metric_definition_id == NULL)
	{
		metric_group_definition_class = NULL;
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	if (metric_group_definition_object)
	{
		metric_definition_object = (*env)->GetObjectField (env, metric_group_definition_object, metric_definition_id);
		size = (*env)->GetArrayLength (env, metric_definition_object);
		if (size > ARM_METRIC_MAX_COUNT)
			return ARM_FAILURE_INVALID_ARGUMENT;
		
		metric_buffer_ptr->count = size;
		for (i = 0; i < size; i++)
		{
			definition_object = (*env)->CallObjectMethod (env, metric_group_definition_object,
													 get_metric_definition_id, i);
			if (definition_object == NULL)
			{
				return ARM_FAILURE_INVALID_ARGUMENT;
			} else {
				status = init_metric_definition (env, app_id_ptr, &(metric_binding_ptr [i]), definition_object, i);
				if (status != ARM_SUCCESS)
					return status;
			}
		}
	}

	return ARM_SUCCESS;
}

JNIEXPORT jstring JNICALL
Java_org_arm4_arm40_jni_ArmNative_getErrorMessage (JNIEnv *env, jclass arm_native, jint error_code)
{
	arm_char_t buffer [ARM_MSG_BUFFER_LENGTH];
	arm_error_t status;

	status = arm_get_error_message(
				ARM_CHARSET_UTF8,
				error_code,
				buffer);
	if (status == ARM_SUCCESS)
		return (*env)->NewStringUTF(env, buffer);

	return NULL;
}

JNIEXPORT jbyteArray JNICALL
Java_org_arm4_arm40_jni_ArmNative_generateUUID (JNIEnv *env, jclass arm_native)
{
	arm_id_t uuid;
	jbyteArray uuid_buffer;

	/* Generate our uuid */
    uuid_generate ((arm_uint8_t *) &uuid);

	/* Copy it to a byte array */
	uuid_buffer = (*env)->NewByteArray(env, sizeof (arm_id_t));
	(*env)->SetByteArrayRegion (env, uuid_buffer, 0, sizeof (arm_id_t), (jbyte *) &uuid);
	
	return uuid_buffer;
}

JNIEXPORT jbyteArray JNICALL
Java_org_arm4_arm40_jni_ArmNative_generateCorrelator (JNIEnv *env, jclass arm_native, jlong app_handle,
	jobject tran_id_object, jobject parent_correlator_object)
{
	arm_correlator_t correlator;
	arm_correlator_t parent_correlator, *parent_ptr;
	arm_id_t tran_id;

	jbyteArray correlator_buffer;

	arm_error_t status;

	/* Get our objects */
	get_id (env, tran_id_object, &tran_id);
	parent_ptr = NULL;
	if (parent_correlator_object != NULL)
	{
		get_correlator (env, parent_correlator_object, &parent_correlator);
		parent_ptr = &parent_correlator;
	}

	status = arm_generate_correlator(
		app_handle,
		&tran_id,
		parent_ptr,
		0,
		NULL, // Should support arm_subbuffer_trancontext_t, and arm_subbuffer_user_t
		&correlator);
	if (status != ARM_SUCCESS)
		return NULL;

	/* Copy it to a byte array */
	correlator_buffer = (*env)->NewByteArray(env, sizeof (arm_correlator_t));
	(*env)->SetByteArrayRegion (env, correlator_buffer, 0, sizeof (arm_correlator_t), (jbyte *) &correlator);
	
	return correlator_buffer;
}


JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_registerApplication (JNIEnv *env, jclass arm_native, jobject app)
{
	/* Information needed to access the java object */
	jclass arm_application_definition_class;
	jfieldID app_name_id;
	jfieldID app_properties_id;
	jfieldID app_id_id;
	jfieldID app_alias_id;
	jobject name_object;
	jobject app_properties_object;
	jobject app_id_object;
	jobject app_alias_object;
	const char *name_ptr;
	
	arm_error_t status;
	arm_id_t input_app_id, *input_app_id_ptr;
    arm_id_t output_app_id;
    
    arm_buffer4_t buffer4, *buffer_ptr;
    arm_subbuffer_t *subbuffer_array [1];
    arm_subbuffer_t *subbuffer_ptr;

    arm_subbuffer_app_identity_t app_identity;
    arm_property_t property_array [ARM_PROPERTY_MAX_COUNT];
    arm_char_t *context_name_array [ARM_PROPERTY_MAX_COUNT];
    java_properties_t property_saver;
    
	arm_application_definition_class = (*env)->GetObjectClass (env, app);
	if (arm_application_definition_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;
	
	app_name_id = (*env)->GetFieldID (env, arm_application_definition_class, "m_name", "Ljava/lang/String;");
	app_properties_id = (*env)->GetFieldID (env, arm_application_definition_class, "m_properties", "Lorg/arm4/arm40/transaction/ArmIdentityProperties;");
	app_id_id = (*env)->GetFieldID (env, arm_application_definition_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");
	app_alias_id = (*env)->GetFieldID (env, arm_application_definition_class, "m_alias", "Lorg/arm4/arm40/transaction/ArmID;");
	
	/* Get the application name */
	name_object = (*env)->GetObjectField (env, app, app_name_id);
	name_ptr = (*env)->GetStringUTFChars(env, name_object, NULL);
	
	/* Get the input_app_id, if any */
	input_app_id_ptr = NULL;
	app_alias_object = (*env)->GetObjectField (env, app, app_alias_id);
	if (app_alias_object)
	{
		get_id (env, app_alias_object, &input_app_id);
		input_app_id_ptr = &input_app_id;
	}

	/* See if there are any properties */
	subbuffer_ptr = NULL;
	app_properties_object = (*env)->GetObjectField (env, app, app_properties_id);
	if (app_properties_object != NULL)
	{
		status = init_identity_properties (env, app_properties_object, &app_identity, property_array, context_name_array, &property_saver);
		if (status != ARM_SUCCESS)
			return status;
		
		subbuffer_ptr = (arm_subbuffer_t *) &app_identity;
	}
	
	/* Build the buffer object */
	if (subbuffer_ptr != NULL)
	{
		buffer_ptr = init_buffer (&buffer4, subbuffer_array);
		add_subbuffer (buffer_ptr, subbuffer_ptr, 1);
	}
	else
	{
		buffer_ptr = NULL;
	}
	
	status = arm_register_application(
					name_ptr,
					input_app_id_ptr,
					0, /* flags */
					buffer_ptr, /* buffer4 */
					&output_app_id);
	
	/* Release the application name */
	(*env)->ReleaseStringUTFChars (env, name_object, name_ptr);
	if (app_properties_object != NULL)
		free_identity_properties (env, &property_saver);
	
	/* Save the app_id to the object */
	app_id_object = create_id (env, &output_app_id);
	if (app_id_object)
	{
		(*env)->SetObjectField (env, app, app_id_id, app_id_object);
	}
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_destroyApplication (JNIEnv *env, jclass arm_native, jobject app)
{
	/* Information needed to access the java object */
	jclass arm_application_definition_class;
	jfieldID app_id_id;
	jobject app_id_object;
	
	arm_error_t status;
	arm_id_t input_app_id;

	arm_application_definition_class = (*env)->GetObjectClass (env, app);
	if (arm_application_definition_class == NULL)
			return ARM_FAILURE_INTERNAL_ERROR;
	
	/* Get the application ID */
	app_id_id = (*env)->GetFieldID (env, arm_application_definition_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");
	app_id_object = (*env)->GetObjectField (env, app, app_id_id);
	if (app_id_object)
		get_id (env, app_id_object, &input_app_id);

	status = arm_destroy_application(
				&input_app_id,
				0,
				NULL);

	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_registerTransaction (JNIEnv *env, jclass arm_native, jobject tran)
{
	arm_error_t status;
	
	jclass transaction_definition_class;

	jfieldID application_definition_id;
	jobject application_definition_object;
	arm_id_t app_id;
	
	jfieldID transaction_name_id;
	jstring tran_name_string;
	const arm_char_t *tran_name;
	
	arm_id_t input_tran_id;
	arm_id_t *input_tran_id_ptr;

	jfieldID tran_id_id;
	jobject tran_id_object;
	arm_id_t output_tran_id;
	
	jfieldID tran_identity_id;
	jobject tran_identity_object;

	jfieldID metric_definition_id;
	jobject metric_definition_object;
    
    arm_buffer4_t buffer4, *buffer_ptr;
    arm_subbuffer_t *subbuffer_array [2]; /* Room for metric subbuffer */

    arm_subbuffer_tran_identity_t tran_identity;
    arm_property_t property_array [ARM_PROPERTY_MAX_COUNT];
    arm_char_t *context_name_array [ARM_PROPERTY_MAX_COUNT];

	arm_subbuffer_metric_bindings_t metric_binding_buffer;
	arm_metric_binding_t metric_binding_array [ARM_METRIC_MAX_COUNT];

    java_properties_t property_saver;

	/* Get the object information */
	transaction_definition_class = (*env)->GetObjectClass (env, tran);
	if (transaction_definition_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	
	/* Get the input parameters */
	application_definition_id = (*env)->GetFieldID (env, transaction_definition_class, "m_appDefinition", "Lorg/arm4/arm40/transaction/ArmApplicationDefinition;");
	if (application_definition_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	application_definition_object = (*env)->GetObjectField (env, tran, application_definition_id);
	if (application_definition_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	get_app_id (env, application_definition_object, &app_id);

	transaction_name_id = (*env)->GetFieldID (env, transaction_definition_class, "m_name", "Ljava/lang/String;");
	tran_name_string = (*env)->GetObjectField (env, tran, transaction_name_id);
	if (tran_name_string == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	tran_name = (*env)->GetStringUTFChars (env, tran_name_string, NULL);

    input_tran_id_ptr = &input_tran_id; // This will be set to NULL if no alias
	get_tran_alias (env, tran, &input_tran_id_ptr);
	tran_id_id = (*env)->GetFieldID (env, transaction_definition_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");

	/* Build the buffer object */
	buffer_ptr = init_buffer (&buffer4, subbuffer_array);

	/* See if there are any identity properties */
	tran_identity_id = (*env)->GetFieldID (env, transaction_definition_class, "m_properties", "Lorg/arm4/arm40/transaction/ArmIdentityPropertiesTransaction;");
	if (tran_identity_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	tran_identity_object = (*env)->GetObjectField (env, tran, tran_identity_id);
	if (tran_identity_object != NULL)
	{
		status = init_tran_identity_properties (env, tran_identity_object, &tran_identity, property_array, context_name_array, &property_saver);
		if (status != ARM_SUCCESS)
			return status;
		
		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &tran_identity, 2);
	}
	
	/* Get the metric definitions */
	metric_definition_id = (*env)->GetFieldID (env, transaction_definition_class, "m_metricGroupDefinition", "Lorg/arm4/arm40/metric/ArmMetricGroupDefinition;");
	metric_definition_object = (*env)->GetObjectField (env, tran, metric_definition_id);
	if (metric_definition_object != NULL)
	{
		status = init_metrics (env, metric_definition_object, &app_id, &metric_binding_buffer,
							   metric_binding_array);
		if (status != ARM_SUCCESS)
			return status;
		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &metric_binding_buffer, 2);
	}

	/* Do the registration */
	status = arm_register_transaction(
				&app_id,
				tran_name,
				input_tran_id_ptr,
				0,
				buffer_ptr,
				&output_tran_id);

	/* Save the transaction ID */
	tran_id_object = create_id (env, &output_tran_id);
	if (tran_id_object)
	{
		(*env)->SetObjectField (env, tran, tran_id_id, tran_id_object);
	}
	
	/* Clean up */
	(*env)->ReleaseStringUTFChars (env, tran_name_string, tran_name);
	if (tran_identity_object != NULL)
		free_identity_properties (env, &property_saver);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_startApplication  (JNIEnv *env, jclass arm_native, jobject app)
{
	arm_error_t status;

	jclass application_class;

	jfieldID app_definition_id;
	jobject app_definition_object;
	arm_id_t app_id;
	
	jfieldID group_id;
	jstring group_string;
	const arm_char_t *group_ptr;
	
	jfieldID instance_id;
	jstring instance_string;
	const arm_char_t *instance_ptr;
	
	arm_app_start_handle_t app_handle;

	arm_buffer4_t buffer4, *buffer_ptr;
    arm_subbuffer_t *subbuffer_array [2];

    arm_subbuffer_app_context_t app_context;
    arm_char_t *context_value_array [ARM_PROPERTY_MAX_COUNT];
    arm_subbuffer_t *context_subbuffer_ptr;

	arm_subbuffer_system_address_t system_address;
    arm_subbuffer_t *system_address_subbuffer_ptr;
	arm_uint8_t system_address_buffer [ARM_PROPERTY_URI_MAX_LENGTH];
    
    jfieldID app_context_id;
    jobject app_context_object;

	jfieldID system_address_id;
	jobject system_address_object;

	/* Get the object information */
	application_class = (*env)->GetObjectClass (env, app);
	if (application_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	
	/* Get the inputs */
	app_definition_id = (*env)->GetFieldID (env, application_class, "m_appDefinition", "Lorg/arm4/arm40/transaction/ArmApplicationDefinition;");
	if (app_definition_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	app_definition_object = (*env)->GetObjectField (env, app, app_definition_id);
	if (app_definition_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	get_app_id (env, app_definition_object, &app_id);

	group_id = (*env)->GetFieldID (env, application_class, "m_group", "Ljava/lang/String;");
	group_string = (*env)->GetObjectField (env, app, group_id);
	if (group_string == NULL)
	{
		/* This is an optional parameter */
		group_ptr = NULL;
	}
	else
	{
		group_ptr = (*env)->GetStringUTFChars (env, group_string, NULL);
	}

	instance_id = (*env)->GetFieldID (env, application_class, "m_instance", "Ljava/lang/String;");
	instance_string = (*env)->GetObjectField (env, app, instance_id);
	if (instance_string == NULL)
	{
		/* This is an optional parameter */
		instance_ptr = NULL;
	}
	else
	{
		instance_ptr = (*env)->GetStringUTFChars (env, instance_string, NULL);
	}

	/* See if there are any context values */
	context_subbuffer_ptr = NULL;
	app_context_id = (*env)->GetFieldID (env, application_class, "m_properties", "Lorg/arm4/arm40/transaction/ArmIdentityProperties;");
	if (app_context_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	app_context_object = (*env)->GetObjectField (env, app, app_context_id);
	if (app_context_object != NULL)
	{
		status = init_app_context (env, app_context_object, &app_context, context_value_array);
		if (status != ARM_SUCCESS)
			return status;
		
		context_subbuffer_ptr = (arm_subbuffer_t *) &app_context;
	}

	/* Get the system address subbuffer */
	system_address_subbuffer_ptr = NULL;
	system_address_id = (*env)->GetFieldID (env, application_class, "m_systemAddress", "Lorg/arm4/arm40/tranreport/ArmSystemAddress;");
	if (system_address_id == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	system_address_object = (*env)->GetObjectField (env, app, system_address_id);
	if (system_address_object != NULL)
	{
		status = init_system_address (env, system_address_object, &system_address, system_address_buffer);
		if (status != ARM_SUCCESS)
			return status;
		
		system_address_subbuffer_ptr = (arm_subbuffer_t *) &system_address;
	}
	
	/* Build the buffer object */
	if ((context_subbuffer_ptr != NULL) || (system_address_subbuffer_ptr != NULL))
	{
		buffer_ptr = init_buffer (&buffer4, subbuffer_array);
		int index = 1;
		if (context_subbuffer_ptr != NULL)
		{
			add_subbuffer (buffer_ptr, context_subbuffer_ptr, index);
			index++;
		}
		if (system_address_subbuffer_ptr != NULL)
		{
			add_subbuffer (buffer_ptr, system_address_subbuffer_ptr, index);
			index++;
		}
	}
	else
	{
		buffer_ptr = NULL;
	}
	
	/* Start the application */
	status = arm_start_application(
				&app_id,
				group_ptr,
				instance_ptr,
				0,
				buffer_ptr,
				&app_handle);
	
	/* Save the application handle */
	set_app_handle (env, app, app_handle);
	
	/* Cleanup */
	if (group_string != NULL)
		(*env)->ReleaseStringUTFChars (env, group_string, group_ptr);
	if (instance_string != NULL)
		(*env)->ReleaseStringUTFChars (env, instance_string, instance_ptr);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_stopApplication (JNIEnv *env, jclass arm_native, jobject app)
{
	arm_error_t status;
	arm_app_start_handle_t app_handle;
	
	app_handle = get_app_handle (env, app);
	
	status = arm_stop_application(app_handle, 0, NULL);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_startTransaction (JNIEnv *env, jclass arm_native, jobject tran)
{
	arm_id_t tran_id;
    arm_app_start_handle_t app_handle;
	arm_tran_start_handle_t tran_handle;
    arm_correlator_t current_correlator;
    arm_correlator_t parent_correlator, *parent_ptr;
	arm_error_t status;
	int flags = 0;

	jclass arm_transaction_class;

	jfieldID application_id;
	jobject application_object;
	
	jfieldID transaction_definition_id;
	jobject transaction_definition_object;
	
	jfieldID parent_id;
	jobject parent_object;
	
	jfieldID trace_id;
	jboolean trace_requested;
	
	jfieldID arrival_time_id;
	jlong arrival_time;

	arm_buffer4_t buffer4, *buffer_ptr;
    arm_subbuffer_t *subbuffer_array [4];

    arm_subbuffer_arrival_time_t arrival_time_buffer;
    arm_subbuffer_tran_context_t tran_context;
    arm_subbuffer_user_t user_buffer;
    arm_char_t *context_value_array [ARM_PROPERTY_MAX_COUNT];

	arm_subbuffer_metric_values_t metric_values;
	arm_metric_t metric_value_array [ARM_METRIC_MAX_COUNT];

	jfieldID metric_group_id;
	jobject metric_group_object;
    
    jfieldID tran_context_id;
    jobject tran_context_object;
    jfieldID tran_context_uri_id;
    jobject tran_context_uri_object;

    jclass user_class;
    jfieldID user_id;
    jobject user_object;
	jfieldID user_name_id;
	jobject user_name_object;
	jfieldID user_id_id;
	jobject user_id_object;
	
	arm_transaction_class = (*env)->GetObjectClass (env, tran);
	if (arm_transaction_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* Get the application handle */
	application_id = (*env)->GetFieldID (env, arm_transaction_class, "m_application", "Lorg/arm4/arm40/transaction/ArmApplication;");
	application_object = (*env)->GetObjectField (env, tran, application_id);
	if (application_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	app_handle = get_app_handle (env, application_object);
	
	/* Get the transaction ID */
	transaction_definition_id = (*env)->GetFieldID (env, arm_transaction_class, "m_definition", "Lorg/arm4/arm40/transaction/ArmTransactionDefinition;");
	transaction_definition_object = (*env)->GetObjectField (env, tran, transaction_definition_id);
	if (transaction_definition_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	get_tran_id (env, transaction_definition_object, &tran_id);
	
	/* Get the parent correlator */
	parent_ptr = NULL;
	parent_id = (*env)->GetFieldID (env, arm_transaction_class, "m_parentCorrelator", "Lorg/arm4/arm40/transaction/ArmCorrelator;");
	parent_object = (*env)->GetObjectField (env, tran, parent_id);
	if (parent_object != NULL)
	{
		get_correlator (env, parent_object, &parent_correlator);
		parent_ptr = &parent_correlator;
	}
	
	/* See if a trace is requested */
	trace_id = (*env)->GetFieldID (env, arm_transaction_class, "m_traceRequested", "Z");
	trace_requested = (*env)->GetBooleanField (env, tran, transaction_definition_id);
	if (trace_requested)
		flags |= ARM_FLAG_TRACE_REQUEST;

	/* Build the buffer object */
	buffer_ptr = init_buffer (&buffer4, subbuffer_array);
	
	/* Get the start time */
	arrival_time_id = (*env)->GetFieldID (env, arm_transaction_class, "m_arrivalTime", "J");
	arrival_time = (*env)->GetLongField (env, tran, arrival_time_id);
	if (arrival_time != 0)
	{
		init_subbuffer ((arm_subbuffer_t *) &arrival_time_buffer, ARM_SUBBUFFER_ARRIVAL_TIME);
		arrival_time_buffer.opaque_time = arrival_time;
		
		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &arrival_time_buffer, 4);
	}

	/* Get the user */
	user_id = (*env)->GetFieldID (env, arm_transaction_class, "m_user", "Lorg/arm4/arm40/transaction/ArmUser;");
	user_object = (*env)->GetObjectField (env, tran, user_id);
	if (user_object != NULL)
	{
		init_subbuffer ((arm_subbuffer_t *) &user_buffer, ARM_SUBBUFFER_USER);

		user_class = (*env)->GetObjectClass (env, user_object);
		user_name_id = (*env)->GetFieldID (env, user_class, "m_name", "Ljava/lang/String;");
		user_name_object = (*env)->GetObjectField (env, user_object, user_name_id);
		if (user_name_object)
			user_buffer.name = (arm_char_t *) (*env)->GetStringUTFChars(env, user_name_object, NULL);
		else
			user_buffer.name = NULL;

		user_id_id = (*env)->GetFieldID (env, user_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");
		user_id_object = (*env)->GetObjectField (env, user_object, user_id_id);
		if (user_id_object)
		{
			user_buffer.id_valid = 1;
			get_id (env, user_id_object, &user_buffer.id);
		}
		else
			user_buffer.id_valid = 0;
		
		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &user_buffer, 4);
	}

	/* See if there are any context values */
	tran_context_id = (*env)->GetFieldID (env, arm_transaction_class, "m_contextValues", "Ljava/util/ArrayList;");
	tran_context_object = (*env)->GetObjectField (env, tran, tran_context_id);
	tran_context_uri_id = (*env)->GetFieldID (env, arm_transaction_class, "m_contextUri", "Ljava/lang/String;");
	tran_context_uri_object = (*env)->GetObjectField (env, tran, tran_context_uri_id);
	if (tran_context_object != NULL)
	{
		status = init_tran_context (env, tran_context_object, tran_context_uri_object, &tran_context, context_value_array);
		if (status != ARM_SUCCESS)
			return status;
		
		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &tran_context, 4);
	}

	/* And finally the metric values */
	metric_group_id = (*env)->GetFieldID (env, arm_transaction_class, "m_metricGroup", "Lorg/arm4/arm40/metric/ArmMetricGroup;");
	metric_group_object = (*env)->GetObjectField (env, tran, metric_group_id);
	if (metric_group_object != NULL)
	{
		status = init_metric_values (env, metric_group_object, &metric_values, metric_value_array);
		if (status != ARM_SUCCESS)
			return status;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &metric_values, 4);
	}
	
	status = arm_start_transaction(
					app_handle,
					&tran_id,
					parent_ptr,
					flags,
					buffer_ptr,
					&tran_handle,
					&current_correlator);
	
	/* Save the transaction handle and correlator to the object */
	set_tran_handle (env, tran, tran_handle);
	set_correlator (env, tran, &current_correlator);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_stopTransaction (JNIEnv *env, jclass arm_native, jobject tran, jint tran_status, jstring detail)
{
	arm_tran_start_handle_t tran_handle;
	arm_error_t status;

	arm_buffer4_t buffer4, *buffer_ptr;
	arm_subbuffer_t *subbuffer_array [1];

	arm_subbuffer_metric_values_t metric_values;
	arm_metric_t metric_value_array [ARM_METRIC_MAX_COUNT];

	jclass arm_transaction_class;

	jfieldID metric_group_id;
	jobject metric_group_object;

	/* Get the object class */
	arm_transaction_class = (*env)->GetObjectClass (env, tran);
	if (arm_transaction_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	tran_handle = get_tran_handle (env, tran);

	/* Build the buffer object */
	buffer_ptr = init_buffer (&buffer4, subbuffer_array);
	metric_group_id = (*env)->GetFieldID (env, arm_transaction_class, "m_metricGroup", "Lorg/arm4/arm40/metric/ArmMetricGroup;");
	metric_group_object = (*env)->GetObjectField (env, tran, metric_group_id);
	if (metric_group_object != NULL)
	{
		status = init_metric_values (env, metric_group_object, &metric_values, metric_value_array);
		if (status != ARM_SUCCESS)
			return status;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &metric_values, 1);
	}

	status = arm_stop_transaction(tran_handle,
					tran_status,
					0, /* Flags */
					buffer_ptr);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_updateTransaction (JNIEnv *env, jclass arm_native, jobject tran)
{
	arm_tran_start_handle_t tran_handle;
	arm_error_t status;

	arm_buffer4_t buffer4, *buffer_ptr;
	arm_subbuffer_t *subbuffer_array [1];

	arm_subbuffer_metric_values_t metric_values;
	arm_metric_t metric_value_array [ARM_METRIC_MAX_COUNT];

	jclass arm_transaction_class;

	jfieldID metric_group_id;
	jobject metric_group_object;

	/* Get the object class */
	arm_transaction_class = (*env)->GetObjectClass (env, tran);
	if (arm_transaction_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	tran_handle = get_tran_handle (env, tran);

	/* Build the buffer object */
	buffer_ptr = init_buffer (&buffer4, subbuffer_array);
	metric_group_id = (*env)->GetFieldID (env, arm_transaction_class, "m_metricGroup", "Lorg/arm4/arm40/metric/ArmMetricGroup;");
	metric_group_object = (*env)->GetObjectField (env, tran, metric_group_id);
	if (metric_group_object != NULL)
	{
		status = init_metric_values (env, metric_group_object, &metric_values, metric_value_array);
		if (status != ARM_SUCCESS)
			return status;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &metric_values, 1);
	}

	status = arm_update_transaction(
		tran_handle,
		0,
		buffer_ptr);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_resetTransaction (JNIEnv *env, jclass arm_native, jlong transaction_handle)
{
	arm_error_t status;
	
	status = arm_discard_transaction(transaction_handle, 0, NULL);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_reportTransaction (JNIEnv *env, jclass arm_native,
	jobject tran_report_object,
	jint tran_status,
	jlong response_time,
	jlong stop_time,
	jstring diag_detail_object)
{
	arm_id_t tran_id;
	arm_app_start_handle_t app_handle;
	arm_correlator_t current_correlator, *current_ptr;
	arm_correlator_t parent_correlator, *parent_ptr;
	arm_error_t status;
	const arm_char_t *diag_detail_ptr;
	int flags = 0;

	jclass arm_tran_report_class;

	jfieldID application_id;
	jobject application_object;

	jfieldID transaction_definition_id;
	jobject transaction_definition_object;

	jfieldID parent_id;
	jobject parent_object;

	jfieldID current_id;
	jobject current_object;

	jfieldID trace_id;
	jboolean trace_requested;

	arm_buffer4_t buffer4, *buffer_ptr;
	arm_subbuffer_t *subbuffer_array [4];

	arm_subbuffer_diag_detail_t diag_detail_buffer;
	arm_subbuffer_tran_context_t tran_context;
	arm_subbuffer_user_t user_buffer;
	arm_subbuffer_metric_values_t metric_values;
	arm_char_t *context_value_array [ARM_PROPERTY_MAX_COUNT];
	arm_metric_t metric_value_array [ARM_METRIC_MAX_COUNT];

	jfieldID tran_context_id;
	jobject tran_context_object;
	jfieldID tran_context_uri_id;
	jobject tran_context_uri_object;

	jclass user_class;
	jfieldID user_id;
	jobject user_object;
	jfieldID user_name_id;
	jobject user_name_object;
	jfieldID user_id_id;
	jobject user_id_object;

	jfieldID metric_group_id;
	jobject metric_group_object;

	/* Get the required inputs */
	arm_tran_report_class = (*env)->GetObjectClass (env, tran_report_object);
	if (arm_tran_report_class == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* Get the application handle */
	application_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_application", "Lorg/arm4/arm40/transaction/ArmApplication;");
	application_object = (*env)->GetObjectField (env, tran_report_object, application_id);
	if (application_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	app_handle = get_app_handle (env, application_object);

	/* Get the transaction ID */
	transaction_definition_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_definition", "Lorg/arm4/arm40/transaction/ArmTransactionDefinition;");
	transaction_definition_object = (*env)->GetObjectField (env, tran_report_object, transaction_definition_id);
	if (transaction_definition_object == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	get_tran_id (env, transaction_definition_object, &tran_id);

	/* Get the parent correlator */
	parent_ptr = NULL;
	parent_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_parentCorrelator", "Lorg/arm4/arm40/transaction/ArmCorrelator;");
	parent_object = (*env)->GetObjectField (env, tran_report_object, parent_id);
	if (parent_object != NULL)
	{
		get_correlator (env, parent_object, &parent_correlator);
		parent_ptr = &parent_correlator;
	}

	/* Get the current correlator */
	current_ptr = NULL;
	current_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_currentCorrelator", "Lorg/arm4/arm40/transaction/ArmCorrelator;");
	current_object = (*env)->GetObjectField (env, tran_report_object, current_id);
	if (current_object != NULL)
	{
		get_correlator (env, current_object, &current_correlator);
		current_ptr = &current_correlator;
	}

	/* See if a trace is requested */
	trace_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_traceRequested", "Z");
	trace_requested = (*env)->GetBooleanField (env, tran_report_object, transaction_definition_id);
	if (trace_requested)
		flags |= ARM_FLAG_TRACE_REQUEST;
	
	/*
		The sub-buffers that may be used are:
			arm_subbuffer_diag_detail_t
			arm_subbuffer_metric_values_t
            arm_subbuffer_tran_context_t
			arm_subbuffer_user_t
    */
	/* Build the buffer object */
	buffer_ptr = init_buffer (&buffer4, subbuffer_array);

	/* Get the arm_subbuffer_diag_detail_t */
	if (diag_detail_object == NULL)
	{
		/* This is an optional parameter */
		diag_detail_ptr = NULL;
	}
	else
	{
		diag_detail_ptr = (*env)->GetStringUTFChars (env, diag_detail_object, NULL);

		init_subbuffer ((arm_subbuffer_t *) &diag_detail_buffer, ARM_SUBBUFFER_DIAG_DETAIL);
		diag_detail_buffer.diag_detail = diag_detail_ptr;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &diag_detail_buffer, 4);
	}

	/* Get the user */
	user_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_user", "Lorg/arm4/arm40/transaction/ArmUser;");
	user_object = (*env)->GetObjectField (env, tran_report_object, user_id);
	if (user_object != NULL)
	{
		init_subbuffer ((arm_subbuffer_t *) &user_buffer, ARM_SUBBUFFER_USER);

		user_class = (*env)->GetObjectClass (env, user_object);
		user_name_id = (*env)->GetFieldID (env, user_class, "m_name", "Ljava/lang/String;");
		user_name_object = (*env)->GetObjectField (env, user_object, user_name_id);
		if (user_name_object)
			user_buffer.name = (arm_char_t *) (*env)->GetStringUTFChars(env, user_name_object, NULL);
		else
			user_buffer.name = NULL;

		user_id_id = (*env)->GetFieldID (env, user_class, "m_id", "Lorg/arm4/arm40/transaction/ArmID;");
		user_id_object = (*env)->GetObjectField (env, user_object, user_id_id);
		if (user_id_object)
		{
			user_buffer.id_valid = 1;
			get_id (env, user_id_object, &user_buffer.id);
		}
		else
			user_buffer.id_valid = 0;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &user_buffer, 4);
	}

	/* See if there are any context values */
	tran_context_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_contextValues", "Ljava/util/ArrayList;");
	tran_context_object = (*env)->GetObjectField (env, tran_report_object, tran_context_id);
	tran_context_uri_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_contextUri", "Ljava/lang/String;");
	tran_context_uri_object = (*env)->GetObjectField (env, tran_report_object, tran_context_uri_id);
	if (tran_context_object != NULL)
	{
		status = init_tran_context (env, tran_context_object, tran_context_uri_object, &tran_context, context_value_array);
		if (status != ARM_SUCCESS)
			return status;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &tran_context, 4);
	}

	/* And finally the metric values */
	metric_group_id = (*env)->GetFieldID (env, arm_tran_report_class, "m_metricGroup", "Lorg/arm4/arm40/metric/ArmMetricGroup;");
	metric_group_object = (*env)->GetObjectField (env, tran_report_object, metric_group_id);
	if (metric_group_object != NULL)
	{
		status = init_metric_values (env, metric_group_object, &metric_values, metric_value_array);
		if (status != ARM_SUCCESS)
			return status;

		add_subbuffer (buffer_ptr, (arm_subbuffer_t *) &metric_values, 4);
	}

	/* Report the application */
	status = arm_report_transaction(
		app_handle,
		&tran_id,
		tran_status,
		response_time,
		stop_time,
		parent_ptr,
		current_ptr,
		flags,
		NULL);

	/* Free any memory */
	if (diag_detail_object != NULL)
		(*env)->ReleaseStringUTFChars (env, diag_detail_object, diag_detail_ptr);

	return status;
}

JNIEXPORT jlong JNICALL
Java_org_arm4_arm40_jni_ArmNative_getArrrivalTime (JNIEnv *env, jclass arm_native)
{
	arm_arrival_time_t	opaque_time;
	arm_error_t status;
	
	status = arm_get_arrival_time(&opaque_time);

	return (jlong) opaque_time;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_bindThread (JNIEnv *env, jclass arm_native, jlong tran_handle)
{
	arm_error_t status;
	
	status = arm_bind_thread(tran_handle, 0, NULL);
	
	return status;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_unbindThread (JNIEnv *env, jclass arm_native, jlong tran_handle)
{
	arm_error_t status;
	
	status = arm_unbind_thread(tran_handle, 0, NULL);
	
	return status;
}

JNIEXPORT jlong JNICALL
Java_org_arm4_arm40_jni_ArmNative_blockTransaction (JNIEnv *env, jclass arm_native, jlong tran_handle)
{
	arm_tran_block_handle_t block_handle;
	arm_error_t status;
	
	status = arm_block_transaction(tran_handle, 0, NULL, &block_handle);
	
	return block_handle;
}

JNIEXPORT jint JNICALL
Java_org_arm4_arm40_jni_ArmNative_unblockTransaction (JNIEnv *env, jclass arm_native, jlong tran_handle, jlong block_handle)
{
	arm_error_t status;
	
	status = arm_unblock_transaction(tran_handle, block_handle, 0, NULL);
	
	return status;
}

JNIEXPORT jboolean JNICALL
Java_org_arm4_arm40_jni_ArmNative_isAgentTrace (JNIEnv *env, jclass arm_native, jobject arm_correlator)
{
	arm_error_t status;
	arm_boolean_t flag;
	arm_correlator_t correlator;
	
	get_correlator (env, arm_correlator, &correlator);
	status = arm_get_correlator_flags(
	    				&correlator,
	    				ARM_CORR_FLAGNUM_AGENT_TRACE,
	    				&flag);

	if (status == ARM_SUCCESS)
		return flag;
	
	return ARM_FALSE;
}

JNIEXPORT jboolean JNICALL
Java_org_arm4_arm40_jni_ArmNative_isApplicationTrace (JNIEnv *env, jclass arm_native, jobject arm_correlator)
{
	arm_error_t status;
	arm_boolean_t flag;
	arm_correlator_t correlator;
	
	get_correlator (env, arm_correlator, &correlator);
	status = arm_get_correlator_flags(
	    				&correlator,
	    				ARM_CORR_FLAGNUM_APP_TRACE,
	    				&flag);
	
	if (status == ARM_SUCCESS)
		return flag;
	
	return ARM_FALSE;
}

JNIEXPORT jlong JNICALL
Java_org_arm4_arm40_jni_ArmNative_getInvalidHandle (JNIEnv *env, jclass arm_native)
{
	/* This returns a handle that should always be considered invalid */
	return 0;
}
