/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <netinet/in.h>

#include <openssl/md5.h>

#include "arm4.h"
#include "arm4agent.h"
#include "libarm4_shm.h"
#include "libarm4_tree.h"
#include "arm4_db.h"

/* Prototypes */
static int time_is_zero (arm_time_t time);
static int is_valid_property (const arm_property_t *property_ptr);
static int is_empty_id (const arm_id_t *id_ptr);
static int is_all_1_id (const arm_id_t *id_ptr);
static int is_valid_address_subbuffer (const arm_subbuffer_system_address_t *address_buffer);
static void set_empty_id (arm_id_t *id_ptr);
static arm_error_t digest_application (const char *application_name_ptr,
	const arm_subbuffer_app_identity_t *identity_buffer_ptr,
	arm_charset_t charset,
	unsigned char *digest_ptr);
static arm_error_t digest_transaction (const arm_id_t *app_id,
	const char *transaction_name_ptr,
	const arm_subbuffer_tran_identity_t *identity_buffer_ptr,
    const arm_subbuffer_metric_bindings_t *metric_buffer,
	unsigned char *digest_ptr);
static int valid_app_handle (const arm_app_start_handle_t app_handle, arm_id_t *app_id);
static int compare_name_value_pairs (const void *pair1, const void *pair2);
static arm_subbuffer_t *get_subbuffer (const arm_buffer4_t *buffer4, arm_subbuffer_format_t type);
static arm_subbuffer_app_identity_t *get_identity_buffer (const arm_buffer4_t *buffer4, arm_error_t *status_ptr);
static arm_subbuffer_tran_identity_t *get_transaction_identity_buffer (const arm_buffer4_t *buffer4, arm_error_t *status_ptr);
static arm_subbuffer_charset_t *get_encoding_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_metric_bindings_t *get_metric_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_app_context_t *get_application_context_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_system_address_t *get_address_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_arrival_time_t *get_arrival_time_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_metric_values_t *get_transaction_metric_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_tran_context_t *get_transaction_context_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_user_t *get_user_buffer (const arm_buffer4_t *buffer4);
static arm_subbuffer_diag_detail_t *get_diag_buffer (const arm_buffer4_t *buffer4);
static void set_invalid_correlator(arm_correlator_t *current_correlator);
static void generate_correlator(
    const arm_app_start_handle_t app_handle,
    const arm_int64_t tran_handle,
    const arm_id_t *tran_id,
    const arm_correlator_t *parent_correlator,
    const arm_int32_t flags,
    arm_correlator_t *current_correlator);
static void get_arrival_time(const arm_arrival_time_t time_handle,
	arm_time_t *output_time_value);
static arm_error_t unblock_transaction (transaction_instance_record_t *transaction_instance_ptr,
	arm_time_t time_value);
static void initialize_trees (void);
arm_error_t verify_alias (tree_id_t *tree_ptr, const arm_id_t *alias_ptr, const unsigned char *digest_ptr);
void add_alias (tree_id_t *tree_ptr, const arm_id_t *alias_ptr, const arm_id_t *actual_ptr, const unsigned char *digest_ptr);
void unalias (tree_id_t *tree_ptr, const arm_id_t *alias_ptr, arm_id_t *actual_ptr);

/* Initialization mutex */
static pthread_once_t g_trees_initialized = PTHREAD_ONCE_INIT;

/* Internal database trees - local to the process */
static tree_t g_application_instance_tree;
static tree_t g_transaction_instance_tree;

static tree_id_t g_application_alias_tree;
static tree_id_t g_transaction_alias_tree;
static tree_id_t g_metric_alias_tree;

static int
time_is_zero (arm_time_t time)
{
	__libarm4_shm_trace_arm_internal ("time_is_zero");

	return ((time.tv_sec == 0) && (time.tv_usec == 0));
}

static int
is_valid_property (const arm_property_t *property_ptr)
{
    int length;

    __libarm4_shm_trace_arm_internal ("is_valid_property");

    if ((property_ptr == NULL) || (property_ptr->name == NULL) || (property_ptr->value == NULL))
        return ARM_FALSE;
    length = strlen (property_ptr->value);
    if ((length > 0) && (length <= ARM_PROPERTY_VALUE_MAX_CHARS))
		return ARM_TRUE;

	return ARM_FALSE;
}

static int
is_valid_binding (const arm_metric_binding_t *binding_ptr)
{
    __libarm4_shm_trace_arm_internal ("is_valid_binding");

    /* Check that we're in a valid metric slot */
    if (binding_ptr->slot > ARM_METRIC_MAX_ARRAY_INDEX) /* This will never be < 0 as it's an unsigned */
        return ARM_FALSE;

    /* While we don't check for a valid ID, we can see if it's empty */
    if (is_empty_id (&(binding_ptr->id)))
        return ARM_FALSE;

    return ARM_TRUE;
}

static int
is_empty_id (const arm_id_t *id_ptr)
{
	static arm_id_t empty = { { {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} } };

    __libarm4_shm_trace_arm_internal ("is_empty_id");

	if (memcmp (id_ptr, &empty, sizeof (arm_id_t)) == 0)
		return ARM_TRUE;

	return ARM_FALSE;
}

static int
is_all_1_id (const arm_id_t *id_ptr)
{
	static arm_id_t all_1 = { { {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1} } };

    __libarm4_shm_trace_arm_internal ("is_all_1_id");

	if (memcmp (id_ptr, &all_1, sizeof (arm_id_t)) == 0)
		return ARM_TRUE;

	return ARM_FALSE;
}

static int
is_valid_address_subbuffer (const arm_subbuffer_system_address_t *address_buffer)
{
	if ((address_buffer != NULL) &&
		(address_buffer->address_length > 0) && /* These values indicate the local system */
		(address_buffer->address != NULL))
		return ARM_TRUE;

	return ARM_FALSE;
}

static void
set_empty_id (arm_id_t *id_ptr)
{
    __libarm4_shm_trace_arm_internal ("set_empty_id");

	memset (id_ptr, 0, sizeof (arm_id_t));
}

static arm_error_t
digest_application (const char *application_name_ptr,
	const arm_subbuffer_app_identity_t *identity_buffer_ptr,
	arm_charset_t charset,
	unsigned char *digest_ptr)
{
	int	i;
	int count;
	MD5_CTX md5_context;

    __libarm4_shm_trace_arm_internal ("digest_application");

	/* Initialize the MD5 checksum */
    MD5_Init (&md5_context);
	MD5_Update (&md5_context, application_name_ptr, strlen (application_name_ptr));
	if (identity_buffer_ptr)
	{
		count = identity_buffer_ptr->identity_property_count;
		for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
		{
			if (is_valid_property (&(identity_buffer_ptr->identity_property_array [i])))
			{
				count--;
				MD5_Update (&md5_context, identity_buffer_ptr->identity_property_array [i].name, strlen (identity_buffer_ptr->identity_property_array [i].name));
				MD5_Update (&md5_context, identity_buffer_ptr->identity_property_array [i].value, strlen (identity_buffer_ptr->identity_property_array [i].value));
			}
		}
		count = identity_buffer_ptr->context_name_count;
		for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
		{
			if (__libarm4_shm_is_valid_identifier (identity_buffer_ptr->context_name_array [i]))
			{
				count--;
				/* Should also check for repeated entries */
				MD5_Update (&md5_context, &i, sizeof (int));	/* Index is part of the signature */
				MD5_Update (&md5_context, identity_buffer_ptr->context_name_array [i], strlen (identity_buffer_ptr->context_name_array [i]));
			}
		}
	}
	MD5_Update (&md5_context, &charset, sizeof (charset));
	MD5_Final (digest_ptr, &md5_context);

	return ARM_SUCCESS;
}

static arm_error_t
digest_transaction (const arm_id_t *app_id,
	const char *transaction_name_ptr,
	const arm_subbuffer_tran_identity_t *identity_buffer_ptr,
    const arm_subbuffer_metric_bindings_t *metric_buffer_ptr,
	unsigned char *digest_ptr)
{
	int	i;
	int count;
	MD5_CTX md5_context;

    __libarm4_shm_trace_arm_internal ("digest_transaction");

	/* Initialize the MD5 checksum */
    MD5_Init (&md5_context);
	MD5_Update (&md5_context, app_id, sizeof (arm_id_t));
	MD5_Update (&md5_context, transaction_name_ptr, strlen (transaction_name_ptr));
	if (identity_buffer_ptr)
	{
		count = identity_buffer_ptr->identity_property_count;
		for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
		{
			if (is_valid_property (&(identity_buffer_ptr->identity_property_array [i])))
			{
				count--;
				MD5_Update (&md5_context, identity_buffer_ptr->identity_property_array [i].name, strlen (identity_buffer_ptr->identity_property_array [i].name));
				MD5_Update (&md5_context, identity_buffer_ptr->identity_property_array [i].value, strlen (identity_buffer_ptr->identity_property_array [i].value));
			}
		}
		count = identity_buffer_ptr->context_name_count;
		for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
		{
			if (__libarm4_shm_is_valid_identifier (identity_buffer_ptr->context_name_array [i]))
			{
				count--;
				/* Should also check for repeated entries */
				MD5_Update (&md5_context, &i, sizeof (int));	/* Index is part of the signature */
				MD5_Update (&md5_context, identity_buffer_ptr->context_name_array [i], strlen (identity_buffer_ptr->context_name_array [i]));
			}
		}
		if (identity_buffer_ptr->uri && (strlen (identity_buffer_ptr->uri) > 0))
			MD5_Update (&md5_context, identity_buffer_ptr->uri, strlen (identity_buffer_ptr->uri));
	}
    if (metric_buffer_ptr)
    {
        for (i = 0; i < metric_buffer_ptr->count; i++)
        {
            MD5_Update (&md5_context,  &(metric_buffer_ptr->metric_binding_array[i].slot), sizeof (arm_metric_slot_t));
            MD5_Update (&md5_context,  &(metric_buffer_ptr->metric_binding_array[i].id), sizeof (arm_id_t));
        }
    }
	MD5_Final (digest_ptr, &md5_context);

	return ARM_SUCCESS;
}

static arm_error_t
digest_metric (const arm_id_t *app_id,
	const char *metric_name_ptr,
    const arm_metric_format_t metric_format,
    const arm_metric_usage_t metric_usage,
    const arm_char_t *metric_unit_ptr,
	unsigned char *digest_ptr)
{
	MD5_CTX md5_context;

    __libarm4_shm_trace_arm_internal ("digest_metric");

	/* Initialize the MD5 checksum */
    MD5_Init (&md5_context);
	MD5_Update (&md5_context, app_id, sizeof (arm_id_t));
	MD5_Update (&md5_context, metric_name_ptr, strlen (metric_name_ptr));
    MD5_Update (&md5_context, &metric_format, sizeof (metric_format));
    MD5_Update (&md5_context, &metric_usage, sizeof (metric_usage));
    if (__libarm4_shm_is_valid_identifier (metric_unit_ptr))
        MD5_Update (&md5_context, metric_unit_ptr, strlen (metric_unit_ptr));
	MD5_Final (digest_ptr, &md5_context);

	return ARM_SUCCESS;
}

static int
valid_app_handle (const arm_app_start_handle_t app_handle, arm_id_t *app_id)
{
	/*
	 * Check that the application handle is valid and active, and if so, return it's associated app_id
	 */
	application_instance_tree_record_t *application_instance_ptr;

    __libarm4_shm_trace_arm_internal ("valid_app_handle");

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Initialize our returned application id to null */
	memset (app_id, 0, sizeof (*app_id));

	if (app_handle == 0)
	{
		/* This is the null id */
		return ARM_FALSE;
	}

	application_instance_ptr = (application_instance_tree_record_t *) __libarm4_tree_find (&g_application_instance_tree, app_handle);
	if (application_instance_ptr == NULL)
		return ARM_FALSE;

	/* The entry must also be marked as active */
	if (application_instance_ptr->active != ARM_ACTIVE)
		return ARM_FALSE;

	/* Return the app_id */
	*app_id = application_instance_ptr->appl_id;

	return ARM_TRUE;
}

static int
compare_name_value_pairs (const void *pair1, const void *pair2)
{
	arm_property_t *p1, *p2;
	int result;

    __libarm4_shm_trace_arm_internal ("compare_name_value_pairs");

	/* Used for sorting name/value pair arrays */
	p1 = (arm_property_t *) pair1;
	p2 = (arm_property_t *) pair2;

	/* Any pairs with null values should be ignored */
	if ((p1->name == NULL) || (p1->value == NULL))
	{
		if ((p2->name == NULL) || (p2->value == NULL))
			return 0; /* Both are identical and should be ignored */
		else
			return 1; /* p1 should be placed at the end */
	}
	if ((p2->name == NULL) || (p2->value == NULL))
		return -1; /* p2 should be placed at the end */

	/* Names have a maximum size of 128 bytes, values 256 */
	result = strncmp (p1->name, p2->name, 128);
	if (result == 0)
		result = strncmp (p1->value, p2->value, 256);

	return result;
}

static int
compare_slot_id (const void *pair1, const void *pair2)
{
	arm_metric_binding_t *p1, *p2;
	int result;

    __libarm4_shm_trace_arm_internal ("compare_slot_id");

	/* Used for sorting metric binding arrays */
	p1 = (arm_metric_binding_t *) pair1;
	p2 = (arm_metric_binding_t *) pair2;

    result = p1->slot - p2->slot;

	return result;
}

static arm_subbuffer_t *
get_subbuffer (const arm_buffer4_t *buffer4, arm_subbuffer_format_t type)
{
	arm_subbuffer_t *subbuffer;
	int i;

    __libarm4_shm_trace_arm_internal ("get_subbuffer");

	if (buffer4 == NULL)
		return NULL;

	for (i = 0; i < buffer4->count; i++)
	{
		subbuffer = buffer4->subbuffer_array [i];
		if (subbuffer && (subbuffer->format == type))
		{
			return subbuffer;
		}
	}
	return NULL;
}

static arm_subbuffer_app_identity_t *
get_identity_buffer (const arm_buffer4_t *buffer4, arm_error_t *status_ptr)
{
	arm_subbuffer_app_identity_t *subbuffer;
	arm_subbuffer_app_identity_t *new_subbuffer;
	int i, j;
	int count;
	int duplicate;

	__libarm4_shm_trace_arm_internal ("get_identity_buffer");

	/* Assume status is success unless indicated otherwise */
	*status_ptr = ARM_SUCCESS;

	/*
	 * Create a new subbuffer based on that passed in to the application. When done,
	 * the subbuffer should be freed, but not the memory pointed to by its members.
	 */

	subbuffer = (arm_subbuffer_app_identity_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_APP_IDENTITY);
    if (subbuffer == NULL)
		new_subbuffer = subbuffer;
	else
	{
		new_subbuffer = (arm_subbuffer_app_identity_t *) malloc (sizeof (arm_subbuffer_app_identity_t));
		if (new_subbuffer)
		{
			/* Copy the subbuffer, leaving the array values as is */
			*new_subbuffer = *subbuffer;

			/* Remove the duplicates before sorting - O(n^2) really bad! */
			count = new_subbuffer->identity_property_count;
			if (count > 0)
			{
				if (new_subbuffer->identity_property_array == NULL)
				{
					*status_ptr = ARM_FAILURE_INVALID_ARGUMENT;
					free (new_subbuffer);
					return NULL;
				}

				for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
				{
					if (is_valid_property (&(new_subbuffer->identity_property_array [i])))
					{
						count--;
	
						/* Scan for any duplicate entries */
						duplicate = 1;
						for (j = 0; (j < i) && (duplicate != 0); j++)
						{
							if (new_subbuffer->identity_property_array [j].name != NULL)
							{
								duplicate = strncmp (new_subbuffer->identity_property_array [i].name,
													 new_subbuffer->identity_property_array [j].name,
													 ARM_PROPERTY_NAME_MAX_CHARS);
								if (duplicate == 0)
								{
									/* These really aren't const since they're a local copy */
									*((char **) &(new_subbuffer->identity_property_array [i].name)) = NULL;
									*((char **) &(new_subbuffer->identity_property_array [i].value)) = NULL;
								}
							}
						}
					}
					else
					{
						*((char **) &(new_subbuffer->identity_property_array [i].name)) = NULL;
						*((char **) &(new_subbuffer->identity_property_array [i].value)) = NULL;
					}
				}
	
				/* This will sort the order of the name/value pairs, without touching the strings */
				qsort ((void *) new_subbuffer->identity_property_array, i /*ARM_PROPERTY_MAX_COUNT*/,
								sizeof (arm_property_t), compare_name_value_pairs);
	
				/* Look for any null values - these should be ignored */
				for (i = 0; i < new_subbuffer->identity_property_count; i++)
				{
					if (!is_valid_property (&(new_subbuffer->identity_property_array [i])))
					{
						/*
						 * Since the array was sorted with all null values at the end, all following
						 * entries can be ignored.
						 */
						new_subbuffer->identity_property_count = i;
						break;
					}
				}
	
				/* Context names remain unsorted */
				for (i = 0; i < new_subbuffer->context_name_count; i++)
				{
					/* Null and empty strings are ignored */
					if ((new_subbuffer->context_name_array [i] != NULL) &&
						(strlen (new_subbuffer->context_name_array [i]) > 0))
					{
						if (!__libarm4_shm_is_valid_identifier (new_subbuffer->context_name_array [i]))
						{
							*status_ptr = ARM_FAILURE_INVALID_ARGUMENT;
							free (new_subbuffer);
						}
					}
	
					/* Should also check for duplicates */
				}
			}
		}
	}
	return new_subbuffer;
}

static arm_subbuffer_tran_identity_t *
get_transaction_identity_buffer (const arm_buffer4_t *buffer4, arm_error_t *status_ptr)
{
	arm_subbuffer_tran_identity_t *subbuffer;
	arm_subbuffer_tran_identity_t *new_subbuffer;
	int i, j;
	int count;
	int duplicate;

	__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer");

	/* Assume status is success unless indicated otherwise */
	*status_ptr = ARM_SUCCESS;

	/*
	 * Create a new subbuffer based on that passed in to the application. When done,
	 * the subbuffer should be freed, but not the memory pointed to by its members.
	 */

	subbuffer = (arm_subbuffer_tran_identity_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_TRAN_IDENTITY);
    if (subbuffer == NULL)
		new_subbuffer = subbuffer;
	else
	{
		new_subbuffer = (arm_subbuffer_tran_identity_t *) malloc (sizeof (arm_subbuffer_tran_identity_t));
		if (new_subbuffer)
		{
			/* Copy the subbuffer, leaving the array values as is */
			*new_subbuffer = *subbuffer;

			/* Remove the duplicates before sorting - O(n^2) really bad! */
			count = new_subbuffer->identity_property_count;
			if (count > 0)
			{
				if (new_subbuffer->identity_property_array == NULL)
				{
					*status_ptr = ARM_FAILURE_INVALID_ARGUMENT;
					free (new_subbuffer);
					return NULL;
				}

				for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
				{
					if (is_valid_property (&(new_subbuffer->identity_property_array [i])))
					{
						count--;
	
						/* Scan for any duplicate entries */
						duplicate = 1;
						for (j = 0; (j < i) && (duplicate != 0); j++)
						{
							if (new_subbuffer->identity_property_array [j].name != NULL)
							{
								duplicate = strncmp (new_subbuffer->identity_property_array [i].name,
													 new_subbuffer->identity_property_array [j].name,
													 ARM_PROPERTY_NAME_MAX_CHARS);
								if (duplicate == 0)
								{
									/* These really aren't const since they're a local copy */
									*((char **) &(new_subbuffer->identity_property_array [i].name)) = NULL;
									*((char **) &(new_subbuffer->identity_property_array [i].value)) = NULL;
								}
							}
						}
					}
					else
					{
						*((char **) &(new_subbuffer->identity_property_array [i].name)) = NULL;
						*((char **) &(new_subbuffer->identity_property_array [i].value)) = NULL;
					}
				}
	
				__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer 1");
				/* This will sort the order of the name/value pairs, without touching the strings */
				qsort ((void *) new_subbuffer->identity_property_array, i /*ARM_PROPERTY_MAX_COUNT*/,
								sizeof (arm_property_t), compare_name_value_pairs);
				__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer 2");
	
				/* Look for any null values - these should be ignored */
				for (i = 0; i < new_subbuffer->identity_property_count; i++)
				{
					if ( (new_subbuffer->identity_property_array [i].name == NULL) ||
						 (new_subbuffer->identity_property_array [i].value == NULL) )
					{
						/*
						 * Since the array was sorted with all null values at the end, all following
						 * entries can be ignored.
						 */
						new_subbuffer->identity_property_count = i;
						break;
					}
				}
			}

			/* Context names remain unsorted */
			__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer 3");
			for (i = 0; i < new_subbuffer->context_name_count; i++)
			{
				/* Null and empty strings are ignored */
				if ((new_subbuffer->context_name_array [i] != NULL) &&
					(strlen (new_subbuffer->context_name_array [i]) > 0))
				{
					if (!__libarm4_shm_is_valid_identifier (new_subbuffer->context_name_array [i]))
					{
						*status_ptr = ARM_FAILURE_INVALID_ARGUMENT;
						free (new_subbuffer);
						return NULL;
					}
				}

				/* Should also check for duplicates */
			}

			__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer 4");
			/* Validate the URI */
			if ((new_subbuffer->uri != NULL) &&
				(strlen (new_subbuffer->uri) > 0))
			{
				/* Ignore null and empty values */

				if (strlen (new_subbuffer->uri) > ARM_PROPERTY_URI_MAX_CHARS)
				{
					*status_ptr = ARM_FAILURE_INVALID_ARGUMENT;
					free (new_subbuffer);
					return NULL;
				}
			}
			__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer 5");
		}
	}

	__libarm4_shm_trace_arm_internal ("get_transaction_identity_buffer 6");
	return new_subbuffer;
}

static arm_subbuffer_charset_t *
get_encoding_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_charset_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_encoding_buffer");

	subbuffer = (arm_subbuffer_charset_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_CHARSET);

	return subbuffer;
}

static arm_subbuffer_metric_bindings_t *
get_metric_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_metric_bindings_t *subbuffer;
	arm_subbuffer_metric_bindings_t *new_subbuffer;
	int i, j;
	int duplicate;
    int count;

    __libarm4_shm_trace_arm_internal ("get_metric_buffer");

    /*
     * Create a new subbuffer based on that passed in to the application. When done,
     * the subbuffer should be freed, but not the memory pointed to by its members.
     */

	subbuffer = (arm_subbuffer_metric_bindings_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_METRIC_BINDINGS);
    if (subbuffer == NULL)
        new_subbuffer = subbuffer;
    else
    {
        new_subbuffer = (arm_subbuffer_metric_bindings_t *) malloc (sizeof (arm_subbuffer_metric_bindings_t));
        if (new_subbuffer)
        {
            /* Copy the subbuffer */
            new_subbuffer->metric_binding_array = calloc (ARM_METRIC_MAX_COUNT, sizeof (arm_metric_binding_t));

            /* Remove the duplicates before sorting - O(n^2) really bad! */
            count = 0;
            for (i = 0; i < subbuffer->count; i++)
            {
                if (is_valid_binding (&(subbuffer->metric_binding_array [i])))
                {
                    /* Scan for any duplicate entries */
                    duplicate = ARM_FALSE;
                    for (j = 0; j < count; j++)
                    {
                        /* Only the first entry is accepted */
                        duplicate = (subbuffer->metric_binding_array [i].slot == new_subbuffer->metric_binding_array [j].slot);
                        if (duplicate)
                            break;
                    }
                    if (!duplicate)
                    {
                        ((arm_metric_binding_t *) new_subbuffer->metric_binding_array) [count].slot = subbuffer->metric_binding_array [i].slot;
                        ((arm_metric_binding_t *) new_subbuffer->metric_binding_array) [count].id = subbuffer->metric_binding_array [i].id;

                        /* Unalias the metric ID */
                        unalias (&g_metric_alias_tree,
                                 &(subbuffer->metric_binding_array [i].id),
                                 &(((arm_metric_binding_t *) new_subbuffer->metric_binding_array) [count].id));

                        count++;
                    }
                }
            }
            new_subbuffer->count = count;

            /* This will sort the order of the name/value pairs, without touching the strings */
            qsort ((void *) new_subbuffer->metric_binding_array, new_subbuffer->count,
                            sizeof (arm_metric_binding_t), compare_slot_id);
        }
    }
    return new_subbuffer;
}

static arm_subbuffer_app_context_t *
get_application_context_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_app_context_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_application_context_buffer");

	subbuffer = (arm_subbuffer_app_context_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_APP_CONTEXT);

	return subbuffer;
}

static arm_subbuffer_system_address_t *
get_address_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_system_address_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_address_buffer");

	subbuffer = (arm_subbuffer_system_address_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_SYSTEM_ADDRESS);

	return subbuffer;
}

static arm_subbuffer_arrival_time_t *
get_arrival_time_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_arrival_time_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_arrival_time_buffer");

	subbuffer = (arm_subbuffer_arrival_time_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_ARRIVAL_TIME);

	return subbuffer;
}

static arm_subbuffer_metric_values_t *
get_transaction_metric_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_metric_values_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_transaction_metric_buffer");

	subbuffer = (arm_subbuffer_metric_values_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_METRIC_VALUES);

	return subbuffer;
}

static arm_subbuffer_tran_context_t *
get_transaction_context_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_tran_context_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_transaction_context_buffer");

	subbuffer = (arm_subbuffer_tran_context_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_TRAN_CONTEXT);

	return subbuffer;
}

static arm_subbuffer_user_t *
get_user_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_user_t *subbuffer;
	int length;

    __libarm4_shm_trace_arm_internal ("get_user_buffer");

	subbuffer = (arm_subbuffer_user_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_USER);

	/* Check for validity */
	if (subbuffer != NULL)
	{
		/* Is a name provided? */
		if (subbuffer->name != NULL)
		{
			length = strlen (subbuffer->name);
#if 0
			/* The name should be truncated if the length is exceeded */
			if ((length < 0) || (length > ARM_NAME_MAX_CHARS))
			{
				/* Invalid name length */
				return NULL;
			}
#endif
			if (length > 0)
				return subbuffer; /* Satisfied the minimum requirements */
		}

		if (subbuffer->id_valid)
			return subbuffer;
	}

	return NULL;	/* We need at least a valid name or id */
}

static arm_subbuffer_diag_detail_t *
get_diag_buffer (const arm_buffer4_t *buffer4)
{
	arm_subbuffer_diag_detail_t *subbuffer;

    __libarm4_shm_trace_arm_internal ("get_diag_buffer");

	subbuffer = (arm_subbuffer_diag_detail_t *) get_subbuffer (buffer4, ARM_SUBBUFFER_DIAG_DETAIL);

	/* Basic validity check */
	if ((subbuffer != NULL) && (subbuffer->diag_detail == NULL))
		return NULL;

	return subbuffer;
}

/*** Aliasing Routines ***/

arm_error_t
verify_alias (tree_id_t *tree_ptr, const arm_id_t *alias_ptr, const unsigned char *digest_ptr)
{
	alias_tree_record_t *record_ptr;

	/* This routine needs to be locked to ensure no search race condition. This is done external to this routine */
	if (alias_ptr == NULL)
		return ARM_SUCCESS;
	if (is_empty_id (alias_ptr) || is_all_1_id (alias_ptr))
		return ARM_FAILURE_BAD_ALIAS;

	record_ptr = __libarm4_tree_id_find (tree_ptr, alias_ptr);
	if (record_ptr != NULL)
		{
		/* Ensure the digest matches */
		if (memcmp (&(record_ptr->digest), digest_ptr, sizeof (record_ptr->digest)) != 0)
			{
			return ARM_FAILURE_DUPLICATE_ALIAS;
			}
		}

	return ARM_SUCCESS;
}

void
add_alias (tree_id_t *tree_ptr, const arm_id_t *alias_ptr, const arm_id_t *actual_ptr, const unsigned char *digest_ptr)
{
	alias_tree_record_t *record_ptr;

	/* This routine needs to be locked to ensure no search race condition. This is done external to this routine */

	if ((alias_ptr == NULL)||(actual_ptr == NULL))
		return;
	if (is_empty_id (alias_ptr))
		return;

	record_ptr = __libarm4_tree_id_find (tree_ptr, alias_ptr);
	if (record_ptr == NULL)
		{
		/* Add a record */
		record_ptr = malloc (sizeof (alias_tree_record_t));
		record_ptr->actual_id = *actual_ptr;
		memcpy (&(record_ptr->digest), digest_ptr, sizeof (record_ptr->digest));
		__libarm4_tree_id_insert (tree_ptr, alias_ptr, record_ptr);
		}
}

void
unalias (tree_id_t *tree_ptr, const arm_id_t *alias_ptr, arm_id_t *actual_ptr)
{
	alias_tree_record_t *record_ptr;

	/* If an alias exists, change the id to the actual id. Otherwise this */
	/* is probably an actual id, so leave it unchanged. */
	if (alias_ptr == NULL)
		return;

	record_ptr = __libarm4_tree_id_find (tree_ptr, alias_ptr);
	if (record_ptr != NULL)
		{
		/* We found an alias */
		memcpy (actual_ptr, &(record_ptr->actual_id), sizeof (arm_id_t));
		}
}

ARM4_API_DYNAMIC(arm_error_t)
arm_register_application(
    const arm_char_t *app_name,
    const arm_id_t *input_app_id,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_id_t *output_app_id)
{
	arm_error_t status;
	arm_subbuffer_app_identity_t *identity_buffer;
	arm_subbuffer_charset_t *encoding_buffer;
	arm_charset_t charset;
	unsigned char digest [MD5_DIGEST_LENGTH];
	arm_boolean_t is_new;
	arm4_shm_header_t *header_ptr;
	int i;

    __libarm4_shm_trace_arm ("arm_register_application");

	if (output_app_id == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	set_empty_id (output_app_id);
	is_new = ARM_FALSE;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Check to see if collection is active */
	if (__libarm4_shm_is_collecting () != ARM_TRUE)
	{
		/* Mimic a NULL library */
		return ARM_SUCCESS;
	}

	if (!__libarm4_shm_is_valid_identifier (app_name))
		return ARM_FAILURE_INVALID_ARGUMENT;

	/*  Get all appropriate subbuffers */
	identity_buffer = get_identity_buffer (buffer4, &status);
	if (status != ARM_SUCCESS)
		return status;
	encoding_buffer = get_encoding_buffer (buffer4);

	/* The default character set is ASCII */
	charset = ARM_CHARSET_ASCII;
	if (encoding_buffer)
		charset = encoding_buffer->charset;

	/* Check for a supported character set */
	if ((charset != ARM_CHARSET_ASCII) && (charset != ARM_CHARSET_UTF8))
		return ARM_FAILURE_UNSUPPORTED_CHARSET;

	/* Use the MD5 digest to look up the application */
	digest_application (app_name, identity_buffer, charset, digest);

	header_ptr = __libarm4_shm_get_header ();
	if (header_ptr)
	{
		__libarm4_shm_lock_application ();

		status = verify_alias (&g_application_alias_tree, input_app_id, digest);
		if (status != ARM_SUCCESS)
		{
			/* Unable to register the application */
			if (header_ptr)
				__libarm4_shm_unlock_application ();

			if (identity_buffer != NULL)
				free (identity_buffer);

			return status;
		}

		for (i = 0; i < header_ptr->application_count; i++)
		{
			if (memcmp (header_ptr->application_table [i].md5_digest, digest, MD5_DIGEST_LENGTH) == 0)
			{
				/* Definition already found! */
				memcpy (output_app_id, &(header_ptr->application_table [i].appl_id), sizeof (arm_id_t));
				add_alias (&g_application_alias_tree, input_app_id, output_app_id, digest);

				__libarm4_shm_unlock_application ();

				if (identity_buffer != NULL)
					free (identity_buffer);

				return ARM_SUCCESS;
			}
		}

		/* For now, use the application table as a ring buffer. This may not be the most efficient */
		memcpy (header_ptr->application_table [header_ptr->next_application].md5_digest, digest, MD5_DIGEST_LENGTH);
		__libarm4_shm_register_application (digest, app_name, charset, header_ptr->next_application); /* Identity properties to be added later */
		/*memcpy (&(header_ptr->application_table [header_ptr->next_application].appl_id), output_app_id, sizeof (arm_id_t));*/

		memcpy (output_app_id, &(header_ptr->application_table [header_ptr->next_application].appl_id), sizeof(arm_id_t));
		is_new = header_ptr->application_table [header_ptr->next_application].is_new;
		header_ptr->application_table [header_ptr->next_application].is_new = ARM_FALSE;

		header_ptr->next_application += 1;
		if (header_ptr->next_application >= header_ptr->max_applications)
			header_ptr->next_application = 0;
		if (header_ptr->application_count < header_ptr->max_applications)
			header_ptr->application_count += 1;
		add_alias (&g_application_alias_tree, input_app_id, output_app_id, digest);

		__libarm4_shm_unlock_application ();
	}

	if (!is_new)
	{
		/* The application has already been registered */
		if (identity_buffer != NULL)
			free (identity_buffer);

		return ARM_SUCCESS;
	}

	/* Register the identity properties */
	if (identity_buffer != NULL)
	{
		status = __libarm4_shm_register_application_identity (output_app_id, identity_buffer);
		free (identity_buffer);

		if (status != ARM_SUCCESS)
			return ARM_FAILURE_INTERNAL_ERROR;
	}

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_destroy_application(
    const arm_id_t *app_id,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
    __libarm4_shm_trace_arm ("arm_destroy_application");

	/* At this point in the implementation, we don't do anything to destroy the application */
	/* We should destroy all active instances */

    /* Destroy metric definitions associated with the application */

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_register_transaction(
    const arm_id_t *app_id,
    const arm_char_t *tran_name,
    const arm_id_t *input_tran_id,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_id_t *output_tran_id)
{
	arm_error_t	status;
	arm_subbuffer_tran_identity_t *identity_buffer;
	arm_subbuffer_metric_bindings_t *metric_buffer;
	unsigned char digest [MD5_DIGEST_LENGTH];
	arm_boolean_t is_new;
	arm4_shm_header_t *header_ptr;
	arm_id_t actual_app_id;
	int i;

    __libarm4_shm_trace_arm ("arm_register_transaction");

	if (output_tran_id == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	set_empty_id (output_tran_id);
	is_new = ARM_FALSE;

	/* If the application ID is empty, then there's no need to continue */
	if (is_empty_id (app_id))
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Check to see if collection is active */
	if (__libarm4_shm_is_collecting () != ARM_TRUE)
	{
		/* Mimic a NULL library */
		return ARM_SUCCESS;
	}

	actual_app_id = *app_id;
	unalias (&g_application_alias_tree, app_id, &actual_app_id);

	if (!__libarm4_shm_is_valid_identifier (tran_name))
		return ARM_FAILURE_INVALID_ARGUMENT;

	/*  Get all appropriate subbuffers */
	identity_buffer = get_transaction_identity_buffer (buffer4, &status);
	if (status != ARM_SUCCESS)
		return status;
	metric_buffer = get_metric_buffer (buffer4);

	/* Use the MD5 digest to look up the transaction */
	digest_transaction (&actual_app_id, tran_name, identity_buffer, metric_buffer, digest);

	header_ptr = __libarm4_shm_get_header ();
	if (header_ptr)
	{
		__libarm4_shm_lock_transaction ();

		status = verify_alias (&g_transaction_alias_tree, input_tran_id, digest);
		if (status != ARM_SUCCESS)
		{
			/* Unable to register the transaction */
			if (header_ptr)
				__libarm4_shm_unlock_transaction ();

			if (identity_buffer != NULL)
				free (identity_buffer);
            if (metric_buffer != NULL)
            {
                free ((arm_metric_binding_t *) metric_buffer->metric_binding_array);
                free (metric_buffer);
            }

			return status;
		}

		for (i = 0; i < header_ptr->transaction_count; i++)
		{
			if (memcmp (header_ptr->transaction_table [i].md5_digest, digest, MD5_DIGEST_LENGTH) == 0)
			{
				/* Transaction already registered */
				memcpy (output_tran_id, &(header_ptr->transaction_table [i].tran_id), sizeof (arm_id_t));
				add_alias (&g_transaction_alias_tree, input_tran_id, output_tran_id, digest);

				__libarm4_shm_unlock_transaction ();

				if (identity_buffer != NULL)
					free (identity_buffer);
                if (metric_buffer != NULL)
                {
                    free ((arm_metric_binding_t *) metric_buffer->metric_binding_array);
                    free (metric_buffer);
                }

				return ARM_SUCCESS;
			}
		}

		/* For now, use the transaction table as a ring buffer. This may not be the most efficient */
		memcpy (header_ptr->transaction_table [header_ptr->next_transaction].md5_digest, digest, MD5_DIGEST_LENGTH);
		__libarm4_shm_register_transaction (digest, &actual_app_id, tran_name, header_ptr->next_transaction); /* Identity properties to be added later */

		memcpy (output_tran_id, &(header_ptr->transaction_table [header_ptr->next_transaction].tran_id), sizeof(arm_id_t));
		is_new = header_ptr->transaction_table [header_ptr->next_transaction].is_new;
		header_ptr->transaction_table [header_ptr->next_transaction].is_new = ARM_FALSE;

		header_ptr->next_transaction += 1;
		if (header_ptr->next_transaction >= header_ptr->max_transactions)
			header_ptr->next_transaction = 0;
		if (header_ptr->transaction_count < header_ptr->max_transactions)
			header_ptr->transaction_count += 1;
		add_alias (&g_transaction_alias_tree, input_tran_id, output_tran_id, digest);

		__libarm4_shm_unlock_transaction ();
	}

	if (!is_new)
	{
		/* The transaction has already been registered */
		if (identity_buffer != NULL)
			free (identity_buffer);
        if (metric_buffer != NULL)
        {
            free ((arm_metric_binding_t *) metric_buffer->metric_binding_array);
            free (metric_buffer);
        }

		return ARM_SUCCESS;
	}

	/* Register the identity properties */
	if (identity_buffer != NULL)
	{
		status = __libarm4_shm_register_transaction_identity (output_tran_id, identity_buffer);

		free (identity_buffer);

		if (status != ARM_SUCCESS)
        {
            if (metric_buffer != NULL)
            {
                free ((arm_metric_binding_t *) metric_buffer->metric_binding_array);
                free (metric_buffer);
            }
			return ARM_FAILURE_INTERNAL_ERROR;
        }
	}

    if (metric_buffer != NULL)
    {
        status = __libarm4_shm_register_transaction_metric_binding (output_tran_id, metric_buffer);

        free ((arm_metric_binding_t *) metric_buffer->metric_binding_array);
        free (metric_buffer);

        if (status != ARM_SUCCESS)
            return ARM_FAILURE_INTERNAL_ERROR;
    }

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_register_metric(
    const arm_id_t *app_id,
    const arm_char_t *metric_name,
    const arm_metric_format_t metric_format,
    const arm_metric_usage_t metric_usage,
    const arm_char_t *metric_unit,
    const arm_id_t *input_metric_id,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_id_t *output_metric_id)
{
	arm_error_t	status;
	unsigned char digest [MD5_DIGEST_LENGTH];
	arm4_shm_header_t *header_ptr;
	int i;
	arm_id_t actual_app_id;

    __libarm4_shm_trace_arm ("arm_register_metric");

    if (output_metric_id == NULL)
        return ARM_FAILURE_NULL_ARGUMENT;

    set_empty_id (output_metric_id);

    /* If the application ID is empty, then there's no need to continue */
    if (is_empty_id (app_id))
        return ARM_SUCCESS;

	actual_app_id = *app_id;
	unalias (&g_application_alias_tree, app_id, &actual_app_id);

    if (!__libarm4_shm_is_valid_identifier (metric_name))
        return ARM_FAILURE_INVALID_ARGUMENT;

    if (metric_unit)
    {
        if (!__libarm4_shm_is_valid_identifier (metric_unit))
            return ARM_FAILURE_INVALID_ARGUMENT;
    }

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Check to see if collection is active */
	if (__libarm4_shm_is_collecting () != ARM_TRUE)
	{
		/* Mimic a NULL library */
		return ARM_SUCCESS;
	}

    /*  There are no sub-buffers defined for metrics. For strict adherence to the standard, this */
    /*  should return an error, but this may not be desired for most applications. */

    /* Use the MD5 digest to look up the transaction */
    digest_metric (&actual_app_id, metric_name, metric_format, metric_usage, metric_unit, digest);

    /* See if we're already registered */
    header_ptr = __libarm4_shm_get_header ();
    if (header_ptr)
    {
        __libarm4_shm_lock_metric ();

		status = verify_alias (&g_metric_alias_tree, input_metric_id, digest);
		if (status != ARM_SUCCESS)
		{
			/* Unable to register the metric */
			if (header_ptr)
				__libarm4_shm_unlock_metric ();

			return status;
		}

        for (i = 0; i < header_ptr->metric_count; i++)
        {
            if (memcmp (header_ptr->metric_table [i].md5_digest, digest, MD5_DIGEST_LENGTH) == 0)
            {
                /* Metric already registered */
                memcpy (output_metric_id, &(header_ptr->metric_table [i].metric_id), sizeof (arm_id_t));
				add_alias (&g_metric_alias_tree, input_metric_id, output_metric_id, digest);

                __libarm4_shm_unlock_metric ();

                return ARM_SUCCESS;
            }
        }

        /* For now, use the metric table as a ring buffer. This may not be the most efficient */
        memcpy (header_ptr->metric_table [header_ptr->next_metric].md5_digest, digest, MD5_DIGEST_LENGTH);
		__libarm4_shm_register_metric (digest, &actual_app_id, metric_name, metric_format, metric_usage, metric_unit, header_ptr->next_metric);

		memcpy (output_metric_id, &(header_ptr->metric_table [header_ptr->next_metric].metric_id), sizeof (arm_id_t));

        header_ptr->next_metric += 1;
        if (header_ptr->next_metric >= header_ptr->max_metrics)
            header_ptr->next_metric = 0;
        if (header_ptr->metric_count < header_ptr->max_metrics)
            header_ptr->metric_count += 1;

		add_alias (&g_metric_alias_tree, input_metric_id, output_metric_id, digest);
        __libarm4_shm_unlock_metric ();
    }

	return ARM_SUCCESS;
}

/* application instance API functions */
ARM4_API_DYNAMIC(arm_error_t)
arm_start_application(
    const arm_id_t *app_id,
    const arm_char_t *app_group_ptr,
    const arm_char_t *app_instance_ptr,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_app_start_handle_t *app_handle)
{
	int	status;
	arm_subbuffer_app_context_t *context_buffer;
	arm_subbuffer_system_address_t *address_buffer;
	application_instance_tree_record_t *application_instance_ptr;
	int i;
	arm_id_t actual_app_id;

    __libarm4_shm_trace_arm ("arm_start_application");

	if (app_handle == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	/* set empty handle */
	*app_handle = 0;

	/* An empty app_id means there's no need to continue */
	if (is_empty_id (app_id))
		return ARM_SUCCESS;

	if ( (app_group_ptr != NULL) &&
		((strlen (app_group_ptr) > 255) ||
	     (strlen (app_group_ptr) == 0) ))
		return ARM_FAILURE_INVALID_ARGUMENT;

	if ( (app_instance_ptr != NULL) &&
		((strlen (app_instance_ptr) > 255) ||
	     (strlen (app_instance_ptr) == 0) ))
		return ARM_FAILURE_INVALID_ARGUMENT;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Check to see if collection is active */
	if (__libarm4_shm_is_collecting () != ARM_TRUE)
	{
		/* Mimic a NULL library */
		return ARM_SUCCESS;
	}

	/* Dereference any aliases */
	actual_app_id = *app_id;
	unalias (&g_application_alias_tree, app_id, &actual_app_id);

	/*  Get all appropriate subbuffers */
	context_buffer = get_application_context_buffer (buffer4);
	address_buffer = get_address_buffer (buffer4);

	/* Initialize the app_handle to NULL */
	memset (app_handle, 0, sizeof (*app_handle));

	/* Get the new sequence number (application handle) */
	status = __libarm4_shm_next_sequence_number (arm4_shm_sequence_application_instance, app_handle);
	if (status != 0)
	{
		/* Error is already reported by the sequence code */
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	/* Save the instance info to shared memory */
	__libarm4_shm_add_app_instance (*app_handle, actual_app_id, ARM_ACTIVE);

	if (app_group_ptr != NULL)
		__libarm4_shm_add_app_group (*app_handle, app_group_ptr);

	if (app_instance_ptr != NULL)
		__libarm4_shm_add_app_instance_id (*app_handle, app_instance_ptr);

	if (context_buffer != NULL)
	{
		for (i = 0; i < context_buffer->context_value_count; i++)
		{
			__libarm4_shm_add_app_context (*app_handle, i, context_buffer->context_value_array [i]);
		}
	}

	if (is_valid_address_subbuffer(address_buffer))
		__libarm4_shm_add_app_address (*app_handle, address_buffer);

	/* Add a new entry to the application tree */
	application_instance_ptr = (application_instance_tree_record_t *) malloc (sizeof (application_instance_tree_record_t));
	application_instance_ptr->appl_id = actual_app_id;
	application_instance_ptr->active = ARM_ACTIVE;
	if (is_valid_address_subbuffer(address_buffer))
	{
		application_instance_ptr->address_format = address_buffer->address_format;
		application_instance_ptr->address_length = address_buffer->address_length;
		application_instance_ptr->address = (arm_uint8_t *) malloc (address_buffer->address_length);
		memcpy (application_instance_ptr->address, address_buffer->address, address_buffer->address_length);
	}
	else
	{
		application_instance_ptr->address_format = ARM_SYSADDR_FORMAT_RESERVED;
		application_instance_ptr->address_length = 0;
		application_instance_ptr->address = NULL;
	}
	__libarm4_tree_init ( &(application_instance_ptr->transaction_tree) );

	__libarm4_tree_insert (&g_application_instance_tree, *app_handle, (void *) application_instance_ptr);

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_stop_application(
    const arm_app_start_handle_t app_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
	application_instance_tree_record_t *application_instance_ptr;
	transaction_instance_record_t *transaction_instance_ptr;
	arm_tran_start_handle_t tran_handle;
	arm_time_t time_value;

	/* Stop this application and all transactions that are part of it */

    __libarm4_shm_trace_arm ("arm_stop_application");

	/* An empty application handle means there's no further operation to perform */
	if (app_handle == 0)
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Find the applicaion in the tree */
	application_instance_ptr = (application_instance_tree_record_t *) __libarm4_tree_remove (&g_application_instance_tree, app_handle);
	if (application_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* Mark the application instance as complete */
	if (application_instance_ptr->active == ARM_ACTIVE)
	{
		/* Save the record back to the database */
		__libarm4_shm_add_app_instance (app_handle, application_instance_ptr->appl_id, ARM_COMPLETE);

		/* Next ensure that all active transactions that are part of this application close */
		while (
			(transaction_instance_ptr = (transaction_instance_record_t *) __libarm4_tree_remove_next (
									&(application_instance_ptr->transaction_tree), &tran_handle))
				!= NULL)
		{
			/* Remove it from the active transaction list */
			transaction_instance_ptr = __libarm4_tree_remove (&g_transaction_instance_tree, tran_handle);

			if (transaction_instance_ptr && (transaction_instance_ptr->active == ARM_ACTIVE))
			{
				/* Unblock the transaction if required. Ignore any error messages  */
				if (!time_is_zero (transaction_instance_ptr->block_start_time))
				{
					gettimeofday (&time_value, NULL);
					unblock_transaction (transaction_instance_ptr, time_value);
				}

				__libarm4_shm_add_tran_instance (
					tran_handle,
					transaction_instance_ptr->appl_id,
					transaction_instance_ptr->tran_id,
					transaction_instance_ptr->app_handle,
					transaction_instance_ptr->arrival_time,
					transaction_instance_ptr->stop_time,
					transaction_instance_ptr->response_time,
					transaction_instance_ptr->blocked_time,
					ARM_STATUS_ABORTED,
					ARM_DISCARD);

				free (transaction_instance_ptr);
			}
		}
	}
	if (application_instance_ptr->address != NULL)
		free (application_instance_ptr->address);
	free (application_instance_ptr);

	return ARM_SUCCESS;
}

/* transaction instance API functions */
ARM4_API_DYNAMIC(arm_error_t)
arm_start_transaction(
    const arm_app_start_handle_t app_handle,
    const arm_id_t *tran_id,
    const arm_correlator_t *parent_correlator,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_tran_start_handle_t *tran_handle,
    arm_correlator_t *current_correlator)
{
	int	status;
	arm_subbuffer_arrival_time_t *arrival_time_buffer;
	arm_subbuffer_metric_values_t *metric_buffer;
	arm_subbuffer_tran_context_t *context_buffer;
	arm_subbuffer_user_t *user_buffer;
	arm_id_t app_id;
	arm_time_t time_value;
	arm_time_t zero_time;
	application_instance_tree_record_t *application_instance_ptr;
	transaction_instance_record_t *transaction_instance_ptr;
	int i;
	arm_id_t actual_tran_id;
	arm_int32_t trace_requested;

    __libarm4_shm_trace_arm ("arm_start_transaction");

	/** Get the current time */
	gettimeofday (&time_value, NULL);

	/* Set an invalid correlator in case of error */
	set_invalid_correlator (current_correlator);

	if (tran_handle == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	/* Initialize the handle as empty */
	*tran_handle = 0;

	/* A tran_id must be provided, even if it's empty */
	if (tran_id == NULL)
		return ARM_FAILURE_INVALID_ARGUMENT;

	/* An empty app_handle or tran_id means there's no need to continue */
	if ((app_handle == 0) || (is_empty_id (tran_id)))
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/* Check to see if collection is active */
	if (__libarm4_shm_is_collecting () != ARM_TRUE)
	{
		/* Mimic a NULL library */
		return ARM_SUCCESS;
	}

	/* Dereference any aliases */
	actual_tran_id = *tran_id;
	unalias (&g_transaction_alias_tree, tran_id, &actual_tran_id);

	/* Check to see if we should trace this transaction, and update it's counter */
	trace_requested = __libarm4_shm_transaction_trace (actual_tran_id, flags, ARM_FALSE);
	if (!trace_requested)
		return ARM_SUCCESS;
	if (flags & ARM_FLAG_TRACE_REQUEST)
		trace_requested |= ARM_CORR_FLAGNUM_APP_TRACE; /* Ensure oure trace flag is set correctly */

	/*  Get all appropriate subbuffers */
	arrival_time_buffer = get_arrival_time_buffer (buffer4);
	metric_buffer = get_transaction_metric_buffer (buffer4);
	context_buffer = get_transaction_context_buffer (buffer4);
	user_buffer = get_user_buffer (buffer4);

	generate_correlator(
		app_handle,
		0, /* tran_handle */
		&actual_tran_id,
		parent_correlator,
		trace_requested,
		current_correlator);

	/* First initialize the tran_handle to NULL */
	memset (tran_handle, 0, sizeof (*tran_handle));

	/* If we're not using a valid app_handle, then there's no point in continuing */
	if (!valid_app_handle (app_handle, &app_id))
		return ARM_FAILURE_INVALID_ARGUMENT;

	/* Get the new sequence number (transaction handle) */
	status = __libarm4_shm_next_sequence_number (arm4_shm_sequence_transaction_instance, tran_handle);
	if (status != 0)
	{
		/* Error is already reported by the sequence code */
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	if (arrival_time_buffer != NULL)
	{
		get_arrival_time (arrival_time_buffer->opaque_time, &time_value);
	}

	memset (&zero_time, 0, sizeof (zero_time));
	if ((flags & ARM_FLAG_TRACE_REQUEST) || (parent_correlator))
	{
		/* If we have a parent correlator, or we've explicitly requested it, then continue the trace */
		__libarm4_shm_add_tran_instance (*tran_handle, app_id, actual_tran_id, app_handle, time_value, zero_time, zero_time, zero_time, ARM_STATUS_UNKNOWN, ARM_ACTIVE);
	}
	else
	{
		status = __libarm4_shm_try_tran_instance (*tran_handle, app_id, actual_tran_id, app_handle, time_value, zero_time, zero_time, zero_time, ARM_STATUS_UNKNOWN, ARM_ACTIVE);
		if (status < 0)
		{
			/* The queue is full, and we've been counted, so just return with no further transaction processing */
			*tran_handle = 0;
			return ARM_SUCCESS;
		}
	}

	if (context_buffer != NULL)
	{
		for (i = 0; i < context_buffer->context_value_count; i++)
		{
			__libarm4_shm_add_tran_context (*tran_handle, i, context_buffer->context_value_array [i]);
		}

		/* Write the URI */
		if (context_buffer->uri != NULL)
		{
			__libarm4_shm_add_tran_uri (*tran_handle, context_buffer->uri);
		}
	}

	if ((current_correlator != NULL)||(parent_correlator != NULL))
	{
		if (current_correlator)
		{
			(((arm3_correlator_t *) current_correlator)->tran_handle) = htonll (*tran_handle);
		}

		__libarm4_shm_add_correlators (*tran_handle, parent_correlator, current_correlator);
	}

	if (metric_buffer != NULL)
	{
		for (i = 0; i < metric_buffer->count; i++)
		{
			__libarm4_shm_add_metric_value (*tran_handle, &(metric_buffer->metric_value_array [i]));
		}
	}
	if (user_buffer != NULL)
		__libarm4_shm_add_user (*tran_handle, user_buffer);

	/* Add a new entry to the transaction tree */
	transaction_instance_ptr = (transaction_instance_record_t *) malloc (sizeof (transaction_instance_record_t));

	memset (transaction_instance_ptr, 0, sizeof (transaction_instance_record_t));
	transaction_instance_ptr->appl_id = app_id;
	transaction_instance_ptr->tran_id = actual_tran_id;
	transaction_instance_ptr->app_handle = app_handle;
	transaction_instance_ptr->arrival_time = time_value;
	transaction_instance_ptr->active = ARM_ACTIVE;

	__libarm4_tree_insert (&g_transaction_instance_tree, *tran_handle, (void *) transaction_instance_ptr);

	/* Also add this to the application instance tree */
	application_instance_ptr = __libarm4_tree_find (&g_application_instance_tree, app_handle);
	if (application_instance_ptr)
		__libarm4_tree_insert ( &(application_instance_ptr->transaction_tree), *tran_handle, (void *) transaction_instance_ptr);

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_stop_transaction(
    const arm_tran_start_handle_t tran_handle,
    const arm_tran_status_t tran_status,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
	arm_subbuffer_diag_detail_t *diag_buffer;
	arm_subbuffer_metric_values_t *metric_buffer;
	arm_time_t time_value;
	arm_time_t arrival_time;
	arm_time_t stop_time;
	arm_time_t response_time;
	transaction_instance_record_t *transaction_instance_ptr;
	application_instance_tree_record_t *application_instance_ptr;
	arm_error_t status;
	int i;

    __libarm4_shm_trace_arm ("arm_stop_transaction");

	/** Get the current time */
	gettimeofday (&time_value, NULL);

	/* If we're not using a valid tran_handle, then there's no point in continuing */
	if (tran_handle == 0)
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/*  Get all appropriate subbuffers */
	diag_buffer = get_diag_buffer (buffer4);
	metric_buffer = get_transaction_metric_buffer (buffer4);

	transaction_instance_ptr = (transaction_instance_record_t *) __libarm4_tree_remove (&g_transaction_instance_tree, tran_handle);
	if (transaction_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* Remove it from the application tree as well */
	application_instance_ptr = (application_instance_tree_record_t *) __libarm4_tree_find (&g_application_instance_tree, transaction_instance_ptr->app_handle);
	if (application_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	__libarm4_tree_remove (&(application_instance_ptr->transaction_tree), tran_handle);

	arrival_time = transaction_instance_ptr->arrival_time;

	/* Update the transaction records, and mark the transaction as complete */
	stop_time = time_value;
	response_time.tv_sec = stop_time.tv_sec - arrival_time.tv_sec;
	response_time.tv_usec = stop_time.tv_usec - arrival_time.tv_usec;
	if (response_time.tv_usec < 0)
	{
		response_time.tv_sec--;
		response_time.tv_usec += 1000000;
	}

	if (!time_is_zero (transaction_instance_ptr->block_start_time))
	{
		/* Unblock the transaction */
		status = unblock_transaction (transaction_instance_ptr, time_value);
		if (status != ARM_SUCCESS)
		{
			return status;
		}
	}

	__libarm4_shm_add_tran_instance (tran_handle,
		transaction_instance_ptr->appl_id,
		transaction_instance_ptr->tran_id,
		transaction_instance_ptr->app_handle,
		arrival_time,
		stop_time,
		response_time,
		transaction_instance_ptr->blocked_time,
		tran_status,
		ARM_COMPLETE);

	if (metric_buffer != NULL)
	{
		for (i = 0; i < metric_buffer->count; i++)
		{
			__libarm4_shm_add_metric_value (tran_handle, &(metric_buffer->metric_value_array [i]));
		}
	}
	if (diag_buffer)
		__libarm4_shm_add_diag (tran_handle, diag_buffer);

	free (transaction_instance_ptr);

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_update_transaction(
    const arm_tran_start_handle_t tran_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
	arm_subbuffer_metric_values_t *metric_buffer;
	arm_time_t time_value;
	arm_time_t arrival_time;
	arm_time_t stop_time;
	arm_time_t response_time;
	transaction_instance_record_t *transaction_instance_ptr;
	int i;

    __libarm4_shm_trace_arm ("arm_update_transaction");

	/** Get the current time */
	gettimeofday (&time_value, NULL);

	/* If we're not using a valid tran_handle, then there's no point in continuing */
	if (tran_handle == 0)
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	/*  Get all appropriate subbuffers */
	metric_buffer = get_transaction_metric_buffer (buffer4);

	transaction_instance_ptr = (transaction_instance_record_t *) __libarm4_tree_find (&g_transaction_instance_tree, tran_handle);
	if (transaction_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* If we're not active, this is an error */
	if (transaction_instance_ptr->active != ARM_ACTIVE)
	{
		return ARM_FAILURE_TRANSACTION_INACTIVE;
	}

	/* If we're blocking, this is an error */
	if (!time_is_zero (transaction_instance_ptr->block_start_time))
	{
		return ARM_FAILURE_TRANSACTION_BLOCKED;
	}

	arrival_time = transaction_instance_ptr->arrival_time;

	/* Update the transaction records, and mark the transaction as complete */
	stop_time = time_value;
	response_time.tv_sec = stop_time.tv_sec - arrival_time.tv_sec;
	response_time.tv_usec = stop_time.tv_usec - arrival_time.tv_usec;
	if (response_time.tv_usec < 0)
	{
		response_time.tv_sec--;
		response_time.tv_usec += 1000000;
	}

	__libarm4_shm_add_tran_instance (tran_handle,
		transaction_instance_ptr->appl_id,
		transaction_instance_ptr->tran_id,
		transaction_instance_ptr->app_handle,
		arrival_time,
		stop_time,
		response_time,
		transaction_instance_ptr->blocked_time,
		transaction_instance_ptr->tran_status,
		ARM_ACTIVE);

	if (metric_buffer != NULL)
	{
		for (i = 0; i < metric_buffer->count; i++)
		{
			__libarm4_shm_add_metric_value (tran_handle, &(metric_buffer->metric_value_array [i]));
		}
	}

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_discard_transaction(
    const arm_tran_start_handle_t tran_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
	transaction_instance_record_t *transaction_instance_ptr;
	application_instance_tree_record_t *application_instance_ptr;

    __libarm4_shm_trace_arm ("arm_discard_transaction");

	/* If we're not using a valid tran_handle, then there's no point in continuing */
	if (tran_handle == 0)
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	transaction_instance_ptr = (transaction_instance_record_t *) __libarm4_tree_remove (&g_transaction_instance_tree, tran_handle);
	if (transaction_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	/* Remove it from the application tree as well */
	application_instance_ptr = (application_instance_tree_record_t *) __libarm4_tree_find (&g_application_instance_tree, transaction_instance_ptr->app_handle);
	if (application_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;
	__libarm4_tree_remove (&(application_instance_ptr->transaction_tree), tran_handle);

	__libarm4_shm_add_tran_instance (tran_handle,
		transaction_instance_ptr->appl_id,
		transaction_instance_ptr->tran_id,
		transaction_instance_ptr->app_handle,
		transaction_instance_ptr->arrival_time,
		transaction_instance_ptr->stop_time,
		transaction_instance_ptr->response_time,
		transaction_instance_ptr->blocked_time,
		ARM_STATUS_ABORTED,
		ARM_DISCARD);

	free (transaction_instance_ptr);

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_block_transaction(
    const arm_tran_start_handle_t tran_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_tran_block_handle_t *block_handle)
{
	int	status;
	arm_time_t time_value;
	transaction_instance_record_t *transaction_instance_ptr;

	/** Get the current time FIRST */
	gettimeofday (&time_value, NULL);

	__libarm4_shm_trace_arm ("arm_block_transaction");

	if (block_handle == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	/* Initialize the handle */
	*block_handle = 0;

	/* If we're not using a valid tran_handle, then there's no point in continuing */
	if (tran_handle == 0)
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	transaction_instance_ptr = (transaction_instance_record_t *) __libarm4_tree_find (&g_transaction_instance_tree, tran_handle);
	if (transaction_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	if (!time_is_zero (transaction_instance_ptr->block_start_time))
	{
		/* Already blocked */
		return ARM_FAILURE_TRANSACTION_BLOCKED;
	}

	/* Create the handle even though it's never used */
	status = __libarm4_shm_next_sequence_number (arm4_shm_sequence_transaction_blocking, block_handle);
	if (status != 0)
	{
		/* Error is already reported by the sequence code */
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	transaction_instance_ptr->block_start_time = time_value;

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_unblock_transaction(
    const arm_tran_start_handle_t tran_handle,
    const arm_tran_block_handle_t block_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
	arm_error_t	status;
	arm_time_t time_value;
	transaction_instance_record_t *transaction_instance_ptr;

	/** Get the current time FIRST */
	gettimeofday (&time_value, NULL);

    __libarm4_shm_trace_arm ("arm_unblock_transaction");

	/* If we're not using a valid tran_handle or block_handle, then there's no point in continuing */
	if ((tran_handle == 0) || (block_handle == 0))
		return ARM_SUCCESS;

	/* Ensure the tree structures are initialized */
	initialize_trees ();

	transaction_instance_ptr = (transaction_instance_record_t *) __libarm4_tree_find (&g_transaction_instance_tree, tran_handle);
	if (transaction_instance_ptr == NULL)
		return ARM_FAILURE_INTERNAL_ERROR;

	status = unblock_transaction (transaction_instance_ptr, time_value);

	return status;
}

/* thread support API functions */
ARM4_API_DYNAMIC(arm_error_t)
arm_bind_thread(
    const arm_tran_start_handle_t tran_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
    __libarm4_shm_trace_arm ("arm_bind_thread");

	/* Process/thread level binding isn't currently supported */
	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_unbind_thread(
    const arm_tran_start_handle_t tran_handle,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
    __libarm4_shm_trace_arm ("arm_unbind_thread");

	/* Process/thread level binding isn't currently supported */
	return ARM_SUCCESS;
}

/* report transaction data API function */
ARM4_API_DYNAMIC(arm_error_t)
arm_report_transaction(
    const arm_app_start_handle_t app_handle,
    const arm_id_t *tran_id,
    const arm_tran_status_t tran_status,
    const arm_response_time_t response_time,
    const arm_stop_time_t stop_time,
    const arm_correlator_t *parent_correlator,
    const arm_correlator_t *current_correlator,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4)
{
	int	status;
	arm_subbuffer_diag_detail_t *diag_buffer;
	arm_subbuffer_metric_values_t *metric_buffer;
	arm_subbuffer_tran_context_t *context_buffer;
	arm_subbuffer_user_t *user_buffer;
	arm_id_t app_id;
	arm_time_t time_value;
	arm_time_t stop_time_value;
	arm_time_t response_time_value;
	arm_time_t block_time_value;
	arm_tran_start_handle_t tran_handle;
	int i;
	arm_id_t actual_tran_id;

    __libarm4_shm_trace_arm ("arm_report_transaction");

	/* tran_id must be provided, even if it's empty */
	if (tran_id == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	/* If our handle or id is empty, then there's no point in continuing */
	if ((app_handle == 0) || (is_empty_id (tran_id)))
		return ARM_SUCCESS;

	/* Dereference any aliases */
	actual_tran_id = *tran_id;
	unalias (&g_transaction_alias_tree, tran_id, &actual_tran_id);

	/* Force a trace request and update it's counter */
	__libarm4_shm_transaction_trace (actual_tran_id, flags, ARM_TRUE);

	/*  Get all appropriate subbuffers */
	diag_buffer = get_diag_buffer (buffer4);
	metric_buffer = get_transaction_metric_buffer (buffer4);
	context_buffer = get_transaction_context_buffer (buffer4);
	user_buffer = get_user_buffer (buffer4);

	/* If we're not using a valid app_handle, then there's no point in continuing */
	if (!valid_app_handle (app_handle, &app_id))
		return ARM_FAILURE_INVALID_ARGUMENT;

	/* Get the new sequence number (transaction handle) */
	status = __libarm4_shm_next_sequence_number (arm4_shm_sequence_transaction_instance, &tran_handle);
	if (status != 0)
	{
		/* Error is already reported by the sequence code */
		return ARM_FAILURE_INTERNAL_ERROR;
	}

	/* Response time is in nanoseconds */
	response_time_value.tv_sec = (response_time / 1000000000);
	response_time_value.tv_usec = ((response_time / 1000) % 1000000);

	/* Stop time is in milliseconds */
	stop_time_value.tv_sec = (stop_time / 1000);
	stop_time_value.tv_usec = (stop_time % 1000) * 1000;

	/* Calculate the arrival time */
	time_value.tv_sec = stop_time_value.tv_sec - response_time_value.tv_sec;
	time_value.tv_usec = stop_time_value.tv_usec - response_time_value.tv_usec;
	if (time_value.tv_usec < 0)
	{
		time_value.tv_sec--;
		time_value.tv_usec += 1000000;
	}

	memset (&block_time_value, 0, sizeof (block_time_value));
	__libarm4_shm_add_tran_instance (tran_handle, app_id, actual_tran_id, app_handle, time_value, stop_time_value, response_time_value, block_time_value, tran_status, ARM_COMPLETE);

	if (context_buffer != NULL)
	{
		for (i = 0; i < context_buffer->context_value_count; i++)
		{
			__libarm4_shm_add_tran_context (tran_handle, i, context_buffer->context_value_array [i]);
		}

		/* Write the URI */
		if (context_buffer->uri != NULL)
		{
			__libarm4_shm_add_tran_uri (tran_handle, context_buffer->uri);
		}
	}

	if ((current_correlator != NULL)||(parent_correlator != NULL))
	{
		if (current_correlator)
		{
			(((arm3_correlator_t *) current_correlator)->tran_handle) = htonll (tran_handle);
		}

		__libarm4_shm_add_correlators (tran_handle, parent_correlator, current_correlator);
	}


	if (metric_buffer != NULL)
	{
		for (i = 0; i < metric_buffer->count; i++)
		{
			__libarm4_shm_add_metric_value (tran_handle, &(metric_buffer->metric_value_array [i]));
		}
	}
	if (user_buffer != NULL)
		__libarm4_shm_add_user (tran_handle, user_buffer);
	if (diag_buffer)
		__libarm4_shm_add_diag (tran_handle, diag_buffer);

	return ARM_SUCCESS;
}

/* correlator API functions */
static void
set_invalid_correlator(arm_correlator_t *current_correlator)
{
	arm_correlator_header_t *correlator_ptr;

    __libarm4_shm_trace_arm_internal ("set_invalid_correlator");

	/* See if we even need a correlator */
	if (current_correlator == NULL)
		return;

	/* Set an empty correlator */
	correlator_ptr = (arm_correlator_header_t *) current_correlator;

	correlator_ptr->length = htons (0);	/* Zero length correlator */
	correlator_ptr->format = ARM_CORR_FORMAT_INVALID;
	correlator_ptr->flags = 0;
}

static void
generate_correlator(
    const arm_app_start_handle_t app_handle,
    const arm_int64_t tran_handle,
    const arm_id_t *tran_id,
    const arm_correlator_t *parent_correlator,
    const arm_int32_t flags,
    arm_correlator_t *current_correlator)
{
	arm3_correlator_t *correlator_ptr;
	int size;
	int major, minor;
	application_instance_tree_record_t *application_instance_ptr;
	arm4_shm_header_t *header_ptr;

    __libarm4_shm_trace_arm_internal ("generate_correlator");

	/* See if we even need a correlator */
	if (current_correlator == NULL)
		return;

	/* Get the major and minor version from config.h */
	sscanf (VERSION, "%d.%d", &major, &minor);

	/* For now we use Version 3 correlators and ignore any subbuffers */
	correlator_ptr = (arm3_correlator_t *) current_correlator;

	correlator_ptr->format = ARM_CORR_FORMAT_VERSION3;
	correlator_ptr->flags = flags; /* This has already been set! | ARM_CORR_FLAG_AGENT_TRACE; */
	correlator_ptr->vendor_id = htons (IANA_VENDOR_ID_ARM4_ORG); /* IANA Private Enterprise Number */
	correlator_ptr->agent_version = htons (major * 256 + minor);
	correlator_ptr->agent_instance = htons (__libarm4_shm_get_instance ());
	(correlator_ptr->tran_handle) = htonll (tran_handle);
	correlator_ptr->tran_id = *tran_id;
	size = 0;

	/* Find the applicaion in the tree */
	application_instance_ptr = (application_instance_tree_record_t *) __libarm4_tree_find (&g_application_instance_tree, app_handle);
	if (application_instance_ptr == NULL)
	{
		/* An invalid app_handle is an error */
		correlator_ptr->length = htons (0);
		return;
	}

	if (application_instance_ptr->address_length > 0)
	{
		/* A system address was passed in */
		correlator_ptr->addr_format = htons(application_instance_ptr->address_format);
		size = application_instance_ptr->address_length;
		if (size > sizeof (correlator_ptr->address))
			size = sizeof (correlator_ptr->address);
		memcpy ((char *) correlator_ptr->address, application_instance_ptr->address, size);
	}
	else
	{
		/* Use the local machine */
		correlator_ptr->addr_format = htons(ARM_SYSADDR_FORMAT_HOSTNAME);
		header_ptr = __libarm4_shm_get_header ();
		if (header_ptr)
		{
			strncpy ((char *) correlator_ptr->address, header_ptr->hostname, sizeof (correlator_ptr->address));
		}
		else
		{
			correlator_ptr->address [0] = '\0';
		}
		size = strlen ((char *) correlator_ptr->address); /* No null terminator */
		if (size > sizeof (correlator_ptr->address))
			size = sizeof (correlator_ptr->address);
	}

	correlator_ptr->addr_length = htons (size);
	size += 38;
	correlator_ptr->length = htons (size);
}

ARM4_API_DYNAMIC(arm_error_t)
arm_generate_correlator(
    const arm_app_start_handle_t app_handle,
    const arm_id_t *tran_id,
    const arm_correlator_t *parent_correlator,
    const arm_int32_t flags,
    const arm_buffer4_t *buffer4,
    arm_correlator_t *current_correlator)
{
	arm_id_t actual_tran_id;

	__libarm4_shm_trace_arm ("arm_generate_correlator");

	if (current_correlator == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	/* Dereference any aliases */
	actual_tran_id = *tran_id;
	unalias (&g_transaction_alias_tree, tran_id, &actual_tran_id);

	generate_correlator(
		app_handle,
		0, /* tran_handle */
		&actual_tran_id,
		parent_correlator,
		flags,
		current_correlator);

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_get_correlator_length(
    const arm_correlator_t *correlator,
    arm_correlator_length_t *length)
{
    __libarm4_shm_trace_arm ("arm_get_correlator_length");

	if (length == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	if (correlator == NULL)
		*length = 0;
	else
		*length = ntohs (*((arm_correlator_length_t *) &(correlator->opaque [0])));

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_get_correlator_flags(
    const arm_correlator_t *correlator,
    const arm_int32_t corr_flag_num,
    arm_boolean_t *flag)
{
	arm_correlator_header_t *header_ptr;

    __libarm4_shm_trace_arm ("arm_get_correlator_flags");

	if ((flag == NULL) || (correlator == NULL))
		return ARM_FAILURE_NULL_ARGUMENT;

	header_ptr = (arm_correlator_header_t *) correlator;

	*flag = ARM_FALSE;
	switch (corr_flag_num)
	{
	case ARM_CORR_FLAGNUM_APP_TRACE:
		if (header_ptr->flags & ARM_CORR_FLAG_APP_TRACE)
			*flag = ARM_TRUE;
		break;

	case ARM_CORR_FLAG_AGENT_TRACE:
		if (header_ptr->flags & ARM_CORR_FLAG_AGENT_TRACE)
			*flag = ARM_TRUE;
		break;

	default:
		return ARM_FAILURE_INVALID_ARGUMENT;
	}

	return ARM_SUCCESS;
}

/* miscellaneous API functions */
ARM4_API_DYNAMIC(arm_error_t)
arm_get_arrival_time(
    arm_arrival_time_t *opaque_time)
{
    __libarm4_shm_trace_arm ("arm_get_arrival_time");

	arm_time_t time_value;
	double fraction;

	/* Get the current time */
	gettimeofday (&time_value, NULL);

	if (opaque_time == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	/*
	 * We use the NTP time format in order to fit into 64 bits. This is
	 * represented as an integral part and a fractional part. This is not
	 * as precise as what our arm_time_t can provide, and this introduces
	 * the NTP concept of the Era. We'll provide the era when the
	 * transaction ends.
	 *
	 * In our case though, we use the UNIX start year of 1970 and not the
	 * NTP start year of 1900.
	 *
	 *    0                   1                   2                   3
	 *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 *   |                         Integral Part                         |
	 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 *   |                         Fractional Part                       |
	 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	 * REF http://www.eecis.udel.edu/~mills/database/rfc/rfc2030.txt
	 */

	fraction = time_value.tv_usec * 0.000001; /* time in microseconds */
	*opaque_time = (time_value.tv_sec * 0x0000000100000000LL) + (fraction * 0x0000000100000000LL);

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_get_error_message(
    const arm_charset_t charset,
    const arm_error_t code,
    arm_message_buffer_t msg)
{
    __libarm4_shm_trace_arm ("arm_get_error_message");

	if (msg == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	if ((charset != ARM_CHARSET_ASCII) && (charset != ARM_CHARSET_UTF8))
	{
		msg [0] = '\0';
		return ARM_FAILURE_UNSUPPORTED_CHARSET;
	}

	switch (code)
	{
		case ARM_SUCCESS:
			strcpy (msg, "");
			break;

		case ARM_FAILURE_INVALID_ARGUMENT:
			strcpy (msg, "Invalid argument");
			break;

		case ARM_FAILURE_NULL_ARGUMENT:
			strcpy (msg, "Return argument is passed a NULL pointer");
			break;

		case ARM_FAILURE_UNSUPPORTED_CHARSET:
			strcpy (msg, "Unsupported character set");
			break;

		case ARM_FAILURE_TRANSACTION_INACTIVE:
			strcpy (msg, "Transaction isn't active");
			break;

		case ARM_FAILURE_TRANSACTION_BLOCKED:
			strcpy (msg, "Transaction blocked");
			break;

		case ARM_FAILURE_TRANSACTION_NOT_BLOCKED:
			strcpy (msg, "Transaction not blocked");
			break;

		case ARM_FAILURE_DUPLICATE_ALIAS:
			strcpy (msg, "Duplicate Alias");
			break;

		case ARM_FAILURE_INTERNAL_ERROR:
			strcpy (msg, "Internal error");
			break;

		default:
			strcpy (msg, "Unknown error");
	}

	return ARM_SUCCESS;
}

ARM4_API_DYNAMIC(arm_error_t)
arm_is_charset_supported(
    const arm_charset_t charset,
    arm_boolean_t *supported)
{
    __libarm4_shm_trace_arm ("arm_is_charset_supported");

	if (supported == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	if ((charset == ARM_CHARSET_ASCII) || (charset == ARM_CHARSET_UTF8))
		*supported = ARM_TRUE;
	else
		*supported = ARM_FALSE;

	return ARM_SUCCESS;
}

/*--- Utility functions ---*/

static void
get_arrival_time(const arm_arrival_time_t time_handle,
	arm_time_t *output_time_value)
{
	arm_time_t time_value;
	arm_arrival_time_t era;
	arm_arrival_time_t era_least;
	arm_arrival_time_t whole;
	double fraction;

	/* Get the current time FIRST */
	gettimeofday (&time_value, NULL);

	__libarm4_shm_trace_arm_internal ("get_arrival_time");

	era = time_value.tv_sec / 0x0000000100000000LL;
	era_least = time_value.tv_sec & 0x0FFFFFFFFLL;

	fraction = (time_handle & 0x0FFFFFFFFLL);
	fraction /= 0x0000000100000000LL;
	whole = time_handle / 0x0000000100000000LL;

	/* Check for era rollover */
	if (era_least < whole)
		era--;

	output_time_value->tv_usec = fraction * 100000;
	output_time_value->tv_sec = whole + era;
}

static arm_error_t
unblock_transaction (transaction_instance_record_t *transaction_instance_ptr,
	arm_time_t time_value)
{
    __libarm4_shm_trace_arm_internal ("unblock_transaction");

	/* This routine assumes the transaction is active */
	if (transaction_instance_ptr == NULL)
		return ARM_FAILURE_NULL_ARGUMENT;

	if (time_is_zero (transaction_instance_ptr->block_start_time))
	{
		/* Not blocked */
		return ARM_FAILURE_TRANSACTION_NOT_BLOCKED;
	}

	/* Add our block time to the total transaction block time */
	transaction_instance_ptr->blocked_time.tv_sec += time_value.tv_sec - transaction_instance_ptr->block_start_time.tv_sec;
	transaction_instance_ptr->blocked_time.tv_usec += time_value.tv_usec - transaction_instance_ptr->block_start_time.tv_usec;
	if (transaction_instance_ptr->blocked_time.tv_usec < 0)
	{
		transaction_instance_ptr->blocked_time.tv_sec--;
		transaction_instance_ptr->blocked_time.tv_usec += 1000000;
	}
	memset (&(transaction_instance_ptr->block_start_time), 0, sizeof (arm_time_t));

	return ARM_SUCCESS;
}

static void
do_tree_initialization (void)
{
    __libarm4_shm_trace_arm_internal ("do_tree_initialization");

	__libarm4_tree_init (&g_application_instance_tree);
	__libarm4_tree_init (&g_transaction_instance_tree);

	__libarm4_tree_id_init (&g_application_alias_tree);
	__libarm4_tree_id_init (&g_transaction_alias_tree);
	__libarm4_tree_id_init (&g_metric_alias_tree);
}

static void
initialize_trees (void)
{
    __libarm4_shm_trace_arm_internal ("initialize_trees");
	pthread_once (&g_trees_initialized, do_tree_initialization);
}
