/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <syslog.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_UUID_UUID_H
#include <uuid/uuid.h>
#else
#ifdef HAVE_UUID_H
#include <uuid.h>
#endif
#endif

#include "arm4.h"
#include "arm4agent.h"
#include "libarm4_shm.h"
#include "libarm4_tree.h"
#include "arm4_db.h"

#define SHM_MODE	0660	/* Shared memory file permissions */

/* Wrappers to add stats to pthreads */
#ifdef HAVE_PTHREAD_STATS
#define PTHREAD_MUTEX_LOCK(lock_name) \
{ \
	int status; \
	int miss; \
	miss = 0; \
	do \
	{ \
		status = pthread_mutex_trylock (&g_shared_memory_ptr->lock_name); \
		if (status == EBUSY) \
			miss = 1; \
	} while (status == EBUSY); \
	if (status == 0) \
	{ \
		g_shared_memory_ptr->lock_name ## _pthread_stat_total++; \
		g_shared_memory_ptr->lock_name ## _pthread_stat_wait += miss; \
		g_shared_memory_ptr->lock_name ## _pthread_stat_active++; \
		if (g_shared_memory_ptr->lock_name ## _pthread_stat_active > g_shared_memory_ptr->lock_name ## _pthread_stat_max_active) \
			g_shared_memory_ptr->lock_name ## _pthread_stat_max_active = g_shared_memory_ptr->lock_name ## _pthread_stat_active; \
	} \
}
#define PTHREAD_MUTEX_UNLOCK(lock_name) \
{ \
	g_shared_memory_ptr->lock_name ## _pthread_stat_active--; \
	pthread_mutex_unlock (&g_shared_memory_ptr->lock_name); \
}
#else
#define PTHREAD_MUTEX_LOCK(lock_name) \
	pthread_mutex_lock (&g_shared_memory_ptr->lock_name);
#define PTHREAD_MUTEX_UNLOCK(lock_name) \
	pthread_mutex_unlock (&g_shared_memory_ptr->lock_name);
#endif

static arm4_shm_header_t *g_shared_memory_ptr = NULL;
static int g_message_queue = (-1);

static pthread_once_t g_attached = PTHREAD_ONCE_INIT;

#ifndef USE_POSIX_PROCESS_SEM
static sem_t *g_sequence_sem_ptr = NULL;
static sem_t *g_application_sem_ptr = NULL;
static sem_t *g_transaction_sem_ptr = NULL;
static sem_t *g_metric_sem_ptr = NULL;

static char *
instanceName (const char *template_ptr, int instance)
{
    static char name [256];
	
    /* Create a name from the template that includes the instance */
    sprintf (name, "%s%d", template_ptr, instance);
	
    return name;
}

#endif /* USE_POSIX_PROCESS_SEM */

static const char *
temp_path (const char *filename_ptr)
{
	static char buffer [4096];
	int instance;

	instance = __libarm4_shm_get_instance ();
	if (instance == 0)
		sprintf (buffer, ARM4DATA_QUEUE_DIR "/%s", filename_ptr);
	else
		sprintf (buffer, "/tmp/arm4-%d-%d/queue/%s", getuid(), instance, filename_ptr);

	return buffer;
}

static void
attach_semaphores (void)
{
	__libarm4_shm_trace_arm_shm_internal ("attach_semaphores");

#ifndef USE_POSIX_PROCESS_SEM
	int instance;
	
	instance = __libarm4_shm_get_instance ();
	if (g_sequence_sem_ptr == NULL)
		g_sequence_sem_ptr = sem_open (instanceName (SEM_SEQUENCE_FILE, instance), 0);
	if (g_sequence_sem_ptr == (sem_t *) SEM_FAILED)
	{
        syslog (LOG_ERR, "g_seqence_sem_ptr(%s) sem_open %s", instanceName (SEM_SEQUENCE_FILE, instance), strerror (errno));
		g_sequence_sem_ptr = NULL;
	}

	if (g_application_sem_ptr == NULL)
		g_application_sem_ptr = sem_open (instanceName (SEM_APPLICATION_FILE, instance), 0);
	if (g_application_sem_ptr == (sem_t *) SEM_FAILED)
	{
        syslog (LOG_ERR, "g_application_sem_ptr(%s) sem_open %s", instanceName (SEM_APPLICATION_FILE, instance), strerror (errno));
		g_application_sem_ptr = NULL;
	}

	if (g_transaction_sem_ptr == NULL)
		g_transaction_sem_ptr = sem_open (instanceName (SEM_TRANSACTION_FILE, instance), 0);
	if (g_transaction_sem_ptr == (sem_t *) SEM_FAILED)
	{
        syslog (LOG_ERR, "g_transaction_sem_ptr(%s) sem_open %s", instanceName (SEM_TRANSACTION_FILE, instance), strerror (errno));
		g_transaction_sem_ptr = NULL;
	}

    if (g_metric_sem_ptr == NULL)
        g_metric_sem_ptr = sem_open (instanceName (SEM_METRIC_FILE, instance), 0);
    if (g_metric_sem_ptr == (sem_t *) SEM_FAILED)
    {
        syslog (LOG_ERR, "g_metric_sem_ptr(%s) sem_open %s", instanceName (SEM_METRIC_FILE, instance), strerror (errno));
        g_metric_sem_ptr = NULL;
    }
#endif /* USE_POSIX_PROCESS_SEM */
}

static void
detach (void)
{
	__libarm4_shm_trace_arm_shm_internal ("detach");

	if (g_shared_memory_ptr)
	{
		if (shmdt ((void *) g_shared_memory_ptr) < 0)
		{
            syslog (LOG_ERR, "shmdt %s", strerror (errno));
		}
		g_shared_memory_ptr = NULL;
	}

	/* Indicate that we're no longer open for business */
	if (g_shared_memory_ptr)
		g_shared_memory_ptr->magic = 0;
}

static void
attach (void)
{
	int shmid;
	key_t token;
	void *ptr;

    /* Initialize the logging facility */
    openlog ("libarm4", LOG_PID, LOG_DAEMON);
	__libarm4_shm_trace_arm_shm_internal ("attach");

	token = ftok (temp_path (MQUEUE_FILE), 1);
	if (token == (key_t)-1)
	{
        syslog (LOG_ERR, "unable to attach to message queue %s: %s", temp_path (MQUEUE_FILE), strerror (errno));
        return;
	}

	g_message_queue = msgget (token, SHM_MODE);
	if (g_message_queue == -1)
	{
        syslog (LOG_ERR, "unable to attach to message queue %s: %s", temp_path (MQUEUE_FILE), strerror (errno));
        return;
	}

	if (g_shared_memory_ptr == NULL)
	{
		token = ftok (temp_path (SHARED_MEMORY_FILE), 1);
		if (token == (key_t)-1)
		{
			syslog (LOG_ERR, "unable to get shared memory token %s: %s", temp_path (SHARED_MEMORY_FILE), strerror (errno));
			return;
		}

		shmid = shmget (token, sizeof (arm4_shm_header_t), SHM_MODE);
		if (shmid < 0)
		{
			syslog (LOG_ERR, "unable to get shared memory %s: %s", temp_path (SHARED_MEMORY_FILE), strerror (errno));
			return;
		}
		ptr = shmat (shmid, NULL, 0);
		if (ptr == NULL)
		{
            syslog (LOG_ERR, "unable to attach to shared memory %s", strerror (errno));
			return;
		}
		g_shared_memory_ptr = ptr;

		/* Check that we've been initialized */
		if (g_shared_memory_ptr->magic != ARM4_SHM_MAGIC)
		{
			detach ();
			return;
		}
	}

	atexit (detach);

	/* Initialize the semaphores */
	attach_semaphores ();

	return;
}

static void *
arm4_shm_attach (void)
{
	__libarm4_shm_trace_arm_shm_internal ("arm4_shm_attach");
	pthread_once (&g_attached, attach);

	return g_shared_memory_ptr;
}

int
__libarm4_shm_get_instance (void)
{
	/* Returns the current instance being used */
	char *env_ptr;
	int instance = 0;

	env_ptr = getenv ("ARM4_INSTANCE");
	if (env_ptr)
		instance = atoi (env_ptr);

	return instance;
}

arm4_shm_header_t *
__libarm4_shm_get_header (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_get_header");
	if (arm4_shm_attach () == NULL)
		return NULL;

	if ((g_shared_memory_ptr == NULL) || (g_shared_memory_ptr->magic != ARM4_SHM_MAGIC))
		return NULL;

	return g_shared_memory_ptr;
}

arm_boolean_t
__libarm4_shm_is_collecting (void)
{
	arm4_shm_header_t *header_ptr;

	/* Check to see if collection is active */
	header_ptr = __libarm4_shm_get_header ();
	if (header_ptr && (header_ptr->null_arm == ARM_FALSE))
		return ARM_TRUE;
	else
		return ARM_FALSE;
}

/*
 * Routines used to isolate the mutex routines
 */

/* Tree mutexes. These always use pthreads, since they're local to the process */
void
__libarm4_shm_init_tree (struct tree *tree_ptr)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_init_tree");
	pthread_mutex_init (&(tree_ptr->mutex), NULL);
}

void
__libarm4_shm_lock_tree (struct tree *tree_ptr)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_lock_tree");
	pthread_mutex_lock (&(tree_ptr->mutex));
}

void
__libarm4_shm_unlock_tree (struct tree *tree_ptr)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_unlock_tree");
	pthread_mutex_unlock (&(tree_ptr->mutex));
}

void
__libarm4_shm_lock_sequence (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_lock_sequence");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_LOCK (sequence_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_wait (g_sequence_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_lock_sequence %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_unlock_sequence (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_unlock_sequence");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_UNLOCK (sequence_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_post (g_sequence_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_unlock_sequence %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_lock_application (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_lock_application");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_LOCK (application_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_wait (g_application_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_lock_application %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_unlock_application (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_unlock_application");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_UNLOCK (application_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_post (g_application_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_unlock_application %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_lock_transaction (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_lock_transaction");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_LOCK (transaction_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_wait (g_transaction_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_lock_transaction %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_unlock_transaction (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_unlock_transaction");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_UNLOCK (transaction_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_post (g_transaction_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_unlock_transaction %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_lock_metric (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_lock_metric");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_LOCK (metric_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_wait (g_metric_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_lock_metric %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
__libarm4_shm_unlock_metric (void)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_unlock_metric");
#ifdef USE_POSIX_PROCESS_SEM
	PTHREAD_MUTEX_UNLOCK (metric_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;

	status = sem_post (g_metric_sem_ptr);
	if (status != 0)
	{
        syslog (LOG_ERR, "__libarm4_shm_unlock_metric %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}


int
__libarm4_shm_is_valid_identifier (const arm_char_t *name_ptr)
{
	/* Validates identifiers passed in as application or transaction names */
	if ( (name_ptr == NULL) ||
		(strlen (name_ptr) > ARM_NAME_MAX_CHARS) ||
	    (strlen (name_ptr) == 0) )
		return ARM_FALSE;

	/* Check for trailing blank or all blanks */
	if (isspace (name_ptr [strlen(name_ptr) - 1]))
		return ARM_FALSE;

	while (*name_ptr && isspace (*name_ptr))
		name_ptr++;
	if (*name_ptr)
		return ARM_TRUE;

	return ARM_FALSE;
}

/*
 * Send data over the message queues
 */

void
__libarm4_shm_add_app_instance (arm_app_start_handle_t app_handle,
	arm_id_t appl_id,
	arm_boolean_t active)
{
	arm4_shm_application_instance_t app_instance;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_app_instance");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	app_instance.mtype = ARM4_SHM_MESSAGE_APPLICATION;
	app_instance.message.app_handle = app_handle;
	app_instance.message.appl_id = appl_id;
	app_instance.message.active = active;

	status = SAFE_SEND (g_message_queue, &app_instance, sizeof (app_instance), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Application msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_app_group (arm_app_start_handle_t app_handle,
	const char *group_name)
{
	arm4_shm_property_t app_group;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_app_group");

	if (group_name == NULL)
		return;

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	app_group.mtype = ARM4_SHM_MESSAGE_APPLICATION_GROUP;
	app_group.message.handle = app_handle;
	strncpy (app_group.message.property_name, group_name, ARM_PROPERTY_VALUE_MAX_LENGTH);

	status = SAFE_SEND (g_message_queue, &app_group, sizeof (app_group), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Application group msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_app_instance_id (arm_app_start_handle_t app_handle,
	const char *id_ptr)
{
	arm4_shm_property_t property;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_app_instance_id");

	if (id_ptr == NULL)
		return;

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	property.mtype = ARM4_SHM_MESSAGE_APPLICATION_INSTANCE_ID;
	property.message.handle = app_handle;
	strncpy (property.message.property_name, id_ptr, ARM_PROPERTY_VALUE_MAX_LENGTH);

	status = SAFE_SEND (g_message_queue, &property, sizeof (property), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Application instance id msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_app_context (arm_app_start_handle_t app_handle,
	int index,
	const char *context_ptr)
{
	arm4_shm_context_property_t property;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_app_context");

    if (context_ptr == NULL)
        return;

    /* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	property.mtype = ARM4_SHM_MESSAGE_APPLICATION_CONTEXT;
	property.message.handle = app_handle;
	property.message.property_index = index;
	strncpy (property.message.property_name, context_ptr, ARM_PROPERTY_VALUE_MAX_LENGTH);

	status = SAFE_SEND (g_message_queue, &property, sizeof (property), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Application context msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_app_address (arm_app_start_handle_t app_handle,
	const arm_subbuffer_system_address_t *address_buffer)
{
	arm4_shm_address_t address;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_app_context");

    if (address_buffer == NULL)
        return;

    /* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	address.mtype = ARM4_SHM_MESSAGE_APPLICATION_ADDRESS;
	address.message.app_handle = app_handle;
	address.message.address_format = address_buffer->address_format;
	address.message.address_length = address_buffer->address_length;
	address.message.id_valid = address_buffer->id_valid;
	address.message.id = address_buffer->id;

	 /* Truncate address that are too long */
	if (address.message.address_length > ARM_PROPERTY_URI_MAX_LENGTH)
		address.message.address_length = ARM_PROPERTY_URI_MAX_LENGTH;
	memcpy (address.message.address, address_buffer->address, address.message.address_length);

	status = SAFE_SEND (g_message_queue, &address,
            sizeof (address) - (ARM_PROPERTY_URI_MAX_LENGTH - address.message.address_length), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Application address msgsnd %s", strerror (errno));
	}
}

int
__libarm4_shm_transaction_trace (const arm_id_t tran_id, const arm_int32_t flags, const arm_boolean_t force_trace)
{
	int i;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_transaction_trace");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return 0;

	__libarm4_shm_lock_transaction ();

	for (i = 0; i < g_shared_memory_ptr->transaction_count; i++)
	{
		if (memcmp (&(g_shared_memory_ptr->transaction_table [i].tran_id), &tran_id, sizeof (tran_id)) == 0)
		{
			// Found the transaction entry
			g_shared_memory_ptr->transaction_table [i].transaction_counter++;

			int trace_flags = (int) g_shared_memory_ptr->transaction_table [i].trace_flags;

			/* Check to see if all are collected */
			if ((trace_flags & arm4_shm_transaction_trace_all) || (force_trace == ARM_TRUE))
			{
				__libarm4_shm_unlock_transaction ();
				return ARM_CORR_FLAGNUM_AGENT_TRACE;
			}

			/* Check for collection based on sample count */
			if (trace_flags & arm4_shm_transaction_trace_n)
			{
				if ((g_shared_memory_ptr->transaction_table [i].transaction_counter %
						g_shared_memory_ptr->transaction_table [i].trace_parameter) == 0)
				{
					__libarm4_shm_unlock_transaction ();
					return ARM_CORR_FLAGNUM_AGENT_TRACE;
				}
			}

			/* Check for interval sampling */
			if (trace_flags & arm4_shm_transaction_trace_interval)
			{
				if (trace_flags & arm4_shm_transaction_trace_next)
				{
					/* clear the trace request bit */
					trace_flags &= (~arm4_shm_transaction_trace_next);
					g_shared_memory_ptr->transaction_table [i].trace_flags = (arm4_shm_transaction_trace_t) trace_flags;

					__libarm4_shm_unlock_transaction ();
					return ARM_CORR_FLAGNUM_AGENT_TRACE;
				}
			}

			/* Check the default setting */

			/* Check for application trace request if all others failed */
			if ((trace_flags & arm4_shm_transaction_trace_honor_request) &&
				(flags & ARM_FLAG_TRACE_REQUEST))
			{
				__libarm4_shm_unlock_transaction ();
				return ARM_CORR_FLAGNUM_APP_TRACE;
			}

			/* No trace required */
			__libarm4_shm_unlock_transaction ();
			return 0;
		}
	}

	/* Not found */
	__libarm4_shm_unlock_transaction ();

	return 0;
}

int
__libarm4_shm_try_tran_instance (arm_tran_start_handle_t tran_handle,
	arm_id_t appl_id,
	arm_id_t tran_id,
	arm_app_start_handle_t app_handle,
	arm_time_t arrival_time,
	arm_time_t stop_time,
	arm_time_t response_time,
	arm_time_t blocked_time,
	arm_tran_status_t tran_status,
	arm_boolean_t active)
{
	arm4_shm_transaction_instance_t tran_instance;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_tran_instance");
	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return -1;

	tran_instance.mtype = ARM4_SHM_MESSAGE_TRANSACTION;
	tran_instance.message.tran_handle = tran_handle;
	tran_instance.message.appl_id = appl_id;
	tran_instance.message.tran_id = tran_id;
	tran_instance.message.app_handle = app_handle;
	tran_instance.message.arrival_time = arrival_time;
	tran_instance.message.stop_time = stop_time;
	tran_instance.message.response_time = response_time;
	tran_instance.message.blocked_time = blocked_time;
	tran_instance.message.tran_status = tran_status;
	tran_instance.message.active = active;

	status = SAFE_SEND (g_message_queue, &tran_instance, sizeof (tran_instance), IPC_NOWAIT);
	if ((status == -1) && (errno != EAGAIN))
	{
        syslog (LOG_ERR, "Transaction msgsnd %s", strerror (errno));
	}

	return status;
}

void
__libarm4_shm_add_tran_instance (arm_tran_start_handle_t tran_handle,
	arm_id_t appl_id,
	arm_id_t tran_id,
	arm_app_start_handle_t app_handle,
	arm_time_t arrival_time,
	arm_time_t stop_time,
	arm_time_t response_time,
	arm_time_t blocked_time,
	arm_tran_status_t tran_status,
	arm_boolean_t active)
{
	arm4_shm_transaction_instance_t tran_instance;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_tran_instance");
	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	tran_instance.mtype = ARM4_SHM_MESSAGE_TRANSACTION;
	tran_instance.message.tran_handle = tran_handle;
	tran_instance.message.appl_id = appl_id;
	tran_instance.message.tran_id = tran_id;
	tran_instance.message.app_handle = app_handle;
	tran_instance.message.arrival_time = arrival_time;
	tran_instance.message.stop_time = stop_time;
	tran_instance.message.response_time = response_time;
	tran_instance.message.blocked_time = blocked_time;
	tran_instance.message.tran_status = tran_status;
	tran_instance.message.active = active;

	status = SAFE_SEND (g_message_queue, &tran_instance, sizeof (tran_instance), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Transaction msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_tran_context (arm_tran_start_handle_t tran_handle,
	int index,
	const char *context_ptr)
{
	arm4_shm_context_property_t property;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_tran_context");

    if (context_ptr == NULL)
        return;

    /* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	property.mtype = ARM4_SHM_MESSAGE_TRANSACTION_CONTEXT;
	property.message.handle = tran_handle;
	property.message.property_index = index;
	strncpy (property.message.property_name, context_ptr, ARM_PROPERTY_VALUE_MAX_LENGTH);

	status = SAFE_SEND (g_message_queue, &property, sizeof (property), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Transaction context msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_tran_uri (arm_tran_start_handle_t tran_handle,
	const char *uri_ptr)
{
	arm4_shm_uri_t uri;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_tran_uri");
	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	uri.mtype = ARM4_SHM_MESSAGE_TRANSACTION_URI;
	uri.message.handle = tran_handle;
	strncpy (uri.message.property_name, uri_ptr, ARM_PROPERTY_URI_MAX_LENGTH);

	status = SAFE_SEND (g_message_queue, &uri, sizeof (uri), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Transaction uri msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_metric_value (arm_tran_start_handle_t tran_handle,
	const arm_metric_t *metric_ptr)
{
	arm4_shm_metric_value_t metric;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_metric_value");

	/* Ensure the metric is valid */
	if (!metric_ptr->valid)
		return;

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	metric.mtype = ARM4_SHM_MESSAGE_TRANSACTION_METRIC_VALUE;
	metric.message.handle = tran_handle;
	memcpy (&(metric.message.value), metric_ptr, sizeof (arm_metric_t));
	if (metric_ptr->format == ARM_METRIC_FORMAT_STRING32)
		{
		strncpy (metric.message.string32, metric_ptr->metric_u.string32, ARM_METRIC_STRING32_MAX_LENGTH);
		metric.message.value.metric_u.string32 = NULL; /* We're not using this entry */
		}
	else
		metric.message.string32 [0] = '\0';

	status = SAFE_SEND (g_message_queue, &metric, sizeof (metric), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Transaction metric value msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_user (arm_tran_start_handle_t tran_handle,
	const arm_subbuffer_user_t *user_ptr)
{
	arm4_shm_user_t user;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_user");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	user.mtype = ARM4_SHM_MESSAGE_TRANSACTION_USER;
	user.message.handle = tran_handle;

	if (user_ptr->name)
		strncpy (user.message.name, user_ptr->name, ARM_NAME_MAX_LENGTH);
	else
		user.message.name [0] = '\0';
	user.message.id_valid = user_ptr->id_valid;
	if (user_ptr->id_valid)
		memcpy (&(user.message.id), &(user_ptr->id), sizeof (arm_id_t));
	else
		memset (&(user.message.id), 0, sizeof (arm_id_t));

	status = SAFE_SEND (g_message_queue, &user, sizeof (user), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Transaction user msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_diag (arm_tran_start_handle_t tran_handle,
	const arm_subbuffer_diag_detail_t *diag_ptr)
{
	arm4_shm_diag_t diag;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_diag");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	diag.mtype = ARM4_SHM_MESSAGE_TRANSACTION_DIAG;
	diag.message.handle = tran_handle;

	strncpy (diag.message.detail, diag_ptr->diag_detail, ARM_DIAG_DETAIL_MAX_LENGTH);

	status = SAFE_SEND (g_message_queue, &diag, sizeof (diag), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Transaction user msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_add_correlators (arm_tran_start_handle_t tran_handle,
	const arm_correlator_t *parent_correlator,
	const arm_correlator_t *current_correlator)
{
	arm4_shm_correlator_t correlator;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_add_correlators");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	correlator.mtype = ARM4_SHM_MESSAGE_CORRELATOR;
	correlator.message.tran_handle = tran_handle;
	if (parent_correlator)
		memcpy (&(correlator.message.parent_correlator), parent_correlator, sizeof (arm_correlator_t));
	else
		memset (&(correlator.message.parent_correlator), 0, sizeof (arm_correlator_t));
	if (current_correlator)
		memcpy (&(correlator.message.current_correlator), current_correlator, sizeof (arm_correlator_t));
	else
		memset (&(correlator.message.current_correlator), 0, sizeof (arm_correlator_t));

	status = SAFE_SEND (g_message_queue, &correlator, sizeof (correlator), 0);
	if (status == -1)
	{
        syslog (LOG_ERR, "Correlator msgsnd %s", strerror (errno));
	}
}

void
__libarm4_shm_save_sequence (arm4_shm_sequence_t index,
	arm_int64_t sequence)
{
	arm4_shm_db_sequence_t	message;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_save_sequence");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	message.mtype = ARM4_SHM_MESSAGE_SEQUENCE;
	message.message.index = index;
	message.message.sequence = sequence;

	status = SAFE_SEND (g_message_queue, &message, sizeof (message), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Sequence msgsnd %s", strerror (errno));
	}
}

int
__libarm4_shm_next_sequence_number (arm4_shm_sequence_t index,
	arm_int64_t *sequence_ptr)
{
	arm4_shm_header_t *header_ptr;
	arm_int64_t	sequence;

    __libarm4_shm_trace_arm_shm ("__libarm4_shm_next_sequence_number");

	/* Get a unique sequence number for the database */
	if (index >= arm4_shm_sequence_max)
		return ARM_FAILURE;

	/* Get the next sequence in memory */
	header_ptr = __libarm4_shm_get_header ();
	if (header_ptr == NULL)
		return ARM_FAILURE;

	__libarm4_shm_lock_sequence (); /* We might be able to improve concurrency if we lock specific sequences? */

	*sequence_ptr = header_ptr->sequence_table [index];

	/* Increment the sequence number */
	*sequence_ptr += 1;

	/* Check for a wrap - don't allow the 0 value */
	if (*sequence_ptr < 1)
		*sequence_ptr = 1;

	header_ptr->sequence_table [index] = *sequence_ptr;

	if ((*sequence_ptr != 1) && ((*sequence_ptr % ARM4_SHM_SEQUENCE_STEP) != 0))
	{
		/* Return right away if we don't need to write to the database */
		__libarm4_shm_unlock_sequence ();
		return ARM_SUCCESS;
	}

	/* Save the sequence */
	sequence = *sequence_ptr;
	__libarm4_shm_unlock_sequence ();
	__libarm4_shm_save_sequence (index, sequence);

	return ARM_SUCCESS;
}

void
__libarm4_shm_register_application (const unsigned char *digest_ptr,
	const arm_char_t *app_name_ptr,
	arm_charset_t charset,
	int application_index)
{
	arm4_shm_register_application_t	message;
	int status;
	struct sembuf sbuf;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_register_application");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	message.mtype = ARM4_SHM_MESSAGE_REGISTER_APPLICATION;
	message.message.application_index = application_index;
	memcpy (&(message.message.digest), digest_ptr, MD5_DIGEST_LENGTH);
	memcpy (&(message.message.app_name), app_name_ptr, ARM_NAME_MAX_LENGTH);
	message.message.charset = charset;

	status = SAFE_SEND (g_message_queue, &message, sizeof (message), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Register application msgsnd %s", strerror (errno));
	}

	/* Wait for registration to complete */
	sbuf.sem_num = SEM_INSTANCE_APPLICATION;
	sbuf.sem_op = -1;
	sbuf.sem_flg = 0;
	if (semop (g_shared_memory_ptr->application_semaphore, & sbuf, 1) != 0)
	{
		perror ("Unable to wait for application semaphore");
	}
}

arm_error_t
__libarm4_shm_register_application_identity (const arm_id_t *app_id_ptr,
	const arm_subbuffer_app_identity_t *identity_buffer_ptr)
{
	arm4_shm_register_application_identity_t	identity;
	arm4_shm_register_application_context_t		context;
	int i;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_register_application_identity");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return ARM_FAILURE;

	/* Send the identity properties */
	for (i = 0; i < identity_buffer_ptr->identity_property_count; i++)
	{
		if ((strlen (identity_buffer_ptr->identity_property_array [i].name) > 0) &&
			(strlen (identity_buffer_ptr->identity_property_array [i].value) > 0))
		{
			identity.mtype = ARM4_SHM_MESSAGE_REGISTER_APPLICATION_IDENTITY;
			identity.message.appid = *app_id_ptr;
			strncpy (identity.message.name, identity_buffer_ptr->identity_property_array [i].name, ARM_PROPERTY_NAME_MAX_LENGTH);
			strncpy (identity.message.value, identity_buffer_ptr->identity_property_array [i].value, ARM_PROPERTY_VALUE_MAX_LENGTH);

			status = SAFE_SEND (g_message_queue, &identity, sizeof (identity), 0);
			if (status == -1)
			{
				syslog (LOG_ERR, "Register application identity msgsnd %s", strerror (errno));
				return ARM_FAILURE;
			}
		}
	}

	/* Save the context properties */
	for (i = 0; i < identity_buffer_ptr->context_name_count; i++)
	{
		if ((identity_buffer_ptr->context_name_array [i]) && (__libarm4_shm_is_valid_identifier (identity_buffer_ptr->context_name_array [i])))
		{
			context.mtype = ARM4_SHM_MESSAGE_REGISTER_APPLICATION_CONTEXT;
			context.message.appid = *app_id_ptr;
			context.message.index = i;
			strncpy (context.message.name, identity_buffer_ptr->context_name_array [i], ARM_PROPERTY_NAME_MAX_LENGTH);

			status = SAFE_SEND (g_message_queue, &context, sizeof (context), 0);
			if (status == -1)
			{
				syslog (LOG_ERR, "Register application context msgsnd %s", strerror (errno));
				return ARM_FAILURE;
			}
		}
	}
	return ARM_SUCCESS;
}

void
__libarm4_shm_register_transaction (const unsigned char *digest_ptr,
	const arm_id_t *app_id_ptr,
	const arm_char_t *tran_name_ptr,
	int transaction_index)
{
	arm4_shm_register_transaction_t	message;
	int status;
	struct sembuf sbuf;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_register_transaction");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	message.mtype = ARM4_SHM_MESSAGE_REGISTER_TRANSACTION;
	message.message.transaction_index = transaction_index;
	memcpy (&(message.message.digest), digest_ptr, MD5_DIGEST_LENGTH);
	strncpy (message.message.tran_name, tran_name_ptr, ARM_NAME_MAX_LENGTH);
	message.message.app_id = *app_id_ptr;

	status = SAFE_SEND (g_message_queue, &message, sizeof (message), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Register transaction msgsnd %s", strerror (errno));
	}

	/* Wait for registration to complete */
	sbuf.sem_num = SEM_INSTANCE_TRANSACTION;
	sbuf.sem_op = -1;
	sbuf.sem_flg = 0;
	if (semop (g_shared_memory_ptr->application_semaphore, & sbuf, 1) != 0)
	{
		perror ("Unable to wait for transaction semaphore");
	}
}

arm_error_t
__libarm4_shm_register_transaction_identity (const arm_id_t *tran_id_ptr,
	const arm_subbuffer_tran_identity_t *identity_buffer_ptr)
{
	arm4_shm_register_application_identity_t	identity;
	arm4_shm_register_application_context_t		context;
	arm4_shm_register_transaction_uri_t			uri;
	int i;
	int count;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_register_transaction_identity");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return ARM_FAILURE;

	/* Send the identity properties */
	count = identity_buffer_ptr->identity_property_count;
	for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
	{
		if ((strlen (identity_buffer_ptr->identity_property_array [i].name) > 0) &&
			(strlen (identity_buffer_ptr->identity_property_array [i].value) > 0))
		{
			count--;

			identity.mtype = ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_IDENTITY;
			identity.message.appid = *tran_id_ptr;
			strncpy (identity.message.name, identity_buffer_ptr->identity_property_array [i].name, ARM_PROPERTY_NAME_MAX_LENGTH);
			strncpy (identity.message.value, identity_buffer_ptr->identity_property_array [i].value, ARM_PROPERTY_VALUE_MAX_LENGTH);

			status = SAFE_SEND (g_message_queue, &identity, sizeof (identity), 0);
			if (status == -1)
			{
				syslog (LOG_ERR, "Register transaction identity msgsnd %s", strerror (errno));
				return ARM_FAILURE;
			}
		}
	}

	/* Save the context properties */
	count = identity_buffer_ptr->context_name_count;
	for (i = 0; (i < ARM_PROPERTY_MAX_COUNT) && count; i++)
	{
		if ((identity_buffer_ptr->context_name_array [i]) && (__libarm4_shm_is_valid_identifier (identity_buffer_ptr->context_name_array [i])))
		{
			count--;

			context.mtype = ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_CONTEXT;
			context.message.appid = *tran_id_ptr;
			context.message.index = i;
			strncpy (context.message.name, identity_buffer_ptr->context_name_array [i], ARM_PROPERTY_NAME_MAX_LENGTH);

			status = SAFE_SEND (g_message_queue, &context, sizeof (context), 0);
			if (status == -1)
			{
				syslog (LOG_ERR, "Register transaction context 1 msgsnd %s", strerror (errno));
				return ARM_FAILURE;
			}
		}
	}

	/* Save the URI */
	if ((identity_buffer_ptr->uri) && (strlen (identity_buffer_ptr->uri) > 0))
	{
		uri.mtype = ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_URI;
		uri.message.tran_id = *tran_id_ptr;
		strncpy (uri.message.property_name, identity_buffer_ptr->uri, ARM_PROPERTY_URI_MAX_LENGTH);

		status = SAFE_SEND (g_message_queue, &uri, sizeof (uri), 0);
		if (status == -1)
		{
			syslog (LOG_ERR, "Register transaction context 2 msgsnd %s", strerror (errno));
			return ARM_FAILURE;
		}
	}

	return ARM_SUCCESS;
}

arm_error_t
__libarm4_shm_register_transaction_metric_binding (const arm_id_t *tran_id_ptr,
	const arm_subbuffer_metric_bindings_t *metric_buffer_ptr)
{
    arm4_shm_register_transaction_metric_t message;
	int i;
	int status;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_register_transaction_metric_binding");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return ARM_FAILURE;

	/* Send the identity properties */
	for (i = 0; i < metric_buffer_ptr->count; i++)
	{
        message.mtype = ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_METRIC;
        message.message.tran_id = *tran_id_ptr;
        message.message.slot = metric_buffer_ptr->metric_binding_array [i].slot;
        message.message.id = metric_buffer_ptr->metric_binding_array [i].id;

        status = SAFE_SEND (g_message_queue, &message, sizeof (message), 0);
        if (status == -1)
        {
            syslog (LOG_ERR, "Register transaction metric msgsnd %s", strerror (errno));
            return ARM_FAILURE;
        }
	}

	return ARM_SUCCESS;
}

void
__libarm4_shm_register_metric (unsigned char *digest_ptr,
    const arm_id_t *app_id_ptr,
	const char *metric_name_ptr,
    const arm_metric_format_t metric_format,
    const arm_metric_usage_t metric_usage,
    const arm_char_t *metric_unit_ptr,
	const int metric_index)
{
	arm4_shm_register_metric_t	message;
	int status;
	struct sembuf sbuf;

	__libarm4_shm_trace_arm_shm ("__libarm4_shm_register_metric");

	/* Ensure we're attached */
	if (arm4_shm_attach () == NULL)
		return;

	message.mtype = ARM4_SHM_MESSAGE_REGISTER_METRIC;
	message.message.metric_index = metric_index;
	memcpy (message.message.digest, digest_ptr, MD5_DIGEST_LENGTH);
	strncpy (message.message.metric_name, metric_name_ptr, ARM_NAME_MAX_LENGTH);
	message.message.app_id = *app_id_ptr;
    message.message.metric_format = metric_format;
    message.message.metric_usage = metric_usage;
    if (metric_unit_ptr)
        strncpy (message.message.metric_unit, metric_unit_ptr,  ARM_NAME_MAX_LENGTH);
    else
        message.message.metric_unit [0] = '\0';

	status = SAFE_SEND (g_message_queue, &message, sizeof (message), 0);
	if (status == -1)
	{
		syslog (LOG_ERR, "Register transaction msgsnd %s", strerror (errno));
	}

	/* Wait for registration to complete */
	sbuf.sem_num = SEM_INSTANCE_METRIC;
	sbuf.sem_op = -1;
	sbuf.sem_flg = 0;
	if (semop (g_shared_memory_ptr->application_semaphore, &sbuf, 1) != 0)
	{
		perror ("Unable to wait for metric semaphore");
	}

}

void
__libarm4_shm_get_uuid (arm_id_t *uuid_ptr)
{
	__libarm4_shm_trace_arm_shm ("__libarm4_shm_get_uuid");

    uuid_generate ((arm_uint8_t *) uuid_ptr);
}

/*
 * Logging functions used to trace execuition
 */
void
__libarm4_shm_trace_arm (const char *string_ptr)
{
    if (g_shared_memory_ptr && (g_shared_memory_ptr->trace_level >= arm4_shm_trace_arm))
    {
        syslog (LOG_DAEMON, "%s", string_ptr);
    }
}

void
__libarm4_shm_trace_arm_internal (const char *string_ptr)
{
    if (g_shared_memory_ptr && (g_shared_memory_ptr->trace_level >= arm4_shm_trace_arm_internal))
    {
        syslog (LOG_DAEMON, "static %s", string_ptr);
    }
}

void
__libarm4_shm_trace_arm_shm (const char *string_ptr)
{
    if (g_shared_memory_ptr && (g_shared_memory_ptr->trace_level >= arm4_shm_trace_arm_shm))
    {
        syslog (LOG_DAEMON, "shared memory %s", string_ptr);
    }
}

void
__libarm4_shm_trace_arm_shm_internal (const char *string_ptr)
{
    if (g_shared_memory_ptr && (g_shared_memory_ptr->trace_level >= arm4_shm_trace_arm_shm_internal))
    {
        syslog (LOG_DAEMON, "shared memory static %s", string_ptr);
    }
}

