/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 
/* Functions */
#include <stdlib.h>
#include <string.h>

#include "libarm4_tree.h"
#include "libarm4_shm.h"

/* Prototypes */
static void initialize_node (tree_t *tree_ptr, tree_record_t *node_ptr, const void *key_ptr, void *data_ptr);
static void insert_child (tree_t *tree_ptr, tree_record_t *node_ptr, tree_record_t *new_node_ptr, const void *key_ptr);
static void insert_child_node (tree_t *tree_ptr, tree_record_t *node_ptr, tree_record_t *child_ptr);
static void *find_child (tree_t *tree_ptr, tree_record_t *node_ptr, const void *key_ptr);
static void *remove_child (tree_t *tree_ptr, tree_record_t *parent_ptr, tree_record_t *node_ptr, const void *key_ptr, int left);
static tree_record_t *remove_first_child (tree_t *tree_ptr, tree_record_t *parent_ptr, tree_record_t *node_ptr);

/* Static functions for the handle tree */
static int
int64_compare (const void *key1_ptr, const void *key2_ptr)
{
	const arm_int64_t	*handle1_ptr;
	const arm_int64_t	*handle2_ptr;

	handle1_ptr = (arm_int64_t *) key1_ptr;
	handle2_ptr = (arm_int64_t *) key2_ptr;

	return (*handle2_ptr - *handle1_ptr);
}

static void *
int64_clone (const void *key_ptr)
{
	arm_int64_t			*new_handle_ptr;

	new_handle_ptr = malloc (sizeof (arm_int64_t));
	*new_handle_ptr = *((arm_int64_t *) key_ptr);

	return new_handle_ptr;
}

/* Static functions for the id tree */
static int
id_compare (const void *key1_ptr, const void *key2_ptr)
{
	return memcmp (key1_ptr, key2_ptr, sizeof (arm_id_t));
}

static void *
id_clone (const void *key_ptr)
{
	arm_id_t			*new_id_ptr;

	new_id_ptr = malloc (sizeof (arm_id_t));
	memcpy (new_id_ptr, key_ptr, sizeof (arm_id_t));

	return new_id_ptr;
}

static void
set_empty_id (arm_id_t *id_ptr)
{
	memset (id_ptr, 0, sizeof (arm_id_t));
}

/* Static functions common to all trees */
static int
key_is_less (tree_t *tree_ptr, const void *key1_ptr, const void *key2_ptr)
{
	return (tree_ptr->compare (key1_ptr, key2_ptr) < 0);
}

static void
initialize_node (tree_t *tree_ptr, tree_record_t *node_ptr, const void *key_ptr, void *data_ptr)
{
	node_ptr->left_ptr = NULL;
	node_ptr->right_ptr = NULL;
	node_ptr->key_ptr = tree_ptr->clone (key_ptr);
	node_ptr->data_ptr = data_ptr;
}

static void
insert_child (tree_t *tree_ptr, tree_record_t *node_ptr, tree_record_t *new_node_ptr, const void *key_ptr)
{
	/* Recursive insert */
	if (key_is_less (tree_ptr, node_ptr->key_ptr, key_ptr))
		{
		/* Insert left */
		if (node_ptr->left_ptr == NULL)
			node_ptr->left_ptr = new_node_ptr;
		else
			insert_child (tree_ptr, node_ptr->left_ptr, new_node_ptr, key_ptr);
		}
	else
		{
		/* Insert right */
		if (node_ptr->right_ptr == NULL)
			node_ptr->right_ptr = new_node_ptr;
		else
			insert_child (tree_ptr, node_ptr->right_ptr, new_node_ptr, key_ptr);
		}
}

static void
insert_child_node (tree_t *tree_ptr, tree_record_t *node_ptr, tree_record_t *child_ptr)
{
	if (child_ptr == NULL)
		return;

	/* Recursive insert */
	if (key_is_less (tree_ptr, node_ptr->key_ptr, child_ptr->key_ptr))
		{
		/* Insert left */
		if (node_ptr->left_ptr == NULL)
			node_ptr->left_ptr = child_ptr;
		else
			insert_child_node (tree_ptr, node_ptr->left_ptr, child_ptr);
		}
	else
		{
		/* Insert right */
		if (node_ptr->right_ptr == NULL)
			node_ptr->right_ptr = child_ptr;
		else
			insert_child_node (tree_ptr, node_ptr->right_ptr, child_ptr);
		}
}

static void *
find_child (tree_t *tree_ptr, tree_record_t *node_ptr, const void *key_ptr)
{
	int compare;

	/* Not found */
	if (node_ptr == NULL)
		return NULL;

	compare = tree_ptr->compare (node_ptr->key_ptr, key_ptr);
	if (compare == 0)
		return node_ptr->data_ptr;
	else if (compare < 0)
		return find_child (tree_ptr, node_ptr->left_ptr, key_ptr);
	else
		return find_child (tree_ptr, node_ptr->right_ptr, key_ptr);
}

static void *
remove_child (tree_t *tree_ptr, tree_record_t *parent_ptr, tree_record_t *node_ptr, const void *key_ptr, int left)
{
	int compare;
	tree_record_t *child_ptr;
	void *data_ptr;
	
	/* Not found */
	if (node_ptr == NULL)
		return NULL;
		
	compare = tree_ptr->compare (node_ptr->key_ptr, key_ptr);
	if (compare == 0)
	{
		/* This is us, create a new subtree */
		if (node_ptr->left_ptr == NULL)
			child_ptr = node_ptr->right_ptr;
		else if (node_ptr->right_ptr == NULL)
			child_ptr = node_ptr->left_ptr;
		else
		{
			/* The node becomes the least of the right tree, which is guaranteed greater than the left tree */
			child_ptr = remove_first_child (tree_ptr, node_ptr, node_ptr->right_ptr);
			child_ptr->left_ptr = node_ptr->left_ptr;
			child_ptr->right_ptr = node_ptr->right_ptr;
		}
		
		if (parent_ptr == NULL)
		{
			/* We're the root node */
			tree_ptr->root.right_ptr = child_ptr;
		}
		else if (left == 1)
		{
			parent_ptr->left_ptr = child_ptr;
		}
		else
		{
			parent_ptr->right_ptr = child_ptr;
		}

		data_ptr = node_ptr->data_ptr;
		free (node_ptr->key_ptr);
		free (node_ptr);
		return data_ptr;
	}
	else if (compare < 0)
		return remove_child (tree_ptr, node_ptr, node_ptr->left_ptr, key_ptr, 1);
	else
		return remove_child (tree_ptr, node_ptr, node_ptr->right_ptr, key_ptr, 0);
}

static tree_record_t *
remove_first_child (tree_t *tree_ptr, tree_record_t *parent_ptr, tree_record_t *node_ptr)
{
	tree_record_t *found_ptr;
	
	/* Not found */
	if (node_ptr == NULL)
		return NULL;
	
	/* If a left sub-tree exists, then it lives there */
	found_ptr = remove_first_child (tree_ptr, node_ptr, node_ptr->left_ptr);
	if (found_ptr != NULL)
		return found_ptr;
		
	/* It must be us */
	if (parent_ptr == NULL)
		{
		if (node_ptr->right_ptr != NULL)
			tree_ptr->root.right_ptr = node_ptr->right_ptr;
		else
			tree_ptr->root.right_ptr = NULL;
		return node_ptr;
		}

	/* We're the leftmost node in this tree */
	if (tree_ptr->compare (parent_ptr->key_ptr, node_ptr->key_ptr) > 0)
		parent_ptr->right_ptr = node_ptr->right_ptr;
	else
		parent_ptr->left_ptr = node_ptr->right_ptr;

	return node_ptr;
}

/* Handle tree implementation */
void
__libarm4_tree_init (tree_t *tree_ptr)
{
	static arm_int64_t key = 0;

	/* Initialize the root node */
	__libarm4_shm_init_tree (tree_ptr);
	tree_ptr->compare = int64_compare;
	tree_ptr->clone = int64_clone;
	initialize_node (tree_ptr, &(tree_ptr->root), &key, NULL);
}

void
__libarm4_tree_insert (tree_t *tree_ptr, arm_int64_t key, void *data_ptr)
{
	tree_record_t	*node_ptr;

	__libarm4_shm_lock_tree (tree_ptr);

	/* Create a new node */
	node_ptr = (tree_record_t *) malloc (sizeof (tree_record_t));
	initialize_node (tree_ptr, node_ptr, &key, data_ptr);

	/* By default, the tree is entirely in the right branch of the root */
	if (tree_ptr->root.right_ptr == NULL)
		tree_ptr->root.right_ptr = node_ptr;
	else
		insert_child (tree_ptr, tree_ptr->root.right_ptr, node_ptr, &key);

	__libarm4_shm_unlock_tree (tree_ptr);
}

void *
__libarm4_tree_find (tree_t *tree_ptr, arm_int64_t key)
{
	void *void_ptr;

	__libarm4_shm_lock_tree (tree_ptr);

	void_ptr = find_child (tree_ptr, tree_ptr->root.right_ptr, &key);

	__libarm4_shm_unlock_tree (tree_ptr);

	return void_ptr;
}

void *
__libarm4_tree_remove (tree_t *tree_ptr, arm_int64_t key)
{
	void *void_ptr;

	__libarm4_shm_lock_tree (tree_ptr);
	void_ptr = remove_child (tree_ptr, NULL, tree_ptr->root.right_ptr, &key, 0);
	__libarm4_shm_unlock_tree (tree_ptr);

	return void_ptr;
}


void *
__libarm4_tree_remove_next (tree_t *tree_ptr, arm_int64_t *key_ptr)
{
	tree_record_t *node_ptr;
	void *data_ptr;
	
	__libarm4_shm_lock_tree (tree_ptr);
	node_ptr = remove_first_child (tree_ptr, NULL, tree_ptr->root.right_ptr);
	if (node_ptr == NULL)
	{
		*key_ptr = 0;
		data_ptr = NULL;
	}
	else
	{
		data_ptr = node_ptr->data_ptr;
		if (key_ptr)
			*key_ptr = *((arm_int64_t *) node_ptr->key_ptr);

		free (node_ptr->key_ptr);
		free (node_ptr);
	}
	__libarm4_shm_unlock_tree (tree_ptr);
	
	return data_ptr;
}


/* ID tree implementation */
void
__libarm4_tree_id_init (tree_id_t *tree_ptr)
{
	static arm_id_t empty = { { {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} } };

	/* Initialize the root node */
	__libarm4_shm_init_tree (tree_ptr);
	tree_ptr->compare = id_compare;
	tree_ptr->clone = id_clone;
	initialize_node (tree_ptr, &(tree_ptr->root), &empty, NULL);
}

void
__libarm4_tree_id_insert (tree_id_t *tree_ptr, const arm_id_t *key_ptr, void *data_ptr)
{
	tree_record_t	*node_ptr;

	__libarm4_shm_lock_tree (tree_ptr);

	/* Create a new node */
	node_ptr = (tree_record_t *) malloc (sizeof (tree_record_t));
	initialize_node (tree_ptr, node_ptr, key_ptr, data_ptr);

	/* By default, the tree is entirely in the right branch of the root */
	if (tree_ptr->root.right_ptr == NULL)
		tree_ptr->root.right_ptr = node_ptr;
	else
		insert_child (tree_ptr, tree_ptr->root.right_ptr, node_ptr, key_ptr);

	__libarm4_shm_unlock_tree (tree_ptr);
}

void *
__libarm4_tree_id_find (tree_id_t *tree_ptr, const arm_id_t *key_ptr)
{
	void *void_ptr;

	__libarm4_shm_lock_tree (tree_ptr);

	void_ptr = find_child (tree_ptr, tree_ptr->root.right_ptr, key_ptr);

	__libarm4_shm_unlock_tree (tree_ptr);

	return void_ptr;
}

void *
__libarm4_tree_id_remove (tree_id_t *tree_ptr, const arm_id_t *key_ptr)
{
	void *void_ptr;

	__libarm4_shm_lock_tree (tree_ptr);
	void_ptr = remove_child (tree_ptr, NULL, tree_ptr->root.right_ptr, key_ptr, 0);
	__libarm4_shm_unlock_tree (tree_ptr);

	return void_ptr;
}


void *
__libarm4_tree_id_remove_next (tree_id_t *tree_ptr, arm_id_t *key_ptr)
{
	tree_record_t *node_ptr;
	void *data_ptr;
	
	__libarm4_shm_lock_tree (tree_ptr);
	node_ptr = remove_first_child (tree_ptr, NULL, tree_ptr->root.right_ptr);
	if (node_ptr == NULL)
	{
		set_empty_id (key_ptr);
		data_ptr = NULL;
	}
	else
	{
		data_ptr = node_ptr->data_ptr;
		if (key_ptr)
			*key_ptr = *((arm_id_t *) node_ptr->key_ptr);

		free (node_ptr->key_ptr);
		free (node_ptr);
	}
	__libarm4_shm_unlock_tree (tree_ptr);
	
	return data_ptr;
}

