/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>
#include <assert.h>
#include <syslog.h>
#include <libgen.h>

#include <iomanip>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Arm4dbDaemonSharedMemory.h"
#include "Arm4dbExceptions.h"

#if defined(HAVE_BDB)
#include "BerkeleyDB.h"
#endif
#if defined(HAVE_SQLITE)
#include "Sqlite.h"
#endif
#if defined(HAVE_MYSQL)
#include "MySql.h"
#endif

Arm4db *Arm4db::m_singletonPtr = NULL;		// The current database instance
Arm4dbBackEnd *Arm4db::m_backendPtr = NULL; // The actual database backend

Arm4db::Arm4db ()
{
	setBackEnd ();
}

Arm4db::~Arm4db ()
{
	m_singletonPtr = NULL;
}

void
Arm4db::setBackEnd (void)
{
	switch (getConfig().getDbType ())
	{
#ifdef HAVE_BDB
	case Arm4dbBerkeleyDB:
		m_backendPtr = new BerkeleyDB ();
		break;
#endif

#ifdef HAVE_SQLITE
	case Arm4dbSqlite:
		m_backendPtr = new Sqlite ();
		break;
#endif

#ifdef HAVE_MYSQL
	case Arm4dbMySql:
		m_backendPtr = new MySql ();
		break;
#endif

	default:
		Logger::getLogger().error ("Unrecognized database type");
		throw new Arm4dbException ("Unrecognized database type");
	}
}

Arm4db *
Arm4db::getArm4db ()
{
	if (m_singletonPtr == NULL)
		m_singletonPtr = new Arm4db ();

	m_singletonPtr->initializeLibrary ();
	return m_singletonPtr;
}

arm_error_t
Arm4db::closeDatabases (void)
{
    /* Ensure we have database access */
    Arm4dbLock lock;

    closeDatabasesNoLock ();

    lock.unlock ();
	return ARM_SUCCESS;	
}

arm_error_t
Arm4db::removeDatabases (void)
{
	DIR *dirp;
	struct dirent *dp;
    struct stat stat_buffer;
	int status;
	char path [2048];

    /* Ensure we have database access */
	Arm4dbWriteLock lock;

	/* Make sure all databases are closed */
	closeDatabasesNoLock ();

	/* Remove all files in the database directory */
	if ((dirp = opendir(getConfig ().getDbHomeC ())) == NULL)
	{
		Logger::getLogger().error ("Can't open %s\n", getConfig ().getDbHomeC ());
        lock.unlock ();
		return ARM_FAILURE;
	}

	for (dp = readdir(dirp);
		  dp != NULL;
		  dp = readdir(dirp))
	{
        snprintf (path, sizeof (path), "%s/%s", getConfig ().getDbHomeC (), dp->d_name);

        /* If this is a directory, then skip it */
        if (stat (path,  &stat_buffer) == 0)
        {
            if (!(S_ISDIR(stat_buffer.st_mode)))
            {
        		if (!(strcmp (dp->d_name, "DB_CONFIG") == 0)) // List of files to keep around
        		{
        			status = unlink (path);
        			if (status != 0)
        			{
        				Logger::getLogger().error ("Unable to remove %s", path);
						closedir(dirp);
                        lock.unlock ();
        				return ARM_FAILURE;
        			}
        		}
            }
        }
	}

	closedir(dirp);

    lock.unlock ();
	return ARM_SUCCESS;	
}

std::ostream&
Arm4db::exportDefinitions (std::ostream &os) const
{
	os << "<?xml version=\"1.0\"?>\n";

	std::string schema = getConfig ().getSchemaLocation () + "/arm4db.xsd";
	os << "<arm4 xmlns=\"http://www.arm4.org/export\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://www.arm4.org/export "
		 << schema << "\">\n";

	os << "<definitions>\n";
	
	exportApplicationDefinitions (os);
	exportMetricDefinitions (os);
	exportTransactionDefinitions (os);

    os << "</definitions>\n";
	os << "</arm4>\n";
    
	return os;
}

std::ostream&
Arm4db::exportAll (std::ostream &os) const
{
	os << "<?xml version=\"1.0\"?>\n";

	std::string schema = getConfig ().getSchemaLocation () + "/arm4db.xsd";
	os << "<arm4 xmlns=\"http://www.arm4.org/export\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://www.arm4.org/export "
		 << schema << "\">\n";

	os << "<definitions>\n";
	exportApplicationDefinitions (os);
	exportMetricDefinitions (os);
	exportTransactionDefinitions (os);
    os << "</definitions>\n";

    os << "<instances>\n";
    exportApplications (os);
    exportTransactions (os);
    exportTraceStats (os);
	os << "</instances>\n";

	os << "</arm4>\n";
    
	return os;
}

