/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <string.h>
#include <grp.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>

#ifdef HAVE_UUID_UUID_H
#include <uuid/uuid.h>
#else
#ifdef HAVE_UUID_H
#include <uuid.h>
#endif
#endif

#include "arm4.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Arm4dbConfig.h"
#include "Arm4dbCommon.h"
#include "Arm4dbDaemonSharedMemory.h"
#include "Arm4dbExceptions.h"

#define SHM_MODE	0666	/* Shared memory file permissions */

Arm4dbDaemonSharedMemory::Arm4dbDaemonSharedMemory () :
	Arm4dbSharedMemory ()
{
}

Arm4dbDaemonSharedMemory::~Arm4dbDaemonSharedMemory ()
{
}

void *
Arm4dbDaemonSharedMemory::shmCreate (arm_boolean_t start_as_null)
{
	void *ptr;
	key_t token;
	int fd;
	int shmid;
	const char *filename_ptr;

	/* Create the message queue */
	int instance = Arm4dbConfig::getConfig ().getInstance ();
	filename_ptr = Arm4dbCommon::tempPath(instance, MQUEUE_FILE);
	fd = creat (filename_ptr, SHM_MODE & (~g_umask));
	if (fd == -1)
	{
		Logger::getLogger().error ("creat %s", filename_ptr);

		throw new Arm4dbSharedMemoryException ("Unable to create message queue file");
	}
	close (fd);
	token = ftok (filename_ptr, 1);
    if (token == -1)
    {
        Logger::getLogger().error ("ftok");

		throw new Arm4dbSharedMemoryException ("Unable to get message queue token");
    }
	g_message_queue = msgget (token, IPC_CREAT | (SHM_MODE & (~g_umask)));
	if (g_message_queue == -1)
	{
		Logger::getLogger().error ("msgget");

		throw new Arm4dbSharedMemoryException ("Unable to create message queue");
	}

	/* Create or attach to a shared memory segment */
	filename_ptr = Arm4dbCommon::tempPath(instance, SHARED_MEMORY_FILE);
	fd = creat (filename_ptr, SHM_MODE & (~g_umask));
	if (fd == -1)
	{
		Logger::getLogger().error ("creat %s", filename_ptr);

		throw new Arm4dbSharedMemoryException ("Unable to create shared memory file");
	}
	close (fd);
	token = ftok (filename_ptr, 1);
    if (token == -1)
    {
        Logger::getLogger().error ("ftok");

		throw new Arm4dbSharedMemoryException ("Unable to get shared memory token");
    }
	shmid = shmget (token, sizeof (arm4_shm_header_t), IPC_CREAT | (SHM_MODE & (~g_umask)));
	if (shmid < 0)
	{
		Logger::getLogger().error ("shmget 2");

		throw new Arm4dbSharedMemoryException ("Unable to create shared memory");
	}
	ptr = shmat (shmid, NULL, 0);
	if (ptr == NULL)
	{
		Logger::getLogger().error ("shmat");

		throw new Arm4dbSharedMemoryException ("Unable to attach to shared memory");
	}
	g_shared_memory_ptr = reinterpret_cast<arm4_shm_header_t *> (ptr);

    /* Check to see if we're already initialized */
    if (g_shared_memory_ptr->magic == ARM4_SHM_MAGIC)
    {
        /* Indicate wether we're acting as a null library */
        g_shared_memory_ptr->null_arm = start_as_null;

        /* Set our instance PID */
        g_shared_memory_ptr->daemon_instance = getpid ();

        return g_shared_memory_ptr;
    }

	/* Initialize the semaphores */
	create_semaphores ();

	/* Initialize the data fields */

	/* Set up the sequence numbers */
	Arm4db::getArm4db ()->initializeSequences (g_shared_memory_ptr);

	/* cached application entries */
	g_shared_memory_ptr->max_applications = ARM4_SHM_APPLICATION_ENTRIES; /* This should be a configurable parameter */
	g_shared_memory_ptr->application_count = 0;
	g_shared_memory_ptr->next_application = 0;
	memset (g_shared_memory_ptr->application_table, 0, sizeof(g_shared_memory_ptr->application_table));

	/* cached transaction entries */
	g_shared_memory_ptr->max_transactions = ARM4_SHM_TRANSACTION_ENTRIES; /* This should be a configurable parameter */
	g_shared_memory_ptr->transaction_count = 0;
	g_shared_memory_ptr->next_transaction = 0;
	memset (g_shared_memory_ptr->transaction_table, 0, sizeof(g_shared_memory_ptr->transaction_table));

	/* cached metric entries */
	g_shared_memory_ptr->max_metrics = ARM4_SHM_METRIC_ENTRIES; /* This should be a configurable parameter */
	g_shared_memory_ptr->metric_count = 0;
	g_shared_memory_ptr->next_metric = 0;
	memset (g_shared_memory_ptr->metric_table, 0, sizeof(g_shared_memory_ptr->metric_table));

	/* Set our hostname */
	std::string hostname = Arm4dbConfig::getConfig ().getHostname ();
	strncpy (g_shared_memory_ptr->hostname, hostname.c_str (), sizeof (g_shared_memory_ptr->hostname));

    /* By default, there's no tracing */
    g_shared_memory_ptr->trace_level = arm4_shm_trace_none;

    /* Indicate wether we're acting as a null library */
    g_shared_memory_ptr->null_arm = start_as_null;

    /* Set our instance PID */
	g_shared_memory_ptr->daemon_instance = getpid ();

	/* Indicate that we're open for business */
	g_shared_memory_ptr->magic = ARM4_SHM_MAGIC;

	return g_shared_memory_ptr;
}

void
Arm4dbDaemonSharedMemory::shmDestroy (void)
{
    int status;
	key_t token;
    int id;
	struct shmid_ds shm_description;
    struct msqid_ds msg_description;
	const char *filename_ptr;

	/* Indicate that we're no longer open for business */
	if (g_shared_memory_ptr)
		g_shared_memory_ptr->magic = 0;

	destroy_semaphores ();

    /* Mark the message queue for removal */
	int instance = Arm4dbConfig::getConfig ().getInstance ();
	filename_ptr = Arm4dbCommon::tempPath(instance, MQUEUE_FILE);
    token = ftok (filename_ptr, 1);
    if (token != -1)
    {
        id = msgget (token, SHM_MODE & (~g_umask));
        if (id != -1)
        {
            status = msgctl (id, IPC_RMID, &msg_description);
            if (status != 0)
            {
                Logger::getLogger().error ("msgctl");
            }
            unlink (filename_ptr);
        }
    }

	/* Mark the region for removal */
	filename_ptr = Arm4dbCommon::tempPath(instance, SHARED_MEMORY_FILE);
	token = ftok (filename_ptr, 1);
    if (token != -1)
    {
        id = shmget (token, sizeof (arm4_shm_header_t), SHM_MODE & (~g_umask));
        if (id != -1)
        {
        	status = shmctl (id, IPC_RMID, &shm_description);
            if (status != 0)
            {
                Logger::getLogger().error ("shmctl");
            }
        	unlink (filename_ptr);
        }
    }

	g_shared_memory_ptr = NULL;
}

void
Arm4dbDaemonSharedMemory::setUmask (int umask)
{
    g_umask = umask;
}

pid_t
Arm4dbDaemonSharedMemory::setDaemon (arm_boolean_t force)
{
    /* Ensure we're attached */
    if (shmAttach () == NULL)
        return (pid_t) 0;

	if ((g_shared_memory_ptr->daemon_instance != (pid_t) 0) && (!force))
		return g_shared_memory_ptr->daemon_instance;

	/* Set our process ID */
	g_shared_memory_ptr->daemon_instance = getpid ();

	return g_shared_memory_ptr->daemon_instance;
}

pid_t
Arm4dbDaemonSharedMemory::getActiveDaemon (void)
{
	int status;

	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return (pid_t) 0;

	if (g_shared_memory_ptr->daemon_instance > (pid_t) 0)
	{
		/* See if the instance is still alive */
		status = kill (g_shared_memory_ptr->daemon_instance, 0);
		if (status != 0)
		{
			if (errno == ESRCH)
			{
				/* The daemon is dead - long live the daemon! */
				g_shared_memory_ptr->daemon_instance = 0;
			}
		}
	}
	return g_shared_memory_ptr->daemon_instance;
}

pid_t
Arm4dbDaemonSharedMemory::clearDaemon (void)
{
	int status;

	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return (pid_t) 0;

	if (g_shared_memory_ptr->daemon_instance > (pid_t) 0)
	{
		/* See if the instance is still alive */
		status = kill (g_shared_memory_ptr->daemon_instance, 0);
		if (status != 0)
		{
			if (errno == ESRCH)
			{
				/* The daemon is dead - long live the daemon! */
				g_shared_memory_ptr->daemon_instance = 0;
			}
		}

		// If we're still active, see if it's us!
		if (g_shared_memory_ptr->daemon_instance == getpid())
		{
			// We'll only clear the daemon setting if we own it!
			g_shared_memory_ptr->daemon_instance = 0;
		}
	}
	return g_shared_memory_ptr->daemon_instance;
}

/*
 * Get data from the message queues
 */

int
Arm4dbDaemonSharedMemory::getAppInstances (arm4_shm_application_instance_t *app_instance_ptr)
{
	return getMessage (app_instance_ptr, sizeof (arm4_shm_application_instance_t), ARM4_SHM_MESSAGE_APPLICATION);
}

int
Arm4dbDaemonSharedMemory::getAppGroup (arm4_shm_property_t *app_group_ptr)
{
	return getMessage (app_group_ptr, sizeof (arm4_shm_property_t), ARM4_SHM_MESSAGE_APPLICATION_GROUP);
}

int
Arm4dbDaemonSharedMemory::getAppInstanceId (arm4_shm_property_t *property_ptr)
{
	return getMessage (property_ptr, sizeof (arm4_shm_property_t), ARM4_SHM_MESSAGE_APPLICATION_INSTANCE_ID);
}

int
Arm4dbDaemonSharedMemory::getAppContext (arm4_shm_context_property_t *property_ptr)
{
	return getMessage (property_ptr, sizeof (arm4_shm_context_property_t), ARM4_SHM_MESSAGE_APPLICATION_CONTEXT);
}

int
Arm4dbDaemonSharedMemory::getAppAddress (arm4_shm_address_t *address_ptr)
{
	return getMessage (address_ptr, sizeof (arm4_shm_address_t), ARM4_SHM_MESSAGE_APPLICATION_ADDRESS);
}

int
Arm4dbDaemonSharedMemory::getTranInstances (arm4_shm_transaction_instance_t *tran_instance_ptr)
{
	return getMessage (tran_instance_ptr, sizeof (arm4_shm_transaction_instance_t), ARM4_SHM_MESSAGE_TRANSACTION);
}

int
Arm4dbDaemonSharedMemory::getTranContext (arm4_shm_context_property_t *property_ptr)
{
	return getMessage (property_ptr, sizeof (arm4_shm_context_property_t), ARM4_SHM_MESSAGE_TRANSACTION_CONTEXT);
}

int
Arm4dbDaemonSharedMemory::getTranUri (arm4_shm_uri_t *uri_ptr)
{
	return getMessage (uri_ptr, sizeof (arm4_shm_uri_t), ARM4_SHM_MESSAGE_TRANSACTION_URI);
}

int
Arm4dbDaemonSharedMemory::getTranMetricValues (arm4_shm_metric_value_t *metric_ptr)
{
	return getMessage (metric_ptr, sizeof (arm4_shm_metric_value_t), ARM4_SHM_MESSAGE_TRANSACTION_METRIC_VALUE);
}

int
Arm4dbDaemonSharedMemory::getTranUser (arm4_shm_user_t *user_ptr)
{
	return getMessage (user_ptr, sizeof (arm4_shm_user_t), ARM4_SHM_MESSAGE_TRANSACTION_USER);
}

int
Arm4dbDaemonSharedMemory::getTranDiag (arm4_shm_diag_t *diag_ptr)
{
	return getMessage (diag_ptr, sizeof (arm4_shm_diag_t), ARM4_SHM_MESSAGE_TRANSACTION_DIAG);
}

int
Arm4dbDaemonSharedMemory::getCorellators (arm4_shm_correlator_t *correlator_ptr)
{
	return getMessage (correlator_ptr, sizeof (arm4_shm_correlator_t), ARM4_SHM_MESSAGE_CORRELATOR);
}

int
Arm4dbDaemonSharedMemory::getSequence (arm4_shm_db_sequence_t *sequence_ptr)
{
	return getMessage (sequence_ptr, sizeof (arm4_shm_db_sequence_t), ARM4_SHM_MESSAGE_SEQUENCE);
}

int
Arm4dbDaemonSharedMemory::getRegisterApplication (arm4_shm_register_application_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_APPLICATION);
}

int
Arm4dbDaemonSharedMemory::getRegisterApplicationIdentity (arm4_shm_register_application_identity_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_APPLICATION_IDENTITY);
}

int
Arm4dbDaemonSharedMemory::getRegisterApplicationContext (arm4_shm_register_application_context_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_APPLICATION_CONTEXT);
}

/* Transaction registration */
int
Arm4dbDaemonSharedMemory::getRegisterTransaction (arm4_shm_register_transaction_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_TRANSACTION);
}

int
Arm4dbDaemonSharedMemory::getRegisterTransactionIdentity (arm4_shm_register_application_identity_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_IDENTITY);
}

int
Arm4dbDaemonSharedMemory::getRegisterTransactionContext (arm4_shm_register_application_context_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_CONTEXT);
}

int
Arm4dbDaemonSharedMemory::getRegisterTransactionUri (arm4_shm_register_transaction_uri_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_URI);
}

int
Arm4dbDaemonSharedMemory::getRegisterTransactionMetricBinding (arm4_shm_register_transaction_metric_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_TRANSACTION_METRIC);
}

/* Metric registration */
int
Arm4dbDaemonSharedMemory::getRegisterMetric (arm4_shm_register_metric_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_REGISTER_METRIC);
}

/* Utility commands */
void
Arm4dbDaemonSharedMemory::utilityCommand (long command)
{
	arm4_shm_utility_t	message;
	int status;

	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	/* Send a reply to the thread that made the request */
	message.mtype = ARM4_SHM_MESSAGE_UTILITY;
	message.message.command = command;

	status = SAFE_SEND (g_message_queue, &message, sizeof (message), 0);
	if (status == -1)
	{
		Logger::getLogger().error ("utilityCommand");
	}
}

int
Arm4dbDaemonSharedMemory::getUtilityCommand (arm4_shm_utility_t *message_ptr)
{
	return getMessage (message_ptr, sizeof (*message_ptr), ARM4_SHM_MESSAGE_UTILITY);
}

void
Arm4dbDaemonSharedMemory::registerApplicationReturn (int application_index, arm_id_t appid, arm_boolean_t is_new)
{
	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	assert ((application_index >= 0) && (application_index < ARM4_SHM_APPLICATION_ENTRIES));
	assert (g_shared_memory_ptr);

	arm4_shm_application_t *app_ptr = &(g_shared_memory_ptr->application_table [application_index]);
	app_ptr->appl_id = appid;
	app_ptr->is_new = is_new;

	// Signal the waiting process
	struct sembuf sbuf;
	sbuf.sem_num = SEM_INSTANCE_APPLICATION;
	sbuf.sem_op = 1;
	sbuf.sem_flg = 0;
	if (semop (g_shared_memory_ptr->application_semaphore, & sbuf, 1) != 0)
	{
		Logger::getLogger().error ("Unable to signal application semaphore");
	}
}

void
Arm4dbDaemonSharedMemory::registerTransactionReturn (int transaction_index, arm_id_t tran_id, arm_boolean_t is_new)
{
	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	assert ((transaction_index >= 0) && (transaction_index < ARM4_SHM_TRANSACTION_ENTRIES));
	assert (g_shared_memory_ptr);

	arm4_shm_transaction_t *tran_ptr = &(g_shared_memory_ptr->transaction_table [transaction_index]);
	tran_ptr->tran_id = tran_id;
	tran_ptr->is_new = is_new;

	// Signal the waiting process
	struct sembuf sbuf;
	sbuf.sem_num = SEM_INSTANCE_TRANSACTION;
	sbuf.sem_op = 1;
	sbuf.sem_flg = 0;
	if (semop (g_shared_memory_ptr->application_semaphore, & sbuf, 1) != 0)
	{
		Logger::getLogger().error ("Unable to signal transaction semaphore");
	}
}

void
Arm4dbDaemonSharedMemory::registerMetricReturn (int metric_index, arm_id_t metric_id)
{
	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	assert ((metric_index >= 0) && (metric_index < ARM4_SHM_METRIC_ENTRIES));
	assert (g_shared_memory_ptr);

	arm4_shm_metric_t *metric_ptr = &(g_shared_memory_ptr->metric_table [metric_index]);
	metric_ptr->metric_id = metric_id;

	// Signal the waiting process
	struct sembuf sbuf;
	sbuf.sem_num = SEM_INSTANCE_METRIC;
	sbuf.sem_op = 1;
	sbuf.sem_flg = 0;
	if (semop (g_shared_memory_ptr->application_semaphore, & sbuf, 1) != 0)
	{
		Logger::getLogger().error ("Unable to signal metric semaphore");
	}
}

void
Arm4dbDaemonSharedMemory::getStats (arm4_shm_stats_t *stats_ptr)
{
	int status;
    struct msqid_ds msg_description;

	memset (stats_ptr, sizeof (*stats_ptr), 0);

	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

    /* Retrieve the message queue stats */
	status = msgctl (g_message_queue, IPC_STAT, &msg_description);
	if (status != 0)
	{
		Logger::getLogger().error ("msgctl");
	}
	else
	{
		stats_ptr->mq_n_messages = msg_description.msg_qnum;
		stats_ptr->mq_n_bytes = msg_description.msg_cbytes;
		stats_ptr->mq_max_bytes = msg_description.msg_qbytes;
	}
}

int
Arm4dbDaemonSharedMemory::waitingMessageCount (void)
{
	// Get the number of messages waiting on the queue
	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return -1;

	struct msqid_ds qinfo;
	int status = msgctl (g_message_queue, IPC_STAT, &qinfo);
	if (status == 0)
		return qinfo.msg_qnum;
	return -1;
}


int
Arm4dbDaemonSharedMemory::getMessage (void *buffer_ptr, int buffer_size, long type)
{
	int	count;

	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return -1;

	count = msgrcv (g_message_queue, buffer_ptr, buffer_size, type, 0);
	if (count < 0)
	{
		Logger::getLogger().error ("msgrcv");
	}

	return count;
}

void
Arm4dbDaemonSharedMemory::sampleFlush (void)
{
	/* This routine gets called very 0.1 seconds */

	char digest [MD5_DIGEST_LENGTH];
	int transaction_count;
	arm_time_t sample_time;
	static int counter = 0;

	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	/* Zero our digest */
	memset (digest, 0, MD5_DIGEST_LENGTH);

	lock_transaction ();

	/* Record the time at which our sample was taken */
	gettimeofday (&sample_time, NULL);

	/* Flush the thread locking stats */
#ifdef HAVE_PTHREAD_STATS
	int total;
	int wait;
	int max_active;
#endif

	counter++;
	transaction_count = g_shared_memory_ptr->transaction_count;
	if (transaction_count == 0)
	{
		// Nothing to do
		unlock_transaction ();
		return;
	}

	/* Allocate memory for storing stats */
	arm4db_transaction_trace_stats_t *stats_ptr =
		(arm4db_transaction_trace_stats_t *) calloc (transaction_count, sizeof (arm4db_transaction_trace_stats_t));

	/* Step through the transaction table */
	bool save_stats = false;
	for (int i = 0; i < transaction_count; i++)
	{
		if (memcmp (&(g_shared_memory_ptr->transaction_table [i].md5_digest), digest, MD5_DIGEST_LENGTH) != 0)
		{
			/* Flush the trace stats once every 30 seconds */
			if ((counter % 300) == 0)
			{
				stats_ptr [i].tran_id = g_shared_memory_ptr->transaction_table [i].tran_id;
				stats_ptr [i].transaction_counter = g_shared_memory_ptr->transaction_table [i].transaction_counter;

				g_shared_memory_ptr->transaction_table [i].transaction_counter = 0;
				save_stats = true;
			}

			/* If we're time sampled, set our time sample flag as required */
			if (((int) g_shared_memory_ptr->transaction_table [i].trace_flags) & arm4_shm_transaction_trace_interval)
			{
				g_shared_memory_ptr->transaction_table [i].trace_counter--;
				if (g_shared_memory_ptr->transaction_table [i].trace_counter == 0)
				{
					/* Flag the transaction for sampling */
					g_shared_memory_ptr->transaction_table [i].trace_flags = (arm4_shm_transaction_trace_t)
						(((int) g_shared_memory_ptr->transaction_table [i].trace_flags) & (int) arm4_shm_transaction_trace_next);

					/* Reset the counter */
					g_shared_memory_ptr->transaction_table [i].trace_counter = g_shared_memory_ptr->transaction_table [i].trace_parameter;
				}
			}
		}
	}

	/* Flush the thread locking stats */
#ifdef HAVE_PTHREAD_STATS
	if ((counter % 600) == 0)
	{
		/* Flush the stats every minute */
		total = g_shared_memory_ptr->transaction_mutex_pthread_stat_total;
		wait = g_shared_memory_ptr->transaction_mutex_pthread_stat_wait;
		max_active = g_shared_memory_ptr->transaction_mutex_pthread_stat_max_active;

		g_shared_memory_ptr->transaction_mutex_pthread_stat_total = 0;
		g_shared_memory_ptr->transaction_mutex_pthread_stat_wait = 0;
		/* Current active stat remains unmodified */
		g_shared_memory_ptr->transaction_mutex_pthread_stat_max_active = 0;
	}
#endif

	// Not found
	unlock_transaction ();

	/* This can be done after the lock is released */
	if (save_stats)
		Arm4db::getArm4db()->processTraceStats (sample_time, stats_ptr, transaction_count);

	// Free our allocated memory
	free (stats_ptr);

#ifdef HAVE_PTHREAD_STATS
	if ((counter % 600) == 0)
	{
		/* Flush the stats every minute */
		printf ("Total transaction locks %d\n", total);
		printf ("Locks requiring a wait %d\n", wait);
		printf ("Active threads %d\n\n", max_active);
	}
#endif
}

void
Arm4dbDaemonSharedMemory::create_semaphores (void)
{
    /* Create the DB access R/W lock */
    pthread_rwlock_init (&g_db_rw_lock, NULL);

#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutexattr_t mattr;

	/* Sequence semaphore */
	pthread_mutexattr_init (&mattr);
	pthread_mutexattr_setpshared (&mattr, PTHREAD_PROCESS_SHARED);
	pthread_mutex_init (&(g_shared_memory_ptr->sequence_mutex), &mattr);
	pthread_mutexattr_destroy (&mattr);

	/* Application semaphore */
	pthread_mutexattr_init (&mattr);
	pthread_mutexattr_setpshared (&mattr, PTHREAD_PROCESS_SHARED);
	pthread_mutex_init (&(g_shared_memory_ptr->application_mutex), &mattr);
	pthread_mutexattr_destroy (&mattr);

	/* Transaction semaphore */
	pthread_mutexattr_init (&mattr);
	pthread_mutexattr_setpshared (&mattr, PTHREAD_PROCESS_SHARED);
	pthread_mutex_init (&(g_shared_memory_ptr->transaction_mutex), &mattr);
	pthread_mutexattr_destroy (&mattr);

    /* Metric semaphore */
    pthread_mutexattr_init (&mattr);
    pthread_mutexattr_setpshared (&mattr, PTHREAD_PROCESS_SHARED);
    pthread_mutex_init (&(g_shared_memory_ptr->metric_mutex), &mattr);
    pthread_mutexattr_destroy (&mattr);

    /* Utility semaphore */
    pthread_mutexattr_init (&mattr);
    pthread_mutexattr_setpshared (&mattr, PTHREAD_PROCESS_SHARED);
    pthread_mutex_init (&(g_shared_memory_ptr->utility_read_mutex), &mattr);
    pthread_mutexattr_destroy (&mattr);

#else /* USE_POSIX_PROCESS_SEM */
	if (g_sequence_sem_ptr == NULL)
		g_sequence_sem_ptr = sem_open (instanceName (SEM_SEQUENCE_FILE), O_CREAT, SHM_MODE & (~g_umask), 1);
	if (g_sequence_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_seqence_sem_ptr sem_open");
		g_sequence_sem_ptr = NULL;

		throw new Arm4dbSharedMemoryException ("Unable to create semaphore");
	}

	if (g_application_sem_ptr == NULL)
		g_application_sem_ptr = sem_open (instanceName (SEM_APPLICATION_FILE), O_CREAT, SHM_MODE & (~g_umask), 1);
	if (g_application_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_application_sem_ptr sem_open");
		g_application_sem_ptr = NULL;

		throw new Arm4dbSharedMemoryException ("Unable to create semaphore");
	}

	if (g_transaction_sem_ptr == NULL)
		g_transaction_sem_ptr = sem_open (instanceName (SEM_TRANSACTION_FILE), O_CREAT, SHM_MODE & (~g_umask), 1);
	if (g_transaction_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_transaction_sem_ptr sem_open");
		g_transaction_sem_ptr = NULL;

		throw new Arm4dbSharedMemoryException ("Unable to create semaphore");
	}

    if (g_metric_sem_ptr == NULL)
        g_metric_sem_ptr = sem_open (instanceName (SEM_METRIC_FILE), O_CREAT, SHM_MODE & (~g_umask), 1);
    if (g_metric_sem_ptr == (sem_t *) SEM_FAILED)
    {
        Logger::getLogger().error ("g_metric_sem_ptr sem_open");
        g_metric_sem_ptr = NULL;

		throw new Arm4dbSharedMemoryException ("Unable to create semaphore");
    }

    if (g_utility_sem_ptr == NULL)
        g_utility_sem_ptr = sem_open (instanceName (SEM_UTILITY_FILE), O_CREAT, SHM_MODE & (~g_umask), 1);
    if (g_utility_sem_ptr == (sem_t *) SEM_FAILED)
    {
        Logger::getLogger().error ("g_utility_sem_ptr sem_open");
        g_utility_sem_ptr = NULL;

		throw new Arm4dbSharedMemoryException ("Unable to create semaphore");
    }

#endif /* USE_POSIX_PROCESS_SEM */

    assert (g_shared_memory_ptr != NULL);

	int instance = Arm4dbConfig::getConfig ().getInstance ();

	// Create the semaphore file if it doesn't exist
	const char *instance_name = Arm4dbCommon::tempPath(instance, SEM_INSTANCE_FILE);
	if (access (instance_name, F_OK) != 0)
	{
		/* Create the directory path */
		int file = creat (instance_name, 0777); // As modified by umask
		if (file == -1)
		{
			Logger::getLogger().error ("Unable to create semaphore file");
		}
		else
		{
			close (file);
		}
	}

	key_t token = ftok (instance_name, 1);
	if (token == (key_t)-1)
	{
		Logger::getLogger().error ("Unable to create application token");
	}
	else
	{
		// Our umask assures the wrong people won't get access
		g_shared_memory_ptr->application_semaphore = semget(token, SEM_INSTANCE_COUNT,
															IPC_CREAT |
															S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
		if (g_shared_memory_ptr->application_semaphore == -1)
		{
			Logger::getLogger().error ("Unable to create application semaphore (token=%lx)", token);
		}
		else
		{
			// Initialize the semaphore value
			struct sembuf sbuf;
			for (int i = 0; i < SEM_INSTANCE_COUNT; i++)
			{
				sbuf.sem_num = i;
				sbuf.sem_op = 0;    // Initial value
				sbuf.sem_flg = 0;
				if (semop (g_shared_memory_ptr->application_semaphore, &sbuf, 1) == -1)
				{
					Logger::getLogger().error ("Unable to initialize semaphore");
				}
			}
		}
	}
}

void
Arm4dbDaemonSharedMemory::destroy_semaphores (void)
{
#ifndef USE_POSIX_PROCESS_SEM
	int status;

	if (g_sequence_sem_ptr != NULL)
	{
		status = sem_close (g_sequence_sem_ptr);
		if (status != 0)
			Logger::getLogger().error (instanceName (SEM_SEQUENCE_FILE));
		g_sequence_sem_ptr = NULL;
	}
	sem_unlink (instanceName (SEM_SEQUENCE_FILE));

	if (g_application_sem_ptr != NULL)
	{
		status = sem_close (g_application_sem_ptr);
		if (status != 0)
			Logger::getLogger().error (instanceName (SEM_APPLICATION_FILE));
		g_application_sem_ptr = NULL;
	}
	sem_unlink (instanceName (SEM_APPLICATION_FILE));

	if (g_transaction_sem_ptr != NULL)
	{
		status = sem_close (g_transaction_sem_ptr);
		if (status != 0)
			Logger::getLogger().error (instanceName (SEM_TRANSACTION_FILE));
		g_transaction_sem_ptr = NULL;
	}
	sem_unlink (instanceName (SEM_TRANSACTION_FILE));

	if (g_metric_sem_ptr != NULL)
	{
		status = sem_close (g_metric_sem_ptr);
		if (status != 0)
			Logger::getLogger().error (instanceName (SEM_METRIC_FILE));
		g_metric_sem_ptr = NULL;
	}
	sem_unlink (instanceName (SEM_METRIC_FILE));

	if (g_utility_sem_ptr != NULL)
	{
		status = sem_close (g_utility_sem_ptr);
		if (status != 0)
			Logger::getLogger().error (instanceName (SEM_UTILITY_FILE));
		g_utility_sem_ptr = NULL;
	}
	sem_unlink (instanceName (SEM_UTILITY_FILE));

#endif /* USE_POSIX_PROCESS_SEM */

    if (g_shared_memory_ptr != NULL)
    {
        if (semctl(g_shared_memory_ptr->application_semaphore, 0, IPC_RMID) != 0)
        {
            Logger::getLogger().error ("Unable to remove application sempaphore");
        }
        g_shared_memory_ptr->application_semaphore = -1;
    }
	int instance = Arm4dbConfig::getConfig ().getInstance ();
	const char *instance_name = Arm4dbCommon::tempPath(instance, SEM_INSTANCE_FILE);
	unlink (instance_name);
}


