/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 
#include <stdio.h>
#include <libxml/xmlreader.h>
#include <cstring>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "Arm4dbDaemonSharedMemory.h"
#include "Arm4dbExceptions.h"
#include "BerkeleyDB.h"

const xmlChar *
Arm4db::getTextNode(xmlTextReaderPtr reader)
{
	// Read a text node from the stream
    int ret;

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if ((type == XML_READER_TYPE_TEXT) || (type == XML_READER_TYPE_CDATA))
    	{
    		return xmlTextReaderConstValue(reader);
    	}
		else if (type == XML_READER_TYPE_END_ELEMENT)
		{
			return NULL;	// This is an empty field
		}
    	else if (
        		(type == XML_READER_TYPE_COMMENT) ||
        		(type == XML_READER_TYPE_WHITESPACE) ||
        		(type == XML_READER_TYPE_SIGNIFICANT_WHITESPACE))
    	{
    		// These types can be ignored
    	}
    	else
    	{
    		fprintf (stderr, "getTextNode: Unrecognized node type %d\n", type);
    		return NULL;
    	}

        ret = xmlTextReaderRead(reader);
    }

	return NULL;
}

arm_int32_t
Arm4db::getIntNode(xmlTextReaderPtr reader)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value == NULL)
	{
		fprintf (stderr, "getIntNode: No value given\n");
		return 0;
	}
	
	return atol ((char *) value);
}

arm_int64_t
Arm4db::getInt64Node(xmlTextReaderPtr reader)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value == NULL)
	{
		fprintf (stderr, "getInt64Node: No value given\n");
		return 0;
	}
	
	return atoll ((char *) value);
}

void
Arm4db::getTimeNode(xmlTextReaderPtr reader, arm_time_t &time_value)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value == NULL)
	{
		fprintf (stderr, "getTimeNode: No value given\n");
		memset (&time_value, 0, sizeof (time_value));
	}
	else
	{
		sscanf ((char *) value, "%ld.%ld", &(time_value.tv_sec), &(time_value.tv_usec));
	}
}

void
Arm4db::getIdNode(xmlTextReaderPtr reader, arm_id_t &id)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value == NULL)
	{
		fprintf (stderr, "getIdNode: No value given\n");
		memset (&id, 0, sizeof (arm_id_t));
	}
	else
	{
		int hex_value;
		for (int i = 0; i < 16; i++)
		{
			sscanf ((char *) &(value [i*2]), "%02x", &hex_value);
			id.id_u.uint8 [i] = hex_value & 0x0FF;
		}
	}
}

void
Arm4db::getDigestNode(xmlTextReaderPtr reader, char *digest_ptr)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value == NULL)
	{
		fprintf (stderr, "getDigestNode: No value given\n");
		memset (digest_ptr, 0, MD5_DIGEST_LENGTH);
	}
	else
	{
		int hex_value;
		for (int i = 0; i < MD5_DIGEST_LENGTH; i++)
		{
			sscanf ((char *) &(value [i*2]), "%02x", &hex_value);
			digest_ptr [i] = hex_value & 0x0FF;
		}
	}
}

void
Arm4db::getAddressNode(xmlTextReaderPtr reader, arm_uint8_t *address_ptr)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value != NULL)
	{
		int hex_value;
		int len = strlen ((char *) value);
		for (int i = 0; i < len; i++)
		{
			sscanf ((char *) &(value [i*2]), "%02x", &hex_value);
			address_ptr [i] = hex_value & 0x0FF;
		}
	}
}

void
Arm4db::getCorrelatorNode(xmlTextReaderPtr reader, arm_correlator_t *correlator_ptr)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value != NULL)
	{
		int hex_value;
		int len = strlen ((char *) value);
		if (len > 2 * ARM_CORR_MAX_LENGTH) // Two hex digits per byte
			len = 2 * ARM_CORR_MAX_LENGTH;
		
		for (int i = 0; i < len; i++)
		{
			sscanf ((char *) &(value [i*2]), "%02x", &hex_value);
			correlator_ptr->opaque [i] = hex_value & 0x0FF;
		}
	}
}

void
Arm4db::getMetricNode(xmlTextReaderPtr reader, metric_value_record_t &metric)
{
	// Get an arm_id_t value from the reader stream
	const xmlChar *value = getTextNode(reader);
	if (value == NULL)
	{
		fprintf (stderr, "getMetricNode: No value given\n");
	}
	else
	{
	    // Format statements need to be made more portable
		switch (metric.value.format)
			{
			case ARM_METRIC_FORMAT_COUNTER32:
				sscanf ((const char *) value, ARM4_FORMAT_INT32, &metric.value.metric_u.counter32);
				break;

			case ARM_METRIC_FORMAT_COUNTER64:
				sscanf ((const char *) value, ARM4_FORMAT_INT64, &metric.value.metric_u.counter64);
				break;

			case ARM_METRIC_FORMAT_CNTRDIVR32:
				sscanf ((const char *) value, ARM4_FORMAT_INT32 ":" ARM4_FORMAT_INT32,
						&metric.value.metric_u.counterdivisor32.counter,
						&metric.value.metric_u.counterdivisor32.divisor);
				break;

			/* Java implementation specific */
			case ARM_METRIC_FORMAT_COUNTERFLOAT32:
				// This is a kludge to put floats in this field, but our float_kludge.m4 macro
				// ensures it fits
				sscanf ((const char *) value, "%f", reinterpret_cast<float *> (&(metric.value.metric_u.counterdivisor32)));
				break;

			case ARM_METRIC_FORMAT_GAUGE32:
				sscanf ((const char *) value, ARM4_FORMAT_INT32, &metric.value.metric_u.gauge32);
				break;

			case ARM_METRIC_FORMAT_GAUGE64:
				sscanf ((const char *) value, ARM4_FORMAT_INT64, &metric.value.metric_u.gauge64);
				break;

			case ARM_METRIC_FORMAT_GAUGEDIVR32:
				sscanf ((const char *) value, ARM4_FORMAT_INT32 ":" ARM4_FORMAT_INT32,
						&metric.value.metric_u.gaugedivisor32.gauge,
						&metric.value.metric_u.gaugedivisor32.divisor);
				break;

			/* Java implementation specific */
			case ARM_METRIC_FORMAT_GAUGEFLOAT32:
				// This is a kludge to put floats in this field, but our float_kludge.m4 macro
				// ensures it fits
				sscanf ((const char *) value, "%f", reinterpret_cast<float *> (&(metric.value.metric_u.gaugedivisor32)));
				break;

			case ARM_METRIC_FORMAT_NUMERICID32:
				sscanf ((const char *) value, ARM4_FORMAT_INT32, &metric.value.metric_u.numericid32);
				break;

			case ARM_METRIC_FORMAT_NUMERICID64:
				sscanf ((const char *) value, ARM4_FORMAT_INT64, &metric.value.metric_u.numericid64);
				break;

			case ARM_METRIC_FORMAT_STRING32:
				strncpy (metric.string32, (const char *) value, ARM_METRIC_STRING32_MAX_LENGTH);
				break;
			}
	}
}

void
Arm4db::processAnyNode(xmlTextReaderPtr reader)
{
	// Used for unimplemented nodes
    const xmlChar *name;
    int ret;

    name = xmlTextReaderConstName(reader);
    printf ("processAnyNode (%s)\n", (char *) name);

    int depth = xmlTextReaderDepth(reader); // This is our exit depth
    
    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    	}
    	else if ((type == XML_READER_TYPE_END_ELEMENT) && (depth == xmlTextReaderDepth(reader)))
    	{
    		return;
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processApplicationIdentity(xmlTextReaderPtr reader, const arm_id_t &appl_id)
{
    const xmlChar *name;
    property_record_t property;
    int ret;
    
    memset (&property, 0, sizeof (property));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *property_name = getTextNode(reader);
    			strncpy (property.name, (const char *) property_name, sizeof (property.name));
    		}
    		else if (strcmp ((char *) name, "value") == 0)
    		{
    			const xmlChar *property_value = getTextNode(reader);
    			strncpy (property.value, (const char *) property_value, sizeof (property.value));
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "identity") == 0)
    		{
	    		// Save the application to the database
    			saveApplicationIdentity (appl_id, property);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processApplicationContext(xmlTextReaderPtr reader, const arm_id_t &appl_id)
{
    const xmlChar *name;
    indexed_property_name_record_t property;
    int ret;
    
    memset (&property, 0, sizeof (property));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *property_name = getTextNode(reader);
    			strncpy (property.name, (const char *) property_name, sizeof (property.name));
    		}
    		else if (strcmp ((char *) name, "slot") == 0)
    		{
    			property.index = getIntNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "context") == 0)
    		{
	    		// Save the application to the database
    			saveApplicationContext (appl_id, property);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4ApplicationDefinitionNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    application_record_t application;
    arm_id_t appl_id;
    int ret;
    
    memset (&appl_id, 0, sizeof (arm_id_t));
    memset (&application, 0, sizeof (application));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "appl_id") == 0)
    		{
    			getIdNode (reader, appl_id);
    		}
    		else if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *application_name = getTextNode(reader);
    			strncpy (application.application_name, (const char *) application_name, sizeof (application.application_name));
    		}
    		else if (strcmp ((char *) name, "digest") == 0)
    		{
    			getDigestNode (reader, application.md5_digest);
    		}
    		else if (strcmp ((char *) name, "charset") == 0)
    		{
    			application.charset = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "default_trace_flags") == 0)
    		{
    			application.default_trace_flags = (arm4_shm_transaction_trace_t) getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "default_trace_parameter") == 0)
    		{
    			application.default_trace_parameter = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "default_appdex_threshold") == 0)
    		{
    			application.default_appdex_threshold = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "identity") == 0)
    		{
    			processApplicationIdentity (reader, appl_id);
    		}
    		else if (strcmp ((char *) name, "context") == 0)
    		{
    			processApplicationContext (reader, appl_id);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "application_definition") == 0)
    		{
	    		// Save the application to the database
    			saveApplicationDefinition (appl_id, application);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4MetricDefinitionNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    metric_record_t metric;
    arm_id_t metric_id;
    int ret;
    
	memset (&metric_id, 0, sizeof (arm_id_t));
    memset (&metric, 0, sizeof (metric));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "metric_id") == 0)
    		{
    			getIdNode (reader, metric_id);
    		}
    		else if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *metric_name = getTextNode(reader);
    			strncpy (metric.metric_name, (const char *) metric_name, sizeof (metric.metric_name));
    		}
    		else if (strcmp ((char *) name, "digest") == 0)
    		{
    			getDigestNode (reader, metric.md5_digest);
    		}
    		else if (strcmp ((char *) name, "appl_id") == 0)
    		{
    			getIdNode (reader, metric.appl_id);
    		}
    		else if (strcmp ((char *) name, "units") == 0)
    		{
    			const xmlChar *metric_units = getTextNode(reader);
				if (metric_units)
					strncpy (metric.metric_units, (const char *) metric_units, sizeof (metric.metric_units));
				else
					metric.metric_units [0] = '\0';
    		}
    		else if (strcmp ((char *) name, "format") == 0)
    		{
    			metric.metric_format = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "usage") == 0)
    		{
    			metric.metric_usage = getIntNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "metric_definition") == 0)
    		{
	    		// Save the metric to the database
    			saveMetricDefinition (metric_id, metric);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processTransactionIdentity(xmlTextReaderPtr reader, const arm_id_t &tran_id)
{
    const xmlChar *name;
    property_record_t property;
    int ret;
    
    memset (&property, 0, sizeof (property));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *property_name = getTextNode(reader);
    			strncpy (property.name, (const char *) property_name, sizeof (property.name));
    		}
    		else if (strcmp ((char *) name, "value") == 0)
    		{
    			const xmlChar *property_value = getTextNode(reader);
    			strncpy (property.value, (const char *) property_value, sizeof (property.value));
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "identity") == 0)
    		{
	    		// Save the application to the database
    			saveTransactionIdentity (tran_id, property);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processTransactionContext(xmlTextReaderPtr reader, const arm_id_t &tran_id)
{
    const xmlChar *name;
    indexed_property_name_record_t property;
    int ret;
    
    memset (&property, 0, sizeof (property));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *property_name = getTextNode(reader);
    			strncpy (property.name, (const char *) property_name, sizeof (property.name));
    		}
    		else if (strcmp ((char *) name, "slot") == 0)
    		{
    			property.index = getIntNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "context") == 0)
    		{
	    		// Save the application to the database
    			saveTransactionContext (tran_id, property);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processTransactionMetric(xmlTextReaderPtr reader, const arm_id_t &tran_id)
{
    const xmlChar *name;
    arm_metric_binding_t binding;
    int ret;
    
    memset (&binding, 0, sizeof (binding));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "metric_id") == 0)
    		{
    			getIdNode (reader, binding.id);
    		}
    		else if (strcmp ((char *) name, "slot") == 0)
    		{
    			binding.slot = getIntNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "metric") == 0)
    		{
	    		// Save the application to the database
    			saveTransactionMetric (tran_id, binding);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4TransactionDefinitionNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    transaction_record_t transaction;
    arm_id_t tran_id;
    int ret;
    
	memset (&tran_id, 0, sizeof (arm_id_t));
    memset (&transaction, 0, sizeof (transaction));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "tran_id") == 0)
    		{
    			getIdNode (reader, tran_id);
    		}
    		else if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *transaction_name = getTextNode(reader);
    			strncpy (transaction.transaction_name, (const char *) transaction_name, sizeof (transaction.transaction_name));
    		}
    		else if (strcmp ((char *) name, "digest") == 0)
    		{
    			getDigestNode (reader, transaction.md5_digest);
    		}
    		else if (strcmp ((char *) name, "appl_id") == 0)
    		{
    			getIdNode (reader, transaction.appl_id);
    		}
    		else if (strcmp ((char *) name, "trace_flags") == 0)
    		{
    			transaction.trace_flags = (arm4_shm_transaction_trace_t) getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "appdex_flags") == 0)
    		{
    			transaction.appdex_flags = (arm4_appdex_t) getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "trace_parameter") == 0)
    		{
    			transaction.trace_parameter = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "appdex_threshold") == 0)
    		{
    			transaction.appdex_threshold = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "identity") == 0)
    		{
    			processTransactionIdentity (reader, tran_id);
    		}
    		else if (strcmp ((char *) name, "context") == 0)
    		{
    			processTransactionContext (reader, tran_id);
    		}
    		else if (strcmp ((char *) name, "uri") == 0)
    		{
    			const xmlChar *uri = getTextNode(reader);
    			saveTransactionUri (tran_id, (const char *) uri);
    		}
    		else if (strcmp ((char *) name, "metric") == 0)
    		{
    			processTransactionMetric (reader, tran_id);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "transaction_definition") == 0)
    		{
	    		// Save the transaction to the database
    			saveTransactionDefinition (tran_id, transaction);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4DefinitionsNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    int ret;

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "application_definition") == 0)
    		{
    			processArm4ApplicationDefinitionNode (reader);
    		}
    		else if (strcmp ((char *) name, "metric_definition") == 0)
    		{
    			processArm4MetricDefinitionNode (reader);
    		}
    		else if (strcmp ((char *) name, "transaction_definition") == 0)
    		{
    			processArm4TransactionDefinitionNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		return;
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processAppInstanceContextValue(xmlTextReaderPtr reader, const arm_app_start_handle_t app_handle)
{
    const xmlChar *name;
    indexed_property_value_record_t property;
    int ret;
    
    memset (&property, 0, sizeof (property));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "value") == 0)
    		{
    			const xmlChar *property_value = getTextNode(reader);
    			strncpy (property.value, (const char *) property_value, sizeof (property.value));
    		}
    		else if (strcmp ((char *) name, "slot") == 0)
    		{
    			property.index = getIntNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "context_value") == 0)
    		{
	    		// Save the application to the database
    			saveAppInstanceContextValue (app_handle, property);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processAppInstanceAddress(xmlTextReaderPtr reader, const arm_app_start_handle_t app_handle)
{
    const xmlChar *name;
	application_address_record_t address_value;
    int ret;
    
    memset (&address_value, 0, sizeof (address_value));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "address_format") == 0)
    		{
				address_value.address_format = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "address_length") == 0)
    		{
    			address_value.address_length = getIntNode (reader);
    		}
			else if (strcmp ((char *) name, "id_valid") == 0)
			{
				address_value.id_valid = getIntNode (reader);
			}
			else if (strcmp ((char *) name, "id") == 0)
			{
				getIdNode (reader, address_value.id);
			}
			else if (strcmp ((char *) name, "address") == 0)
			{
				getAddressNode (reader, address_value.address);
			}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "address") == 0)
    		{
	    		// Save the application to the database
    			saveAppInstanceAddress (app_handle, address_value);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4ApplicationInstanceNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    application_instance_record_t application;
    arm_app_start_handle_t app_handle;
    int ret;
    
	app_handle = 0;
    memset (&application, 0, sizeof (application));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "handle") == 0)
    		{
    			app_handle = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "appl_id") == 0)
    		{
    			getIdNode (reader, application.appl_id);
    		}
    		else if (strcmp ((char *) name, "active") == 0)
    		{
    			application.active = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "group") == 0)
    		{
    			const xmlChar *group = getTextNode(reader);
    			saveAppInstanceGroup (app_handle, (const char *) group);
    		}
    		else if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *instance_name = getTextNode(reader);
    			saveAppInstanceName (app_handle, (const char *) instance_name);
    		}
    		else if (strcmp ((char *) name, "context_value") == 0)
    		{
    			processAppInstanceContextValue (reader, app_handle);
    		}
			else if (strcmp ((char *) name, "address") == 0)
			{
				processAppInstanceAddress (reader, app_handle);
			}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "application_instance") == 0)
    		{
	    		// Save the application to the database
    			saveApplicationInstance (app_handle, application);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processTranInstanceContextValue(xmlTextReaderPtr reader, const arm_tran_start_handle_t tran_handle)
{
    const xmlChar *name;
    indexed_property_value_record_t property;
    int ret;
    
    memset (&property, 0, sizeof (property));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "value") == 0)
    		{
    			const xmlChar *property_value = getTextNode(reader);
    			strncpy (property.value, (const char *) property_value, sizeof (property.value));
    		}
    		else if (strcmp ((char *) name, "slot") == 0)
    		{
    			property.index = getIntNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "context_value") == 0)
    		{
	    		// Save the application to the database
    			saveTranInstanceContextValue (tran_handle, property);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processTranInstanceMetricValue(xmlTextReaderPtr reader, const arm_tran_start_handle_t tran_handle)
{
    const xmlChar *name;
	metric_value_record_t metric;
    int ret;
    
    memset (&metric, 0, sizeof (metric));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "slot") == 0)
    		{
    			metric.value.slot = (arm_metric_slot_t) getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "format") == 0)
    		{
    			metric.value.format = (arm_metric_format_t) getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "usage") == 0)
    		{
    			metric.value.usage = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "valid") == 0)
    		{
    			metric.value.valid = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "value") == 0)
    		{
    			getMetricNode (reader, metric);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "metric_value") == 0)
    		{
	    		// Save the application to the database
    			saveTranInstanceMetricValue (tran_handle, metric);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processTranInstanceUser(xmlTextReaderPtr reader, const arm_tran_start_handle_t tran_handle)
{
    const xmlChar *name;
    user_record_t user;
    int ret;
    
    memset (&user, 0, sizeof (user));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "name") == 0)
    		{
    			const xmlChar *username = getTextNode(reader);
    			strncpy (user.name, (const char *) username, sizeof (user.name));
    		}
    		else if (strcmp ((char *) name, "id_valid") == 0)
    		{
    			user.id_valid = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "id") == 0)
    		{
    			getIdNode (reader, user.id);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "user") == 0)
    		{
	    		// Save the application to the database
    			saveTranInstanceUser (tran_handle, user);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4TransactionInstanceNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    transaction_instance_record_t transaction;
    arm_tran_start_handle_t tran_handle;
    int ret;
    
	tran_handle = 0;
    memset (&transaction, 0, sizeof (transaction));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "handle") == 0)
    		{
    			tran_handle = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "appl_id") == 0)
    		{
    			getIdNode (reader, transaction.appl_id);
    		}
    		else if (strcmp ((char *) name, "tran_id") == 0)
    		{
    			getIdNode (reader, transaction.tran_id);
    		}
    		else if (strcmp ((char *) name, "app_handle") == 0)
    		{
    			transaction.app_handle = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "arrival_time") == 0)
    		{
    			getTimeNode (reader, transaction.arrival_time);
    		}
    		else if (strcmp ((char *) name, "stop_time") == 0)
    		{
    			getTimeNode (reader, transaction.stop_time);
    		}
    		else if (strcmp ((char *) name, "response_time") == 0)
    		{
    			getTimeNode (reader, transaction.response_time);
    		}
    		else if (strcmp ((char *) name, "block_start_time") == 0)
    		{
    			getTimeNode (reader, transaction.block_start_time);
    		}
    		else if (strcmp ((char *) name, "blocked_time") == 0)
    		{
    			getTimeNode (reader, transaction.blocked_time);
    		}
    		else if (strcmp ((char *) name, "status") == 0)
    		{
    			transaction.tran_status = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "active") == 0)
    		{
    			transaction.active = getIntNode (reader);
    		}
    		else if (strcmp ((char *) name, "context_value") == 0)
    		{
    			processTranInstanceContextValue (reader, tran_handle);
    		}
    		else if (strcmp ((char *) name, "context_uri") == 0)
    		{
    			const xmlChar *uri = getTextNode(reader);
    			saveTranInstanceUri (tran_handle, (const char *) uri);
    		}
    		else if (strcmp ((char *) name, "metric_value") == 0)
    		{
    			processTranInstanceMetricValue (reader, tran_handle);
    		}
    		else if (strcmp ((char *) name, "user") == 0)
    		{
    			processTranInstanceUser (reader, tran_handle);
    		}
    		else if (strcmp ((char *) name, "diag") == 0)
    		{
    			const xmlChar *diag = getTextNode(reader);
    			saveTranInstanceDiag (tran_handle, (const char *) diag);
    		}
    		else if (strcmp ((char *) name, "correlator") == 0)
    		{
    			arm_correlator_t correlator;
    			
    			getCorrelatorNode (reader, &correlator);
    			saveTranCorrelator (tran_handle, &correlator);
    		}
    		else if (strcmp ((char *) name, "parent_correlator") == 0)
    		{
    			arm_correlator_t correlator;
    			
    			getCorrelatorNode (reader, &correlator);
    			saveTranParentCorrelator (tran_handle, &correlator);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "transaction_instance") == 0)
    		{
	    		// Save the transaction to the database
    			saveTransactionInstance (tran_handle, transaction);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4TransactionTraceNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    arm4db_transaction_trace_stats_t trace_stats;
    arm_time_t sample_time;
    int ret;
    
    memset (&sample_time, 0, sizeof (sample_time));
    memset (&trace_stats, 0, sizeof (trace_stats));

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "transaction_trace_sample_time") == 0)
    		{
    			sample_time.tv_sec = getInt64Node (reader);
    		}
    		else if (strcmp ((char *) name, "transaction_trace_tran_id") == 0)
    		{
    			getIdNode (reader, trace_stats.tran_id);
    		}
    		else if (strcmp ((char *) name, "transaction_trace_counter") == 0)
    		{
    			trace_stats.transaction_counter = getInt64Node (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "transaction_trace") == 0)
    		{
	    		// Save the trace information to the database
    			saveTraceStats (sample_time, trace_stats);
	    		return;
    		}
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4InstancesNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    int ret;

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "application_instance") == 0)
    		{
    			processArm4ApplicationInstanceNode (reader);
    		}
    		else if (strcmp ((char *) name, "transaction_instance") == 0)
    		{
    			processArm4TransactionInstanceNode (reader);
    		}
    		else if (strcmp ((char *) name, "transaction_trace") == 0)
    		{
    			processArm4TransactionTraceNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		return;
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processArm4Node(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    int ret;

    // Get the next node item
    ret = xmlTextReaderRead(reader);
 
    while (ret == 1)
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "definitions") == 0)
    		{
    			processArm4DefinitionsNode (reader);
    		}
    		else if (strcmp ((char *) name, "instances") == 0)
    		{
    			processArm4InstancesNode (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		return;
    	}
        ret = xmlTextReaderRead(reader);
    }
}

void
Arm4db::processRootNode(xmlTextReaderPtr reader)
{
    const xmlChar *name;
    int ret;

    do
    {
    	int type = xmlTextReaderNodeType(reader);
    	if (type == XML_READER_TYPE_ELEMENT)
    	{
    		name = xmlTextReaderConstName(reader);
    		if (strcmp ((char *) name, "arm4") == 0)
    		{
    			processArm4Node (reader);
    		}
    		else
    		{
    			fprintf (stderr, "Unknown node name %s\n", name);
    			processAnyNode (reader);
    		}
    	}
    	else if (type == XML_READER_TYPE_END_ELEMENT)
    	{
    		return;
    	}
        ret = xmlTextReaderRead(reader);
    } while (ret == 1);
}

void
Arm4db::streamReader(xmlTextReaderPtr reader)
{
    if (reader != NULL)
    {
        int ret = xmlTextReaderRead(reader);
        while (ret == 1) {
        	processRootNode(reader);
            ret = xmlTextReaderRead(reader);
        }
        xmlFreeTextReader(reader);
        if (ret != 0)
        {
            fprintf(stderr, "Unable to read XML\n");
        }
    }
    else
    {
        fprintf(stderr, "Unable to open XML\n");
    }
}

void
Arm4db::streamFile(const char *filename)
{
    xmlTextReaderPtr reader = xmlReaderForFile(filename, NULL, 0);
	streamReader (reader);
}

void
Arm4db::streamFileFromMemory(const std::string &xml)
{
    xmlTextReaderPtr reader = xmlReaderForMemory(xml.c_str(), xml.length (), NULL, NULL, 0);
	streamReader (reader);
}

void
Arm4db::import (const char *filename_ptr)
{
    /*
     * this initializes the library and checks for potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
    LIBXML_TEST_VERSION

    streamFile(filename_ptr);

    /*
     * Cleanup function for the XML library.
     */
    xmlCleanupParser();
}

void
Arm4db::importFromMemory (const std::string &xml)
{
    /*
     * this initializes the library and checks for potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
    LIBXML_TEST_VERSION

	Arm4dbUtilityLock lock;

	Arm4db *arm4db_ptr = Arm4db::getArm4db ();
	arm_error_t status = arm4db_ptr->openDatabases ();
	if (status != ARM_SUCCESS)
	{
		lock.unlock ();

		throw new Arm4dbException ("Unable to open databases");
	}

	streamFileFromMemory(xml);
		
	arm4db_ptr->closeDatabases ();

	lock.unlock ();

    /*
     * Cleanup function for the XML library.
     */
    xmlCleanupParser();
}

