/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "Logger.h"
#include "BerkeleyDB.h"
#include "Arm4dbCommon.h"
#include "Arm4dbExceptions.h"
#include "Arm4dbSharedMemory.h"

/*
 *	C++ Implementation
 */

void
BerkeleyDB::panic (const DbRunRecoveryException &r, const char *file, int line)
{
    r.get_env()->err(r.get_errno (), "PANIC: Recovery error %s(%d)", file, line);

	throw new Arm4dbException ("PANIC: Recovery error %s(%d)", file, line);
}


void
BerkeleyDB::panic (const DbException &e, const char *file, int line)
{
    e.get_env()->err(e.get_errno (), "PANIC: Exception error %s(%d)", file, line);

	throw new Arm4dbException ("PANIC: Exception error %s(%d)", file, line);
}

BerkeleyDB::BerkeleyDB () :
	m_environmentPtr (NULL),
	db_sequence_ptr (NULL),

	db_application_ptr (NULL),
	db_application_name_ptr (NULL),
	db_application_identity_ptr (NULL),
	db_application_context_names_ptr (NULL),
	db_application_instance_ptr (NULL),
	db_application_instance_app_ptr (NULL),
	db_application_group_ptr (NULL),
	db_application_instance_names_ptr (NULL),
	db_application_context_values_ptr (NULL),
	db_application_address_ptr (NULL),
	
	db_transaction_ptr (NULL),
	db_transaction_name_ptr (NULL),
	db_transaction_appid_ptr (NULL),
	db_transaction_identity_ptr (NULL),
	db_transaction_context_names_ptr (NULL),
	db_transaction_identity_uri_ptr (NULL),
	db_transaction_metric_binding_ptr (NULL),
	db_transaction_instance_ptr (NULL),
	db_transaction_instance_tranid_ptr (NULL),
	db_transaction_instance_app_handle_ptr (NULL),
	db_transaction_context_values_ptr (NULL),
	db_transaction_context_uri_ptr (NULL),
	db_transaction_context_metric_value_ptr (NULL),
	db_transaction_context_user_ptr (NULL),
	db_transaction_context_diag_ptr (NULL),
	
	db_metric_ptr (NULL),
	db_metric_name_ptr (NULL),
	db_metric_appid_ptr (NULL),
	
	db_parent_correlator_ptr (NULL),
	db_current_correlator_ptr (NULL),

	db_trace_stat_ptr (NULL),
	db_appdex_stat_ptr (NULL)
{
}

BerkeleyDB::~BerkeleyDB ()
{
	// Ensure everything is closed up
	closeDatabasesNoLock ();

	int status = m_environmentPtr->close (0);
	if (status != 0)
		throw new Arm4dbException ("Environment close error %d", status);
	delete m_environmentPtr;

	// Ensure we don't reaccess it
	m_environmentPtr = NULL;
}

int
BerkeleyDB::openEnvironment (int flags)
{
    // If we're already open then return
    if (m_environmentPtr)
        return 0;

	m_environmentPtr = new DbEnv (0);

	/* Put all transaction logs in memory to improve performance */
#if defined(DB_LOG_INMEMORY) // pre-4.7 API
	m_environmentPtr->set_flags(DB_LOG_INMEMORY, 1);
#else
	m_environmentPtr->log_set_config(DB_LOG_IN_MEMORY, 1);
#endif
	m_environmentPtr->set_lg_bsize (getConfig ().getDbLogBufferSize ());
	m_environmentPtr->set_lg_max (getConfig ().getDbLogFileSize ());
	
	/* Set the shared memory key */
	long key = getConfig ().getDbKey ();
	if (key == 0)
		m_environmentPtr->get_shm_key (&key);
	m_environmentPtr->set_shm_key (key + (getConfig ().getInstance () * 100));

	/* Create a new environment if needed */
	int status = 0;
	try
	{
		if (flags & DB_RDONLY)
		{
			status = m_environmentPtr->open (
				getConfig ().getDbHomeC (),
				DB_INIT_LOCK | 
				DB_INIT_LOG | 
				DB_INIT_MPOOL | 
				DB_INIT_TXN | 
				DB_THREAD, 
				0644 /* Mode - this will be adjusted with our umask */
			);
		}
		else
		{
			status = m_environmentPtr->open (
				getConfig ().getDbHomeC (),
				DB_INIT_LOCK | 
				DB_INIT_LOG | 
				DB_INIT_MPOOL | 
				DB_INIT_TXN | 
				DB_CREATE | 
				DB_THREAD, 
				0644 /* Mode - this will be adjusted with our umask */
			);
		}
	}
	catch (DbException e)
	{
		m_environmentPtr->err (e.get_errno (), "Unable to open or create environment");
		return e.get_errno ();
	}
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to open or create environment");
		return status;
	}

	/* Send additonal debugging information to stderr for now */
#ifdef DEBUG2
	m_environmentPtr->set_errfile (stdout);
#else
	m_environmentPtr->set_errfile (NULL);
#endif

	return 0;
}

int
BerkeleyDB::openDatabaseWithTxn (Db **db_ptr,
	const char *db_name_ptr,
	int flags,
	DbTxn *transaction_ptr)
{
	int status;

	/* If we're already open, there's nothing to do */
	if ((*db_ptr) != NULL)
		return 0;

	try
	{
		*db_ptr = new Db (m_environmentPtr, 0);
	}
	catch (DbException e)
	{
		m_environmentPtr->err (e.get_errno (), "Unable to create database");
		return e.get_errno ();
	}
	if (flags)
	{
		/* Set flags as appropriate for the database */
		status = (*db_ptr)->set_flags (flags);
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to set database flags");
			return status;
		}
	}

	/* Open the database, creating it if necessary. To allow transactions, */
	/* this operation needs to be protected by a transaction as well */
	char full_db_name [2048];
	snprintf (full_db_name, sizeof(full_db_name), "%s/%s", getConfig().getDbHomeC (), db_name_ptr);
    WRAP_EXCEPTION(status, (*db_ptr)->open (
        transaction_ptr,
        full_db_name,
        NULL,
        DB_BTREE,
        DB_THREAD | DB_CREATE,
        0644));  /* Mode will be adjusted with our umask */
	if (status != 0)
	{
		char error [2048];

		sprintf (error, "Unable to open or create database '%s'", db_name_ptr);
		m_environmentPtr->err (status, error);
	}

	return status;
}

int
BerkeleyDB::openDatabase (Db **db_ptr,
	const char *db_name_ptr,
	int flags)
{
	int status;
	int failures;
	DbTxn *transaction_ptr;

	/* If we're already open, there's nothing to do */
	if ((*db_ptr) != NULL)
		return 0;

    // Ensure our environment is open
    status = openEnvironment (flags);
	if (status != 0)
		return status;

	for (failures = 0; failures < MAX_FAILURES; failures++)
	{
		/* Open the database, creating it if necessary. To allow transactions, */
		/* this operation needs to be protected by a transaction as well */
		WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to begin transaction");
			return status;
		}
#if 0
        transaction_ptr->set_name ("openDatabase");
#endif
	
		status = openDatabaseWithTxn (db_ptr, db_name_ptr, flags, transaction_ptr);
		if (status != 0)
		{
			transaction_ptr->abort ();
            if (*db_ptr)
            {
                (*db_ptr)->close (0);
                *db_ptr = NULL; /* Reset before next loop */
            }
			continue;
		}
	
		WRAP_EXCEPTION(status, transaction_ptr->commit (0));
		if (status != 0)
			{
			m_environmentPtr->err (status, "Unable to commit database open transaction");
			return status;
			}
		break; /* Operation successful */
	}
	return status;
}

void
BerkeleyDB::close_database (Db **database_pptr)
{
	if (*database_pptr)
	{
		(*database_pptr)->close (0);
		delete *database_pptr;

		*database_pptr = NULL;
	}
}

int
BerkeleyDB::get_application_name (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (((application_record_t *) primary_data_ptr->get_data ())->application_name);
	secondary_key_ptr->set_size (strlen (((application_record_t *) primary_data_ptr->get_data ())->application_name) + 1);
	return 0;
}

int
BerkeleyDB::get_application_instance_app (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (&(((application_instance_record_t *) primary_data_ptr->get_data ())->appl_id));
	secondary_key_ptr->set_size (sizeof (((application_instance_record_t *) primary_data_ptr->get_data ())->appl_id));
	return 0;
}

int
BerkeleyDB::get_transaction_name (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (((transaction_record_t *) primary_data_ptr->get_data ())->transaction_name);
	secondary_key_ptr->set_size (strlen (((transaction_record_t *) primary_data_ptr->get_data ())->transaction_name) + 1);
	return 0;
}

int
BerkeleyDB::get_transaction_appid (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (&(((transaction_record_t *) primary_data_ptr->get_data ())->appl_id));
	secondary_key_ptr->set_size (sizeof (((transaction_record_t *) primary_data_ptr->get_data ())->appl_id));
	return 0;
}

int
BerkeleyDB::get_tran_instance_tranid (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (&(((transaction_instance_record_t *) primary_data_ptr->get_data ())->tran_id));
	secondary_key_ptr->set_size (sizeof (((transaction_instance_record_t *) primary_data_ptr->get_data ())->tran_id));
	return 0;
}

int
BerkeleyDB::get_tran_instance_app_handle (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (&(((transaction_instance_record_t *) primary_data_ptr->get_data ())->app_handle));
	secondary_key_ptr->set_size (sizeof (((transaction_instance_record_t *) primary_data_ptr->get_data ())->app_handle));
	return 0;
}

int
BerkeleyDB::get_metric_name (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (((metric_record_t *) primary_data_ptr->get_data ())->metric_name);
	secondary_key_ptr->set_size (strlen (((metric_record_t *) primary_data_ptr->get_data ())->metric_name) + 1);
	return 0;
}

int
BerkeleyDB::get_metric_appid (Db *db_ptr,
	const Dbt *primary_key_ptr,
	const Dbt *primary_data_ptr,
	Dbt *secondary_key_ptr)
{
	/*
	 * Extract the secondary key from the primary data structure
	 */
	secondary_key_ptr->set_data (&(((metric_record_t *) primary_data_ptr->get_data ())->appl_id));
	secondary_key_ptr->set_size (sizeof (((metric_record_t *) primary_data_ptr->get_data ())->appl_id));
	return 0;
}

int
BerkeleyDB::openApplicationDatabase (int flags)
{
	int	status = ARM_FAILURE;
	int failures;
	DbTxn *transaction_ptr;

	/* Check to see if we're already opened */
	if (db_application_ptr)
		return 0;

    // Ensure our environment is open
	openEnvironment (flags);

	for (failures = 0; failures < MAX_FAILURES; failures++)
	{
		/* Open the database, creating it if necessary. To allow transactions, */
		/* this operation needs to be protected by a transaction as well */
		WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to begin transaction");
			return status;
		}
#if 0
        transaction_ptr->set_name ("open_application_database");
#endif

		int step = 0;
		WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_application_ptr, APPLICATION_DB_NAME, flags, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_application_name_ptr, APPLICATION_NAME_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_application_ptr->associate (transaction_ptr, db_application_name_ptr, get_application_name, DB_CREATE), step);
	
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to create application database at step %d", step);
			transaction_ptr->abort ();

			/* Mark the databases as unopened */
            close_database (&db_application_ptr);
            close_database (&db_application_name_ptr);

			continue;
		}
	
		WRAP_EXCEPTION(status, transaction_ptr->commit (0));
		if (status != 0)
			{
			m_environmentPtr->err (status, "Unable to commit database open transaction");
			return status;
			}
		break; /* Operation successful */
	}
	return status;
}

int
BerkeleyDB::openApplicationInstanceDatabase (int flags)
{
	int status;
	int failures;
	DbTxn *transaction_ptr;

	/* Check to see if we're already opened */
	if (db_application_instance_ptr)
		return 0;

    // Ensure our environment is open
	status = openEnvironment (flags);
	if (status != 0)
		return status;

	for (failures = 0; failures < MAX_FAILURES; failures++)
	{
		/* Open the database, creating it if necessary. To allow transactions, */
		/* this operation needs to be protected by a transaction as well */
		WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to begin transaction");
			return status;
		}
#if 0
        transaction_ptr->set_name ("open_application_instance_database");
#endif
	
		WRAP_EXCEPTION(status, openDatabaseWithTxn (&db_application_instance_ptr, APPLICATION_INSTANCE_DB_NAME, flags, transaction_ptr));
		if (status == 0) WRAP_EXCEPTION(status, openDatabaseWithTxn (&db_application_instance_app_ptr, APPLICATION_INSTANCE_APP_DB_NAME, flags | DB_DUPSORT, transaction_ptr));
		if (status == 0) WRAP_EXCEPTION(status, db_application_instance_ptr->associate (transaction_ptr, db_application_instance_app_ptr, get_application_instance_app, DB_CREATE));
	
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to create application database");
			transaction_ptr->abort ();

			/* Mark the databases as unopened */
            close_database (&db_application_instance_ptr);
            close_database (&db_application_instance_app_ptr);

			continue;
		}
	
		WRAP_EXCEPTION(status, transaction_ptr->commit (0));
		if (status != 0)
			{
			m_environmentPtr->err (status, "Unable to commit database open transaction");
			return status;
			}
		break; /* Operation successful */
	}
	return status;
}

int
BerkeleyDB::openTransactionDatabase (int flags)
{
	int status;
	int failures;
	DbTxn *transaction_ptr;

	/* Check to see if we're already opened */
	if (db_transaction_ptr)
		return 0;

    // Ensure our environment is open
	status = openEnvironment (flags);
	if (status != 0)
		return status;

	for (failures = 0; failures < MAX_FAILURES; failures++)
	{
		/* Open the database, creating it if necessary. To allow transactions, */
		/* this operation needs to be protected by a transaction as well */
		WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to begin transaction");
			return status;
		}
#if 0
        transaction_ptr->set_name ("open_transaction_database");
#endif

		int step = 0;
		WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_transaction_ptr, TRANSACTION_DB_NAME, flags, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_transaction_name_ptr, TRANSACTION_NAME_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_transaction_appid_ptr, TRANSACTION_APPID_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_transaction_ptr->associate (transaction_ptr, db_transaction_name_ptr, get_transaction_name, DB_CREATE), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_transaction_ptr->associate (transaction_ptr, db_transaction_appid_ptr, get_transaction_appid, DB_CREATE), step);
	
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to create transaction database at step %d", step);
			transaction_ptr->abort ();

			/* Mark the databases as unopened */
            close_database (&db_transaction_ptr);
            close_database (&db_transaction_name_ptr);
            close_database (&db_transaction_appid_ptr);

			continue;
		}
	
		WRAP_EXCEPTION(status, transaction_ptr->commit (0));
		if (status != 0)
			{
			m_environmentPtr->err (status, "Unable to commit database open transaction");
			return status;
			}
		break; /* Operation successful */
	}
	return status;
}

int
BerkeleyDB::openTransactionInstanceDatabase (int flags)
{
	int status;
	int failures;
	DbTxn *transaction_ptr;

	/* Check to see if we're already opened */
	if (db_transaction_instance_ptr)
		return 0;

    // Ensure our environment is open
	status = openEnvironment (flags);
	if (status != 0)
		return status;

	for (failures = 0; failures < MAX_FAILURES; failures++)
	{
		/* Open the database, creating it if necessary. To allow transactions, */
		/* this operation needs to be protected by a transaction as well */
		WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to begin transaction");
			return status;
		}
#if 0
        transaction_ptr->set_name ("open_transaction_instance_database");
#endif

		int step = 0;
		WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_transaction_instance_ptr, TRANSACTION_INSTANCE_DB_NAME, flags, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_transaction_instance_tranid_ptr, TRANSACTION_INSTANCE_TRANID_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_transaction_instance_app_handle_ptr, TRANSACTION_INSTANCE_APP_HANDLE_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_transaction_instance_ptr->associate (transaction_ptr, db_transaction_instance_tranid_ptr, get_tran_instance_tranid, DB_CREATE), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_transaction_instance_ptr->associate (transaction_ptr, db_transaction_instance_app_handle_ptr, get_tran_instance_app_handle, DB_CREATE), step);
	
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to create transaction instance database at step %d", step);
			transaction_ptr->abort ();

			/* Mark the databases as unopened */
            close_database (&db_transaction_instance_ptr);
            close_database (&db_transaction_instance_tranid_ptr);
            close_database (&db_transaction_instance_app_handle_ptr);

			continue;
		}
	
		WRAP_EXCEPTION(status, transaction_ptr->commit (0));
		if (status != 0)
			{
			m_environmentPtr->err (status, "Unable to commit database open transaction");
			return status;
			}
		break; /* Operation successful */
	}
	return status;
}

int
BerkeleyDB::openMetricDatabase (int flags)
{
	int status;
	int failures;
	DbTxn *transaction_ptr;

	/* Check to see if we're already opened */
	if (db_metric_ptr)
		return 0;

    // Ensure our environment is open
	status = openEnvironment (flags);
	if (status != 0)
		return status;

	for (failures = 0; failures < MAX_FAILURES; failures++)
	{
		/* Open the database, creating it if necessary. To allow transactions, */
		/* this operation needs to be protected by a transaction as well */
		WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to begin transaction");
			return status;
		}
#if 0
        transaction_ptr->set_name ("open_metric_database");
#endif

		int step = 0;
		WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_metric_ptr, METRIC_DB_NAME, flags, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_metric_name_ptr, METRIC_NAME_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, openDatabaseWithTxn (&db_metric_appid_ptr, METRIC_APPID_DB_NAME, flags | DB_DUPSORT, transaction_ptr), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_metric_ptr->associate (transaction_ptr, db_metric_name_ptr, get_metric_name, DB_CREATE), step);
		if (status == 0) WRAP_EXCEPTION_STEP(status, db_metric_ptr->associate (transaction_ptr, db_metric_appid_ptr, get_metric_appid, DB_CREATE), step);
	
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to create metric database at step %d", step);
			transaction_ptr->abort ();

			/* Mark the databases as unopened */
            close_database (&db_metric_ptr);
            close_database (&db_metric_name_ptr);
            close_database (&db_metric_appid_ptr);

			continue;
		}
	
		WRAP_EXCEPTION(status, transaction_ptr->commit (0));
		if (status != 0)
			{
			m_environmentPtr->err (status, "Unable to commit database open transaction");
			return status;
			}
		break; /* Operation successful */
	}
	return status;
}

arm_error_t
BerkeleyDB::openDatabasesNoLock (int flags)
{
	int	status;

	/* Open the primary databases */
	status = openDatabase (&db_sequence_ptr, SEQUENCE_DB_NAME, flags);
	if (status != 0)
		return ARM_FAILURE;

	status = openApplicationDatabase (flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_application_identity_ptr, APPLICATION_IDENTITY_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_application_context_names_ptr, APPLICATION_CONTEXT_NAMES_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openApplicationInstanceDatabase (flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_application_group_ptr, APPLICATION_GROUP_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_application_instance_names_ptr, APPLICATION_INSTANCE_NAMES_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_application_context_values_ptr, APPLICATION_CONTEXT_VALUES_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_application_address_ptr, APPLICATION_ADDRESS_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;

	status = openTransactionDatabase (flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_identity_ptr, TRANSACTION_IDENTITY_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_context_names_ptr, TRANSACTION_CONTEXT_NAMES_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_identity_uri_ptr, TRANSACTION_IDENTITY_URI_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_metric_binding_ptr, TRANSACTION_METRIC_BINDING_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openTransactionInstanceDatabase (flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_context_values_ptr, TRANSACTION_CONTEXT_VALUES_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_context_uri_ptr, TRANSACTION_CONTEXT_URI_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_context_metric_value_ptr, TRANSACTION_CONTEXT_METRIC_VALUE_DB_NAME, flags | DB_DUPSORT);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_context_user_ptr, TRANSACTION_CONTEXT_USER_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_transaction_context_diag_ptr, TRANSACTION_CONTEXT_DIAG_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;

    status = openMetricDatabase (flags);
    if (status != 0)
        return ARM_FAILURE;

	status = openDatabase (&db_parent_correlator_ptr, CORRELATOR_PARENT_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_current_correlator_ptr, CORRELATOR_CURRENT_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;

	status = openDatabase (&db_trace_stat_ptr, TRACE_STAT_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;
	status = openDatabase (&db_appdex_stat_ptr, APPDEX_STAT_DB_NAME, flags);
	if (status != 0)
        return ARM_FAILURE;

	return ARM_SUCCESS;	
}

void
BerkeleyDB::initializeLibrary (void)
{
	
}

void
BerkeleyDB::initializeThread (void)
{
	
}

arm_error_t
BerkeleyDB::openDatabasesReadOnly (void)
{
	int	status;

	/* Ensure we have database access */
	Arm4dbLock lock;

	/* Open the primary databases */
	//status = openDatabasesNoLock (DB_RDONLY);
	status = openDatabasesNoLock (0); // See ticket 77

	lock.unlock ();

	return status;	
}

arm_error_t
BerkeleyDB::openDatabases (void)
{
	int	status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	/* Open the primary databases */
	status = openDatabasesNoLock (0);

    lock.unlock ();

	return status;	
}

arm_error_t
BerkeleyDB::closeDatabasesNoLock (void)
{
    /* Close the secondary indexes */
	close_database (&db_application_name_ptr);
	close_database (&db_application_instance_app_ptr);

	close_database (&db_transaction_name_ptr);
	close_database (&db_transaction_appid_ptr);
	close_database (&db_transaction_instance_tranid_ptr);
	close_database (&db_transaction_instance_app_handle_ptr);

    close_database (&db_metric_name_ptr);
    close_database (&db_metric_appid_ptr);

	/* Close the primary databases */
	close_database (&db_sequence_ptr);

	close_database (&db_application_ptr);
	close_database (&db_application_identity_ptr);
	close_database (&db_application_context_names_ptr);
	close_database (&db_application_instance_ptr);
	close_database (&db_application_group_ptr);
	close_database (&db_application_instance_names_ptr);
	close_database (&db_application_context_values_ptr);
	close_database (&db_application_address_ptr);

	close_database (&db_transaction_ptr);
	close_database (&db_transaction_identity_ptr);
	close_database (&db_transaction_context_names_ptr);
	close_database (&db_transaction_identity_uri_ptr);
    close_database (&db_transaction_metric_binding_ptr);
	close_database (&db_transaction_instance_ptr);
	close_database (&db_transaction_context_values_ptr);
	close_database (&db_transaction_context_uri_ptr);
	close_database (&db_transaction_context_metric_value_ptr);
	close_database (&db_transaction_context_user_ptr);
	close_database (&db_transaction_context_diag_ptr);

    close_database (&db_metric_ptr);

	close_database (&db_parent_correlator_ptr);
	close_database (&db_current_correlator_ptr);

	close_database (&db_trace_stat_ptr);
	close_database (&db_appdex_stat_ptr);

	if (m_environmentPtr)
	{
		int status = m_environmentPtr->close (0);
		if (status != 0)
			throw new Arm4dbException ("Environment close error %d", status);
		delete m_environmentPtr;

		// Ensure we don't reaccess it
		m_environmentPtr = NULL;
	}

    return ARM_SUCCESS;	
}

/* Transactions that extract from the queues into the instance databases */
int
BerkeleyDB::process_application_instances_txn (const arm4_shm_application_instance_t *entry_ptr)
{
	int status;
	application_instance_record_t application_instance;

	/* Update the record */
	application_instance.appl_id = entry_ptr->message.appl_id;
	application_instance.active = entry_ptr->message.active;

	status = save (db_application_instance_ptr, &(entry_ptr->message.app_handle), sizeof(entry_ptr->message.app_handle),
			&application_instance, sizeof (application_instance));
	return status;
}

int
BerkeleyDB::process_application_group_txn (const arm4_shm_property_t *group_ptr)
{
	int status;

	status = save (db_application_group_ptr, &(group_ptr->message.handle), sizeof(group_ptr->message.handle),
			group_ptr->message.property_name);
	return status;
}

int
BerkeleyDB::process_application_instance_id_txn (const arm4_shm_property_t *property_ptr)
{
	int status;

	status = save (db_application_instance_names_ptr, &(property_ptr->message.handle), sizeof(property_ptr->message.handle),
			property_ptr->message.property_name);
	return status;
}

int
BerkeleyDB::process_application_context_txn (const arm4_shm_context_property_t *property_ptr)
{
	int status;
	indexed_property_value_record_t context_value;

	context_value.index = property_ptr->message.property_index;
	strncpy (context_value.value,  property_ptr->message.property_name, ARM_PROPERTY_VALUE_MAX_LENGTH);

	status = save (db_application_context_values_ptr,
				   &(property_ptr->message.handle), sizeof(property_ptr->message.handle),
				   &context_value, sizeof (context_value));
	return status;
}

int
BerkeleyDB::process_application_address_txn (const arm4_shm_address_t *address_ptr)
{
	int status;
	application_address_record_t address_value;

	address_value.address_format = address_ptr->message.address_format;
	address_value.address_length = address_ptr->message.address_length;
	address_value.id_valid = address_ptr->message.id_valid;
	address_value.id = address_ptr->message.id;
	memcpy (address_value.address, address_ptr->message.address, ARM_PROPERTY_URI_MAX_LENGTH);

	status = save (db_application_address_ptr,
				   &(address_ptr->message.app_handle), sizeof(address_ptr->message.app_handle),
				   &address_value, sizeof (address_value));
	return status;
}

int
BerkeleyDB::process_transaction_instances_txn (const arm4_shm_transaction_instance_t *entry_ptr)
{
	int status;
	transaction_instance_record_t transaction_instance;

	memset (&(transaction_instance), 0, sizeof (transaction_instance));
	transaction_instance.appl_id = entry_ptr->message.appl_id;
	transaction_instance.tran_id = entry_ptr->message.tran_id;
	transaction_instance.app_handle = entry_ptr->message.app_handle;
	transaction_instance.arrival_time = entry_ptr->message.arrival_time;
	transaction_instance.stop_time = entry_ptr->message.stop_time;
	transaction_instance.response_time = entry_ptr->message.response_time;
	transaction_instance.blocked_time = entry_ptr->message.blocked_time;
	transaction_instance.tran_status = entry_ptr->message.tran_status;
	transaction_instance.active = entry_ptr->message.active;

	status = save (db_transaction_instance_ptr,
				   &(entry_ptr->message.tran_handle), sizeof(entry_ptr->message.tran_handle),
				   &transaction_instance, sizeof (transaction_instance));
	return status;
}

int
BerkeleyDB::process_transaction_context_txn (const arm4_shm_context_property_t *property_ptr)
{
	int status;
	indexed_property_value_record_t context_value;

	context_value.index = property_ptr->message.property_index;
	strncpy (context_value.value,  property_ptr->message.property_name, ARM_PROPERTY_VALUE_MAX_LENGTH);

	status = save (db_transaction_context_values_ptr,
				   &(property_ptr->message.handle), sizeof(property_ptr->message.handle),
				   &context_value, sizeof (context_value));
	return status;
}

int
BerkeleyDB::process_transaction_uri_txn (const arm4_shm_uri_t *property_ptr)
{
	int status;

	status = save (db_transaction_context_uri_ptr,
				   &(property_ptr->message.handle), sizeof(property_ptr->message.handle),
				   property_ptr->message.property_name);
	return status;
}

int
BerkeleyDB::process_transaction_metric_value_txn (const arm4_shm_metric_value_t *metric_ptr)
{
	int status;
	metric_value_record_t record;

	memcpy (&(record.value), &(metric_ptr->message.value), sizeof (record.value));
	strncpy (record.string32, metric_ptr->message.string32, ARM_METRIC_STRING32_MAX_CHARS);

	status = save (db_transaction_context_metric_value_ptr,
				   &(metric_ptr->message.handle), sizeof(metric_ptr->message.handle),
				   &record, sizeof (record));
	return status;
}

int
BerkeleyDB::process_transaction_user_txn (const arm4_shm_user_t *user_ptr)
{
	int status;
	user_record_t record;

	strncpy (record.name, user_ptr->message.name, ARM_NAME_MAX_CHARS);
	record.id_valid = user_ptr->message.id_valid;
	memcpy (&(record.id), &(user_ptr->message.id), sizeof (record.id));

	status = save (db_transaction_context_user_ptr,
				   &(user_ptr->message.handle), sizeof(user_ptr->message.handle),
				   &record, sizeof (record));
	return status;
}

int
BerkeleyDB::process_transaction_diag_txn (const arm4_shm_diag_t *diag_ptr)
{
	int status;

	status = save (db_transaction_context_diag_ptr,
				   &(diag_ptr->message.handle), sizeof(diag_ptr->message.handle),
				   diag_ptr->message.detail);
	return status;
}

int
BerkeleyDB::process_correlator_txn (const arm4_shm_correlator_t *correlator_ptr)
{
	int status;
	arm3_correlator_t *arm3_correlator_ptr;

	arm3_correlator_ptr = (arm3_correlator_t *) &correlator_ptr->message.parent_correlator;
	if (arm3_correlator_ptr->length != 0)
	{
		status = save (db_parent_correlator_ptr,
					   &(correlator_ptr->message.tran_handle), sizeof(correlator_ptr->message.tran_handle),
					   &correlator_ptr->message.parent_correlator, ntohs(arm3_correlator_ptr->length));
		if (status != 0)
			return status;
	}

	arm3_correlator_ptr = (arm3_correlator_t *) &correlator_ptr->message.current_correlator;
    if (arm3_correlator_ptr->length != 0)
	{
		status = save (db_current_correlator_ptr,
					   &(correlator_ptr->message.tran_handle), sizeof(correlator_ptr->message.tran_handle),
					   &correlator_ptr->message.current_correlator, ntohs(arm3_correlator_ptr->length));
	}

	return status;
}

int
BerkeleyDB::process_sequence_txn (const arm4_shm_db_sequence_t *sequence_ptr)
{
	int	status = ARM_FAILURE;
	Dbt	key, data;
	const char *sequence_name_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_sequence_ptr != NULL);

	sequence_name_ptr = Arm4dbCommon::getSequenceName (sequence_ptr->message.index);
	key.set_data ((char *) sequence_name_ptr);
	key.set_size (strlen (sequence_name_ptr) + 1);

	data.set_data ((void *) &(sequence_ptr->message.sequence));
	data.set_size (sizeof(sequence_ptr->message.sequence));

	/* Save in the database */
	/* Since we used transactions previously, we need to continue using them here */
	WRAP_EXCEPTION(status, db_sequence_ptr->put (NULL, &key, &data, DB_AUTO_COMMIT));
	if (status != 0)
	{
		/* Error */
		m_environmentPtr->err (status, "Unable to write sequence record");
		return status;
	}

	return 0;
}

int
BerkeleyDB::process_register_application_txn (const unsigned char *digest_ptr,
	const char *application_name_ptr,
	const arm_charset_t charset,
	arm_id_t *output_app_id,
	arm_boolean_t *is_new_ptr)
{
	int	status = ARM_FAILURE;
	DbTxn	*transaction_ptr;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;
	arm_id_t	appl_id;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_name_ptr != NULL);

	/* Initialize an invalid handle */
	memset (output_app_id, 0, sizeof (arm_id_t));
	memset (&appl_id, 0, sizeof (arm_id_t));
	*is_new_ptr = ARM_FALSE;

	WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to begin transaction");
		return status;
	}
#if 0
    transaction_ptr->set_name ("process_register_application_txn");
#endif
	
	/*
	 * See if an entry already exists by iterating through all applications with the same name
	 */
	WRAP_EXCEPTION(status, db_application_name_ptr->cursor (transaction_ptr, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		transaction_ptr->abort ();
		return status;
	}

	secondary_key.set_data ((char *) application_name_ptr);
	secondary_key.set_size (strlen (application_name_ptr) + 1);

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION(status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			appl_id =*((arm_id_t *) primary_key.get_data ());
			if (memcmp (application.md5_digest, digest_ptr, MD5_DIGEST_LENGTH) == 0)
			{
				/* We've found our application */
				cursor_ptr->close ();
				transaction_ptr->commit (0);
	
				/* Return the application id to the user */
				*output_app_id = appl_id;
				*is_new_ptr = ARM_FALSE;
				return status;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION(status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();
	if (status == 0)
	{
		appl_id =*((arm_id_t *) primary_key.get_data ());
	}
	if (status == DB_NOTFOUND)
	{
#ifdef DEBUG2
		m_environmentPtr->err (status, "Unable to get application record, creating");
#endif

		/* Get the new sequence number (application id) */
		Arm4dbCommon::getUuid (&appl_id);

		/* Use the new application id to create a new entry */
		primary_key.set_data (&appl_id);
		primary_key.set_size (sizeof (appl_id));
		
		strncpy (application.application_name, application_name_ptr, sizeof (application.application_name));
		memcpy (application.md5_digest, digest_ptr, MD5_DIGEST_LENGTH);
        application.charset = charset;
        application.default_trace_flags = arm4_shm_transaction_trace_all;
        application.default_trace_parameter = 0;
        application.default_appdex_threshold = ARM4_APPDEX_DEFAULT_RESPONSE_TIME;
		
		data.set_data (&application);
		data.set_size (sizeof (application));
		
		WRAP_EXCEPTION(status, db_application_ptr->put (transaction_ptr, &primary_key, &data, DB_NOOVERWRITE));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to write application record");
			transaction_ptr->abort ();
			return status;
		}
		*is_new_ptr = ARM_TRUE;
	}
	else if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to get application record");
		transaction_ptr->abort ();
		return status;
	}

	transaction_ptr->commit (0);

	/* Return the application id to the user */
	*output_app_id = appl_id;

	return 0;
}

arm_error_t
BerkeleyDB::process_register_application_identity_txn (const arm4_shm_register_application_identity_t *identity_ptr)
{
	int status;
	property_record_t identity_property;

	strncpy (identity_property.name, identity_ptr->message.name, ARM_PROPERTY_NAME_MAX_CHARS);
	strncpy (identity_property.value, identity_ptr->message.value, ARM_PROPERTY_VALUE_MAX_CHARS);

	status = save (db_application_identity_ptr,
				   &(identity_ptr->message.appid), sizeof(identity_ptr->message.appid),
				   &identity_property, sizeof (identity_property));
	return status;
}

int
BerkeleyDB::process_register_application_context_txn (const arm4_shm_register_application_context_t *context_ptr)
{
	int status;
	indexed_property_name_record_t identity_name;

	identity_name.index = context_ptr->message.index;
	strncpy (identity_name.name, context_ptr->message.name, ARM_PROPERTY_NAME_MAX_CHARS);

	status = save (db_application_context_names_ptr,
				   &(context_ptr->message.appid), sizeof(context_ptr->message.appid),
				   &identity_name, sizeof (identity_name));
	return status;
}

int
BerkeleyDB::process_register_transaction_txn (const arm4_shm_register_transaction_t *transaction_message_ptr,
	arm_id_t *output_tran_id,
	arm_boolean_t *is_new_ptr)
{
	int	status = ARM_FAILURE;
	DbTxn	*transaction_ptr;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;
	arm_id_t	tran_id;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_name_ptr != NULL);

	/* Initialize an invalid handle */
	memset (output_tran_id, 0, sizeof (arm_id_t));
	memset (&tran_id, 0, sizeof (arm_id_t));
	*is_new_ptr = ARM_FALSE;
	
	WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to begin transaction");
		return status;
	}
#if 0
    transaction_ptr->set_name ("process_register_transaction_txn");
#endif
		
	/*
	 * See if an entry already exists by iterating through all transactions with the same name
	 */
	WRAP_EXCEPTION(status, db_transaction_name_ptr->cursor (transaction_ptr, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		transaction_ptr->abort ();
		return status;
	}

	secondary_key.set_data ((char *) transaction_message_ptr->message.tran_name);
	secondary_key.set_size (strlen (transaction_message_ptr->message.tran_name) + 1);

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION(status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			tran_id =*((arm_id_t *) primary_key.get_data ());
			if (memcmp (transaction.md5_digest, transaction_message_ptr->message.digest, MD5_DIGEST_LENGTH) == 0)
			{
				/* We've found our application */
				cursor_ptr->close ();
				transaction_ptr->commit (0);
	
				/* Return the application id to the user */
				*output_tran_id = tran_id;
				*is_new_ptr = ARM_FALSE;
				return status;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION(status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();
	
	if (status == 0)
	{
		tran_id =*((arm_id_t *) primary_key.get_data ());
	}
	if (status == DB_NOTFOUND)
	{
#ifdef DEBUG2
		m_environmentPtr->err (status, "Unable to get transaction record, creating");
#endif

		/* Get the new sequence number (transaction id) */
		Arm4dbCommon::getUuid (&tran_id);

		/* Use the new transaction id to create a new entry */
		primary_key.set_data (&tran_id);
		primary_key.set_size (sizeof (tran_id));
		
		strncpy (transaction.transaction_name, transaction_message_ptr->message.tran_name, sizeof (transaction.transaction_name));
		memcpy (transaction.md5_digest, transaction_message_ptr->message.digest, MD5_DIGEST_LENGTH);
		transaction.appl_id = transaction_message_ptr->message.app_id;
		transaction.trace_flags = arm4_shm_transaction_trace_all;
		transaction.appdex_flags = ARM4_APPDEX_DEFAULT_FLAGS;
		transaction.trace_parameter = 0;
		transaction.appdex_threshold = ARM4_APPDEX_DEFAULT_RESPONSE_TIME;
		
		data.set_data (&transaction);
		data.set_size (sizeof (transaction));
		
		WRAP_EXCEPTION(status, db_transaction_ptr->put (transaction_ptr, &primary_key, &data, DB_NOOVERWRITE));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to write transaction record");
			transaction_ptr->abort ();
			return ARM_FAILURE;
		}
		*is_new_ptr = ARM_TRUE;
	}
	else if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to get application record");
		transaction_ptr->abort ();
		return status;
	}

	transaction_ptr->commit (0);
	
	// Initialize the in-memory values
	Arm4dbSharedMemory::setNewTransactionTraceInformation (transaction_message_ptr->message.transaction_index);

	// Return the application id to the user
	*output_tran_id = tran_id;

	return 0;
}

int
BerkeleyDB::process_register_transaction_identity_txn (const arm4_shm_register_application_identity_t *identity_ptr)
{
	int status;
	property_record_t identity_property;

	strncpy (identity_property.name, identity_ptr->message.name, ARM_PROPERTY_NAME_MAX_CHARS);
	strncpy (identity_property.value, identity_ptr->message.value, ARM_PROPERTY_VALUE_MAX_CHARS);

	status = save (db_transaction_identity_ptr,
				   &(identity_ptr->message.appid), sizeof(identity_ptr->message.appid),
				   &identity_property, sizeof (identity_property));
	return status;
}

int
BerkeleyDB::process_register_transaction_context_txn (const arm4_shm_register_application_context_t *context_ptr)
{
	int status;
	indexed_property_name_record_t identity_name;

	identity_name.index = context_ptr->message.index;
	strncpy (identity_name.name, context_ptr->message.name, ARM_PROPERTY_NAME_MAX_CHARS);

	status = save (db_transaction_context_names_ptr,
				   &(context_ptr->message.appid), sizeof(context_ptr->message.appid),
				   &identity_name, sizeof (identity_name));
	return status;
}

int
BerkeleyDB::process_register_transaction_uri_txn (const arm4_shm_register_transaction_uri_t *uri_ptr)
{
	int status;

	status = save (db_transaction_identity_uri_ptr,
				   &(uri_ptr->message.tran_id), sizeof(uri_ptr->message.tran_id),
				   uri_ptr->message.property_name);
	return status;
}

int
BerkeleyDB::process_register_transaction_metric_binding_txn (const arm4_shm_register_transaction_metric_t *metric_ptr)
{
	int status;
	arm_metric_binding_t binding;

	binding.slot = metric_ptr->message.slot;
	binding.id = metric_ptr->message.id;

	status = save (db_transaction_metric_binding_ptr,
				   &(metric_ptr->message.tran_id), sizeof(metric_ptr->message.tran_id),
				   &binding, sizeof (binding));
	return status;
}

int
BerkeleyDB::process_register_metric_txn (const unsigned char *digest_ptr,
	const char *metric_name_ptr,
	const arm_id_t *app_id,
    const char *metric_units_ptr,
    const arm_metric_format_t metric_format,
    const arm_metric_usage_t metric_usage,
	arm_id_t *output_metric_id)
{
	int	status = ARM_FAILURE;
	DbTxn	*transaction_ptr;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	metric_record_t metric;
	arm_id_t	metric_id;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_metric_name_ptr != NULL);

	/* Initialize an invalid handle */
	memset (output_metric_id, 0, sizeof (arm_id_t));
	memset (&metric_id, 0, sizeof (arm_id_t));
	
	WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to begin transaction");
		return status;
	}
#if 0
    transaction_ptr->set_name ("process_register_metric_txn");
#endif
		
	/*
	 * See if an entry already exists by iterating through all metrics with the same name
	 */
	WRAP_EXCEPTION(status, db_metric_name_ptr->cursor (transaction_ptr, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		transaction_ptr->abort ();
		return status;
	}

	secondary_key.set_data ((char *) metric_name_ptr);
	secondary_key.set_size (Arm4dbCommon::stringLength (metric_name_ptr, ARM_NAME_MAX_CHARS));

	data.set_data (&metric);
	data.set_size (sizeof(metric));
	data.set_ulen (sizeof(metric));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION(status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			metric_id =*((arm_id_t *) primary_key.get_data ());
			if (memcmp (metric.md5_digest, digest_ptr, MD5_DIGEST_LENGTH) == 0)
			{
				/* We've found our application */
				cursor_ptr->close ();
				transaction_ptr->commit (0);
	
				/* Return the application id to the user */
				*output_metric_id = metric_id;
				return status;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION(status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();
	
	if (status == 0)
	{
		metric_id =*((arm_id_t *) primary_key.get_data ());
	}
	if (status == DB_NOTFOUND)
	{
#ifdef DEBUG2
		m_environmentPtr->err (status, "Unable to get metric record, creating");
#endif

		/* Get the new sequence number (metric id) */
		Arm4dbCommon::getUuid (&metric_id);

		/* Use the new metric id to create a new entry */
		primary_key.set_data (&metric_id);
		primary_key.set_size (sizeof (metric_id));
		
		strncpy (metric.metric_name, metric_name_ptr, ARM_NAME_MAX_CHARS);
        strncpy (metric.metric_units, metric_units_ptr, ARM_NAME_MAX_CHARS);
		memcpy (metric.md5_digest, digest_ptr, MD5_DIGEST_LENGTH);
		metric.appl_id = *app_id;
        metric.metric_format = metric_format;
        metric.metric_usage = metric_usage;
		
		data.set_data (&metric);
		data.set_size (sizeof (metric));
		
		WRAP_EXCEPTION(status, db_metric_ptr->put (transaction_ptr, &primary_key, &data, DB_NOOVERWRITE));
		if (status != 0)
		{
			m_environmentPtr->err (status, "Unable to write metric record");
			transaction_ptr->abort ();
			return ARM_FAILURE;
		}
	}
	else if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to get application record");
		transaction_ptr->abort ();
		return status;
	}

	transaction_ptr->commit (0);

	/* Return the application id to the user */
	*output_metric_id = metric_id;

	return 0;
}

int
BerkeleyDB::process_trace_stats_txn (arm_time_t sample_time, arm4db_transaction_trace_stats_t *stats_ptr)
{
	int status;

	status = save (db_trace_stat_ptr,
				   &(sample_time), sizeof(sample_time),
				   stats_ptr, sizeof (*stats_ptr));
	return status;
}

int
BerkeleyDB::initialize_sequence (int index, arm4_shm_header_t *header_ptr)
{
	int	status = ARM_FAILURE;
	DbTxn		*transaction_ptr;
	Dbt			key, data;
	const char	*sequence_name_ptr;
	arm_int64_t *sequence_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_sequence_ptr != NULL);

	/* Get a unique sequence number for the database */
	if (index >= arm4_shm_sequence_max)
		return ARM_FAILURE;

	sequence_ptr = &(header_ptr->sequence_table [index]);
	sequence_name_ptr = Arm4dbCommon::getSequenceName (index);

	/* Create a local transaction */
	WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create local transaction");
		return ARM_FAILURE;
	}
#if 0
    transaction_ptr->set_name ("initialize_sequence");
#endif

	key.set_data ((char *) sequence_name_ptr);
	key.set_size (strlen (sequence_name_ptr) + 1);

	data.set_data (sequence_ptr);
	data.set_size (sizeof(*sequence_ptr));
	data.set_ulen (sizeof(*sequence_ptr));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION(status, db_sequence_ptr->get (transaction_ptr, &key, &data, DB_RMW));
	if (status == 0)
	{
		/*
		 * Increment the sequence number in case we crashed since the last increment
		 */

		*sequence_ptr += ARM4_SHM_SEQUENCE_STEP;

		/* Check for a wrap - don't allow the 0 value */
		if (*sequence_ptr < 1)
			*sequence_ptr = 1;
	}
	else if (status == DB_NOTFOUND)
	{
		/* The initial value is 1 */
		*sequence_ptr = 1;
	}
	else
	{
		/* Error */
		m_environmentPtr->err (status, "Unable to get sequence record");
		transaction_ptr->abort ();
		return ARM_FAILURE;
	}
	if (*sequence_ptr <= 0)
		*sequence_ptr = 1; /* Wrap to 1 */

#ifdef DEBUG2
	fprintf (stderr, "Sequence number %lld\n", *sequence_ptr);
#endif

	/* Save in the database */
	WRAP_EXCEPTION(status, db_sequence_ptr->put (transaction_ptr, &key, &data, 0));
	if (status != 0)
	{
		/* Error */
		m_environmentPtr->err (status, "Unable to write sequence record");
		transaction_ptr->abort ();
		return ARM_FAILURE;
	}


	transaction_ptr->commit (0);
	return ARM_SUCCESS;
}

/*
 *
 *	External interfaces
 *
 */

arm_error_t
BerkeleyDB::initializeSequences (arm4_shm_header_t *header_ptr)
{
	int	status = ARM_FAILURE;
	int			i;

    Arm4dbWriteLock lock;

	/* Ensure the databases have been opened */
	WRAP_EXCEPTION(status, openDatabase (&db_sequence_ptr, SEQUENCE_DB_NAME, 0));
	if (status != 0)
    {
        lock.unlock ();
		return ARM_FAILURE;
    }

	for (i = 0; i < arm4_shm_sequence_max; i++)
	{
		status = initialize_sequence (i, header_ptr);
		if (status != ARM_SUCCESS)
        {
            lock.unlock ();
			return status;
        }
	}
    lock.unlock ();

	return ARM_SUCCESS;
}

void
BerkeleyDB::checkpoint (void)
{
	/* Checkpoint the database. This is something that is normally called from a thread on a regular basis */
	/* We force a checkpoint every time. */

    /* Ensure we have database access */
    Arm4dbLock lock;

	Arm4dbSharedMemory shm;
	arm_boolean_t current_null = shm.setNullCollector (ARM_TRUE);

    if (m_environmentPtr == NULL)
    {
		printf ("Done checkpoint\n");

		/* Resume collection as before */
		shm.setNullCollector (current_null);
        lock.unlock ();
        return;
    }

    try
    {
        m_environmentPtr->txn_checkpoint(0, 0, DB_FORCE);
    }
    catch (DbException e)
    {
        m_environmentPtr->err(e.get_errno (), "PANIC: Checkpoint error");

		throw new Arm4dbException ("PANIC(%d): Checkpoint error", e.get_errno ());
    }

    // Remove any unused log files. This may prevent catastrophic recovery
    try
    {
        m_environmentPtr->log_archive(NULL, DB_ARCH_REMOVE);
    }
    catch (DbException e)
    {
        m_environmentPtr->err(e.get_errno (), "PANIC: Log archive error");

		throw new Arm4dbException ("PANIC(%d): Checkpoint error", e.get_errno ());
    }

	/* Resume collection as before */
	shm.setNullCollector (current_null);
    lock.unlock ();
}

void *
BerkeleyDB::getStats (void)
{
    DB_TXN_STAT *statp;

    DB_ASSERT (m_environmentPtr != NULL);

    /* Ensure we have database access */
    Arm4dbLock lock;

    /* The statp structure is dynamically allocated and should be freed when done */
    statp = NULL;
	m_environmentPtr->txn_stat(&statp, 0);

    lock.unlock ();

    return statp;
}


arm_error_t
BerkeleyDB::processApplicationInstances (const arm4_shm_application_instance_t *app_instance_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_instance_ptr, APPLICATION_INSTANCE_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_application_instances_txn (app_instance_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processApplicationGroup (const arm4_shm_property_t *group_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_group_ptr, APPLICATION_GROUP_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_application_group_txn (group_ptr);

    lock.unlock ();

	return 0;
}

arm_error_t
BerkeleyDB::processApplicationInstanceId (const arm4_shm_property_t *property_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_instance_names_ptr, APPLICATION_INSTANCE_NAMES_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();

        return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_application_instance_id_txn (property_ptr);

    lock.unlock ();

	return 0;
}

arm_error_t
BerkeleyDB::processApplicationContext (const arm4_shm_context_property_t *property_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_context_values_ptr, APPLICATION_CONTEXT_VALUES_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_application_context_txn (property_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processApplicationAddress (const arm4_shm_address_t *address_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_address_ptr, APPLICATION_ADDRESS_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_application_address_txn (address_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processTransactionInstances (const arm4_shm_transaction_instance_t *tran_instance_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openTransactionInstanceDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_transaction_instances_txn (tran_instance_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processTransactionContext (const arm4_shm_context_property_t *property_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_context_values_ptr, TRANSACTION_CONTEXT_VALUES_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_transaction_context_txn (property_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processTransactionUri (const arm4_shm_uri_t *uri_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_context_uri_ptr, TRANSACTION_CONTEXT_URI_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_transaction_uri_txn (uri_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processTransactionMetricValue (const arm4_shm_metric_value_t *metric_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_context_metric_value_ptr, TRANSACTION_CONTEXT_METRIC_VALUE_DB_NAME, DB_DUPSORT);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_transaction_metric_value_txn (metric_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processTransactionUser (const arm4_shm_user_t *user_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_context_user_ptr, TRANSACTION_CONTEXT_USER_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_transaction_user_txn (user_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processTransactionDiag (const arm4_shm_diag_t *diag_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_context_diag_ptr, TRANSACTION_CONTEXT_DIAG_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_transaction_diag_txn (diag_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processCorrelators (const arm4_shm_correlator_t *correlator_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_parent_correlator_ptr, CORRELATOR_PARENT_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }
	status = openDatabase (&db_current_correlator_ptr, CORRELATOR_CURRENT_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_correlator_txn (correlator_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processSequence (const arm4_shm_db_sequence_t *sequence_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_sequence_ptr, SEQUENCE_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_sequence_txn (sequence_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterApplication (const arm4_shm_register_application_t *application_ptr,
	arm_id_t *appid_ptr,
	arm_boolean_t *is_new_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openApplicationDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	status = process_register_application_txn ((application_ptr->message.digest),
									 (application_ptr->message.app_name),
									 application_ptr->message.charset,
									 appid_ptr,
									 is_new_ptr);
    lock.unlock ();

    return status;
}

arm_error_t
BerkeleyDB::processRegisterApplicationIdentity (const arm4_shm_register_application_identity_t *identity_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_identity_ptr, APPLICATION_IDENTITY_DB_NAME, DB_DUPSORT);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_register_application_identity_txn (identity_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterApplicationContext (const arm4_shm_register_application_context_t *context_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_application_context_names_ptr, APPLICATION_CONTEXT_NAMES_DB_NAME, DB_DUPSORT);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_register_application_context_txn (context_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterTransaction (const arm4_shm_register_transaction_t *transaction_ptr,
	arm_id_t *tranid_ptr,
	arm_boolean_t *is_new_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openTransactionDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	status = process_register_transaction_txn (transaction_ptr,
									 tranid_ptr,
									 is_new_ptr);

    lock.unlock ();

    return status;
}


arm_error_t
BerkeleyDB::processRegisterTransactionIdentity (const arm4_shm_register_application_identity_t *identity_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_identity_ptr, TRANSACTION_IDENTITY_DB_NAME, DB_DUPSORT);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_register_transaction_identity_txn (identity_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterTransactionContext (const arm4_shm_register_application_context_t *context_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_context_names_ptr, TRANSACTION_CONTEXT_NAMES_DB_NAME, DB_DUPSORT);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_register_transaction_context_txn (context_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterTransactionUri (const arm4_shm_register_transaction_uri_t *uri_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_identity_uri_ptr, TRANSACTION_IDENTITY_URI_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_register_transaction_uri_txn (uri_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterTransactionMetricBinding (const arm4_shm_register_transaction_metric_t *metric_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_transaction_metric_binding_ptr, TRANSACTION_METRIC_BINDING_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	/* Ignore the status since our behaviour doesn't change */
	process_register_transaction_metric_binding_txn (metric_ptr);

    lock.unlock ();

    return 0;
}

arm_error_t
BerkeleyDB::processRegisterMetric (arm4_shm_register_metric_t *metric_ptr,
	arm_id_t *metricid_ptr)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openMetricDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	status = process_register_metric_txn ((metric_ptr->message.digest),
									 (metric_ptr->message.metric_name),
									 &(metric_ptr->message.app_id),
                                     (metric_ptr->message.metric_unit),
                                     (metric_ptr->message.metric_format),
                                     (metric_ptr->message.metric_usage),
									 metricid_ptr);
    lock.unlock ();

    return status;
}

arm_error_t
BerkeleyDB::processTraceStats (arm_time_t sample_time, arm4db_transaction_trace_stats_t *stats_ptr, int transaction_count)
{
	int status;

    /* Ensure we have database access */
    Arm4dbLock lock;

	status = openDatabase (&db_trace_stat_ptr, TRACE_STAT_DB_NAME, 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	for (int i = 0; i < transaction_count; i++)
	{
		if (stats_ptr [i].transaction_counter > 0)
			status = process_trace_stats_txn (sample_time, &(stats_ptr [i]));
	}
	
    lock.unlock ();

    return status;
}

std::list<arm_id_t> *
BerkeleyDB::getApplicationList (void)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;

	// Create the list
	std::list<arm_id_t> *list_ptr = new std::list<arm_id_t> ();

    Arm4dbLock lock;
	
	status = openApplicationDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return list_ptr;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_application_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return list_ptr;
	}

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			list_ptr->push_back (*((arm_id_t *) primary_key.get_data ()));
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return list_ptr;
}

int
BerkeleyDB::getApplicationCount (void)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;

    Arm4dbLock lock;
	
	status = openApplicationDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return 0;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_application_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return 0;
	}

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	int count = 0;
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			count++;
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return count;
}

std::list<arm_id_t> *
BerkeleyDB::getTransactionList (void)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;

	// Create the list
	std::list<arm_id_t> *list_ptr = new std::list<arm_id_t> ();

    Arm4dbLock lock;
	
	status = openTransactionDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return list_ptr;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return list_ptr;
	}
	
	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			list_ptr->push_back (*((arm_id_t *) primary_key.get_data ()));
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return list_ptr;
}

int
BerkeleyDB::getTransactionCount (void)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;

    Arm4dbLock lock;
	
	status = openTransactionDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return 0;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return 0;
	}
	
	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	int count = 0;
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			count++;
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return count;
}

std::list<arm_id_t> *
BerkeleyDB::getTransactionList (const arm_id_t &app_id)
{
	int	status;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;

	// Create the list
	std::list<arm_id_t> *list_ptr = new std::list<arm_id_t> ();

    Arm4dbLock lock;
	
	status = openTransactionDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return list_ptr;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_appid_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_appid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return list_ptr;
	}
	
	secondary_key.set_data ((void *) &app_id);
	secondary_key.set_size (sizeof (app_id));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			list_ptr->push_back (*((arm_id_t *) primary_key.get_data ()));
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return list_ptr;
}

std::list<arm_tran_start_handle_t> *
BerkeleyDB::getTransactionInstanceList (const arm_id_t &tran_id)
{
	int	status;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_instance_record_t transaction;

	// Create the list
	std::list<arm_tran_start_handle_t> *list_ptr = new std::list<arm_tran_start_handle_t> ();

    Arm4dbLock lock;
	
	status = openTransactionDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return list_ptr;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_instance_tranid_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_instance_tranid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return list_ptr;
	}
	
	secondary_key.set_data ((void *) &tran_id);
	secondary_key.set_size (sizeof (tran_id));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			list_ptr->push_back (*((arm_tran_start_handle_t *) primary_key.get_data ()));
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return list_ptr;
}

int
BerkeleyDB::getTransactionCount (const arm_id_t &app_id)
{
	int	status;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;

    Arm4dbLock lock;
	
	status = openTransactionDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return 0;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_appid_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_appid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return 0;
	}
	
	secondary_key.set_data ((void *) &app_id);
	secondary_key.set_size (sizeof (app_id));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	int count = 0;
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			count++;
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();

    return count;
}

arm_error_t
BerkeleyDB::getApplicationInformation (const arm_id_t &app_id, arm4db_application_summary_t *summary_ptr)
{
	int	status;
	Dbt	primary_key, data;
	application_record_t application;
	property_record_t identity_property;
	indexed_property_name_record_t identity_name;
	Dbc *cursor_ptr;

    Arm4dbLock lock;
	
	status = openApplicationDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);
	
    primary_key.set_data ((void *) &app_id);
    primary_key.set_size (sizeof (app_id));

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, db_application_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
	{
		summary_ptr->appl_id = app_id;
		strncpy (summary_ptr->application_name, application.application_name, ARM_NAME_MAX_LENGTH);
		summary_ptr->charset = application.charset;
		summary_ptr->default_trace_flags = application.default_trace_flags;
		summary_ptr->default_trace_parameter = application.default_trace_parameter;
		summary_ptr->default_appdex_threshold = application.default_appdex_threshold;
	}

	//	identity_properties
	summary_ptr->identity_properties.clear ();

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_application_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		lock.unlock ();
		return status;
	}

	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			property_record_t *identity_ptr = new property_record_t;
			memcpy (identity_ptr, &identity_property, sizeof (property_record_t));
			summary_ptr->identity_properties.push_back (identity_ptr);

			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

	//	context_names
	summary_ptr->context_names.clear ();

	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_application_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		lock.unlock ();
		return status;
	}

	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			indexed_property_name_record_t *identity_ptr = new indexed_property_name_record_t;
			memcpy (identity_ptr, &identity_name, sizeof (indexed_property_name_record_t));
			summary_ptr->context_names.push_back (identity_ptr);

			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

    lock.unlock ();
    
    return status;
}

arm_error_t
BerkeleyDB::getTransactionInformation (const arm_id_t &tran_id, arm4db_transaction_summary_t *summary_ptr)
{
	int	status;
	Dbt	primary_key, data;
	transaction_record_t transaction;
	property_record_t identity_property;
	indexed_property_name_record_t identity_name;
	arm_char_t uri [ARM_PROPERTY_URI_MAX_LENGTH];
	Dbc *cursor_ptr;

    Arm4dbLock lock;
	
	status = openTransactionDatabase (/*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }
	status = openDatabase (&db_transaction_identity_uri_ptr, TRANSACTION_IDENTITY_URI_DB_NAME, /*DB_RDONLY*/ 0);
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);
	
    primary_key.set_data ((void *) &tran_id);
    primary_key.set_size (sizeof (tran_id));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Get the basic transaction data */
	WRAP_EXCEPTION (status, db_transaction_ptr->get (NULL, &primary_key, &data, 0));
	if (status != 0)
    {
        lock.unlock ();
		return status;
    }

	summary_ptr->tran_id = tran_id;
	summary_ptr->appl_id = transaction.appl_id;
	strncpy (summary_ptr->transaction_name, transaction.transaction_name, ARM_NAME_MAX_LENGTH);
	summary_ptr->trace_flags = transaction.trace_flags;
	summary_ptr->appdex_flags = transaction.appdex_flags;
	summary_ptr->trace_parameter = transaction.trace_parameter;
	summary_ptr->appdex_threshold = transaction.appdex_threshold;
	summary_ptr->identity_properties.clear ();
	summary_ptr->context_names.clear ();
	
	//	identity_properties
	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_transaction_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return status;
	}
	
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			property_record_t *identity_ptr = new property_record_t;
			memcpy (identity_ptr, &identity_property, sizeof (property_record_t));
			summary_ptr->identity_properties.push_back (identity_ptr);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

	//	context_names
	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_transaction_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        lock.unlock ();
		return status;
	}
	
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			indexed_property_name_record_t *identity_ptr = new indexed_property_name_record_t;
			memcpy (identity_ptr, &identity_name, sizeof (indexed_property_name_record_t));
			summary_ptr->context_names.push_back (identity_ptr);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

	// Get the URI data
	data.set_data (uri);
	data.set_size (sizeof(uri));
	data.set_ulen (sizeof(uri));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_transaction_identity_uri_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
		strncpy (summary_ptr->uri, uri, ARM_PROPERTY_URI_MAX_LENGTH);
	else if (status == DB_NOTFOUND) // This is optional data
	{
		summary_ptr->uri [0] = '\0';
		status = 0;
	}
	else
    {
		printf ("status = %d\n", status);
        lock.unlock ();
		return status;
    }

    lock.unlock ();
    
    return status;
}

const arm_correlator_t *
BerkeleyDB::getCorrelator (arm_tran_start_handle_t tran_handle, arm_correlator_t *correlator_ptr)
{
	int	status;
	Dbt	primary_key, data;

    Arm4dbLock lock;

	status = openDatabase (&db_current_correlator_ptr, CORRELATOR_CURRENT_DB_NAME, /*DB_RDONLY*/ 0);
	if (status != 0)
	{
		std::cerr << "Unable to open correlator database, status = " << status << "\n";
		lock.unlock ();
		return NULL;
	}

	DB_ASSERT (m_environmentPtr != NULL);
	DB_ASSERT (db_current_correlator_ptr != NULL);

	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (tran_handle));

	data.set_data (correlator_ptr);
	data.set_size (sizeof(arm_correlator_t));
	data.set_ulen (sizeof(arm_correlator_t));
	data.set_flags (DB_DBT_USERMEM);

	/* Get the basic transaction data */
	WRAP_EXCEPTION (status, db_current_correlator_ptr->get (NULL, &primary_key, &data, 0));
	if (status != 0)
	{
		std::cerr << "Unable to get correlator, status = " << status << "\n";
		lock.unlock ();
		return NULL;
	}

	lock.unlock ();

	return correlator_ptr;
}

const arm_correlator_t *
BerkeleyDB::getParentCorrelator (arm_tran_start_handle_t tran_handle, arm_correlator_t *correlator_ptr)
{
	int	status;
	Dbt	primary_key, data;

    Arm4dbLock lock;

	status = openDatabase (&db_parent_correlator_ptr, CORRELATOR_CURRENT_DB_NAME, /*DB_RDONLY*/ 0);
	if (status != 0)
	{
		lock.unlock ();
		return NULL;
	}

	DB_ASSERT (m_environmentPtr != NULL);
	DB_ASSERT (db_parent_correlator_ptr != NULL);

	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (tran_handle));

	data.set_data (correlator_ptr);
	data.set_size (sizeof(arm_correlator_t));
	data.set_ulen (sizeof(arm_correlator_t));
	data.set_flags (DB_DBT_USERMEM);

	/* Get the basic transaction data */
	WRAP_EXCEPTION (status, db_parent_correlator_ptr->get (NULL, &primary_key, &data, 0));
	if (status != 0)
	{
		lock.unlock ();
		return NULL;
	}

	lock.unlock ();

	return correlator_ptr;
}

void
BerkeleyDB::setTransactionTraceInformation (const arm_id_t &tran_id, const arm4db_transaction_summary_t &summary)
{
	int	status;
	DbTxn *transaction_ptr;
	Dbt	primary_key, data;
	transaction_record_t transaction;

    Arm4dbLock lock;

	status = openTransactionDatabase (0);
	if (status != 0)
    {
        lock.unlock ();
		return;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);

	WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to begin transaction");
		return;
	}
	
    primary_key.set_data ((void *) &tran_id);
    primary_key.set_size (sizeof (tran_id));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Get the basic transaction data */
	WRAP_EXCEPTION (status, db_transaction_ptr->get (transaction_ptr, &primary_key, &data, 0));
	if (status != 0)
    {
        lock.unlock ();
		return;
    }
	
	// update the trace information
	transaction.trace_flags = summary.trace_flags;
	transaction.trace_parameter = summary.trace_parameter;
	
	WRAP_EXCEPTION(status, db_transaction_ptr->put (transaction_ptr, &primary_key, &data, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to write transaction record");
		transaction_ptr->abort ();
		return;
	}

	transaction_ptr->commit (0);
}

void
BerkeleyDB::saveApplicationDefinition (const arm_id_t &appl_id, const application_record_t &application)
{
	save (db_application_ptr, &appl_id, sizeof(appl_id), &application, sizeof (application));
}

void
BerkeleyDB::saveApplicationIdentity (const arm_id_t &appl_id, const property_record_t &property)
{
	save (db_application_identity_ptr, &appl_id, sizeof(appl_id), &property, sizeof (property));
}

void
BerkeleyDB::saveApplicationContext (const arm_id_t &appl_id, const indexed_property_name_record_t &property)
{
	save (db_application_context_names_ptr, &appl_id, sizeof(appl_id), &property, sizeof (property));
}

void
BerkeleyDB::saveMetricDefinition (const arm_id_t &metric_id, const metric_record_t &metric)
{
	save (db_metric_ptr, &metric_id, sizeof(metric_id), &metric, sizeof (metric));
}

void
BerkeleyDB::saveTransactionDefinition (const arm_id_t &tran_id, const transaction_record_t &transaction)
{
	save (db_transaction_ptr, &tran_id, sizeof(tran_id), &transaction, sizeof (transaction));
}

void
BerkeleyDB::saveTransactionIdentity (const arm_id_t &tran_id, const property_record_t &property)
{
	save (db_transaction_identity_ptr, &tran_id, sizeof(tran_id), &property, sizeof (property));
}

void
BerkeleyDB::saveTransactionContext (const arm_id_t &tran_id, const indexed_property_name_record_t &property)
{
	save (db_transaction_context_names_ptr, &tran_id, sizeof(tran_id), &property, sizeof (property));
}

void
BerkeleyDB::saveTransactionUri (const arm_id_t &tran_id, const char *uri_ptr)
{
	save (db_transaction_identity_uri_ptr, &tran_id, sizeof(tran_id), uri_ptr);
}

void
BerkeleyDB::saveTransactionMetric (const arm_id_t &tran_id, const arm_metric_binding_t &binding)
{
	save (db_transaction_metric_binding_ptr, &tran_id, sizeof(tran_id), &binding, sizeof (binding));
}

void
BerkeleyDB::saveApplicationInstance (const arm_app_start_handle_t app_handle, const application_instance_record_t &application)
{
	save (db_application_instance_ptr, &app_handle, sizeof(app_handle), &application, sizeof (application));
}

void
BerkeleyDB::saveAppInstanceContextValue (const arm_app_start_handle_t app_handle, const indexed_property_value_record_t &property)
{
	save (db_application_group_ptr, &app_handle, sizeof(app_handle), &property, sizeof (property));
}

void
BerkeleyDB::saveAppInstanceGroup (const arm_app_start_handle_t app_handle, const char *group_ptr)
{
	save (db_application_group_ptr, &app_handle, sizeof(app_handle), group_ptr);
}

void
BerkeleyDB::saveAppInstanceName (const arm_app_start_handle_t app_handle, const char *name_ptr)
{
	save (db_application_instance_names_ptr, &app_handle, sizeof(app_handle), name_ptr);
}

void
BerkeleyDB::saveAppInstanceAddress (const arm_app_start_handle_t app_handle, const application_address_record_t &address_value)
{
	save (db_application_address_ptr, &app_handle, sizeof(app_handle), &address_value, sizeof (address_value));
}

void
BerkeleyDB::saveTransactionInstance (const arm_tran_start_handle_t tran_handle, const transaction_instance_record_t &transaction)
{
	save (db_transaction_instance_ptr, &tran_handle, sizeof(tran_handle), &transaction, sizeof (transaction));
}

void
BerkeleyDB::saveTranInstanceContextValue (const arm_tran_start_handle_t tran_handle, const indexed_property_value_record_t &property)
{
	save (db_transaction_context_values_ptr, &tran_handle, sizeof(tran_handle), &property, sizeof (property));
}

void
BerkeleyDB::saveTranInstanceMetricValue (const arm_tran_start_handle_t tran_handle, const metric_value_record_t &metric)
{
	save (db_transaction_context_metric_value_ptr, &tran_handle, sizeof(tran_handle), &metric, sizeof (metric));
}

void
BerkeleyDB::saveTranInstanceUri (const arm_tran_start_handle_t tran_handle, const char *uri_ptr)
{
	save (db_transaction_context_uri_ptr, &tran_handle, sizeof(tran_handle), uri_ptr);
}

void
BerkeleyDB::saveTranInstanceUser (const arm_tran_start_handle_t tran_handle, const user_record_t &user)
{
	save (db_transaction_context_user_ptr, &tran_handle, sizeof(tran_handle), &user, sizeof (user));
}

void
BerkeleyDB::saveTranInstanceDiag (const arm_tran_start_handle_t tran_handle, const char *diag_ptr)
{
	save (db_transaction_context_diag_ptr, &tran_handle, sizeof(tran_handle), diag_ptr);
}

void
BerkeleyDB::saveTranCorrelator (const arm_tran_start_handle_t tran_handle, const arm_correlator_t *correlator_ptr)
{
	save (db_current_correlator_ptr, &tran_handle, sizeof(tran_handle), correlator_ptr, Arm4dbCommon::correlatorSize (correlator_ptr));
}

void
BerkeleyDB::saveTranParentCorrelator (const arm_tran_start_handle_t tran_handle, const arm_correlator_t *correlator_ptr)
{
	save (db_parent_correlator_ptr, &tran_handle, sizeof(tran_handle), correlator_ptr, Arm4dbCommon::correlatorSize (correlator_ptr));
}

void
BerkeleyDB::saveTraceStats (const arm_time_t sample_time, const arm4db_transaction_trace_stats_t &trace_stats)
{
	save (db_trace_stat_ptr, &sample_time, sizeof(sample_time), &trace_stats, sizeof (trace_stats));
}

int
BerkeleyDB::save (Db *db_ptr, const void *key_ptr, size_t key_size, const char *data_ptr)
{
	return save (db_ptr, key_ptr, key_size, data_ptr, strlen (data_ptr) + 1);
}

int
BerkeleyDB::save (Db *db_ptr, const void *key_ptr, size_t key_size, const void *data_ptr, size_t data_size)
{
	int	status = ARM_FAILURE;
	DbTxn	*transaction_ptr;
	Dbt	primary_key, data;

	DB_ASSERT (m_environmentPtr != NULL);
	DB_ASSERT (db_ptr != NULL);

	WRAP_EXCEPTION(status, m_environmentPtr->txn_begin (NULL, &transaction_ptr, DB_TXN_SYNC));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to begin transaction");
		return status;
	}

	primary_key.set_data ((void *) key_ptr);
	primary_key.set_size (key_size);

	data.set_data ((void *) data_ptr);
	data.set_size (data_size);

	WRAP_EXCEPTION(status, db_ptr->put (transaction_ptr, &primary_key, &data, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to write record");
		transaction_ptr->abort ();
		return status;
	}
	transaction_ptr->commit (0);

	return 0;
}
