/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>
#include <math.h>
#include <errno.h>
#include <time.h>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "Logger.h"
#include "BerkeleyDB.h"
#include "Arm4dbSharedMemory.h"
#include "Arm4dbCommon.h"

arm_error_t
BerkeleyDB::copyDatabases (void)
{
	DIR *dirp;
	struct dirent *dp;
    struct stat stat_buffer;
	int status;
	char old_path [2048];
	char new_path [2048];
    char backup_dir [2048];
    time_t current_time;
    struct tm *time_ptr;
	Arm4dbSharedMemory shm;

    current_time = time (NULL);
    time_ptr = localtime (&current_time);

    /* Get the backup directory name */
    sprintf (backup_dir, "%s/%04d%02d%02d%02d%02d%02d",
             getConfig ().getDbBackupC (),
             time_ptr->tm_year + 1900,
             time_ptr->tm_mon + 1,
             time_ptr->tm_mday,
             time_ptr->tm_hour,
             time_ptr->tm_min,
             time_ptr->tm_sec
             );

	/* If the database already exists, we've done a backup very recently so exit */
	if (access (backup_dir, F_OK) == 0)
		return ARM_FAILURE;

	/* Create the backup directory */
    if (Arm4dbCommon::createPath (backup_dir) != ARM_SUCCESS)
    {
        Logger::getLogger().error ("Unable to create backup directory %s", backup_dir);
        return ARM_FAILURE;
    }

	/* Copy all files in the database directory */
	if ((dirp = opendir(getConfig ().getDbHomeC ())) == NULL)
	{
        Logger::getLogger().error ("Can't open %s: %s\n", getConfig ().getDbHomeC (), strerror (errno));
		return ARM_FAILURE;
	}

	for (dp = readdir(dirp);
		  dp != NULL;
		  dp = readdir(dirp))
	{
        snprintf (old_path, sizeof (old_path), "%s/%s", getConfig ().getDbHomeC (), dp->d_name);

        /* If this is a directory, then skip it */
        if (stat (old_path,  &stat_buffer) == 0)
        {
            if (!(S_ISDIR(stat_buffer.st_mode)))
            {
        		if (!((strcmp (dp->d_name, "arm4.shm") == 0) ||
        			  (strcmp (dp->d_name, "arm4.mq") == 0)))
        		{
        			snprintf (new_path, sizeof (new_path), "%s/%s", backup_dir, dp->d_name);
        			status = link (old_path, new_path);

        			if (status != 0)
        			{
                        Logger::getLogger().error ("Unable to copy %s to %s: %s", old_path, new_path, strerror (errno));
						closedir (dirp);
        				return ARM_FAILURE;
        			}
        		}
            }
        }
	}

	closedir(dirp);
    return ARM_SUCCESS;
}

void
BerkeleyDB::destroyInstanceDatabase (const char *name_ptr)
{
    int status;
    char path [2048];
	Db *db_ptr;

    // Ensure our environment is open
    status = openEnvironment (0);
	if (status != 0)
		return;

	// We need a new database handle to remove one. How odd is that?
	try
	{
		db_ptr = new Db (m_environmentPtr, 0);
	}
	catch (DbException e)
	{
		m_environmentPtr->err (e.get_errno (), "Unable to create database");
		return;
	}

    /* Get the full path */
    sprintf (path, "%s/%s",  getConfig ().getDbHomeC (), name_ptr);

	WRAP_EXCEPTION(status, db_ptr->remove (path, NULL, 0))
	if (status != 0)
	{
		char error [2048];

		sprintf (error, "Unable to remove database '%s'", path);
		m_environmentPtr->err (status, error);
	}

	delete db_ptr;
}

void
BerkeleyDB::destroyInstanceDatabases (void)
{
    destroyInstanceDatabase (APPLICATION_INSTANCE_DB_NAME);
    destroyInstanceDatabase (APPLICATION_INSTANCE_APP_DB_NAME);
    destroyInstanceDatabase (APPLICATION_GROUP_DB_NAME);
    destroyInstanceDatabase (APPLICATION_INSTANCE_NAMES_DB_NAME);
    destroyInstanceDatabase (APPLICATION_CONTEXT_VALUES_DB_NAME);
	destroyInstanceDatabase (APPLICATION_ADDRESS_DB_NAME);

    destroyInstanceDatabase (TRANSACTION_INSTANCE_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_INSTANCE_TRANID_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_INSTANCE_APP_HANDLE_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_CONTEXT_VALUES_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_CONTEXT_URI_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_CONTEXT_METRIC_VALUE_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_CONTEXT_USER_DB_NAME);
    destroyInstanceDatabase (TRANSACTION_CONTEXT_DIAG_DB_NAME);

    destroyInstanceDatabase (CORRELATOR_PARENT_DB_NAME);
    destroyInstanceDatabase (CORRELATOR_CURRENT_DB_NAME);

    destroyInstanceDatabase (TRACE_STAT_DB_NAME);
    destroyInstanceDatabase (APPDEX_STAT_DB_NAME);
}

void
BerkeleyDB::backup (void)
{
    arm_boolean_t current_null;
    arm_error_t status;
	Arm4dbSharedMemory shm;

	/* Ensure we have exclusive access */
	Arm4dbUtilityLock utilityLock;
	Arm4dbWriteLock lock;

    /* Turn collection off so that programs can continue */
    current_null = shm.setNullCollector (ARM_TRUE);

    /* Rollover the database */

    /* Step 1. Close all databases */
	closeDatabasesNoLock ();

    /* Step 2. Copy all databases to the backup directory */
    status = copyDatabases ();

    /* Step 3. Destroy the instance databases */
    if (status == ARM_SUCCESS)
	{
        destroyInstanceDatabases ();
	}

    /* Step 4. Re-create the instance databases */
	openDatabasesNoLock (0);

    /* Resume collection as before */
    shm.setNullCollector (current_null);

    lock.unlock ();
	utilityLock.unlock ();
}

