/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>
#include <math.h>

#include <iomanip>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "BerkeleyDB.h"
#include "Arm4dbCommon.h"
#include "Arm4dbSharedMemory.h"

static void
print_metric_value (const metric_value_record_t *metric_ptr)
{
	switch (metric_ptr->value.format)
		{
		case ARM_METRIC_FORMAT_COUNTER32:
			printf ("%d", metric_ptr->value.metric_u.counter32);
			break;

		case ARM_METRIC_FORMAT_COUNTER64:
			printf ("%lld", metric_ptr->value.metric_u.counter64);
			break;

		case ARM_METRIC_FORMAT_CNTRDIVR32:
			printf ("%d:", metric_ptr->value.metric_u.counterdivisor32.counter);
			printf ("%d", metric_ptr->value.metric_u.counterdivisor32.divisor);
			break;

		/* Java implementation specific */
		case ARM_METRIC_FORMAT_COUNTERFLOAT32:
			// This is a kludge to put floats in this field, but our float_kludge.m4 macro
			// ensures it fits
			printf ("%f:", *(reinterpret_cast<const float *> (&(metric_ptr->value.metric_u.counterdivisor32))));
			break;

		case ARM_METRIC_FORMAT_GAUGE32:
			printf ("%d", metric_ptr->value.metric_u.gauge32);
			break;

		case ARM_METRIC_FORMAT_GAUGE64:
			printf ("%lld", metric_ptr->value.metric_u.gauge64);
			break;

		case ARM_METRIC_FORMAT_GAUGEDIVR32:
			printf ("%d:", metric_ptr->value.metric_u.gaugedivisor32.gauge);
			printf ("%d", metric_ptr->value.metric_u.gaugedivisor32.divisor);
			break;

		/* Java implementation specific */
		case ARM_METRIC_FORMAT_GAUGEFLOAT32:
			// This is a kludge to put floats in this field, but our float_kludge.m4 macro
			// ensures it fits
			printf ("%f:", *(reinterpret_cast<const float *> (&(metric_ptr->value.metric_u.gaugedivisor32))));
			break;

		case ARM_METRIC_FORMAT_NUMERICID32:
			printf ("%d", metric_ptr->value.metric_u.numericid32);
			break;

		case ARM_METRIC_FORMAT_NUMERICID64:
			printf ("%lld", metric_ptr->value.metric_u.numericid64);
			break;

		case ARM_METRIC_FORMAT_STRING32:
			printf ("%s", metric_ptr->string32);
			break;
		}
}

void
BerkeleyDB::dump_sequence (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	arm_int64_t sequence;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_sequence_ptr != NULL);

	WRAP_EXCEPTION (status, db_sequence_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&sequence);
	data.set_size (sizeof(sequence));
	data.set_ulen (sizeof(sequence));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nSequence Table\n");
	printf ("Sequence\tValue\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%s\t%lld\n", (const char *) primary_key.get_data (), sequence);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nApplication Table\n");
	printf ("ID\tName\tDigest\tCharset\tDefault Trace Flags\tDefault Trace Parameter\tDefault Appdex Threshold\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId ((arm_id_t *) primary_key.get_data ());
			printf ("\t%s\t", application.application_name);
			Arm4dbCommon::printDigest (application.md5_digest);
			printf ("\t%d", application.charset);
			printf ("\t%d", application.default_trace_flags);
			printf ("\t%lld", application.default_trace_parameter);
			printf ("\t%lld\n", application.default_appdex_threshold);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_identity (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	property_record_t identity_property;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_identity_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nApplication Identity Table\n");
	printf ("ID\tName\tValue\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId ((arm_id_t *) primary_key.get_data ());
			printf ("\t%s\t", identity_property.name);
			printf ("\t%s\n", identity_property.value);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_context_names (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_name_record_t identity_name;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_context_names_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nApplication Context Names Table\n");
	printf ("ID\tIndex\tName\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId ((arm_id_t *) primary_key.get_data ());
			printf ("\t%d\t", identity_name.index);
			printf ("\t%s\n", identity_name.name);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_instance (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_instance_record_t application_instance;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_instance_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_instance_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&application_instance);
	data.set_size (sizeof(application_instance));
	data.set_ulen (sizeof(application_instance));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nApplication Instance Table\n");
	printf ("Handle\tApp ID\tActive\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_app_start_handle_t *) primary_key.get_data ()));
			Arm4dbCommon::printId (&(application_instance.appl_id));
			printf ("\t%d\n", application_instance.active);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_group (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_group_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_group_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	printf ("\nApplication Group Table\n");
	printf ("Handle\tGroup\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_app_start_handle_t *) primary_key.get_data ()));
			printf ("%s\n", (char *) data.get_data ());

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_instance_names (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_instance_names_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_instance_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	printf ("\nApplication Instance Names Table\n");
	printf ("Handle\tName\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_app_start_handle_t *) primary_key.get_data ()));
			printf ("%s\n", (char *) data.get_data ());

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_context_values (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_value_record_t context_value;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_context_values_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_context_values_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&context_value);
	data.set_size (sizeof(context_value));
	data.set_ulen (sizeof(context_value));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nApplication Context Values Table\n");
	printf ("Handle\tIndex\tValue\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_app_start_handle_t *) primary_key.get_data ()));
			printf ("%d\t", context_value.index);
			printf ("%s\n", context_value.value);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_application_address (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_address_record_t address_value;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_address_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_address_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&address_value);
	data.set_size (sizeof(address_value));
	data.set_ulen (sizeof(address_value));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nApplication Address Table\n");
	printf ("Handle\tFormat\tLength\tId Valid\tId\tAddress\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_app_start_handle_t *) primary_key.get_data ()));
			printf ("%d\t", address_value.address_format);
			printf ("%d\t", address_value.address_length);
			printf ("%d\t", address_value.id_valid);
			Arm4dbCommon::printId (&(address_value.id));
			printf ("\t");
			Arm4dbCommon::printAddress (&address_value);
			printf ("\n");

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Table\n");
	printf ("ID\tName\tDigest\tApp ID\tTrace Flags\tAppdex Flags\tTrace Parameter\tAppdex Threshold\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId (static_cast<const arm_id_t *> (primary_key.get_data ()));
			printf ("\t%s\t", transaction.transaction_name);
			Arm4dbCommon::printDigest (transaction.md5_digest);
			printf ("\t");
			Arm4dbCommon::printId (&(transaction.appl_id));
			printf ("\t%d", transaction.trace_flags);
			printf ("\t%d", transaction.appdex_flags);
			printf ("\t%lld", transaction.trace_parameter);
			printf ("\t%lld", transaction.appdex_threshold);
			printf ("\n");

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_identity (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	property_record_t identity_property;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_identity_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Identity Table\n");
	printf ("ID\tName\tValue\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId (static_cast<const arm_id_t *> (primary_key.get_data ()));
			printf ("\t%s\t", identity_property.name);
			printf ("%s\n", identity_property.value);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_context_names (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_name_record_t identity_name;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_names_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Context Names Table\n");
	printf ("ID\tIndex\tName\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId (static_cast<const arm_id_t *> (primary_key.get_data ()));
			printf ("\t%d\t", identity_name.index);
			printf ("%s\n", identity_name.name);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_identity_uri (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_identity_uri_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_identity_uri_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	printf ("\nTransaction Identity URI Table\n");
	printf ("ID\tURI\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId (static_cast<const arm_id_t *> (primary_key.get_data ()));
			printf ("\t%s\n", (char *) data.get_data ());

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_metric_binding (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
    arm_metric_binding_t binding;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_metric_binding_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_metric_binding_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&binding);
	data.set_size (sizeof(binding));
	data.set_ulen (sizeof(binding));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Metric Binding Table\n");
	printf ("ID\tIndex\tMetric ID\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId (static_cast<const arm_id_t *> (primary_key.get_data ()));
			printf ("\t%d\t", binding.slot);
			Arm4dbCommon::printId (&(binding.id));
			printf ("\n");

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_instance (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_instance_record_t transaction_instance;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_instance_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_instance_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&transaction_instance);
	data.set_size (sizeof(transaction_instance));
	data.set_ulen (sizeof(transaction_instance));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Instance Table\n");
	printf ("Handle\tApp ID\tTran ID\tApp Handle\tArrival Time\tStop Time\tResponse Time\tStatus\tActive\tBlock Start\tBlocked Time\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_tran_start_handle_t *) primary_key.get_data ()));
			Arm4dbCommon::printId (&(transaction_instance.appl_id));
			printf ("\t");
			Arm4dbCommon::printId (&(transaction_instance.tran_id));
			printf ("\t%lld\t", transaction_instance.app_handle);
			printf ("%ld.%06ld\t", transaction_instance.arrival_time.tv_sec, transaction_instance.arrival_time.tv_usec);
			printf ("%ld.%06ld\t", transaction_instance.stop_time.tv_sec, transaction_instance.stop_time.tv_usec);
			printf ("%ld.%06ld\t", transaction_instance.response_time.tv_sec, transaction_instance.response_time.tv_usec);
			printf ("%d\t", transaction_instance.tran_status);
			printf ("%d\t", transaction_instance.active);
			printf ("%ld.%06ld\t", transaction_instance.block_start_time.tv_sec, transaction_instance.block_start_time.tv_usec);
			printf ("%ld.%06ld\n", transaction_instance.blocked_time.tv_sec, transaction_instance.blocked_time.tv_usec);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_context_values (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_value_record_t context_value;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_values_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_values_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&context_value);
	data.set_size (sizeof(context_value));
	data.set_ulen (sizeof(context_value));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Context Values Table\n");
	printf ("Handle\tIndex\tValue\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld", *((arm_tran_start_handle_t *) primary_key.get_data ()));
			printf ("\t%d\t%s\n", context_value.index, context_value.value);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_context_uri (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_uri_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_uri_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	printf ("\nTransaction Context URI Table\n");
	printf ("Handle\tURI\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld", *((arm_tran_start_handle_t *) primary_key.get_data ()));
			printf ("\t%s\n", (char *) data.get_data ());

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_context_metric_value (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	metric_value_record_t record;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_metric_value_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_metric_value_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&record);
	data.set_size (sizeof(record));
	data.set_ulen (sizeof(record));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Context Metric Values Table\n");
	printf ("Handle\tSlot\tFormat\tUsage\tValid\tValue\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_tran_start_handle_t *) primary_key.get_data()));
			printf( "%d\t", record.value.slot);
			printf( "%d\t", record.value.format);
			printf ("%d\t", record.value.usage);
			printf ("%d\t", record.value.valid);
			print_metric_value (&record);
			printf ("\n");

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_context_user (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	user_record_t record;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_user_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_user_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&record);
	data.set_size (sizeof(record));
	data.set_ulen (sizeof(record));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nTransaction Context User Table\n");
	printf ("Handle\tName\tID Valid\tID\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld\t", *((arm_tran_start_handle_t *) primary_key.get_data ()));
			printf( "%s\t%d\t", record.name, record.id_valid);
			if (record.id_valid)
				Arm4dbCommon::printId (&(record.id));
			printf ("\n");

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_transaction_context_diag (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_diag_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_diag_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	printf ("\nTransaction Context Diagnostic Detail Table\n");
	printf ("Handle\tDetail\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			printf ("%lld", *((arm_tran_start_handle_t *) primary_key.get_data ()));
			printf ("\t%s\n", (char *) data.get_data ());

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump_metric (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	metric_record_t metric;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_metric_ptr != NULL);

	WRAP_EXCEPTION (status, db_metric_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&metric);
	data.set_size (sizeof(metric));
	data.set_ulen (sizeof(metric));
	data.set_flags (DB_DBT_USERMEM);

	printf ("\nMetric Table\n");
	printf ("ID\tName\tUnits\tDigest\tApp ID\tFormat\tUsage\n");

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			Arm4dbCommon::printId ((arm_id_t *) primary_key.get_data ());
			printf ("\t%s\t", metric.metric_name);
			printf ("%s\t", metric.metric_units);
			Arm4dbCommon::printDigest (metric.md5_digest);
			printf ("\t");
			Arm4dbCommon::printId (&(metric.appl_id));
			printf ("\t%d\t", metric.metric_format);
			printf ("%d\n", metric.metric_usage);

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No records\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::dump (void) const
{
	Arm4dbLock lock;

    dump_sequence ();

	dump_application ();
	dump_application_identity ();
	dump_application_context_names ();

	dump_application_instance ();
	dump_application_group ();
	dump_application_instance_names ();
	dump_application_context_values ();
	dump_application_address ();

	dump_transaction ();
	dump_transaction_identity ();
	dump_transaction_context_names ();
	dump_transaction_identity_uri ();
	dump_transaction_metric_binding ();

	dump_transaction_instance ();
	dump_transaction_context_values ();
	dump_transaction_context_uri ();
	dump_transaction_context_metric_value ();
	dump_transaction_context_user ();
	dump_transaction_context_diag ();

	dump_metric ();

    lock.unlock ();
}

std::ostream&
BerkeleyDB::export_application_identity (std::ostream &os, const arm_id_t &appl_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	property_record_t identity_property;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_identity_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &appl_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<identity>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, identity_property.name, ARM_PROPERTY_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, identity_property.value, ARM_PROPERTY_VALUE_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</identity>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream& 
BerkeleyDB::export_application_context_names (std::ostream &os, const arm_id_t &appl_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_name_record_t identity_name;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_context_names_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &appl_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<context>\n";
			os << "\t\t\t<slot>" << identity_name.index << "</slot>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, identity_name.name, ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t</context>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::export_application_definition (std::ostream &os, const arm_id_t &appl_id, const application_record_t &application) const
{
	os << "\t<application_definition>\n";
	os << "\t\t<appl_id>";
		Arm4dbCommon::printId (os, appl_id);
	os << "</appl_id>\n";
	os << "\t\t<name>";
		Arm4dbCommon::escapeForXml(os, application.application_name, ARM_NAME_MAX_CHARS);
	os << "</name>\n";
	os << "\t\t<digest>";
		Arm4dbCommon::printDigest (os, application.md5_digest);
	os << "</digest>\n";
	os << "\t\t<charset>" << application.charset << "</charset>\n";
	os << "\t\t<default_trace_flags>" << application.default_trace_flags << "</default_trace_flags>\n";
	os << "\t\t<default_trace_parameter>" << application.default_trace_parameter << "</default_trace_parameter>\n";
	os << "\t\t<default_appdex_threshold>" << application.default_appdex_threshold << "</default_appdex_threshold>\n";
	
	export_application_identity (os, appl_id);
	export_application_context_names (os, appl_id);
	
	os << "\t</application_definition>\n";
	return os;
}

std::ostream&
BerkeleyDB::exportApplicationDefinitions (std::ostream &os) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			arm_id_t appl_id = *((arm_id_t *) primary_key.get_data ());
			export_application_definition (os, appl_id, application);
			
			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();
	return os;
}


std::ostream&
BerkeleyDB::export_transaction_identity (std::ostream &os, const arm_id_t &tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	property_record_t identity_property;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_identity_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<identity>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, identity_property.name, ARM_PROPERTY_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, identity_property.value, ARM_PROPERTY_VALUE_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</identity>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::export_transaction_context (std::ostream &os, const arm_id_t &tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_name_record_t identity_name;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_names_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<context>\n";
			os << "\t\t\t<slot>" << identity_name.index << "</slot>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, identity_name.name, ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t</context>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::export_transaction_uri (std::ostream &os, const arm_id_t &tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_identity_uri_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_identity_uri_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_id);
	primary_key.set_size (sizeof (arm_id_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<uri>";
				Arm4dbCommon::escapeForXml(os, (const char *) data.get_data (), ARM_PROPERTY_URI_MAX_CHARS);
			os << "</uri>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::export_transaction_metric_binding (std::ostream &os, const arm_id_t &tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
    arm_metric_binding_t binding;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_metric_binding_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_metric_binding_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&binding);
	data.set_size (sizeof(binding));
	data.set_ulen (sizeof(binding));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<metric>\n";
			os << "\t\t\t<slot>" << (int) binding.slot << "</slot>\n";
			os << "\t\t\t<metric_id>";
				Arm4dbCommon::printId (os, binding.id);
			os << "</metric_id>\n";
			os << "\t\t</metric>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTransactionDefinitions (std::ostream &os) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			arm_id_t tran_id = *((arm_id_t *) primary_key.get_data ());

			os << "\t<transaction_definition>\n";
			os << "\t\t<tran_id>";
				Arm4dbCommon::printId (os, tran_id);
			os << "</tran_id>\n";
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, transaction.transaction_name, ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t<digest>";
				Arm4dbCommon::printDigest (os, transaction.md5_digest);
			os << "</digest>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, transaction.appl_id);
			os << "</appl_id>\n";
			os << "\t\t<trace_flags>" << transaction.trace_flags << "</trace_flags>\n";
			os << "\t\t<appdex_flags>" << transaction.appdex_flags << "</appdex_flags>\n";
			os << "\t\t<trace_parameter>" << transaction.trace_parameter << "</trace_parameter>\n";
			os << "\t\t<appdex_threshold>" << transaction.appdex_threshold << "</appdex_threshold>\n";

			export_transaction_identity (os, tran_id);
			export_transaction_context (os, tran_id);
			export_transaction_uri (os, tran_id);
			export_transaction_metric_binding (os, tran_id);
			
			os << "\t</transaction_definition>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportMetricDefinitions (std::ostream &os) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	metric_record_t metric;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_metric_ptr != NULL);

	WRAP_EXCEPTION (status, db_metric_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}

	data.set_data (&metric);
	data.set_size (sizeof(metric));
	data.set_ulen (sizeof(metric));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			os << "\t<metric_definition>\n";
			os << "\t\t<metric_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) primary_key.get_data ());
			os << "</metric_id>\n";
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, metric.metric_name, ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t<digest>";
				Arm4dbCommon::printDigest (os, metric.md5_digest);
			os << "</digest>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, metric.appl_id);
			os << "</appl_id>\n";
			os << "\t\t<units>";
				Arm4dbCommon::escapeForXml(os, metric.metric_units, ARM_NAME_MAX_CHARS);
			os << "</units>\n";
			os << "\t\t<format>" << (int) metric.metric_format << "</format>\n";
			os << "\t\t<usage>" << metric.metric_usage << "</usage>\n";
			os << "\t</metric_definition>\n";
			
			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportApplicationGroup (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_group_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_group_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &app_handle);
	primary_key.set_size (sizeof (arm_app_start_handle_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<group>";
				Arm4dbCommon::escapeForXml(os, (char *) data.get_data (), ARM_NAME_MAX_CHARS);
			os << "</group>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportApplicationInstanceNames (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_instance_names_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_instance_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &app_handle);
	primary_key.set_size (sizeof (arm_app_start_handle_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, (char *) data.get_data (), ARM_NAME_MAX_CHARS);
			os << "\t\t</name>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportApplicationContextValues (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_value_record_t context_value;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_context_values_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_context_values_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &app_handle);
	primary_key.set_size (sizeof (arm_app_start_handle_t));

	data.set_data (&context_value);
	data.set_size (sizeof(context_value));
	data.set_ulen (sizeof(context_value));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<context_value>\n";
			os << "\t\t\t<slot>" << context_value.index << "</slot>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, context_value.value, ARM_NAME_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</context_value>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportApplicationAddress (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_address_record_t address_value;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_address_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_address_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &app_handle);
	primary_key.set_size (sizeof (arm_app_start_handle_t));

	data.set_data (&address_value);
	data.set_size (sizeof(address_value));
	data.set_ulen (sizeof(address_value));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<address>\n";
			os << "\t\t\t<address_format>" << address_value.address_format << "</address_format>\n";
			os << "\t\t\t<address_length>" << address_value.address_length << "</address_length>\n";
			os << "\t\t\t<id_valid>" << address_value.id_valid << "</id_valid>\n";
			os << "\t\t\t<id>";
				Arm4dbCommon::printId (os, address_value.id);
			os << "</id>\n";
			os << "\t\t\t<address>";
				Arm4dbCommon::printAddress (os, address_value.address, address_value.address_length);
			os << "</address>\n";
			os << "\t\t</address>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportApplications (std::ostream &os) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_instance_record_t application_instance;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_instance_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_instance_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}

	data.set_data (&application_instance);
	data.set_size (sizeof(application_instance));
	data.set_ulen (sizeof(application_instance));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			arm_app_start_handle_t app_handle = *((arm_app_start_handle_t *) primary_key.get_data ());
			
			os << "\t<application_instance>\n";
			os << "\t\t<handle>" << app_handle << "</handle>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, application_instance.appl_id);
			os << "</appl_id>\n";
			os << "\t\t<active>" << application_instance.active << "</active>\n";
			
			exportApplicationGroup (os, app_handle);
			exportApplicationInstanceNames (os, app_handle);
			exportApplicationContextValues (os, app_handle);
			exportApplicationAddress (os, app_handle);
			
			os << "\t</application_instance>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}


std::ostream&
BerkeleyDB::exportTransactionContextValues (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	indexed_property_value_record_t context_value;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_values_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_values_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	data.set_data (&context_value);
	data.set_size (sizeof(context_value));
	data.set_ulen (sizeof(context_value));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<context_value>\n";
			os << "\t\t\t<slot>" << context_value.index << "</slot>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, context_value.value, ARM_NAME_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</context_value>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTransactionContextUri (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_uri_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_uri_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<context_uri>";
				Arm4dbCommon::escapeForXml(os, (char *) data.get_data (), ARM_PROPERTY_URI_MAX_CHARS);
			os << "\t\t</context_uri>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTransactionContextMetricValue (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	metric_value_record_t record;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_metric_value_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_metric_value_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	data.set_data (&record);
	data.set_size (sizeof(record));
	data.set_ulen (sizeof(record));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<metric_value>\n";
			os << "\t\t\t<slot>" << (int) record.value.slot << "</slot>\n";
			os << "\t\t\t<format>" << (int) record.value.format << "</format>\n";
			os << "\t\t\t<usage>" << record.value.usage << "</usage>\n";
			os << "\t\t\t<valid>" << record.value.valid << "</valid>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::printMetricValue (os, record.value, record.string32);
			os << "</value>\n";
			os << "\t\t</metric_value>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTransactionContextUser (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	user_record_t record;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_user_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_user_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	data.set_data (&record);
	data.set_size (sizeof(record));
	data.set_ulen (sizeof(record));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<user>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, record.name, ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t\t<id_valid>" << record.id_valid << "</id_valid>\n";
			os << "\t\t\t<id>";
				Arm4dbCommon::printId (os, record.id);
			os << "</id>\n";
			os << "\t\t</user>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTransactionContextDiag (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_diag_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_diag_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<diag>";
				Arm4dbCommon::escapeForXml(os, (char *) data.get_data (), ARM_DIAG_DETAIL_MAX_CHARS);
			os << "</diag>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportParentCorrelator (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_parent_correlator_ptr != NULL);

	WRAP_EXCEPTION (status, db_parent_correlator_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<parent_correlator>";
				Arm4dbCommon::printCorrelator(os, (arm_correlator_t *) data.get_data ());
			os << "\t\t</parent_correlator>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportCorrelator (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_parent_correlator_ptr != NULL);
    DB_ASSERT (db_current_correlator_ptr != NULL);

	WRAP_EXCEPTION (status, db_current_correlator_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}
	
	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			os << "\t\t<correlator>";
				Arm4dbCommon::printCorrelator(os, (arm_correlator_t *) data.get_data ());
			os << "\t\t</correlator>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTransactions (std::ostream &os) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_instance_record_t transaction_instance;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_instance_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_instance_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}

	data.set_data (&transaction_instance);
	data.set_size (sizeof(transaction_instance));
	data.set_ulen (sizeof(transaction_instance));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			arm_tran_start_handle_t tran_handle = *((arm_tran_start_handle_t *) primary_key.get_data ());
			
			os << "\t<transaction_instance>\n";
			os << "\t\t<handle>" << tran_handle << "</handle>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, transaction_instance.appl_id);
			os << "</appl_id>\n";
			os << "\t\t<tran_id>";
				Arm4dbCommon::printId (os, transaction_instance.tran_id);
			os << "</tran_id>\n";
			os << "\t\t<app_handle>" << transaction_instance.app_handle << "</app_handle>\n";
			os << "\t\t<arrival_time>" << transaction_instance.arrival_time.tv_sec << "."
					<< std::setfill ('0') << std::setw (6) << transaction_instance.arrival_time.tv_usec << "</arrival_time>\n";
			os << "\t\t<stop_time>" << transaction_instance.stop_time.tv_sec << "."
					<< std::setfill ('0') << std::setw (6) << transaction_instance.stop_time.tv_usec << "</stop_time>\n";
			os << "\t\t<response_time>" << transaction_instance.response_time.tv_sec << "."
					<< std::setfill ('0') << std::setw (6) << transaction_instance.response_time.tv_usec << "</response_time>\n";
			os << "\t\t<blocked_time>" << transaction_instance.blocked_time.tv_sec << "."
					<< std::setfill ('0') << std::setw (6) << transaction_instance.blocked_time.tv_usec << "</blocked_time>\n";
			
			os << "\t\t<status>" << transaction_instance.tran_status << "</status>\n";
			os << "\t\t<active>" << transaction_instance.active << "</active>\n";
			
			exportTransactionContextValues (os, tran_handle);
			exportTransactionContextUri (os, tran_handle);
			exportTransactionContextMetricValue (os, tran_handle);
			exportTransactionContextUser (os, tran_handle);
			exportTransactionContextDiag (os, tran_handle);
			exportParentCorrelator (os, tran_handle);
			exportCorrelator (os, tran_handle);
			
			os << "\t</transaction_instance>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}

std::ostream&
BerkeleyDB::exportTraceStats (std::ostream &os) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	arm4db_transaction_trace_stats_t trace_stats;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_trace_stat_ptr != NULL);

	WRAP_EXCEPTION (status, db_trace_stat_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return os;
	}

	data.set_data (&trace_stats);
	data.set_size (sizeof(trace_stats));
	data.set_ulen (sizeof(trace_stats));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			arm_time_t sample_time = *((arm_time_t *) primary_key.get_data ());
			
			os << "\t<transaction_trace>\n";
			os << "\t\t<transaction_trace_sample_time>" << sample_time.tv_sec << "</transaction_trace_sample_time>\n";
			os << "\t\t<transaction_trace_tran_id>";
				Arm4dbCommon::printId (os, trace_stats.tran_id);
			os << "</transaction_trace_tran_id>\n";
			os << "\t\t<transaction_trace_counter>" << trace_stats.transaction_counter << "</transaction_trace_counter>\n";
			os << "\t</transaction_trace>\n";

			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}

	cursor_ptr->close ();
	return os;
}
