/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>
#include <math.h>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "BerkeleyDB.h"
#include "Arm4dbCommon.h"

/*
 *
 * TO DO: What happens to the memory returned as the key? Does this need to be freed?
 *
 */

/* Prototypes */

static int
time_less (const arm_time_t *time1_ptr, const arm_time_t *time2_ptr)
{
	if (time1_ptr->tv_sec < time2_ptr->tv_sec)
		return ARM_TRUE;
	else if ((time1_ptr->tv_sec == time2_ptr->tv_sec) &&
				(time1_ptr->tv_usec < time2_ptr->tv_usec))
		return ARM_TRUE;
	
	return ARM_FALSE;
}

static int
time_greater (const arm_time_t *time1_ptr, const arm_time_t *time2_ptr)
{
	if (time1_ptr->tv_sec > time2_ptr->tv_sec)
		return ARM_TRUE;
	else if ((time1_ptr->tv_sec == time2_ptr->tv_sec) &&
				(time1_ptr->tv_usec > time2_ptr->tv_usec))
		return ARM_TRUE;
	
	return ARM_FALSE;
}

static void
print_id (const arm_id_t *id_ptr)
{
	int i;
	for (i  = 0; i < 16; i++)
		printf ("%02x", id_ptr->id_u.uint8 [i]);
}

static void
print_correlator (const arm_correlator_t *correlator_ptr)
{
	arm_correlator_header_t *header_ptr;
	arm_uint8_t *hex_ptr;
	int length;
	int i;
	
	header_ptr = (arm_correlator_header_t *) correlator_ptr;
	hex_ptr = (arm_uint8_t *) correlator_ptr;
	
	length = ntohs (header_ptr->length);
	printf ("\t\tlength = %d\n", length);
	printf ("\t\tformat = %d\n", header_ptr->format);
	printf ("\t\tflags = %d\n", header_ptr->flags);
	printf ("\t\t");
	for (i = 0; i < length; i++)
	{
		printf ("%02x", (hex_ptr [i] & 0x0FF));
	}
	printf ("\n");
}

static void
print_metric_value (const metric_value_record_t *metric_ptr)
{
	printf( "\t\tSlot %d\n", metric_ptr->value.slot);
	printf( "\t\tFormat: ");
	switch (metric_ptr->value.format)
		{
		case ARM_METRIC_FORMAT_RESERVED:
			printf ("Reserved\n");
			break;

		case ARM_METRIC_FORMAT_COUNTER32:
			printf ("Counter 32\n");
			break;

		case ARM_METRIC_FORMAT_COUNTER64:
			printf ("Counter 64\n");
			break;

		case ARM_METRIC_FORMAT_CNTRDIVR32:
			printf ("CNTRDIVR32\n");
			break;

		case ARM_METRIC_FORMAT_COUNTERFLOAT32:
			printf ("COUNTERFLOAT32\n");
			break;

		case ARM_METRIC_FORMAT_GAUGE32:
			printf ("Guage 32\n");
			break;

		case ARM_METRIC_FORMAT_GAUGE64:
			printf ("Guage 64\n");
			break;

		case ARM_METRIC_FORMAT_GAUGEDIVR32:
			printf ("GAUGEDIVR32\n");
			break;

		case ARM_METRIC_FORMAT_GAUGEFLOAT32:
			printf ("GAUGEFLOAT32\n");
			break;
	
		case ARM_METRIC_FORMAT_NUMERICID32:
			printf ("Numeric ID 32\n");
			break;

		case ARM_METRIC_FORMAT_NUMERICID64:
			printf ("Numeric ID 64\n");
			break;

		case ARM_METRIC_FORMAT_STRING32:
			printf ("String 32\n");
			break;

		default:
			printf ("Unknown (%d)\n", metric_ptr->value.format);
		}

	printf ("\t\tUsage: ");
	switch (metric_ptr->value.usage)
		{
		case ARM_METRIC_USE_GENERAL:
			printf ("General\n");
			break;

		case ARM_METRIC_USE_TRAN_SIZE:
			printf ("Transaction Size\n");
			break;

		case ARM_METRIC_USE_TRAN_STATUS:
			printf ("Transaction Status\n");
			break;

		default:
			printf ("Unknown (%d)\n", metric_ptr->value.usage);
		}

	if (metric_ptr->value.valid)
		printf ("\t\tValid: Yes\n");
	else
		printf ("\t\tValid: No\n");
}

void
BerkeleyDB::find_and_print_correlators (arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	arm_correlator_t correlator;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_parent_correlator_ptr != NULL);
    DB_ASSERT (db_current_correlator_ptr != NULL);

	primary_key.set_data (&tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));
	
	data.set_data (&correlator);
	data.set_size (sizeof(arm_correlator_t));
	data.set_ulen (sizeof(arm_correlator_t));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_parent_correlator_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
	{
		printf ("\tParent Correlator\n");
		print_correlator (&correlator);
	}
	else if (status != DB_NOTFOUND)
		return;

	WRAP_EXCEPTION (status, db_current_correlator_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
	{
		printf ("\tCurrent Correlator\n");
		print_correlator (&correlator);
	}
	else if (status != DB_NOTFOUND)
		return;
}

void
BerkeleyDB::find_and_print_metric_values (arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	metric_value_record_t metric;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_metric_value_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_context_metric_value_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) &tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));

	data.set_data (&metric);
	data.set_size (sizeof(metric));
	data.set_ulen (sizeof(metric));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		printf ("\tMetric Values:\n");
		do
		{
			print_metric_value (&metric);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	else if (status != DB_NOTFOUND)
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::find_and_print_user_values (arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	user_record_t user;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_user_ptr != NULL);

	primary_key.set_data (&tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));
	
	data.set_data (&user);
	data.set_size (sizeof(user));
	data.set_ulen (sizeof(user));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_transaction_context_user_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
	{
		if (strlen (user.name) > 0)
			printf ("\tUser Name: %s\n", user.name);
		if (user.id_valid)
		{
			printf ("\tUser ID: ");
			print_id (&user.id);
			printf ("\n");
		}
	}
	else if (status != DB_NOTFOUND)
		m_environmentPtr->err (status, "No records found");
}

void
BerkeleyDB::find_and_print_diag_values (arm_tran_start_handle_t tran_handle) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	char detail [ARM_DIAG_DETAIL_MAX_LENGTH];

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_context_diag_ptr != NULL);

	primary_key.set_data (&tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));
	
	data.set_data ((void *) detail);
	data.set_size (ARM_DIAG_DETAIL_MAX_LENGTH);
	data.set_ulen (ARM_DIAG_DETAIL_MAX_LENGTH);
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_transaction_context_diag_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
	{
		if (strlen (detail) > 0)
			printf ("\tDiag Detail: %s\n", detail);
	}
	else if (status != DB_NOTFOUND)
		m_environmentPtr->err (status, "No records found");
}

int
BerkeleyDB::find_application (application_record_t *application_ptr,
	const arm_id_t *appl_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);

	primary_key.set_data ((void *) appl_id);
	primary_key.set_size (sizeof (arm_id_t));
	
	data.set_data (application_ptr);
	data.set_size (sizeof(application_record_t));
	data.set_ulen (sizeof(application_record_t));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_application_ptr->get (NULL, &primary_key, &data, 0));
	if (status != 0)
		return ARM_FAILURE;
	
	return ARM_SUCCESS;
}

int
BerkeleyDB::find_transaction (transaction_record_t *transaction_ptr,
	const arm_id_t *transaction_instance_ptr) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);

	primary_key.set_data ((void *) transaction_instance_ptr);
	primary_key.set_size (sizeof (arm_id_t));
	
	data.set_data (transaction_ptr);
	data.set_size (sizeof(transaction_record_t));
	data.set_ulen (sizeof(transaction_record_t));
	data.set_flags (DB_DBT_USERMEM);

	WRAP_EXCEPTION (status, db_transaction_ptr->get (NULL, &primary_key, &data, 0));
	if (status != 0)
		return ARM_FAILURE;
	
	return ARM_SUCCESS;
}

void
BerkeleyDB::print_transaction_metadata (const arm_id_t *tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	property_record_t identity_property;
	indexed_property_name_record_t context_name;
	arm_char_t uri [ARM_PROPERTY_URI_MAX_LENGTH];

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_identity_ptr != NULL);
    DB_ASSERT (db_transaction_context_names_ptr != NULL);
    DB_ASSERT (db_transaction_identity_uri_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) tran_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			printf ("\t\tIdentity: %s=%s\n", identity_property.name, identity_property.value);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	else if (status != DB_NOTFOUND)
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();

	/* Display context data */
	WRAP_EXCEPTION (status, db_transaction_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&context_name);
	data.set_size (sizeof(context_name));
	data.set_ulen (sizeof(context_name));
	data.set_flags (DB_DBT_USERMEM);
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			printf ("\t\tContext Name [%d]: %s\n", context_name.index, context_name.name);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	else if (status != DB_NOTFOUND)
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();

	/* Display URI data */
	data.set_data (&uri);
	data.set_size (sizeof(uri));
	data.set_ulen (sizeof(uri));
	data.set_flags (DB_DBT_USERMEM);
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, db_transaction_identity_uri_ptr->get (NULL, &primary_key, &data, 0));
	if (status == 0)
	{
		printf ("\t\tURI: %s\n", uri);
	}
	else if (status != DB_NOTFOUND)
	{
		m_environmentPtr->err (status, "No records found");
	}
}

void
BerkeleyDB::print_trace_stats (const arm_id_t *tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	arm_time_t sample_time;
	arm4db_transaction_trace_stats_t stats;
	arm_int64_t count;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_trace_stat_ptr != NULL);

	WRAP_EXCEPTION (status, db_trace_stat_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) &sample_time);
	primary_key.set_size (sizeof (arm_time_t));
	primary_key.set_ulen (sizeof (arm_time_t));
	primary_key.set_flags (DB_DBT_USERMEM);
	
	data.set_data (&stats);
	data.set_size (sizeof(stats));
	data.set_ulen (sizeof(stats));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	count = 0;
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			if (memcmp (&(stats.tran_id), tran_id, sizeof (arm_id_t)) == 0)
			{
				count += stats.transaction_counter;
			}
			
			/* Move on to the next entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

	printf ("\t\t" ARM4_FORMAT_INT64 " total transactions\n", count);
}

arm_error_t
BerkeleyDB::summarize_transaction (const arm_id_t *tran_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_instance_record_t transaction_instance;
	arm_tran_start_handle_t transaction_instance_id;
	arm_time_t	min_time, max_time;
	double raw_time;
	double total_time;
	double total_blocked;
	double total_time_squared;
	double deviation;
	int count;
	int status_good, status_aborted, status_failed, status_unknown;
	int discard;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_instance_tranid_ptr != NULL);

	print_transaction_metadata (tran_id);
	
	/* This lookup is based on the secondary key of tran_id */
	WRAP_EXCEPTION (status, db_transaction_instance_tranid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status,     "Unable to create cursor");
		return ARM_FAILURE;
	}

	primary_key.set_data ((void *) &transaction_instance_id);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));
	primary_key.set_ulen (sizeof (arm_tran_start_handle_t));
	primary_key.set_flags (DB_DBT_USERMEM);

	secondary_key.set_data ((void *) tran_id);
	secondary_key.set_size (sizeof (arm_id_t));

	data.set_data (&transaction_instance);
	data.set_size (sizeof(transaction_instance));
	data.set_ulen (sizeof(transaction_instance));
	data.set_flags (DB_DBT_USERMEM);

	/* Initialize the stats, except min_time which comes later */
	max_time.tv_sec = max_time.tv_usec = 0;
	total_time = 0.0;
	total_blocked = 0.0;
	total_time_squared = 0.0;
	count = 0;
	status_good = status_aborted = status_failed = status_unknown = discard = 0;

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		min_time = transaction_instance.response_time;
		do
		{
			count++;
			raw_time = transaction_instance.response_time.tv_sec +
				(transaction_instance.response_time.tv_usec * 1e-6);
			total_time += raw_time;
			total_time_squared += (raw_time * raw_time);
			if (time_greater (&transaction_instance.response_time, &max_time))
				max_time = transaction_instance.response_time;
			if (time_less (&transaction_instance.response_time, &min_time))
				min_time = transaction_instance.response_time;
			raw_time = transaction_instance.blocked_time.tv_sec +
				(transaction_instance.blocked_time.tv_usec * 1e-6);
			total_blocked += raw_time;
			
			if (transaction_instance.active == ARM_DISCARD)
				discard++;
			else
			{
				switch (transaction_instance.tran_status)
				{
					case ARM_STATUS_GOOD:
						status_good++;
						break;
					
					case ARM_STATUS_ABORTED:
						status_aborted++;
						break;
					
					case ARM_STATUS_FAILED:
						status_aborted++;
						break;
					
					case ARM_STATUS_UNKNOWN:
					default:
						status_unknown++;
				}
			}

			/* Move on to the next transaction instance entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
        if (count <= 1)
        {
            deviation = 0.0;
        }
        else
        {
            deviation = sqrt ((total_time_squared - (total_time * total_time / count)) / count);
        }
		printf ("\t\tMin: %ld.%06ld, Max: %ld.%06ld, Mean: %g, Deviation: %g, Blocked: %g\n",
					min_time.tv_sec, min_time.tv_usec,
					max_time.tv_sec, max_time.tv_usec,
					(total_time / count),
					deviation,
					(total_blocked / count));
		printf ("\t\tExit status: Good %d, Aborted %d, Failed %d, Unknown %d, Total %d\n",
					status_good,
					status_aborted,
					status_failed,
					status_unknown,
					count);
		printf ("\t\t%d discards\n", discard);
		print_trace_stats (tran_id);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("\t\tNo transactions recorded\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
		cursor_ptr->close ();
		return ARM_FAILURE;
	}
	cursor_ptr->close ();
	return ARM_SUCCESS;
}

void
BerkeleyDB::print_application_metadata (const arm_id_t *appl_id) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	property_record_t identity_property;
	indexed_property_name_record_t context_name;

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_identity_ptr != NULL);
    DB_ASSERT (db_application_context_names_ptr != NULL);

	/* Display any identity data */
	WRAP_EXCEPTION (status, db_application_identity_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) appl_id);
	primary_key.set_size (sizeof (arm_id_t));

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			printf ("\tIdentity: %s=%s\n", identity_property.name, identity_property.value);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	else if (status != DB_NOTFOUND)
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();

	/* Display context data */
	WRAP_EXCEPTION (status, db_application_context_names_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&context_name);
	data.set_size (sizeof(context_name));
	data.set_ulen (sizeof(context_name));
	data.set_flags (DB_DBT_USERMEM);
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			printf ("\tContext Name [%d]: %s\n", context_name.index, context_name.name);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	else if (status != DB_NOTFOUND)
	{
		m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();
}

arm_error_t
BerkeleyDB::summarize_application (const arm_id_t *appl_id, const application_record_t *application_ptr) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;
	arm_id_t tran_id;
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_appid_ptr != NULL);

	/* This lookup is based on the secondary key of app_id, not tran_id */
	WRAP_EXCEPTION (status, db_transaction_appid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return ARM_FAILURE;
	}

	primary_key.set_data ((void *) &tran_id);
	primary_key.set_size (sizeof (arm_id_t));
	primary_key.set_ulen (sizeof (arm_id_t));
	primary_key.set_flags (DB_DBT_USERMEM);
	
	secondary_key.set_data ((void *) appl_id);
	secondary_key.set_size (sizeof (arm_id_t));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	printf ("Application: %s\n", application_ptr->application_name);
	printf ("\tID:");
		print_id (appl_id);
		printf ("\n");
	print_application_metadata (appl_id);
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			printf ("\tTransaction: %s\n", transaction.transaction_name);
			WRAP_EXCEPTION (status, summarize_transaction (&tran_id));
			if (status != ARM_SUCCESS)
			{
				cursor_ptr->close ();
				return ARM_FAILURE;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("\tNo transactions defined\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
		cursor_ptr->close ();
		return ARM_FAILURE;
	}
	cursor_ptr->close ();
	return ARM_SUCCESS;
}

void
BerkeleyDB::printSummaries (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);

	WRAP_EXCEPTION (status, db_application_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			WRAP_EXCEPTION (status, summarize_application ((const arm_id_t *) primary_key.get_data (), &application));
			if (status != ARM_SUCCESS)
			{
				cursor_ptr->close ();
				return;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else if (status == DB_NOTFOUND)
	{
		printf ("No applications defined\n");
	}
	else
	{
		m_environmentPtr->err (status, "No records found");
		cursor_ptr->close ();
		return;
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::printTransactions (void) const
{
	int	status = ARM_FAILURE;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;
	transaction_record_t transaction;
	transaction_instance_record_t transaction_instance;
	arm_tran_start_handle_t tran_handle;
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_instance_ptr != NULL);

	WRAP_EXCEPTION (status, db_transaction_instance_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data (&tran_handle);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));
	primary_key.set_ulen (sizeof (arm_tran_start_handle_t));
	primary_key.set_flags (DB_DBT_USERMEM);

	data.set_data (&transaction_instance);
	data.set_size (sizeof(transaction_instance));
	data.set_ulen (sizeof(transaction_instance));
	data.set_flags (DB_DBT_USERMEM);

	printf ("Application\tTransaction\tResponse Time\tBlocked Time\tStatus\n");
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			WRAP_EXCEPTION (status, find_transaction (&transaction, &(transaction_instance.tran_id)));
			if (status != ARM_SUCCESS)
			{
				cursor_ptr->close ();
				return;
			}
			
			WRAP_EXCEPTION (status, find_application (&application, &(transaction.appl_id)));
			if (status != ARM_SUCCESS)
			{
				cursor_ptr->close ();
				return;
			}
			
			printf ("%s\t%s\t%ld.%06ld sec\t%ld.%06ld sec\t%d\n",
				application.application_name,
				transaction.transaction_name,
				transaction_instance.response_time.tv_sec,
				transaction_instance.response_time.tv_usec,
				transaction_instance.blocked_time.tv_sec,
				transaction_instance.blocked_time.tv_usec,
				transaction_instance.tran_status);
			find_and_print_correlators (tran_handle);
			find_and_print_metric_values (tran_handle);
			find_and_print_user_values (tran_handle);
			find_and_print_diag_values (tran_handle);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	else
	{
		m_environmentPtr->err (status,     "No records found");
		cursor_ptr->close ();
		return;
	}
	cursor_ptr->close ();
}


#if 0
void
BerkeleyDB::list_identity_properties (DB *db_ptr, list_t *list_ptr, arm_id_t *id_ptr)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	list_t next;
	property_record_t identity_property;
	property_record_t *identity_ptr;
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_ptr != NULL);

	WRAP_EXCEPTION (status, db_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) id_ptr);
	primary_key.set_size (sizeof (*id_ptr));

	data.set_data (&identity_property);
	data.set_size (sizeof(identity_property));
	data.set_ulen (sizeof(identity_property));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			identity_ptr = (property_record_t *) malloc (sizeof (property_record_t));
			memcpy (identity_ptr, &identity_property, sizeof (property_record_t));
			
			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) identity_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) identity_ptr;
				next->next_ptr = NULL;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::list_application_identity_properties (list_t *list_ptr, arm_id_t *appl_id)
{
	list_identity_properties (db_application_identity_ptr, list_ptr, appl_id);
}

void
BerkeleyDB::list_transaction_identity_properties (list_t *list_ptr, arm_id_t *tran_id)
{
	list_identity_properties (db_transaction_identity_ptr, list_ptr, tran_id);
}

void
BerkeleyDB::list_context_names (DB *db_ptr, list_t *list_ptr, arm_id_t *id_ptr)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	list_t next;
	indexed_property_name_record_t identity_name;
	indexed_property_name_record_t *context_ptr;
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_ptr != NULL);

	WRAP_EXCEPTION (status, db_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) id_ptr);
	primary_key.set_size (sizeof (*id_ptr));

	data.set_data (&identity_name);
	data.set_size (sizeof(identity_name));
	data.set_ulen (sizeof(identity_name));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			context_ptr = (indexed_property_name_record_t *) malloc (sizeof (indexed_property_name_record_t));
			context_ptr->index = identity_name.index;
			strncpy (context_ptr->name, identity_name.name, ARM_PROPERTY_NAME_MAX_LENGTH);

			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) context_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) context_ptr;
				next->next_ptr = NULL;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::list_application_context_names (list_t *list_ptr, arm_id_t *appl_id)
{
	list_context_names (db_application_context_names_ptr, list_ptr, appl_id);
}

void
BerkeleyDB::list_transaction_context_names (list_t *list_ptr, arm_id_t *tran_id)
{
	list_context_names (db_transaction_context_names_ptr, list_ptr, tran_id);
}

void
BerkeleyDB::list_context_values (DB *db_ptr, list_t *list_ptr, arm_int64_t handle)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	list_t next;
	indexed_property_value_record_t context_value;
	indexed_property_value_record_t *context_ptr;
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_ptr != NULL);

	WRAP_EXCEPTION (status, db_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
		return;
	}

	primary_key.set_data ((void *) &handle);
	primary_key.set_size (sizeof (handle));

	data.set_data (&context_value);
	data.set_size (sizeof(context_value));
	data.set_ulen (sizeof(context_value));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			context_ptr = (indexed_property_value_record_t *) malloc (sizeof (indexed_property_value_record_t));
			context_ptr->index = context_value.index;
			strncpy (context_ptr->value, context_value.value, ARM_PROPERTY_VALUE_MAX_LENGTH);

			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) context_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) context_ptr;
				next->next_ptr = NULL;
			}
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (cursor_ptr, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();
}

void
BerkeleyDB::list_application_context_values (list_t *list_ptr, arm_int64_t handle)
{
	list_context_values (db_application_context_values_ptr, list_ptr, handle);
}

void
BerkeleyDB::list_transaction_context_values (list_t *list_ptr, arm_int64_t handle)
{
	list_context_values (db_transaction_context_values_ptr, list_ptr, handle);
}

void
BerkeleyDB::list_application_group_name (arm_char_t **group_ptr, arm_int64_t handle)
{
	int	status;
	Dbt	primary_key, data;
	char buffer [ARM_PROPERTY_VALUE_MAX_LENGTH];
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_group_ptr != NULL);

	primary_key.set_data ((void *) &handle);
	primary_key.set_size (sizeof (handle));

	data.set_data (&buffer);
	data.set_size (sizeof(buffer));
	data.set_ulen (sizeof(buffer));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*group_ptr = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, db_application_group_ptr->get (NULL, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		*group_ptr = (arm_char_t *) malloc (data.get_size ()); /* Includes the trailing null */
		strncpy (*group_ptr, buffer, data.get_size ());
	}
}

void
BerkeleyDB::list_application_instance_id (arm_char_t **id_ptr, arm_int64_t handle)
{
	int	status;
	Dbt	primary_key, data;
	char buffer [ARM_PROPERTY_VALUE_MAX_LENGTH];
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_instance_names_ptr != NULL);

	primary_key.set_data ((void *) &handle);
	primary_key.set_size (sizeof (handle));

	data.set_data (&buffer);
	data.set_size (sizeof(buffer));
	data.set_ulen (sizeof(buffer));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*id_ptr = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, db_application_instance_names_ptr->get (NULL, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		*id_ptr = (arm_char_t *) malloc (data.get_size ()); /* Includes the trailing null */
		strncpy (*id_ptr, buffer, data.get_size ());
	}
}

void
BerkeleyDB::list_transaction_uri (arm_char_t **uri_ptr, arm_id_t *tran_id)
{
	int	status;
	Dbt	primary_key, data;
	char buffer [ARM_PROPERTY_URI_MAX_LENGTH];
	
    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_identity_uri_ptr != NULL);

	primary_key.set_data ((void *) tran_id);
	primary_key.set_size (sizeof (*tran_id));

	data.set_data (&buffer);
	data.set_size (sizeof(buffer));
	data.set_ulen (sizeof(buffer));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*uri_ptr = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, db_transaction_identity_uri_ptr->get (NULL, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		*uri_ptr = (arm_char_t *) malloc (data.get_size ()); /* Includes the trailing null */
		strncpy (*uri_ptr, buffer, data.get_size ());
	}
}

void
arm4db_list_applications (list_t *list_ptr)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	application_record_t application;
	list_t next;
	application_summary_t *application_ptr;

    arm4db_shm_db_read_lock ();
	
	WRAP_EXCEPTION (status, arm4db_open_databases ());
	if (status != ARM_SUCCESS)
    {
        arm4db_shm_db_unlock ();
		return;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_application_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        arm4db_shm_db_unlock ();
		return;
	}

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			application_ptr = (application_summary_t *) malloc (sizeof (application_summary_t));
			application_ptr->appl_id = *((const arm_id_t *) primary_key.set_data);
			memcpy (&(application_ptr->application), &application, sizeof (application));
			
			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) application_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) application_ptr;
				next->next_ptr = NULL;
			}

			list_application_identity_properties (&(application_ptr->identity_properties), &(application_ptr->appl_id));
			list_application_context_names (&(application_ptr->context_names), &(application_ptr->appl_id));
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

    arm4db_shm_db_unlock ();
}

void
arm4db_list_application_instances (const arm_id_t *app_id_ptr, list_t *list_ptr)
{
	int	status;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	application_instance_record_t application;
	list_t next;
	application_instance_summary_t *application_ptr;

    arm4db_shm_db_read_lock ();
	
	WRAP_EXCEPTION (status, arm4db_open_databases ());
	if (status != ARM_SUCCESS)
    {
        arm4db_shm_db_unlock ();
        return;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_application_instance_app_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_application_instance_app_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        arm4db_shm_db_unlock ();
		return;
	}

	secondary_key.set_data ((void *) app_id_ptr);
	secondary_key.set_size (sizeof (*app_id_ptr));

	data.set_data (&application);
	data.set_size (sizeof(application));
	data.set_ulen (sizeof(application));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			application_ptr = (application_instance_summary_t *) malloc (sizeof (application_instance_summary_t));
			application_ptr->app_handle = *((const arm_app_start_handle_t *) primary_key.set_data);
			memcpy (&(application_ptr->application), &application, sizeof (application));
			
			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) application_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) application_ptr;
				next->next_ptr = NULL;
			}

			list_application_group_name (&(application_ptr->application_group), application_ptr->app_handle);
			list_application_instance_id (&(application_ptr->application_instance_id), application_ptr->app_handle);
			list_application_context_values (&(application_ptr->context_values), application_ptr->app_handle);
			
			/* Move on to the next application entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

    arm4db_shm_db_unlock ();
}

void
arm4db_list_transactions (list_t *list_ptr)
{
	int	status;
	Dbt	primary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;
	list_t next;
	transaction_summary_t *transaction_ptr;

    arm4db_shm_db_read_lock ();
	
	WRAP_EXCEPTION (status, arm4db_open_databases ());
	if (status != ARM_SUCCESS)
    {
        arm4db_shm_db_unlock ();
		return;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        arm4db_shm_db_unlock ();
		return;
	}

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_FIRST));
	if (status == 0)
	{
		do
		{
			transaction_ptr = (transaction_summary_t *) malloc (sizeof (transaction_summary_t));
			transaction_ptr->tran_id = *((const arm_id_t *) primary_key.set_data);
			memcpy (&(transaction_ptr->transaction), &transaction, sizeof (transaction));
			
			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) transaction_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) transaction_ptr;
				next->next_ptr = NULL;
			}

			list_transaction_identity_properties (&(transaction_ptr->identity_properties), &(transaction_ptr->tran_id));
			list_transaction_context_names (&(transaction_ptr->context_names), &(transaction_ptr->tran_id));
			transaction_ptr->uri = NULL;
			
			/* Move on to the next transaction entry */
			WRAP_EXCEPTION (status, cursor_ptr->get (&primary_key, &data, DB_NEXT));
		} while (status == 0);
	}
	cursor_ptr->close ();

    arm4db_shm_db_unlock ();
}

void
arm4db_list_app_transactions (const arm_id_t *app_id_ptr, list_t *list_ptr)
{
	int	status;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_record_t transaction;
	list_t next;
	transaction_summary_t *transaction_ptr;

    arm4db_shm_db_read_lock ();
	
	WRAP_EXCEPTION (status, arm4db_open_databases ());
	if (status != ARM_SUCCESS)
    {
        arm4db_shm_db_unlock ();
		return;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_appid_ptr != NULL);
	
	WRAP_EXCEPTION (status, db_transaction_appid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (status, "Unable to create cursor");
        arm4db_shm_db_unlock ();
		return;
	}

	secondary_key.set_data ((void *) app_id_ptr);
	secondary_key.set_size (sizeof (*app_id_ptr));

	data.set_data (&transaction);
	data.set_size (sizeof(transaction));
	data.set_ulen (sizeof(transaction));
	data.set_flags (DB_DBT_USERMEM);

	/* Start the list */
	*list_ptr = NULL;
	next = NULL;
	
	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		do
		{
			transaction_ptr = (transaction_summary_t *) malloc (sizeof (transaction_summary_t));
			transaction_ptr->tran_id = *((const arm_id_t *) primary_key.set_data);
			memcpy (&(transaction_ptr->transaction), &transaction, sizeof (transaction));
			
			if (next == NULL)
			{
				next = (list_entry_t *) malloc (sizeof (list_entry_t));
				next->data_ptr = (void *) transaction_ptr;
				next->next_ptr = NULL;
				*list_ptr = next;
			}
			else
			{
				next->next_ptr = (list_entry_t *) malloc (sizeof (list_entry_t));
				next = next->next_ptr;
				next->data_ptr = (void *) transaction_ptr;
				next->next_ptr = NULL;
			}

			list_transaction_identity_properties (&(transaction_ptr->identity_properties), &(transaction_ptr->tran_id));
			list_transaction_context_names (&(transaction_ptr->context_names), &(transaction_ptr->tran_id));
			list_transaction_uri (&(transaction_ptr->uri), &(transaction_ptr->tran_id));
			
			/* Move on to the next transaction entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
	}
	cursor_ptr->close ();

    arm4db_shm_db_unlock ();
}

void
arm4db_summarize_transaction (const arm_id_t *tran_id, transaction_stat_summary_t *summary_ptr)
{
	int	status;
	Dbt	primary_key, secondary_key, data;
	Dbc *cursor_ptr;
	transaction_instance_record_t transaction_instance;
	arm_tran_start_handle_t transaction_instance_id;
	arm_time_t	min_time, max_time;
	double raw_time;
	double total_time;
	double total_blocked;
	double total_time_squared;
	double deviation;
	int count;
	int status_good, status_aborted, status_failed, status_unknown;
	int discard;

    arm4db_shm_db_read_lock ();

    WRAP_EXCEPTION (status, arm4db_open_databases ());
    if (status != ARM_SUCCESS)
    {
        arm4db_shm_db_unlock ();
        return;
    }

    DB_ASSERT (m_environmentPtr != NULL);
    DB_ASSERT (db_transaction_instance_tranid_ptr != NULL);
	
	/* Inititalize our stats */
	memset (summary_ptr, sizeof (*summary_ptr), 0);

	/* This lookup is based on the secondary key of tran_id */
	WRAP_EXCEPTION (status, db_transaction_instance_tranid_ptr->cursor (NULL, &cursor_ptr, 0));
	if (status != 0)
	{
		m_environmentPtr->err (environment_ptr, status,     "Unable to create cursor");
        arm4db_shm_db_unlock ();
		return;
	}

	primary_key.set_data ((void *) &transaction_instance_id);
	primary_key.set_size (sizeof (arm_tran_start_handle_t));
	primary_key.set_ulen (sizeof (arm_tran_start_handle_t));
	primary_key.set_flags (DB_DBT_USERMEM);

	secondary_key.set_data ((void *) tran_id);
	secondary_key.set_size (sizeof (arm_id_t));

	data.set_data (&transaction_instance);
	data.set_size (sizeof(transaction_instance));
	data.set_ulen (sizeof(transaction_instance));
	data.set_flags (DB_DBT_USERMEM);

	/* Initialize the stats, except min_time which comes later */
	max_time.tv_sec = max_time.tv_usec = 0;
	total_time = 0.0;
	total_blocked = 0.0;
	total_time_squared = 0.0;
	count = 0;
	status_good = status_aborted = status_failed = status_unknown = discard = 0;

	/* Iterate through the cursor */
	WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_SET));
	if (status == 0)
	{
		min_time = transaction_instance.response_time;
		do
		{
			count++;
			raw_time = transaction_instance.response_time.tv_sec +
				(transaction_instance.response_time.tv_usec * 1e-6);
			total_time += raw_time;
			total_time_squared += (raw_time * raw_time);
			if (time_greater (&transaction_instance.response_time, &max_time))
				max_time = transaction_instance.response_time;
			if (time_less (&transaction_instance.response_time, &min_time))
				min_time = transaction_instance.response_time;
			raw_time = transaction_instance.blocked_time.tv_sec +
				(transaction_instance.blocked_time.tv_usec * 1e-6);
			total_blocked += raw_time;
			
			if (transaction_instance.active == ARM_DISCARD)
				discard++;
			else
			{
				switch (transaction_instance.tran_status)
				{
					case ARM_STATUS_GOOD:
						status_good++;
						break;
					
					case ARM_STATUS_ABORTED:
						status_aborted++;
						break;
					
					case ARM_STATUS_FAILED:
						status_aborted++;
						break;
					
					case ARM_STATUS_UNKNOWN:
					default:
						status_unknown++;
				}
			}

			/* Move on to the next transaction instance entry */
			WRAP_EXCEPTION (status, cursor_ptr->pget (&secondary_key, &primary_key, &data, DB_NEXT_DUP));
		} while (status == 0);
        if (count <= 1)
        {
            deviation = 0.0;
        }
        else
        {
            deviation = sqrt ((total_time_squared - (total_time * total_time / count)) / count);
        }
		summary_ptr->min_time = min_time;
		summary_ptr->max_time = max_time;
		summary_ptr->total_time = total_time;
		summary_ptr->total_blocked = total_blocked;
		summary_ptr->deviation = deviation;
		summary_ptr->status_good = status_good;
		summary_ptr->status_aborted = status_aborted;
		summary_ptr->status_failed = status_failed;
		summary_ptr->status_unknown = status_unknown;
		summary_ptr->discard = discard;
		summary_ptr->count = count;
	}
	else
	{
		if (status != DB_NOTFOUND)
			m_environmentPtr->err (status, "No records found");
	}
	cursor_ptr->close ();

    arm4db_shm_db_unlock ();
}
#endif

