/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"
#if defined(HAVE_EXECINFO_H)
#include <execinfo.h>
#endif
#include <iomanip>
#include <netinet/in.h>
#include <libgen.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#ifdef HAVE_UUID_UUID_H
#include <uuid/uuid.h>
#else
#ifdef HAVE_UUID_H
#include <uuid.h>
#endif
#endif

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4dbCommon.h"
#include "Arm4dbExceptions.h"
#include "Logger.h"

/* This needs to match the definition of the arm4_shm_sequence_t enum */
char Arm4dbCommon::m_sequence_name_table [][64] = {
	"APPLICATION_SEQUENCE",
	"TRANSACTION_SEQUENCE",
	"APPLICATION_INSTANCE_SEQUENCE",
	"TRANSACTION_INSTANCE_SEQUENCE",
	"TRANSACTION_BLOCKING_SEQUENCE"
};

Arm4dbCommon::~Arm4dbCommon ()
{
}

void
Arm4dbCommon::db_assert (int assert, const char *file, int line)
{
    if (!assert)
    {
        fprintf (stderr, "Assertion failed File (%s) Line %d\n", file, line);
		Logger::getLogger().log ("Assertion failed File (%s) Line %d\n", file, line);

		// Show the call stack
#if defined(HAVE_EXECINFO_H)
		void* callstack[128];
		int i, frames = backtrace(callstack, 128);
		char** strs = backtrace_symbols(callstack, frames);
		for (i = 0; i < frames; ++i) {
			Logger::getLogger().log ("%s\n", strs[i]);
		}
		free(strs);
#endif

		throw new Arm4dbException ("Assertion failed File (%s) Line %d\n", file, line);
    }
}

std::ostream&
Arm4dbCommon::printId (std::ostream &os, const arm_id_t *id_ptr)
{
	os.setf ( std::ios::hex, std::ios::basefield );
	os.fill ('0');
	os.width (2);

	for (int i = 0; i < 16; i++)
		os << std::setfill ('0') << std::setw (2) << (int) (id_ptr->id_u.uint8 [i]);

	os.setf ( std::ios::dec, std::ios::basefield );
	return os;
}

std::ostream&
Arm4dbCommon::printDigest (std::ostream &os, const char *digest_ptr)
{
	os.setf ( std::ios::hex, std::ios::basefield );
	os.fill ('0');
	os.width (2);

	for (int i = 0; i < MD5_DIGEST_LENGTH; i++)
		os << std::setfill ('0') << std::setw (2) << (int) (digest_ptr [i] & 0x0FF);

	os.setf ( std::ios::dec, std::ios::basefield );
	return os;
}

std::ostream&
Arm4dbCommon::printMetricValue (std::ostream &os, const arm_metric_t &value, const std::string &string_value)
{
	switch (value.format)
		{
		case ARM_METRIC_FORMAT_COUNTER32:
			os << value.metric_u.counter32;
			break;

		case ARM_METRIC_FORMAT_COUNTER64:
			os << value.metric_u.counter64;
			break;

		case ARM_METRIC_FORMAT_CNTRDIVR32:
			os << value.metric_u.counterdivisor32.counter << ":"
				<< value.metric_u.counterdivisor32.divisor;
			break;

		/* Java implementation specific */
		case ARM_METRIC_FORMAT_COUNTERFLOAT32:
			// This is a kludge to put floats in this field, but our float_kludge.m4 macro
			// ensures it fits
			os << *(reinterpret_cast<const float *> (&(value.metric_u.counterdivisor32)));
			break;

		case ARM_METRIC_FORMAT_GAUGE32:
			os << value.metric_u.gauge32;
			break;

		case ARM_METRIC_FORMAT_GAUGE64:
			os << value.metric_u.gauge64;
			break;

		case ARM_METRIC_FORMAT_GAUGEDIVR32:
			os << value.metric_u.gaugedivisor32.gauge << ":"
				<< value.metric_u.gaugedivisor32.divisor;
			break;

		/* Java implementation specific */
		case ARM_METRIC_FORMAT_GAUGEFLOAT32:
			// This is a kludge to put floats in this field, but our float_kludge.m4 macro
			// ensures it fits
			os << *(reinterpret_cast<const float *> (&(value.metric_u.gaugedivisor32)));
			break;

		case ARM_METRIC_FORMAT_NUMERICID32:
			os << value.metric_u.numericid32;
			break;

		case ARM_METRIC_FORMAT_NUMERICID64:
			os << value.metric_u.numericid64;
			break;

		case ARM_METRIC_FORMAT_STRING32:
			escapeForXml(os, string_value);
			break;
		}

	return os;
}

std::ostream&
Arm4dbCommon::printCorrelator(std::ostream &os, const arm_correlator_t *correlator_ptr)
{
	os.setf ( std::ios::hex, std::ios::basefield );
	os.fill ('0');
	os.width (2);

	int size = correlatorSize (correlator_ptr);
	for (int i = 0; i < size; i++)
		os << std::setfill ('0') << std::setw (2) << (int) (correlator_ptr->opaque [i] & 0x0FF);

	os.setf ( std::ios::dec, std::ios::basefield );
	return os;
}

std::ostream&
Arm4dbCommon::printIpV4Address (std::ostream &os, const arm_uint8_t *address_ptr)
{
	os << "IPv4: "
		<< address_ptr [0] << "."
		<< address_ptr [1] << "."
		<< address_ptr [2] << "."
		<< address_ptr [3];
	return os;
}

std::ostream&
Arm4dbCommon::printIpV4PortAddress (std::ostream &os, const arm_uint8_t *address_ptr)
{
	os << "IPv4: "
		<< address_ptr [0] << "."
		<< address_ptr [1] << "."
		<< address_ptr [2] << "."
		<< address_ptr [3] << " Port: "
		<< *((arm_uint16_t *) &(address_ptr [4]));
	return os;
}

std::ostream&
Arm4dbCommon::printAddress (std::ostream &os, const arm_uint8_t *address_ptr, int length)
{
	os.setf ( std::ios::hex, std::ios::basefield );
	os.fill ('0');
	os.width (2);

	for (int i = 0; i < length; i++)
		os << std::setfill ('0') << std::setw (2) << ((int) address_ptr [i] & 0x0FF);

	os.setf ( std::ios::dec, std::ios::basefield );

	return os;
}

std::ostream&
Arm4dbCommon::printCharacterAddress (std::ostream &os, const arm_uint8_t *address_ptr, int length)
{
	for (int i = 0; i < length; i++)
		os << (char) (address_ptr [i]);

	return os;
}

std::ostream&
Arm4dbCommon::printAddress (std::ostream &os, const application_address_record_t *address_ptr)
{
	switch (address_ptr->address_format)
		{
		case ARM_SYSADDR_FORMAT_IPV4:
			printIpV4Address (os, address_ptr->address);
			break;

		case ARM_SYSADDR_FORMAT_IPV4PORT:
			os << printIpV4PortAddress (os, address_ptr->address);
			break;

		case ARM_SYSADDR_FORMAT_IPV6:
			os << "IPv6: "
				<< printAddress (os, address_ptr->address, address_ptr->address_length);
			break;

		case ARM_SYSADDR_FORMAT_IPV6PORT:
			os << "IPv6/Port: "
				<< printAddress (os, address_ptr->address, address_ptr->address_length);
			break;

		case ARM_SYSADDR_FORMAT_SNA:
			os << "SNA: "
				<< printAddress (os, address_ptr->address, address_ptr->address_length);
			break;

		case ARM_SYSADDR_FORMAT_X25:
			os << "X.25: "
				<< printCharacterAddress (os, address_ptr->address, address_ptr->address_length);
			break;

		case ARM_SYSADDR_FORMAT_HOSTNAME:
			os << "Hostname: ";
			os << printCharacterAddress (os, address_ptr->address, address_ptr->address_length);
			break;

		case ARM_SYSADDR_FORMAT_UUID:
			os << "UUID: "
				<< printAddress (os, address_ptr->address, address_ptr->address_length);
			break;

		default:
			os << "Unknown address format: "
				<< printAddress (os, address_ptr->address, address_ptr->address_length);
		}

	return os;
}

std::ostream&
Arm4dbCommon::escapeForXml (std::ostream &os, const std::string &string_value)
{
	int count = string_value.length ();
	const char *data_ptr = string_value.data ();
	for (int i = 0; i < count; i++)
	{
		char c = data_ptr [i];
		switch (c)
		{
		case '&':
			os << "&amp;";
			break;

		case '"':
			os << "&quot;";
			break;

		case '<':
			os << "&lt;";
			break;

		case '>':
			os << "&gt;";
			break;

		default:
			os << c;
		}
	}

	return os;
}

std::ostream&
Arm4dbCommon::escapeForXml (std::ostream &os, const char *string_ptr, int max_length)
{
	// This ensures the string remains in scope
	int length = stringLength (string_ptr, max_length);
	std::string new_string (string_ptr, length);

	return escapeForXml (os, new_string);
}

int
Arm4dbCommon::correlatorSize (const arm_correlator_t *correlator_ptr)
{
	if (correlator_ptr == NULL)
		return 0;

	const arm_correlator_header_t *header_ptr = (arm_correlator_header_t *) correlator_ptr;
	int size = ntohs (header_ptr->length);
	if (size < ARM_CORR_MIN_LENGTH || size > ARM_CORR_MAX_LENGTH)
		return 0;

	return size;
}

const char *
Arm4dbCommon::tempPath (int instance, const char *filename_ptr)
{
	if (instance == 0)
		return createRootTempPath (filename_ptr);
	else
		return createUserTempPath (instance, filename_ptr);
}

const char *
Arm4dbCommon::createRootTempPath (const char *filename_ptr)
{
	static char buffer [4096];

	createPath (ARM4DATA_QUEUE_DIR);
	sprintf (buffer, ARM4DATA_QUEUE_DIR "/%s", filename_ptr);

	return buffer;
}

const char *
Arm4dbCommon::createUserTempPath (int instance, const char *filename_ptr)
{
	static char base_buffer [4096];
	static char dir_buffer [4096];

	sprintf (base_buffer, "/tmp/arm4-%d-%d", getuid(), instance);

	createPath (base_buffer);
	sprintf (dir_buffer, "%s/queue", base_buffer);
	if (access (dir_buffer, F_OK) != 0)
	{
		/* Create the directory path */
		mkdir (dir_buffer, 0777);
	}

	sprintf (dir_buffer, "%s/queue/%s", base_buffer, filename_ptr);

	return dir_buffer;
}

arm_error_t
Arm4dbCommon::createPath (const char *path_ptr)
{
    char *pathname_ptr;
    char *dirname_ptr;
    char *basename_ptr;

    /* Ensure the pathname exists, or create it if it doesn't */
    if (access (path_ptr, F_OK) == 0)
        return ARM_SUCCESS;

    /* First create a copy since the functions used later may modify the original */
    pathname_ptr = reinterpret_cast<char *> (malloc (strlen (path_ptr) + 1));
    strcpy (pathname_ptr,  path_ptr);

    /* Now break it down into it's parts */
    dirname_ptr = dirname (pathname_ptr);
    basename_ptr = basename (pathname_ptr);

    if ((strcmp (basename_ptr,  ".") == 0) || (strcmp (basename_ptr, "..") == 0))
    {
        /* If we're at the current or parent directory pointers, we're done */
		free (pathname_ptr);
        return ARM_SUCCESS;
    }

    /* Create the parent */
    createPath (dirname_ptr);
    if (access (dirname_ptr, W_OK) != 0)
	{
		Logger::getLogger().log ("Unable to create directory (%s)\n", dirname_ptr);

		free (pathname_ptr);
        return ARM_FAILURE; /* Can't create */
	}
	free (pathname_ptr);

    /* Create the directory */
    /* Use the original unmodified string value, and rely on umask to modify the permissions correctly */
    if (mkdir (path_ptr, 0777) != 0)
    {
		Logger::getLogger().log ("Unable to create directory (%s)\n", path_ptr);
        return ARM_FAILURE;
    }

    return ARM_SUCCESS;
}

void
Arm4dbCommon::getUuid (arm_id_t *uuid_ptr)
{
    uuid_generate ((arm_uint8_t *) uuid_ptr);
}

