/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include "arm4agent.h"
#include "Arm4db.h"
#include "Arm4dbConfig.h"
#include "Logger.h"
#include "BerkeleyDB.h"

Arm4dbConfig *Arm4dbConfig::m_singletonPtr = NULL;	// Configuration object

const Arm4dbConfig &
Arm4dbConfig::getConfig (const std::string &filename)
{
	// Replace the current config object
	if (m_singletonPtr != NULL)
		delete m_singletonPtr;
    m_singletonPtr = new Arm4dbConfig (filename);
	Logger::getLogger().setFile (m_singletonPtr->getLogFilename ());

	return *m_singletonPtr;
}

const Arm4dbConfig &
Arm4dbConfig::getConfig (void)
{
	if (m_singletonPtr == NULL)
	{
		m_singletonPtr = new Arm4dbConfig ();
		Logger::getLogger().setFile (m_singletonPtr->getLogFilename ());
	}

	return *m_singletonPtr;
}

void
Arm4dbConfig::setConfig (const std::string &filename)
{
	// Replace the current config object
	if (m_singletonPtr != NULL)
		delete m_singletonPtr;

    m_singletonPtr = new Arm4dbConfig (filename);
	Logger::getLogger().setFile (m_singletonPtr->getLogFilename ());
}

Arm4dbConfig::Arm4dbConfig () :
	m_filename (""),
	m_hostname (""),
	m_logFilename (""),
	m_dbType (DEFAULT_ARM4DB_TYPE),
	m_instance (0),
	m_uid ((uid_t) -1),
	m_gid ((gid_t) -1),
	m_umask (0077),   /* Block group and world access by default */
	m_schemaLocation (DEFAULT_SCHEMA_LOCATION),
	m_dbHome (ARM4DATA_DB_DIR),
	m_dbBackup (ARM4DATA_DB_BACKUP_DIR),
	m_dbKey (0), /* No key by default */
	m_dbLogFileSize (DEFAULT_LOG_FILE_SIZE),
	m_dbLogBufferSize (DEFAULT_LOG_BUFFER_SIZE),
	m_checkpointInterval (120),
	m_mysqlHost ("localhost"),
	m_mysqlUser ("arm4"),
	m_mysqlPassword (""), // No default is used here
	m_mysqlDb ("arm4"),
	m_mysqlSocket (""),
	m_mysqlPort (0),
	m_dbHomeC (NULL),
	m_dbBackupC (NULL),
	m_mysqlHostC (NULL),
	m_mysqlUserC (NULL),
	m_mysqlPasswordC (NULL),
	m_mysqlDbC (NULL),
	m_mysqlSocketC (NULL)
{
	setDefaultHostname ();
	adjustDbHome ();
}

Arm4dbConfig::Arm4dbConfig (const Arm4dbConfig &config) :
	m_filename (config.m_filename),
	m_hostname (config.m_hostname),
	m_logFilename (config.m_logFilename),
	m_dbType (config.m_dbType),
	m_instance (config.m_instance),
	m_uid (config.m_uid),
	m_gid (config.m_gid),
	m_umask (config.m_umask),
	m_schemaLocation (config.m_schemaLocation),
	m_dbHome (config.m_dbHome),
	m_dbBackup (config.m_dbBackup),
	m_dbKey (config.m_dbKey),
	m_dbLogFileSize (config.m_dbLogFileSize),
	m_dbLogBufferSize (config.m_dbLogBufferSize),
	m_checkpointInterval (config.m_checkpointInterval),
	m_mysqlHost (config.m_mysqlHost),
	m_mysqlUser (config.m_mysqlUser),
	m_mysqlPassword (config.m_mysqlPassword),
	m_mysqlDb (config.m_mysqlDb),
	m_mysqlSocket (config.m_mysqlSocket),
	m_mysqlPort (config.m_mysqlPort),
	m_dbHomeC (NULL),
	m_dbBackupC (NULL),
	m_mysqlHostC (NULL),
	m_mysqlUserC (NULL),
	m_mysqlPasswordC (NULL),
	m_mysqlDbC (NULL),
	m_mysqlSocketC (NULL)
{
	adjustDbHome ();
}


Arm4dbConfig::Arm4dbConfig (const std::string &filename) :
	m_filename (""),
	m_hostname (""),
	m_logFilename (""),
	m_dbType (DEFAULT_ARM4DB_TYPE),
	m_instance (0),
	m_uid ((uid_t) -1),
	m_gid ((gid_t) -1),
	m_umask (0077),   /* Block group and world access by default */
	m_schemaLocation (DEFAULT_SCHEMA_LOCATION),
	m_dbHome (ARM4DATA_DB_DIR),
	m_dbBackup (ARM4DATA_DB_BACKUP_DIR),
	m_dbKey (0), /* No key by default */
	m_dbLogFileSize (DEFAULT_LOG_FILE_SIZE),
	m_dbLogBufferSize (DEFAULT_LOG_BUFFER_SIZE),
	m_checkpointInterval (120),
	m_mysqlHost ("localhost"),
	m_mysqlUser ("arm4"),
	m_mysqlPassword (""), // No default is used here
	m_mysqlDb ("arm4"),
	m_mysqlSocket (""),
	m_mysqlPort (0),
	m_dbHomeC (NULL),
	m_dbBackupC (NULL),
	m_mysqlHostC (NULL),
	m_mysqlUserC (NULL),
	m_mysqlPasswordC (NULL),
	m_mysqlDbC (NULL),
	m_mysqlSocketC (NULL)
{
	setDefaultHostname ();
	read (filename.c_str());
	adjustDbHome ();
}

Arm4dbConfig::~Arm4dbConfig ()
{
	if (m_dbHomeC)
		delete [] m_dbHomeC;
	if (m_dbBackupC)
		delete [] m_dbBackupC;
	if (m_mysqlHostC)
		delete [] m_mysqlHostC;
	if (m_mysqlUserC)
		delete [] m_mysqlUserC;
	if (m_mysqlPasswordC)
		delete [] m_mysqlPasswordC;
	if (m_mysqlDbC)
		delete [] m_mysqlDbC;
	if (m_mysqlSocketC)
		delete [] m_mysqlSocketC;
}

Arm4dbConfig &
Arm4dbConfig::operator= (const Arm4dbConfig &config)
{
	if (&config != this)
	{
		m_filename = config.m_filename;
		m_hostname = config.m_hostname;
		m_logFilename = config.m_logFilename;
		m_dbType = config.m_dbType;
		m_instance = config.m_instance;
		m_uid = config.m_uid;
		m_gid = config.m_gid;
		m_umask = config.m_umask;
		m_schemaLocation = config.m_schemaLocation;
		m_dbHome = config.m_dbHome;
		m_dbBackup = config.m_dbBackup;
		m_dbKey = config.m_dbKey;
		m_dbLogFileSize = config.m_dbLogFileSize;
		m_dbLogBufferSize = config.m_dbLogBufferSize;
		m_checkpointInterval = config.m_checkpointInterval;
		m_mysqlHost = config.m_mysqlHost;
		m_mysqlUser = config.m_mysqlUser;
		m_mysqlPassword = config.m_mysqlPassword;
		m_mysqlDb = config.m_mysqlDb;
		m_mysqlSocket = config.m_mysqlSocket;
		m_mysqlPort = config.m_mysqlPort;
		clearC (&m_dbHomeC);
		clearC (&m_dbBackupC);
		clearC (&m_mysqlHostC);
		clearC (&m_mysqlUserC);
		clearC (&m_mysqlPasswordC);
		clearC (&m_mysqlDbC);
		clearC (&m_mysqlSocketC);

		adjustDbHome ();
	}

	return *this;
}

void
Arm4dbConfig::setDefaultHostname (void)
{
	char hostname [HOST_NAME_MAX];
	gethostname (hostname, sizeof (hostname));

	setHostname (hostname);
}

void
Arm4dbConfig::setDBHome (const std::string &home)
{
	if (m_singletonPtr != NULL)
		m_singletonPtr->setDbHome (home);
}

void
Arm4dbConfig::read (const char *config_filename_ptr)
{
	FILE *file_ptr;
	char buffer [4096];
	char *parameter_ptr;
	char *value_ptr;

	file_ptr = fopen (config_filename_ptr, "r");
	if (file_ptr == NULL)
		return;

	m_filename = config_filename_ptr;
	while (!feof (file_ptr))
	{
		read_line (file_ptr, buffer, sizeof (buffer));
		parameter_ptr = uncomment (buffer);
		if (*parameter_ptr != '\0')
		{
			get_parameters (parameter_ptr, &value_ptr);
			if (strcasecmp (parameter_ptr, "hostname") == 0)
			{
				parse_hostname (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "log_filename") == 0)
			{
				parse_log_filename (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "user") == 0)
			{
				parse_uid (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "group") == 0)
			{
				parse_gid (value_ptr);
			}
            else if (strcasecmp (parameter_ptr, "umask") == 0)
            {
                parse_umask (value_ptr);
            }
			else if (strcasecmp (parameter_ptr, "schema_location") == 0)
			{
				parse_schema_location (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "db") == 0)
			{
				parse_db (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "db_home_dir") == 0)
			{
				parse_db_home (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "db_backup_dir") == 0)
			{
				parse_db_backup (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "db_key") == 0)
			{
				parse_db_key (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "db_log_buffer_size") == 0)
			{
				parse_db_log_buffer_size (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "db_log_file_size") == 0)
			{
				parse_db_log_file_size (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "instance") == 0)
			{
				parse_instance (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "checkpoint") == 0)
			{
				parse_checkpoint_interval (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "mysql_host") == 0)
			{
				parse_mysql_host (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "mysql_user") == 0)
			{
				parse_mysql_user (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "mysql_password") == 0)
			{
				parse_mysql_password (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "mysql_db") == 0)
			{
				parse_mysql_db (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "mysql_socket") == 0)
			{
				parse_mysql_socket (value_ptr);
			}
			else if (strcasecmp (parameter_ptr, "mysql_port") == 0)
			{
				parse_mysql_port (value_ptr);
			}
			else
			{
				fprintf (stderr, "Unrecognized configuration parameter '%s'\n", parameter_ptr);
			}
		}
	}

    fclose (file_ptr);
}

int
Arm4dbConfig::parse_integer (const char *value_ptr)
{
    int value;

    /* Using the %i for sscanf allows us to scan octal and hex values as well */
    sscanf (value_ptr, "%i", &value);

    return value;
}

long
Arm4dbConfig::parse_long (const char *value_ptr)
{
    long value;

    /* Using the %i for sscanf allows us to scan octal and hex values as well */
    sscanf (value_ptr, "%li", &value);

    return value;
}

void
Arm4dbConfig::read_line (FILE *file, char *buffer_ptr, int size)
{
	int c;

	/* Allow for a null terminator */
	size--;
	while (size--)
	{
		c = fgetc(file);
		switch (c)
		{
		case EOF:
		case '\r':
		case '\n':
			*buffer_ptr++ = '\0';
			return;

		default:
			*buffer_ptr++ = c;
		}
	}
	*buffer_ptr++ = '\0';
}

char *
Arm4dbConfig::uncomment (char *buffer_ptr)
{
	char *comment_ptr;

	comment_ptr = index(buffer_ptr, '#');
	if (comment_ptr != NULL)
		*comment_ptr = '\0';

	while (isspace (*buffer_ptr))
		buffer_ptr++;
	return buffer_ptr;
}

void
Arm4dbConfig::trim_tail (char *buffer_ptr)
{
	while (*buffer_ptr && !isspace (*buffer_ptr))
		buffer_ptr++;
	*buffer_ptr = '\0';
}

void
Arm4dbConfig::get_parameters (char *line_ptr, char **value_pptr)
{
	*value_pptr = line_ptr;
	*value_pptr = strchr (*value_pptr, '=');
    if (*value_pptr)
    {
        **value_pptr = '\0';
        (*value_pptr)++;
    }

	if (!(*value_pptr))
	{
		/* try a space separator */
		*value_pptr = line_ptr;
        *value_pptr = strchr (*value_pptr, ' ');
        if (*value_pptr)
        {
            **value_pptr = '\0';
            (*value_pptr)++;
        }
	}

	if (!(*value_pptr))
	{
		/* try a tab separator */
		*value_pptr = line_ptr;
        *value_pptr = strchr (*value_pptr, '\t');
        if (*value_pptr)
        {
            **value_pptr = '\0';
            (*value_pptr)++;
        }
	}

	trim_tail (line_ptr);

	if (*value_pptr)
	{
		while (isspace (**value_pptr))
			(*value_pptr)++;
		trim_tail (*value_pptr);
	}
}

void
Arm4dbConfig::parse_hostname (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for hostname\n");
		return; /* Nothing to parse */
	}

	setHostname (value_ptr);
}

void
Arm4dbConfig::parse_log_filename (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for hostname\n");
		return; /* Nothing to parse */
	}

	setLogFilename (value_ptr);
}

void
Arm4dbConfig::parse_uid (const char *value_ptr)
{
    struct passwd *password_ptr;

	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for uid\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_uid = (uid_t) parse_integer (value_ptr);
	}
	else
	{
		password_ptr = getpwnam (value_ptr);
		if (password_ptr == NULL)
		{
			Logger::getLogger().error ("Unable to find user name");
		}
		else
			m_uid = password_ptr->pw_uid;
	}
}

void
Arm4dbConfig::parse_gid (const char *value_ptr)
{
	struct group *group_ptr;

	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for gid\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_gid = (gid_t) parse_integer (value_ptr);
	}
	else
	{
		group_ptr = getgrnam (value_ptr);
		if (group_ptr == NULL)
		{
			Logger::getLogger().error ("Unable to find group name");
		}
		else
			m_gid = group_ptr->gr_gid;
	}
}

void
Arm4dbConfig::parse_umask (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for umask\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_umask = (gid_t) parse_integer (value_ptr);
	}
	else
	{
        fprintf (stderr, "umask value must be numeric\n");
	}
}

void
Arm4dbConfig::parse_db (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for db\n");
		return; /* Nothing to parse */
	}

	if (strcasecmp (value_ptr, "bdb") == 0)
	{
		m_dbType = Arm4dbBerkeleyDB;
	}
	else if (strcasecmp (value_ptr, "sqlite") == 0)
	{
		m_dbType = Arm4dbSqlite;
	}
	else if (strcasecmp (value_ptr, "mysql") == 0)
	{
		m_dbType = Arm4dbMySql;
	}
	else
	{
		fprintf (stderr, "Unrecognized value for db\n");
	}
}

void
Arm4dbConfig::parse_schema_location (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for schema_location\n");
		return; /* Nothing to parse */
	}

	setSchemaLocation (value_ptr);
}

void
Arm4dbConfig::parse_db_home (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for db_home_dir\n");
		return; /* Nothing to parse */
	}

	setDbHome (value_ptr);
}

void
Arm4dbConfig::parse_db_backup (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for db_backup_dir\n");
		return; /* Nothing to parse */
	}

	setDbBackup (value_ptr);
}

void
Arm4dbConfig::parse_db_key (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for db_key\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_dbKey = parse_long (value_ptr);
	}
	else
	{
		fprintf (stderr, "db_key must be a numeric value\n");
	}
}

void
Arm4dbConfig::parse_db_log_buffer_size (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for db_log_buffer_size\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_dbLogBufferSize = parse_long (value_ptr);
	}
	else
	{
		fprintf (stderr, "db_log_buffer_size must be a numeric value\n");
	}
}

void
Arm4dbConfig::parse_db_log_file_size (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for db_log_file_size\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_dbLogFileSize = parse_long (value_ptr);
	}
	else
	{
		fprintf (stderr, "db_log_file_size must be a numeric value\n");
	}
}

void
Arm4dbConfig::parse_instance (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for instance\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_instance = parse_integer (value_ptr);
	}
	else
	{
		fprintf (stderr, "Instance must be a numeric value\n");
	}
}

void
Arm4dbConfig::parse_checkpoint_interval (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for checkpoint interval\n");
		return; /* Nothing to parse */
	}

	if ((isdigit (*value_ptr)) || (*value_ptr == '-')) // Allow for negative values
	{
		m_checkpointInterval = parse_integer (value_ptr);
	}
	else
	{
		fprintf (stderr, "Checkpoint must be a numeric value\n");
	}
}

void
Arm4dbConfig::parse_mysql_host (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for mysql_host\n");
		return; /* Nothing to parse */
	}

	setMysqlHost (value_ptr);
}

void
Arm4dbConfig::parse_mysql_user (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for mysql_user\n");
		return; /* Nothing to parse */
	}

	setMysqlUser (value_ptr);
}

void
Arm4dbConfig::parse_mysql_password (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for mysql_password\n");
		return; /* Nothing to parse */
	}

	setMysqlPassword (value_ptr);
}

void
Arm4dbConfig::parse_mysql_db (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for mysql_db\n");
		return; /* Nothing to parse */
	}

	setMysqlDb (value_ptr);
}

void
Arm4dbConfig::parse_mysql_socket (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for mysql_socket\n");
		return; /* Nothing to parse */
	}

	setMysqlSocket (value_ptr);
}

void
Arm4dbConfig::parse_mysql_port (const char *value_ptr)
{
	if (!value_ptr)
	{
		fprintf (stderr, "No value specified for mysql_port\n");
		return; /* Nothing to parse */
	}

	if (isdigit (*value_ptr))
	{
		m_mysqlPort = parse_integer (value_ptr);
	}
	else
	{
		fprintf (stderr, "MySql port number must be a numeric value\n");
	}
}

void
Arm4dbConfig::adjustDbHome (void)
{
	// Change the default locations for user space instances
	if (getInstance () != 0)
	{
		if (m_dbHome == ARM4DATA_DB_DIR)
		{
			char buffer [4096];
			sprintf (buffer, "/tmp/arm4-%d-%d/data", getuid (), getInstance ());

			setDbHome (buffer);
		}

		if (m_dbBackup == ARM4DATA_DB_BACKUP_DIR)
		{
			char buffer [4096];
			sprintf (buffer, "/tmp/arm4-%d-%d/data/backup", getuid (), getInstance ());

			setDbBackup (buffer);
		}
	}
}

const char *
Arm4dbConfig::toC (const std::string &cppString, char **cString)
{
	if (*cString == NULL)
	{
		size_t bufferSize = cppString.size() + 1;
		*cString = new char [bufferSize];
		::strncpy (*cString, cppString.c_str (), bufferSize);
	}

	return *cString;
}

void
Arm4dbConfig::clearC (char **cString)
{
	if (*cString != NULL)
	{
		delete [] *cString;
		*cString = NULL;
	}
}

void
Arm4dbConfig::setFilename (const std::string &filename)
{
	m_filename = filename;
	//clearC (&m_filenameC);
}

void
Arm4dbConfig::setHostname (const std::string &hostname)
{
	m_hostname = hostname;
	//clearC (&m_hostnameC);
}

void
Arm4dbConfig::setLogFilename (const std::string &logFilename)
{
	m_logFilename = logFilename;
	//clearC (&m_logFilenameC);
}

void
Arm4dbConfig::setSchemaLocation (const std::string &schemaLocation)
{
	m_schemaLocation = schemaLocation;
	//clearC (&m_schemaLocationC);
}

void
Arm4dbConfig::setDbHome (const std::string &dbHome)
{
	m_dbHome = dbHome;
	clearC (&m_dbHomeC);
}

void
Arm4dbConfig::setDbBackup (const std::string &dbBackup)
{
	m_dbBackup = dbBackup;
	clearC (&m_dbBackupC);
}

void
Arm4dbConfig::setMysqlHost (const std::string &host)
{
	m_mysqlHost = host;
	clearC (&m_mysqlHostC);
}

void
Arm4dbConfig::setMysqlUser (const std::string &user)
{
	m_mysqlUser = user;
	clearC (&m_mysqlUserC);
}

void
Arm4dbConfig::setMysqlPassword (const std::string &password)
{
	m_mysqlPassword = password;
	clearC (&m_mysqlPasswordC);
}

void
Arm4dbConfig::setMysqlDb (const std::string &db)
{
	m_mysqlDb = db;
	clearC (&m_mysqlDbC);
}

void
Arm4dbConfig::setMysqlSocket (const std::string &socket)
{
	m_mysqlSocket = socket;
	clearC (&m_mysqlSocketC);
}

