/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <string.h>
#include <grp.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>

#include "arm4.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Arm4dbConfig.h"
#include "Arm4dbCommon.h"
#include "Arm4dbExceptions.h"
#include "Arm4dbSharedMemory.h"

#define SHM_MODE	0666	/* Shared memory file permissions */

arm4_shm_header_t *Arm4dbSharedMemory::g_shared_memory_ptr = NULL;
int Arm4dbSharedMemory::g_message_queue = (-1);
int Arm4dbSharedMemory::g_umask;

pthread_once_t Arm4dbSharedMemory::g_attached = PTHREAD_ONCE_INIT;

/* This is a library local (in process) lock */
pthread_rwlock_t Arm4dbSharedMemory::g_db_rw_lock;

#ifndef USE_POSIX_PROCESS_SEM
sem_t *Arm4dbSharedMemory::g_sequence_sem_ptr = NULL;
sem_t *Arm4dbSharedMemory::g_application_sem_ptr = NULL;
sem_t *Arm4dbSharedMemory::g_transaction_sem_ptr = NULL;
sem_t *Arm4dbSharedMemory::g_metric_sem_ptr = NULL;
sem_t *Arm4dbSharedMemory::g_utility_sem_ptr = NULL;
#endif /* USE_POSIX_PROCESS_SEM */


Arm4dbSharedMemory::Arm4dbSharedMemory ()
{
	// Ensure we're attached
	//doAttach (); -- this is already done in the individual routines
}

Arm4dbSharedMemory::~Arm4dbSharedMemory ()
{
	// Detaching is one explicitly elsewhere
}

#ifndef USE_POSIX_PROCESS_SEM

const char *
Arm4dbSharedMemory::instanceName (const char *template_ptr)
{
    static char name [256];

    /* Create a name from the template that includes the instance */
	int instance = Arm4dbConfig::getConfig ().getInstance ();
    sprintf (name, "%s%d", template_ptr, instance);

    return name;
}

#endif /* USE_POSIX_PROCESS_SEM */

void
Arm4dbSharedMemory::attach_semaphores (void)
{
#ifndef USE_POSIX_PROCESS_SEM
	if (g_sequence_sem_ptr == NULL)
		g_sequence_sem_ptr = sem_open (instanceName (SEM_SEQUENCE_FILE), 0);
	if (g_sequence_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_seqence_sem_ptr sem_open");
		g_sequence_sem_ptr = NULL;
	}

	if (g_application_sem_ptr == NULL)
		g_application_sem_ptr = sem_open (instanceName (SEM_APPLICATION_FILE), 0);
	if (g_application_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_application_sem_ptr sem_open");
		g_application_sem_ptr = NULL;
	}

	if (g_transaction_sem_ptr == NULL)
		g_transaction_sem_ptr = sem_open (instanceName (SEM_TRANSACTION_FILE), 0);
	if (g_transaction_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_transaction_sem_ptr sem_open");
		g_transaction_sem_ptr = NULL;
	}

	if (g_metric_sem_ptr == NULL)
		g_metric_sem_ptr = sem_open (instanceName (SEM_METRIC_FILE), 0);
	if (g_metric_sem_ptr == (sem_t *) SEM_FAILED)
	{
		Logger::getLogger().error ("g_metric_sem_ptr sem_open");
		g_metric_sem_ptr = NULL;
	}

    if (g_utility_sem_ptr == NULL)
        g_utility_sem_ptr = sem_open (instanceName (SEM_UTILITY_FILE), 0);
    if (g_utility_sem_ptr == (sem_t *) SEM_FAILED)
    {
        Logger::getLogger().error ("g_utility_sem_ptr sem_open");
        g_utility_sem_ptr = NULL;

		throw new Arm4dbSharedMemoryException ("Unable to attach semaphores");
    }

#endif /* USE_POSIX_PROCESS_SEM */
}

void
Arm4dbSharedMemory::doDetach (void)
{
	if (g_shared_memory_ptr)
	{
		if (shmdt ((void *) g_shared_memory_ptr) < 0)
		{
			Logger::getLogger().error ("shmdt");
		}
		g_shared_memory_ptr = NULL;
	}
}

void
Arm4dbSharedMemory::doAttach (void)
{
	int shmid;
	key_t token;
	void *ptr;

	g_umask = Arm4dbConfig::getConfig ().getUmask ();
	int instance = Arm4dbConfig::getConfig ().getInstance ();
	token = ftok (Arm4dbCommon::tempPath(instance, MQUEUE_FILE), 1);
	g_message_queue = msgget (token, SHM_MODE & (~g_umask));
	if (g_message_queue == -1)
	{
		// This should be an error when we create, but not when we attach
		//Logger::getLogger().error ("Unable to attach to message queue");
		return;
	}

	if (g_shared_memory_ptr == NULL)
	{
		token = ftok (Arm4dbCommon::tempPath(instance, SHARED_MEMORY_FILE), 1);
		shmid = shmget (token, sizeof (arm4_shm_header_t), SHM_MODE & (~g_umask));
		if (shmid < 0)
		{
			Logger::getLogger().error ("shmget 1");
			throw new Arm4dbSharedMemoryException ("Unable to get shared memory id");
		}
		ptr = shmat (shmid, NULL, 0);
		if (ptr == NULL)
		{
			Logger::getLogger().error ("shmat");
			throw new Arm4dbSharedMemoryException ("Unable to attach to shared memory");
		}
		g_shared_memory_ptr = reinterpret_cast<arm4_shm_header_t *> (ptr);

		/* Check that we've been initialized */
		if (g_shared_memory_ptr->magic != ARM4_SHM_MAGIC)
		{
			doDetach ();
			//exit (1);
            return;
		}
	}

	atexit (doDetach);
	
	/* Initialize the semaphores */
	attach_semaphores ();
		
	return;
}

void
Arm4dbSharedMemory::lock_transaction (void)
{
#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_lock (&g_shared_memory_ptr->transaction_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;
	
	status = sem_wait (g_transaction_sem_ptr);
	if (status != 0)
	{
		Logger::getLogger().error ("lock_transaction %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

void
Arm4dbSharedMemory::unlock_transaction (void)
{
#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_unlock (&g_shared_memory_ptr->transaction_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;
	
	status = sem_post (g_transaction_sem_ptr);
	if (status != 0)
	{
		Logger::getLogger().error ("unlock_transaction %s", strerror (errno));
	}
#endif /* USE_POSIX_PROCESS_SEM */
}

arm4_shm_header_t *
Arm4dbSharedMemory::shmAttach (void)
{
	pthread_once (&g_attached, doAttach);
		
	return g_shared_memory_ptr;
}

arm_boolean_t
Arm4dbSharedMemory::setNullCollector (arm_boolean_t set_as_null)
{
    arm_boolean_t current_state;

    /* Ensure we're attached */
    if (shmAttach () == NULL)
        return ARM_FALSE;

    /* Indicate wether we're acting as a null library */
    current_state = g_shared_memory_ptr->null_arm;
    if (set_as_null)
        g_shared_memory_ptr->null_arm = ARM_TRUE;
    else
        g_shared_memory_ptr->null_arm = ARM_FALSE;

    return current_state;
}

arm_boolean_t
Arm4dbSharedMemory::getNullCollector (void)
{
    arm_boolean_t current_state;

    /* Ensure we're attached */
    if (shmAttach () == NULL)
        return ARM_TRUE; // If we can't attach then we're a null collector!

    /* Determine wether we're acting as a null library */
    current_state = g_shared_memory_ptr->null_arm;

    return current_state;
}

void
Arm4dbSharedMemory::setDebugLevel (int level)
{
    /* Ensure we're attached */
    if (shmAttach () == NULL)
        return;

	g_shared_memory_ptr->trace_level = (arm4_shm_tracing_t) level;
}

void
Arm4dbSharedMemory::setTransactionTraceInformation (const arm_id_t &tran_id, const arm4db_transaction_summary_t &summary)
{
	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	lock_transaction ();

	for (int i = 0; i < g_shared_memory_ptr->transaction_count; i++)
	{
		if (memcmp (&(g_shared_memory_ptr->transaction_table [i].tran_id), &tran_id, sizeof (tran_id)) == 0)
		{
			// Found the transaction entry
			g_shared_memory_ptr->transaction_table [i].trace_flags = summary.trace_flags;
			g_shared_memory_ptr->transaction_table [i].trace_parameter = summary.trace_parameter;
			g_shared_memory_ptr->transaction_table [i].trace_counter = summary.trace_parameter;

			unlock_transaction ();
			return;
		}
	}

	// Not found
	unlock_transaction ();
}


void
Arm4dbSharedMemory::setNewTransactionTraceInformation (int index)
{
	/* Ensure we're attached */
	if (shmAttach () == NULL)
		return;

	arm4_shm_transaction_t *tran_ptr = &(g_shared_memory_ptr->transaction_table [index]);

	// Set our default values
	tran_ptr->trace_flags = arm4_shm_transaction_trace_all;
	tran_ptr->appdex_flags = ARM4_APPDEX_DEFAULT_FLAGS;
	tran_ptr->trace_parameter = 0; // should really look this up if default
	tran_ptr->appdex_threshold = ARM4_APPDEX_DEFAULT_RESPONSE_TIME; // should really look this up if default
	
	// Initialize any counters
	tran_ptr->trace_counter = tran_ptr->trace_parameter;
	tran_ptr->transaction_counter = 0;
	tran_ptr->appdex_satisfactory = 0;
	tran_ptr->appdex_tolerable = 0;
	tran_ptr->appdex_frustrating = 0;
}

/*
 * Class Arm4dbLock
 */

Arm4dbLock::Arm4dbLock () :
	m_locked (false)
{
	lock ();
}

Arm4dbLock::~Arm4dbLock ()
{
	// Ensure our program has unlocked this
	DB_ASSERT (m_locked == false);
}


void
Arm4dbLock::lock (void)
{
	DB_ASSERT (m_locked == false);

    int status = pthread_rwlock_rdlock (&g_db_rw_lock);
    if (status != 0 && status != EINVAL)
    {
		DB_ASSERT (status != EDEADLK); // Prevent nested locking

        Logger::getLogger().error ("Arm4dbLock::lock");
		throw new Arm4dbLockException ("Unable to lock");
    }
	else
	{
		// Ignore an EINVAL
		m_locked = true;
	}
}

void
Arm4dbLock::unlock (void)
{
	DB_ASSERT (m_locked == true);

    int status = pthread_rwlock_unlock (&g_db_rw_lock);
    if (status != 0 && status != EINVAL)
    {
		DB_ASSERT (status != EDEADLK); // Prevent nested locking

		Logger::getLogger().error ("Arm4dbLock::unlock");
		throw new Arm4dbLockException ("Unable to unlock");
    }
	else
	{
		// Ignore an EINVAL
		m_locked = false;
	}
}

/*
 * Class Arm4dbWriteLock
 */

Arm4dbWriteLock::Arm4dbWriteLock () :
	m_locked (false)
{
	lock ();
}

Arm4dbWriteLock::~Arm4dbWriteLock ()
{
	// Ensure our program has unlocked this
	DB_ASSERT (m_locked == false);
}

void
Arm4dbWriteLock::lock (void)
{
	DB_ASSERT (m_locked == false);

    int status = pthread_rwlock_wrlock (&g_db_rw_lock);
    if (status != 0 && status != EINVAL)
    {
		DB_ASSERT (status != EDEADLK); // Prevent nested locking

		Logger::getLogger().error ("Arm4dbWriteLock::lock");
		throw new Arm4dbLockException ("Unable to lock");
    }
	else
	{
		// Ignore an EINVAL
		m_locked = true;
	}
}

void
Arm4dbWriteLock::unlock (void)
{
	DB_ASSERT (m_locked == true);

	int status = pthread_rwlock_unlock (&g_db_rw_lock);
    if (status != 0 && status != EINVAL)
    {
		DB_ASSERT (status != EDEADLK); // Prevent nested locking

		Logger::getLogger().error ("Arm4dbWriteLock::unlock");
		throw new Arm4dbLockException ("Unable to unlock");
    }
	else
	{
		// Ignore an EINVAL
		m_locked = false;
	}
}

/*
 * Class Arm4dbUtilityLock
 */

Arm4dbUtilityLock::Arm4dbUtilityLock () :
	m_locked (false)
{
	lock ();
}

Arm4dbUtilityLock::~Arm4dbUtilityLock ()
{
	// Ensure our program has unlocked this
	DB_ASSERT (m_locked == false);
}

void
Arm4dbUtilityLock::lock (void)
{
	DB_ASSERT (m_locked == false);

	/* Ensure we're attached */
	if (shmAttach () == NULL)
	{
		// Act like everything is normal, as this may be an expected condition
		m_locked = true;
		return;
	}

#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_lock (&g_shared_memory_ptr->utility_read_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;
	
	status = sem_wait (g_utility_sem_ptr);
	if (status != 0)
	{
        Logger::getLogger().error ("Arm4dbUtilityLock::lock %s", strerror (errno));
		throw new Arm4dbLockException ("Unable to lock");
	}
#endif /* USE_POSIX_PROCESS_SEM */

	m_locked = true;
}

void
Arm4dbUtilityLock::unlock (void)
{
	DB_ASSERT (m_locked == true);

	/* Ensure we're attached */
	if (shmAttach () == NULL)
	{
		// Act like everything is normal, as this may be an expected condition
		m_locked = false;
		return;
	}

#ifdef USE_POSIX_PROCESS_SEM
	pthread_mutex_unlock (&g_shared_memory_ptr->utility_read_mutex);
#else /* USE_POSIX_PROCESS_SEM */
	int status;
	
	status = sem_post (g_utility_sem_ptr);
	if (status != 0)
	{
        Logger::getLogger().error ("Arm4dbUtilityLock::unlock %s", strerror (errno));
		throw new Arm4dbLockException ("Unable to unlock");
	}
#endif /* USE_POSIX_PROCESS_SEM */

	m_locked = false;
}

