/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Sqlite.h"
#include "Arm4dbCommon.h"
#include "Arm4dbSharedMemory.h"

Sqlite::Sqlite () :
	m_db (NULL)
{
}

Sqlite::~Sqlite ()
{
	if (m_db)
		closeDatabasesNoLock ();
}

std::string
Sqlite::fullDbName (void)
{
	std::string filename = getConfig ().getDbHome () + "/" + SQLITE_DB_NAME;

	return filename;
}

bool
Sqlite::dbExists (void)
{
	std::string filename = fullDbName ();
	if (access (filename.c_str(), F_OK) == 0)
		return true;
	
	return false;
}

void
Sqlite::logError (void) const
{
	Logger::getLogger().log ("Sqlite %s\n", sqlite3_errmsg(m_db));
	Logger::getLogger().log ("Sqlite errno = %d\n", sqlite3_errcode(m_db));
	DB_ASSERT (false);
}

void
Sqlite::logError (const char *error_point) const
{
	Logger::getLogger().log ("Sqlite (%s) %s\n", error_point, sqlite3_errmsg(m_db));
	Logger::getLogger().log ("Sqlite errno = %d\n", sqlite3_errcode(m_db));
	DB_ASSERT (false);
}

arm_error_t
Sqlite::openDatabasesReadOnlyNoLock (void)
{
	int	status;

	// Check if we're already open
	if (m_db)
		return ARM_SUCCESS;
	
	std::string filename = fullDbName ();
#if defined(HAVE_SQLITE_OPEN_V2)
	status = sqlite3_open_v2 (filename.c_str (), &m_db, SQLITE_OPEN_READONLY, NULL);
#else
	// Our library version doesn't support read only opens
	status = sqlite3_open(filename.c_str (), &m_db);
#endif
	if (status != SQLITE_OK)
	{
		logError (filename);
		return ARM_FAILURE;
	}

	return ARM_SUCCESS;	
}

arm_error_t
Sqlite::openDatabasesNoLock (void)
{
	int	status;

	// Check if we're already open
	if (m_db)
		return ARM_SUCCESS;
	
	std::string filename = fullDbName ();
	status = sqlite3_open(filename.c_str (), &m_db);
	if (status != SQLITE_OK)
	{
		logError (filename);
		return ARM_FAILURE;
	}

	return ARM_SUCCESS;	
}

arm_error_t
Sqlite::openOrCreateDatabasesNoLock (void)
{
	int	status;
	char *error_ptr;
	
	/* Open the primary database */
    bool exists = dbExists ();
	status = openDatabasesNoLock ();
	
	/* Create the tables */
	if (!exists)
	{
		/*
		 * Application table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application ("
				"  appl_id BLOB PRIMARY KEY UNIQUE NOT NULL,"
				"  application_name TEXT NOT NULL,"
				"  md5_digest BLOB UNIQUE NOT NULL,"
				"  charset INT NOT NULL,"
				"  default_trace_flags INT DEFAULT 1," // arm4_shm_transaction_trace_all
				"  default_trace_parameter INT DEFAULT 0,"
				"  default_appdex_threshold INT DEFAULT 4)", // ARM4_APPDEX_DEFAULT_RESPONSE_TIME
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS application_application_name_index"
				"  ON application (application_name)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS application_md5_digest_index"
				"  ON application (md5_digest)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application identity table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_identity ("
				"  appl_id BLOB PRIMARY KEY NOT NULL,"
				"  name TEXT NOT NULL,"
				"  value TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application context table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_context ("
				"  appl_id BLOB NOT NULL," // Duplicates prevent this from being a primary key
				"  slot INT NOT NULL,"
				"  name TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS application_context_appl_id_index"
				"  ON application_context (appl_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS trans ("
				"  tran_id BLOB PRIMARY KEY UNIQUE NOT NULL,"
				"  transaction_name TEXT NOT NULL,"
				"  md5_digest BLOB UNIQUE NOT NULL,"
				"  appl_id BLOB NOT NULL,"
				"  trace_flags INT DEFAULT 1," // arm4_shm_transaction_trace_all
				"  appdex_flags INT DEFAULT 2," // arm4_appdex_use_default
				"  trace_parameter INT DEFAULT 0,"
				"  appdex_threshold INT DEFAULT 4)", // ARM4_APPDEX_DEFAULT_RESPONSE_TIME
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_transaction_name_index"
				"  ON trans (transaction_name)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_md5_digest_index"
				"  ON trans (md5_digest)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_appl_id_index"
				"  ON trans (appl_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction identity table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_identity ("
				"  tran_id BLOB PRIMARY KEY NOT NULL,"
				"  name TEXT NOT NULL,"
				"  value TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction context table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_context ("
				"  tran_id BLOB NOT NULL," // Duplicates prevent this from being a primary key
				"  slot INT NOT NULL,"
				"  name TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_context_tran_id_index"
				"  ON transaction_context (tran_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction uri table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_uri ("
				"  tran_id BLOB PRIMARY KEY NOT NULL,"
				"  uri TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction identity table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_metric_binding ("
				"  tran_id BLOB NOT NULL," // Duplicates prevent this from being a primary key
				"  slot INT NOT NULL,"
				"  id BLOB NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_metric_binding_tran_id_index"
				"  ON transaction_metric_binding (tran_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Metric table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS metric ("
				"  metric_id BLOB PRIMARY KEY UNIQUE NOT NULL,"
				"  metric_name TEXT NOT NULL,"
				"  metric_units TEXT,"
				"  md5_digest BLOB UNIQUE NOT NULL,"
				"  appl_id BLOB NOT NULL,"
				"  metric_format INT NOT NULL,"
				"  metric_usage INT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS metric_metric_name_index"
				"  ON metric (metric_name)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS metric_md5_digest_index"
				"  ON metric (md5_digest)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS metric_appl_id_index"
				"  ON metric (appl_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application instance table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_instance ("
				"  app_handle INT PRIMARY KEY NOT NULL,"
				"  appl_id BLOB NOT NULL,"
				"  active INT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS application_instance_appl_id_index"
				"  ON application_instance (appl_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application group table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_group ("
				"  app_handle INT PRIMARY KEY NOT NULL,"
				"  app_group TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application instance table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_instance_id ("
				"  app_handle INT PRIMARY KEY NOT NULL,"
				"  instance_name TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application context table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_context_values ("
				"  app_handle INT NOT NULL," // Duplicates prevent this from being a primary key
				"  slot INT NOT NULL,"
				"  property_name TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS application_context_values_app_handle_index"
				"  ON application_context_values (app_handle)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Application address table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS application_address ("
				"  app_handle INT NOT NULL," // Duplicates prevent this from being a primary key
				"  address_format INT NOT NULL,"
				"  address_length INT NOT NULL,"
				"  id_valid INT NOT NULL,"
				"  id BLOB NOT NULL,"
				"  address BLOB NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
			return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS application_address_app_handle_index"
				"  ON application_address (app_handle)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
			return ARM_FAILURE;
		}

		/*
		 * Transaction instance table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_instance ("
				"  tran_handle INT PRIMARY KEY NOT NULL,"
				"  appl_id BLOB NOT NULL,"
				"  tran_id BLOB NOT NULL,"
				"  app_handle INT NOT NULL,"
				"  arrival_time INT NOT NULL,"
				"  arrival_time_usec INT NOT NULL,"
				"  stop_time INT NOT NULL,"
				"  stop_time_usec INT NOT NULL,"
				"  response_time INT NOT NULL,"
				"  response_time_usec INT NOT NULL,"
				"  blocked_time INT NOT NULL,"
				"  blocked_time_usec INT NOT NULL,"
				"  tran_status INT NOT NULL,"
				"  active INT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction context table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_context_values ("
				"  tran_handle INT NOT NULL," // Duplicates prevent this from being a primary key
				"  slot INT NOT NULL,"
				"  value TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_context_values_tran_handle_index"
				"  ON transaction_context_values (tran_handle)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction uri table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_context_uri ("
				"  tran_handle INT PRIMARY KEY NOT NULL,"
				"  uri TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction metric value table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_metric_value ("
				"  tran_handle INT NOT NULL," // Duplicates prevent this from being a primary key
				"  slot INT NOT NULL,"
				"  format INT NOT NULL,"
				"  usage INT NOT NULL,"
				"  valid INT NOT NULL,"
				"  metric_value BLOB NOT NULL,"
				"  string_value TEXT)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS transaction_metric_value_tran_handle_index"
				"  ON transaction_metric_value (tran_handle)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction user table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_user ("
				"  tran_handle INT PRIMARY KEY NOT NULL,"
				"  name TEXT NOT NULL,"
				"  valid INT NOT NULL,"
				"  id BLOB NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Transaction diagnostic table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS transaction_diag ("
				"  tran_handle INT PRIMARY KEY NOT NULL,"
				"  detail TEXT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Correlator tables
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS current_correlator ("
				"  tran_handle INT PRIMARY KEY NOT NULL,"
				"  current_correlator BLOB NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS parent_correlator ("
				"  tran_handle INT PRIMARY KEY NOT NULL,"
				"  parent_correlator BLOB NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Sequence table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS sequence ("
				"  sequence_name TEXT PRIMARY KEY NOT NULL,"
				"  sequence_value INT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}

		/*
		 * Trace stats table
		 */
		sqlite3_exec (m_db,
				"CREATE TABLE IF NOT EXISTS trace_stats ("
				"  time_seconds INT NOT NULL," // Making this the primary key would disallow duplicates
				"  tran_id BLOB NOT NULL,"
				"  counter INT NOT NULL)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS trace_stats_time_seconds_index"
				"  ON trace_stats (time_seconds)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
		sqlite3_exec (m_db,
				"CREATE INDEX IF NOT EXISTS trace_stats_tran_id_index"
				"  ON trace_stats (tran_id)",
				NULL,
				NULL,
				&error_ptr);
		if (error_ptr != NULL)
		{
			logError ();
		    return ARM_FAILURE;
		}
	}

	return status;	
}

void
Sqlite::initializeLibrary (void)
{
	
}

void
Sqlite::initializeThread (void)
{
	
}

arm_error_t
Sqlite::openDatabasesReadOnly (void)
{
	int	status;

	/* Ensure we have database access */
	Arm4dbLock lock;

	/* Open the primary database */
	status = openDatabasesReadOnlyNoLock ();

	lock.unlock ();

	return status;	
}

arm_error_t
Sqlite::openDatabases (void)
{
	int	status;
	
    /* Ensure we have database access */
    Arm4dbLock lock;

	/* Open the primary database */
	status = openOrCreateDatabasesNoLock ();
	
    lock.unlock ();

	return status;	
}

arm_error_t
Sqlite::closeDatabasesNoLock (void)
{
	sqlite3_close (m_db);
	m_db = NULL;
	
    return ARM_SUCCESS;	
}

/* Transactions that extract from the queues into the instance databases */
/*
 *
 *	External interfaces
 *
 */

arm_error_t
Sqlite::initializeSequence (int index, arm4_shm_header_t *header_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	arm_int64_t *sequence_ptr = &(header_ptr->sequence_table [index]);

	// Find the sequence value
	std::string sql = "SELECT sequence_value FROM sequence"
		" WHERE sequence_name = '" + ((std::string) Arm4dbCommon::getSequenceName (index)) + "'";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		*sequence_ptr =  sqlite3_column_int64(statement_ptr, 0);
		/*
		 * Increment the sequence number in case we crashed since the last increment
		 */

		*sequence_ptr += ARM4_SHM_SEQUENCE_STEP;

		// Check for a wrap - don't allow the 0 value
		if (*sequence_ptr < 1)
			*sequence_ptr = 1;
	}
	else
	{
		/* The initial value is 1 */
		*sequence_ptr = 1;
	}
	sqlite3_finalize(statement_ptr);

	// Wrap the sequence if needed
	if (*sequence_ptr <= 0)
		*sequence_ptr = 1;
	
	// Save the sequence back to the database
	sql = "INSERT OR REPLACE INTO sequence"
		" (sequence_name, sequence_value)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return ARM_FAILURE;
	}
	const char *sequence_name_ptr = Arm4dbCommon::getSequenceName (index);
	status = sqlite3_bind_text(statement_ptr, 1, sequence_name_ptr, strlen (sequence_name_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 2, *sequence_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	return ARM_SUCCESS;
}

arm_error_t
Sqlite::initializeSequences (arm4_shm_header_t *header_ptr)
{
	int			status;
	int			i;

    Arm4dbWriteLock lock;

	/* Open the primary database */
	status = openOrCreateDatabasesNoLock ();
	if (status != ARM_SUCCESS)
	{
        lock.unlock ();
		return status;
	}

	for (i = 0; i < arm4_shm_sequence_max; i++)
	{
		status = initializeSequence (i, header_ptr);
		if (status != ARM_SUCCESS)
        {
            lock.unlock ();
			return status;
        }
	}
    lock.unlock ();

	return ARM_SUCCESS;
}

arm_error_t
Sqlite::processApplicationInstances (const arm4_shm_application_instance_t *app_instance_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processApplicationInstances\n");
	
    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT OR REPLACE INTO application_instance"
		" (app_handle, appl_id, active)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, app_instance_ptr->message.app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, &(app_instance_ptr->message.appl_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 3, app_instance_ptr->message.active);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processApplicationGroup (const arm4_shm_property_t *group_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processApplicationGroup\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_group"
		" (app_handle, app_group)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, group_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, group_ptr->message.property_name, strlen (group_ptr->message.property_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processApplicationInstanceId (const arm4_shm_property_t *property_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processApplicationInstanceId\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_instance_id"
		" (app_handle, instance_name)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, property_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, property_ptr->message.property_name, strlen (property_ptr->message.property_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processApplicationContext (const arm4_shm_context_property_t *property_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processApplicationContext\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_context_values"
		" (app_handle, slot, property_name)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, property_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, property_ptr->message.property_index);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property_ptr->message.property_name, strlen (property_ptr->message.property_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processApplicationAddress (const arm4_shm_address_t *address_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_address"
		" (app_handle, address_format, address_length, id_valid, id, address)"
		" VALUES (?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, address_ptr->message.app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, address_ptr->message.address_format);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 3, address_ptr->message.address_length);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 4, address_ptr->message.id_valid);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 5, &(address_ptr->message.id), sizeof (address_ptr->message.id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 6, address_ptr->message.address, address_ptr->message.address_length, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTransactionInstances (const arm4_shm_transaction_instance_t *tran_instance_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionInstances\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT OR REPLACE INTO transaction_instance"
		" (tran_handle, appl_id, tran_id, app_handle,"
		"  arrival_time, arrival_time_usec, stop_time, stop_time_usec, response_time, response_time_usec, blocked_time, blocked_time_usec,"
		"  tran_status, active)"
		" VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_instance_ptr->message.tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, &(tran_instance_ptr->message.appl_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, &(tran_instance_ptr->message.tran_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 4, tran_instance_ptr->message.app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 5, tran_instance_ptr->message.arrival_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 6, tran_instance_ptr->message.arrival_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 7, tran_instance_ptr->message.stop_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 8, tran_instance_ptr->message.stop_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 9, tran_instance_ptr->message.response_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 10, tran_instance_ptr->message.response_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 11, tran_instance_ptr->message.blocked_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 12, tran_instance_ptr->message.blocked_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 13, tran_instance_ptr->message.tran_status);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 14, tran_instance_ptr->message.active);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}

	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTransactionContext (const arm4_shm_context_property_t *property_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionContext\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_context_values"
		" (tran_handle, slot, value)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, property_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, property_ptr->message.property_index);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property_ptr->message.property_name, strlen (property_ptr->message.property_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTransactionUri (const arm4_shm_uri_t *uri_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionUri\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_context_uri"
		" (tran_handle, uri)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, uri_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, uri_ptr->message.property_name, strlen (uri_ptr->message.property_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTransactionMetricValue (const arm4_shm_metric_value_t *metric_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionMetricValue\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_metric_value"
		" (tran_handle, slot, format, usage, valid, metric_value, string_value)"
		" VALUES (?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, metric_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, metric_ptr->message.value.slot);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 3, metric_ptr->message.value.format);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 4, metric_ptr->message.value.usage);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 5, metric_ptr->message.value.valid);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 6, &(metric_ptr->message.value.metric_u), sizeof (metric_ptr->message.value.metric_u), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 7, metric_ptr->message.string32, strlen (metric_ptr->message.string32), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTransactionUser (const arm4_shm_user_t *user_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionUser\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_user"
		" (tran_handle, name, valid, id)"
		" VALUES (?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, user_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, user_ptr->message.name, strlen (user_ptr->message.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 3, user_ptr->message.id_valid);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 4, &(user_ptr->message.id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTransactionDiag (const arm4_shm_diag_t *diag_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionDiag\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_diag"
		" (tran_handle, detail)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, diag_ptr->message.handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, diag_ptr->message.detail, strlen (diag_ptr->message.detail), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processCorrelators (const arm4_shm_correlator_t *correlator_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processCorrelators\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create new entries
	int length = Arm4dbCommon::correlatorSize (correlator_ptr->message.current_correlator);
	if (length > 0)
	{
		std::string sql = "INSERT INTO current_correlator"
			" (tran_handle, current_correlator)"
			" VALUES (?,?)";
		status = prepare(
				m_db,
				sql.c_str (), sql.length (),
				&statement_ptr,
				&unused_ptr);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_bind_int64(statement_ptr, 1, correlator_ptr->message.tran_handle);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_bind_blob(statement_ptr, 2, &(correlator_ptr->message.current_correlator),
				length, SQLITE_STATIC);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_step(statement_ptr);
		if (status != SQLITE_DONE)
		{
			logError ();
			sqlite3_finalize(statement_ptr);
			lock.unlock ();
		
			return ARM_FAILURE;
		}
		sqlite3_finalize(statement_ptr);
	}

	// Now process the parent
	length = Arm4dbCommon::correlatorSize (correlator_ptr->message.parent_correlator);
	if (length > 0)
	{
		std::string sql = "INSERT INTO parent_correlator"
			" (tran_handle, parent_correlator)"
			" VALUES (?,?)";
		status = prepare(
				m_db,
				sql.c_str (), sql.length (),
				&statement_ptr,
				&unused_ptr);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_bind_int64(statement_ptr, 1, correlator_ptr->message.tran_handle);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_bind_blob(statement_ptr, 2, &(correlator_ptr->message.parent_correlator),
				length, SQLITE_STATIC);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_step(statement_ptr);
		if (status != SQLITE_DONE)
		{
			logError ();
			sqlite3_finalize(statement_ptr);
			lock.unlock ();

			return ARM_FAILURE;
		}
		sqlite3_finalize(statement_ptr);
	}

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processSequence (const arm4_shm_db_sequence_t *sequence_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processSequence\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

    const char *sequence_name_ptr = Arm4dbCommon::getSequenceName (sequence_ptr->message.index);

	// Create a new entry
	std::string sql = "INSERT OR REPLACE INTO sequence"
		" (sequence_name, sequence_value)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 1, sequence_name_ptr, strlen (sequence_name_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 2, sequence_ptr->message.sequence);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterApplication (const arm4_shm_register_application_t *application_ptr,
	arm_id_t *appid_ptr,
	arm_boolean_t *is_new_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;
	arm_id_t appl_id;

	//printf ("processRegisterApplication\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

    /* Initialize an invalid handle */
	memset (appid_ptr, 0, sizeof (arm_id_t));
	memset (&appl_id, 0, sizeof (arm_id_t));
	*is_new_ptr = ARM_FALSE;

	/*
	 * See if an entry already exists by iterating through all applications with the same name
	 */
	std::string sql = "SELECT appl_id FROM application WHERE application.md5_digest = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, application_ptr->message.digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		memcpy (appid_ptr, sqlite3_column_blob(statement_ptr, 0), sizeof (arm_id_t));
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_SUCCESS;
	}
	sqlite3_finalize(statement_ptr);

	// Create a new entry

	// Get the new sequence number (application id)
	Arm4dbCommon::getUuid (&appl_id);

	sql = "INSERT INTO application"
		" (appl_id, application_name, md5_digest, charset, default_trace_flags, default_trace_parameter, default_appdex_threshold)"
		" VALUES (?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &appl_id, sizeof (appl_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, application_ptr->message.app_name, strlen (application_ptr->message.app_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, application_ptr->message.digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 4, application_ptr->message.charset);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 5, arm4_shm_transaction_trace_all);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 6, 0);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 7, ARM4_APPDEX_DEFAULT_RESPONSE_TIME);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	/* Return the application id to the user */
	*appid_ptr = appl_id;
	*is_new_ptr = ARM_TRUE;
	
	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterApplicationIdentity (const arm4_shm_register_application_identity_t *identity_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processRegisterApplicationIdentity\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_identity"
		" (appl_id, name, value)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &(identity_ptr->message.appid), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, identity_ptr->message.name, strlen (identity_ptr->message.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, identity_ptr->message.value, strlen (identity_ptr->message.value), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterApplicationContext (const arm4_shm_register_application_context_t *context_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processRegisterApplicationContext\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_context"
		" (appl_id, slot, name)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &(context_ptr->message.appid), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, context_ptr->message.index);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, context_ptr->message.name, strlen (context_ptr->message.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterTransaction (const arm4_shm_register_transaction_t *transaction_ptr,
	arm_id_t *tranid_ptr,
	arm_boolean_t *is_new_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;
	arm_id_t tran_id;

	//printf ("processRegisterTransaction\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

    /* Initialize an invalid handle */
	memset (tranid_ptr, 0, sizeof (arm_id_t));
	memset (&tran_id, 0, sizeof (arm_id_t));
	*is_new_ptr = ARM_FALSE;

	/*
	 * See if an entry already exists by iterating through all applications with the same name
	 */
	std::string sql = "SELECT tran_id FROM trans WHERE trans.md5_digest = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, transaction_ptr->message.digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		memcpy (tranid_ptr, sqlite3_column_blob(statement_ptr, 0), sizeof (arm_id_t));
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_SUCCESS;
	}
	sqlite3_finalize(statement_ptr);

	// Create a new entry

	// Get the new sequence number (transaction id)
	Arm4dbCommon::getUuid (&tran_id);

	sql = "INSERT INTO trans"
		" (tran_id, transaction_name, md5_digest, appl_id, trace_flags, appdex_flags, trace_parameter, appdex_threshold)"
		" VALUES (?,?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, transaction_ptr->message.tran_name, strlen (transaction_ptr->message.tran_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, transaction_ptr->message.digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 4, &(transaction_ptr->message.app_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 5, arm4_shm_transaction_trace_all);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 6, ARM4_APPDEX_DEFAULT_FLAGS);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 7, 0);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int64(statement_ptr, 8, ARM4_APPDEX_DEFAULT_RESPONSE_TIME);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);
	
	// Initialize the in-memory values
	Arm4dbSharedMemory::setNewTransactionTraceInformation (transaction_ptr->message.transaction_index);
	
	// Return the application id to the user
	*tranid_ptr = tran_id;
	*is_new_ptr = ARM_TRUE;
	
	lock.unlock ();

    return ARM_SUCCESS;
}


arm_error_t
Sqlite::processRegisterTransactionIdentity (const arm4_shm_register_application_identity_t *identity_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processRegisterTransactionIdentity\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_identity"
		" (tran_id, name, value)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &(identity_ptr->message.appid), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, identity_ptr->message.name, strlen (identity_ptr->message.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, identity_ptr->message.value, strlen (identity_ptr->message.value), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterTransactionContext (const arm4_shm_register_application_context_t *context_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processRegisterTransactionContext\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_context"
		" (tran_id, slot, name)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &(context_ptr->message.appid), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, context_ptr->message.index);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, context_ptr->message.name, strlen (context_ptr->message.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterTransactionUri (const arm4_shm_register_transaction_uri_t *uri_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processRegisterTransactionUri\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_uri"
		" (tran_id, uri)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &(uri_ptr->message.tran_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, uri_ptr->message.property_name, strlen (uri_ptr->message.property_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterTransactionMetricBinding (const arm4_shm_register_transaction_metric_t *metric_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processRegisterTransactionMetricBinding\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_metric_binding"
		" (tran_id, slot, id)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &(metric_ptr->message.tran_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 2, metric_ptr->message.slot);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, &(metric_ptr->message.id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processRegisterMetric (arm4_shm_register_metric_t *metric_ptr,
	arm_id_t *metricid_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;
	arm_id_t metric_id;

	//printf ("processRegisterMetric\n");

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

    /* Initialize an invalid handle */
	memset (metricid_ptr, 0, sizeof (arm_id_t));
	memset (&metric_id, 0, sizeof (arm_id_t));

	/*
	 * See if an entry already exists by iterating through all applications with the same name
	 */
	std::string sql = "SELECT metric_id FROM metric WHERE metric.md5_digest = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, metric_ptr->message.digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		memcpy (metricid_ptr, sqlite3_column_blob(statement_ptr, 0), sizeof (arm_id_t));
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_SUCCESS;
	}
	sqlite3_finalize(statement_ptr);

	// Create a new entry

	// Get the new sequence number (transaction id)
	Arm4dbCommon::getUuid (&metric_id);

	sql = "INSERT INTO metric"
		" (metric_id, metric_name, metric_units, md5_digest, appl_id, metric_format, metric_usage)"
		" VALUES (?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &metric_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 2, metric_ptr->message.metric_name, strlen (metric_ptr->message.metric_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_text(statement_ptr, 3, metric_ptr->message.metric_unit, strlen (metric_ptr->message.metric_unit), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 4, metric_ptr->message.digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 5, &(metric_ptr->message.app_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 6, metric_ptr->message.metric_format);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_int(statement_ptr, 7, metric_ptr->message.metric_usage);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		lock.unlock ();

	    return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	/* Return the application id to the user */
	*metricid_ptr = metric_id;
	
	lock.unlock ();

    return ARM_SUCCESS;
}

arm_error_t
Sqlite::processTraceStats (arm_time_t sample_time, arm4db_transaction_trace_stats_t *stats_ptr, int transaction_count)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTraceStats (%d)\n", transaction_count);

    /* Ensure we have database access */
    Arm4dbLock lock;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO trace_stats"
		" (time_seconds, tran_id, counter)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	for (int i = 0; i < transaction_count; i++)
	{
		status = sqlite3_bind_int64(statement_ptr, 1, sample_time.tv_sec);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_bind_blob(statement_ptr, 2, &(stats_ptr[i].tran_id), sizeof (arm_id_t), SQLITE_STATIC);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_bind_int64(statement_ptr, 3, stats_ptr[i].transaction_counter);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
		status = sqlite3_step(statement_ptr);
		if (status != SQLITE_DONE)
		{
			logError ();
			sqlite3_finalize(statement_ptr);
			lock.unlock ();
	
		    return ARM_FAILURE;
		}

#if defined(HAVE_SQLITE_CLEAR_BINDINGS)
		status = sqlite3_clear_bindings(statement_ptr);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
#endif
		status = sqlite3_reset(statement_ptr);
		if (status != SQLITE_OK)
		{
			logError ();
			lock.unlock ();
			return ARM_FAILURE;
		}
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return ARM_SUCCESS;
}

std::list<arm_id_t> *
Sqlite::getApplicationList (void)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	// Create the list
	std::list<arm_id_t> *list_ptr = new std::list<arm_id_t> ();

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return list_ptr;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT appl_id FROM application";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return list_ptr;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			list_ptr->push_back (*((arm_id_t *) sqlite3_column_blob(statement_ptr, 0)));
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

	return list_ptr;
}

int
Sqlite::getApplicationCount (void)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return 0;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT count(*) FROM application";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return 0;
	}

	int count = 0;
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			count = sqlite3_column_int (statement_ptr, 0);
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

	return count;
}

std::list<arm_id_t> *
Sqlite::getTransactionList (void)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	// Create the list
	std::list<arm_id_t> *list_ptr = new std::list<arm_id_t> ();

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return list_ptr;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT tran_id FROM trans";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return list_ptr;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			list_ptr->push_back (*((arm_id_t *) sqlite3_column_blob(statement_ptr, 0)));
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return list_ptr;
}

int
Sqlite::getTransactionCount (void)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return 0;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT count(*) FROM trans";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return 0;
	}

	int count = 0;
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			count = sqlite3_column_int (statement_ptr, 0);
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return count;
}

std::list<arm_id_t> *
Sqlite::getTransactionList (const arm_id_t &app_id)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	// Create the list
	std::list<arm_id_t> *list_ptr = new std::list<arm_id_t> ();

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return list_ptr;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT tran_id FROM trans WHERE trans.appl_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return list_ptr;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &app_id, sizeof (app_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return list_ptr;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			list_ptr->push_back (*((arm_id_t *) sqlite3_column_blob(statement_ptr, 0)));
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return list_ptr;
}

std::list<arm_tran_start_handle_t> *
Sqlite::getTransactionInstanceList (const arm_id_t &tran_id)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	// Create the list
	std::list<arm_tran_start_handle_t> *list_ptr = new std::list<arm_tran_start_handle_t> ();

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return list_ptr;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT tran_handle FROM transaction_instance WHERE transaction_instance.tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return list_ptr;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (tran_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return list_ptr;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			list_ptr->push_back (*((arm_tran_start_handle_t *) sqlite3_column_int64(statement_ptr, 0)));
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return list_ptr;
}

int
Sqlite::getTransactionCount (const arm_id_t &app_id)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return 0;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT count(*) FROM trans WHERE trans.appl_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return 0;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &app_id, sizeof (app_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return 0;
	}

	int count = 0;
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			// We found an entry
			count = sqlite3_column_int (statement_ptr, 0);
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	lock.unlock ();

    return count;
}

arm_error_t
Sqlite::getApplicationInformation (const arm_id_t &app_id, arm4db_application_summary_t *summary_ptr)
{
	int	status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return ARM_FAILURE;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT appl_id,application_name,charset,default_trace_flags,default_trace_parameter,default_appdex_threshold"
		" FROM application WHERE appl_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &app_id, sizeof (app_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		summary_ptr->appl_id = *((arm_id_t *) sqlite3_column_blob(statement_ptr, 0));
		strncpy (summary_ptr->application_name, (const char *) sqlite3_column_text(statement_ptr, 1), ARM_NAME_MAX_LENGTH);
		summary_ptr->charset = sqlite3_column_int(statement_ptr, 2);
		summary_ptr->default_trace_flags = (arm4_shm_transaction_trace_t) sqlite3_column_int(statement_ptr, 3);
		summary_ptr->default_trace_parameter = sqlite3_column_int64(statement_ptr, 4);
		summary_ptr->default_appdex_threshold = sqlite3_column_int64(statement_ptr, 5);
	}
	else
	{
		logError ();
		status = ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	//	identity_properties
	summary_ptr->identity_properties.clear ();
	sql = "SELECT name,value"
		" FROM application_identity WHERE appl_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &app_id, sizeof (app_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}

	status = sqlite3_step(statement_ptr);
	while (status == SQLITE_ROW)
	{
		// We found an entry
		property_record_t *identity_ptr = new property_record_t;
		strncpy (identity_ptr->name, (const char *) sqlite3_column_text(statement_ptr, 0), ARM_NAME_MAX_LENGTH);
		strncpy (identity_ptr->value, (const char *) sqlite3_column_text(statement_ptr, 1), ARM_NAME_MAX_LENGTH);
		summary_ptr->identity_properties.push_back (identity_ptr);

		status = sqlite3_step(statement_ptr);
	}
	sqlite3_finalize(statement_ptr);

	//	context_names
	summary_ptr->context_names.clear ();
	sql = "SELECT slot,name"
		" FROM application_context WHERE appl_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &app_id, sizeof (app_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}

	status = sqlite3_step(statement_ptr);
	while (status == SQLITE_ROW)
	{
		// We found an entry
		indexed_property_name_record_t *identity_ptr = new indexed_property_name_record_t;
		identity_ptr->index = sqlite3_column_int(statement_ptr, 0);
		strncpy (identity_ptr->name, (const char *) sqlite3_column_text(statement_ptr, 1), ARM_NAME_MAX_LENGTH);
		summary_ptr->context_names.push_back (identity_ptr);

		status = sqlite3_step(statement_ptr);
	}
	sqlite3_finalize(statement_ptr);


    lock.unlock ();
    
    return ARM_SUCCESS;
}

arm_error_t
Sqlite::getTransactionInformation (const arm_id_t &tran_id, arm4db_transaction_summary_t *summary_ptr)
{
	int	status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return ARM_FAILURE;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT tran_id,appl_id,transaction_name,trace_flags,appdex_flags,trace_parameter,appdex_threshold"
		" FROM trans WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (tran_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		summary_ptr->tran_id = *((arm_id_t *) sqlite3_column_blob(statement_ptr, 0));
		summary_ptr->appl_id = *((arm_id_t *) sqlite3_column_blob(statement_ptr, 1));
		strncpy (summary_ptr->transaction_name, (const char *) sqlite3_column_text(statement_ptr, 2), ARM_NAME_MAX_LENGTH);
		summary_ptr->trace_flags = (arm4_shm_transaction_trace_t) sqlite3_column_int(statement_ptr, 3);
		summary_ptr->appdex_flags = (arm4_appdex_t) sqlite3_column_int(statement_ptr, 4);
		summary_ptr->trace_parameter = sqlite3_column_int64(statement_ptr, 5);
		summary_ptr->appdex_threshold = sqlite3_column_int64(statement_ptr, 6);
	}
	else
	{
		logError ();
		return ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

	//	identity_properties
	summary_ptr->identity_properties.clear ();
	sql = "SELECT name,value"
		" FROM transaction_identity WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (tran_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	
	status = sqlite3_step(statement_ptr);
	while (status == SQLITE_ROW)
	{
		// We found an entry
		property_record_t *identity_ptr = new property_record_t;
		strncpy (identity_ptr->name, (const char *) sqlite3_column_text(statement_ptr, 0), ARM_NAME_MAX_LENGTH);
		strncpy (identity_ptr->value, (const char *) sqlite3_column_text(statement_ptr, 1), ARM_NAME_MAX_LENGTH);
		summary_ptr->identity_properties.push_back (identity_ptr);

		status = sqlite3_step(statement_ptr);
	}
	sqlite3_finalize(statement_ptr);

	//	context_names
	summary_ptr->context_names.clear ();
	sql = "SELECT slot,name"
		" FROM transaction_context WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (tran_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	
	status = sqlite3_step(statement_ptr);
	while (status == SQLITE_ROW)
	{
		// We found an entry
		indexed_property_name_record_t *identity_ptr = new indexed_property_name_record_t;
		identity_ptr->index = sqlite3_column_int(statement_ptr, 0);
		strncpy (identity_ptr->name, (const char *) sqlite3_column_text(statement_ptr, 1), ARM_NAME_MAX_LENGTH);
		summary_ptr->context_names.push_back (identity_ptr);

		status = sqlite3_step(statement_ptr);
	}
	sqlite3_finalize(statement_ptr);

	// Get the URI data
	sql = "SELECT uri"
		" FROM transaction_uri WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (tran_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return ARM_FAILURE;
	}
	
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		strncpy (summary_ptr->uri, (const char *) sqlite3_column_text(statement_ptr, 0), ARM_PROPERTY_URI_MAX_LENGTH);
	}
	else
	{
		summary_ptr->uri [0] = '\0';
	}
	sqlite3_finalize(statement_ptr);

	lock.unlock ();
    
    return ARM_SUCCESS;
}

const arm_correlator_t *
Sqlite::getCorrelator (arm_tran_start_handle_t tran_handle, arm_correlator_t *correlator_ptr)
{
	int	status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	if (correlator_ptr == NULL)
		return NULL;

    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return NULL;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT current_correlator"
		" FROM current_correlator WHERE current_correlator.tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return NULL;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return NULL;
	}
	
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		int size = sqlite3_column_bytes(statement_ptr, 0);
		if (size > 0)
		{
			memcpy (correlator_ptr, sqlite3_column_blob(statement_ptr, 0), size);
	
			status = ARM_SUCCESS;
		}
		else
		{
			status = ARM_FAILURE;
		}
	}
	else
	{
		status = ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

    lock.unlock ();

	if (status == ARM_SUCCESS)
		return correlator_ptr;
    return NULL;
}

const arm_correlator_t *
Sqlite::getParentCorrelator (arm_tran_start_handle_t tran_handle, arm_correlator_t *correlator_ptr)
{
	int	status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	if (correlator_ptr == NULL)
		return NULL;

    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return NULL;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT parent_correlator"
		" FROM parent_correlator WHERE parent_correlator.tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return NULL;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		lock.unlock ();
		return NULL;
	}
	
	status = sqlite3_step(statement_ptr);
	if (status == SQLITE_ROW)
	{
		// We found an entry
		int size = sqlite3_column_bytes(statement_ptr, 0);
		if (size > 0)
		{
			memcpy (correlator_ptr, sqlite3_column_blob(statement_ptr, 0), size);

			status = ARM_SUCCESS;
		}
		else
		{
			status = ARM_FAILURE;
		}
	}
	else
	{
		status = ARM_FAILURE;
	}
	sqlite3_finalize(statement_ptr);

    lock.unlock ();

	if (status == ARM_SUCCESS)
		return correlator_ptr;
    return NULL;
}

void
Sqlite::setTransactionTraceInformation (const arm_id_t &tran_id, const arm4db_transaction_summary_t &summary)
{
	int	status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    Arm4dbLock lock;

    status = openOrCreateDatabasesNoLock ();
    if (status != ARM_SUCCESS)
    {
		lock.unlock ();
		return;
    }
    DB_ASSERT (m_db != NULL);

	std::string sql = "INSERT OR REPLACE INTO trans"
		" (trace_flags, trace_parameter)"
		" VALUES (?,?)"
		" WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 1, summary.trace_flags);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 2, summary.trace_parameter);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, &tran_id, sizeof (tran_id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		lock.unlock ();
		return;
	}
	
	sqlite3_step(statement_ptr);
	sqlite3_finalize(statement_ptr);

    lock.unlock ();
    
    return;
}

void *
Sqlite::getStats (void)
{
	/* No stats to return */
	return NULL;
}

void
Sqlite::dump (void) const
{
	std::cerr << "Dump routines aren't supported for this database type\n";
}

void
Sqlite::backup (void)
{
	// Like dump, this is not supported for all databases, but this is probably OK
}

void
Sqlite::checkpoint (void)
{
	// Like dump, this is not supported for all databases, but this is probably OK
}

void
Sqlite::printSummaries (void) const
{
	std::cerr << "Not supported for this database\n"; 
}

void
Sqlite::printTransactions (void) const
{
	std::cerr << "Not supported for this database\n"; 
}

void
Sqlite::saveApplicationDefinition (const arm_id_t &appl_id, const application_record_t &application)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry

	std::string sql = "INSERT OR REPLACE INTO application"
		" (appl_id, application_name, md5_digest, charset, default_trace_flags, default_trace_parameter, default_appdex_threshold)"
		" VALUES (?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &appl_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, application.application_name, strlen (application.application_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, application.md5_digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 4, application.charset);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 5, application.default_trace_flags);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 6, application.default_trace_parameter);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 7, application.default_appdex_threshold);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

	    return;
	}
	sqlite3_finalize(statement_ptr);

    return;
}

void
Sqlite::saveApplicationIdentity (const arm_id_t &appl_id, const property_record_t &property)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_identity"
		" (appl_id, name, value)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &appl_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, property.name, strlen (property.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property.value, strlen (property.value), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveApplicationContext (const arm_id_t &appl_id, const indexed_property_name_record_t &property)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_context"
		" (appl_id, slot, name)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &appl_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, property.index);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property.name, strlen (property.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveMetricDefinition (const arm_id_t &metric_id, const metric_record_t &metric)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	std::string sql = "INSERT OR REPLACE INTO metric"
		" (metric_id, metric_name, metric_units, md5_digest, appl_id, metric_format, metric_usage)"
		" VALUES (?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &metric_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, metric.metric_name, strlen (metric.metric_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, metric.metric_units, strlen (metric.metric_units), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 4, metric.md5_digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 5, &(metric.appl_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 6, metric.metric_format);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 7, metric.metric_usage);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
		return;
	}
	sqlite3_finalize(statement_ptr);
	
    return;
}

void
Sqlite::saveTransactionDefinition (const arm_id_t &tran_id, const transaction_record_t &transaction)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	std::string sql = "INSERT INTO trans"
		" (tran_id, transaction_name, md5_digest, appl_id, trace_flags, appdex_flags, trace_parameter, appdex_threshold)"
		" VALUES (?,?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, transaction.transaction_name, strlen (transaction.transaction_name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, transaction.md5_digest, MD5_DIGEST_LENGTH, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 4, &(transaction.appl_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 5, transaction.trace_flags);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 6, transaction.appdex_flags);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 7, transaction.trace_parameter);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 8, transaction.appdex_threshold);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return;
	}
	sqlite3_finalize(statement_ptr);

    return;
}

void
Sqlite::saveTransactionIdentity (const arm_id_t &tran_id, const property_record_t &property)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_identity"
		" (tran_id, name, value)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, property.name, strlen (property.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property.value, strlen (property.value), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTransactionContext (const arm_id_t &tran_id, const indexed_property_name_record_t &property)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_context"
		" (tran_id, slot, name)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, property.index);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property.name, strlen (property.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTransactionUri (const arm_id_t &tran_id, const char *uri_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_uri"
		" (tran_id, uri)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, uri_ptr, strlen (uri_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTransactionMetric (const arm_id_t &tran_id, const arm_metric_binding_t &binding)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_metric_binding"
		" (tran_id, slot, id)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, binding.slot);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, &(binding.id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveApplicationInstance (const arm_app_start_handle_t app_handle, const application_instance_record_t &application)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT OR REPLACE INTO application_instance"
		" (app_handle, appl_id, active)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, &(application.appl_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 3, application.active);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return;
	}
	sqlite3_finalize(statement_ptr);

    return;
}

void
Sqlite::saveAppInstanceContextValue (const arm_app_start_handle_t app_handle, const indexed_property_value_record_t &property)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_context_values"
		" (app_handle, slot, property_name)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, property.index);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property.value, strlen (property.value), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveAppInstanceGroup (const arm_app_start_handle_t app_handle, const char *group_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_group"
		" (app_handle, app_group)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, group_ptr, strlen (group_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveAppInstanceName (const arm_app_start_handle_t app_handle, const char *name_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_instance_id"
		" (app_handle, instance_name)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, name_ptr, strlen (name_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveAppInstanceAddress (const arm_app_start_handle_t app_handle, const application_address_record_t &address_value)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO application_address"
		" (app_handle, address_format, address_length, id_valid, id, address)"
		" VALUES (?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, address_value.address_format);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 3, address_value.address_length);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 4, address_value.id_valid);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 5, &(address_value.id), sizeof (address_value.id), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 6, address_value.address, address_value.address_length, SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTransactionInstance (const arm_tran_start_handle_t tran_handle, const transaction_instance_record_t &transaction)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT OR REPLACE INTO transaction_instance"
		" (tran_handle, appl_id, tran_id, app_handle,"
		"  arrival_time, arrival_time_usec, stop_time, stop_time_usec, response_time, response_time_usec, blocked_time, blocked_time_usec,"
		"  tran_status, active)"
		" VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, &(transaction.appl_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 3, &(transaction.tran_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 4, transaction.app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 5, transaction.arrival_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 6, transaction.arrival_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 7, transaction.stop_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 8, transaction.stop_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 9, transaction.response_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 10, transaction.response_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 11, transaction.blocked_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 12, transaction.blocked_time.tv_usec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 13, transaction.tran_status);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int(statement_ptr, 14, transaction.active);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}

	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return;
	}
	sqlite3_finalize(statement_ptr);

    return;
}

void
Sqlite::saveTranInstanceContextValue (const arm_tran_start_handle_t tran_handle, const indexed_property_value_record_t &property)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_context_values"
		" (tran_handle, slot, value)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, property.index);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 3, property.value, strlen (property.value), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTranInstanceMetricValue (const arm_tran_start_handle_t tran_handle, const metric_value_record_t &metric)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_metric_value"
		" (tran_handle, slot, format, usage, valid, metric_value, string_value)"
		" VALUES (?,?,?,?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 2, metric.value.slot);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 3, metric.value.format);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 4, metric.value.usage);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 5, metric.value.valid);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 6, &(metric.value.metric_u), sizeof (metric.value.metric_u), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 7, metric.string32, strlen (metric.string32), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTranInstanceUri (const arm_tran_start_handle_t tran_handle, const char *uri_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_context_uri"
		" (tran_handle, uri)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, uri_ptr, strlen (uri_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTranInstanceUser (const arm_tran_start_handle_t tran_handle, const user_record_t &user)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	//printf ("processTransactionUser\n");

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_user"
		" (tran_handle, name, valid, id)"
		" VALUES (?,?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, user.name, strlen (user.name), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int(statement_ptr, 3, user.id_valid);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_blob(statement_ptr, 4, &(user.id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTranInstanceDiag (const arm_tran_start_handle_t tran_handle, const char *diag_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

	DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO transaction_diag"
		" (tran_handle, detail)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_bind_text(statement_ptr, 2, diag_ptr, strlen (diag_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);

		return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTranCorrelator (const arm_tran_start_handle_t tran_handle, const arm_correlator_t *correlator_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO current_correlator"
		" (tran_handle, current_correlator)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, correlator_ptr,
			Arm4dbCommon::correlatorSize (correlator_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTranParentCorrelator (const arm_tran_start_handle_t tran_handle, const arm_correlator_t *correlator_ptr)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO parent_correlator"
		" (tran_handle, parent_correlator)"
		" VALUES (?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, correlator_ptr,
			Arm4dbCommon::correlatorSize (correlator_ptr), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return;
	}
	sqlite3_finalize(statement_ptr);
}

void
Sqlite::saveTraceStats (const arm_time_t sample_time, const arm4db_transaction_trace_stats_t &trace_stats)
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    DB_ASSERT (m_db != NULL);

	// Create a new entry
	std::string sql = "INSERT INTO trace_stats"
		" (time_seconds, tran_id, counter)"
		" VALUES (?,?,?)";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}

	status = sqlite3_bind_int64(statement_ptr, 1, sample_time.tv_sec);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_blob(statement_ptr, 2, &(trace_stats.tran_id), sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_bind_int64(statement_ptr, 3, trace_stats.transaction_counter);
	if (status != SQLITE_OK)
	{
		logError ();
	    return;
	}
	status = sqlite3_step(statement_ptr);
	if (status != SQLITE_DONE)
	{
		logError ();
		sqlite3_finalize(statement_ptr);
	    return;
	}
	sqlite3_finalize(statement_ptr);

    return;
}
