/**********************************************************************
 * Copyright (c) 2005-2008 David Carter <dcarter@arm4.org> and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 * David Carter - Initial API and implementation
 **********************************************************************/
 

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <netinet/in.h>

#include <iomanip>

#include "arm4.h"
#include "arm4agent.h"
#include "Arm4db.h"
#include "Logger.h"
#include "Sqlite.h"
#include "Arm4dbCommon.h"

std::ostream&
Sqlite::exportApplicationIdentity (std::ostream &os, const arm_id_t &appl_id) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT name,value"
		" FROM application_identity"
		" WHERE appl_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_blob(statement_ptr, 1, &appl_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<identity>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_PROPERTY_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_PROPERTY_VALUE_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</identity>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplicationContext (std::ostream &os, const arm_id_t &appl_id) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT slot,name"
		" FROM application_context"
		" WHERE appl_id = ?"
		" ORDER BY slot";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_blob(statement_ptr, 1, &appl_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<context>\n";
			os << "\t\t\t<slot>" << sqlite3_column_int (statement_ptr, 0) << "</slot>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t</context>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplicationDefinitions (std::ostream &os) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT appl_id,application_name,md5_digest,charset,default_trace_flags,default_trace_parameter,default_appdex_threshold"
		" FROM application";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			arm_id_t appl_id = *((arm_id_t *) sqlite3_column_blob (statement_ptr, 0));
			os << "\t<application_definition>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, appl_id);
			os << "</appl_id>\n";
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t<digest>";
				Arm4dbCommon::printDigest (os, (const char *) sqlite3_column_blob (statement_ptr, 2));
			os << "</digest>\n";
			os << "\t\t<charset>" << sqlite3_column_int (statement_ptr, 3) << "</charset>\n";
			os << "\t\t<default_trace_flags>" << sqlite3_column_int (statement_ptr, 4) << "</default_trace_flags>\n";
			os << "\t\t<default_trace_parameter>" << sqlite3_column_int64 (statement_ptr, 5) << "</default_trace_parameter>\n";
			os << "\t\t<default_appdex_threshold>" << sqlite3_column_int64 (statement_ptr, 6) << "</default_appdex_threshold>\n";
			
			exportApplicationIdentity (os, appl_id);
			exportApplicationContext (os, appl_id);

		    os << "\t</application_definition>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionIdentity (std::ostream &os, const arm_id_t &tran_id) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT name,value"
		" FROM transaction_identity"
		" WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<identity>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_PROPERTY_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_PROPERTY_VALUE_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</identity>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionContext (std::ostream &os, const arm_id_t &tran_id) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT slot,name"
		" FROM transaction_context"
		" WHERE tran_id = ?"
		" ORDER BY slot";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<context>\n";
			os << "\t\t\t<slot>" << sqlite3_column_int (statement_ptr, 0) << "</slot>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t</context>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionUri (std::ostream &os, const arm_id_t &tran_id) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT uri"
		" FROM transaction_uri"
		" WHERE tran_id = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<uri>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_PROPERTY_URI_MAX_CHARS);
			os << "</uri>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionMetricBinding (std::ostream &os, const arm_id_t &tran_id) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT slot,id"
		" FROM transaction_metric_binding"
		" WHERE tran_id = ?"
		" ORDER BY slot";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_blob(statement_ptr, 1, &tran_id, sizeof (arm_id_t), SQLITE_STATIC);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<metric>\n";
			os << "\t\t\t<slot>" << sqlite3_column_int (statement_ptr, 0) << "</slot>\n";
			os << "\t\t\t<metric_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 1));
			os << "</metric_id>\n";
			os << "\t\t</metric>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionDefinitions (std::ostream &os) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);

	std::string sql = "SELECT tran_id,transaction_name,md5_digest,appl_id,trace_flags,appdex_flags,trace_parameter,appdex_threshold"
		" FROM trans";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			arm_id_t tran_id = *((arm_id_t *) sqlite3_column_blob (statement_ptr, 0));
			os << "\t<transaction_definition>\n";
			os << "\t\t<tran_id>";
				Arm4dbCommon::printId (os, tran_id);
			os << "</tran_id>\n";
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t<digest>";
				Arm4dbCommon::printDigest (os, (const char *) sqlite3_column_blob (statement_ptr, 2));
			os << "</digest>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 3));
			os << "</appl_id>\n";
			os << "\t\t<trace_flags>" << sqlite3_column_int (statement_ptr, 4) << "</trace_flags>\n";
			os << "\t\t<appdex_flags>" << sqlite3_column_int (statement_ptr, 5) << "</appdex_flags>\n";
			os << "\t\t<trace_parameter>" << sqlite3_column_int64 (statement_ptr, 6) << "</trace_parameter>\n";
			os << "\t\t<appdex_threshold>" << sqlite3_column_int64 (statement_ptr, 7) << "</appdex_threshold>\n";
			
			exportTransactionIdentity (os, tran_id);
			exportTransactionContext (os, tran_id);
			exportTransactionUri (os, tran_id);
			exportTransactionMetricBinding (os, tran_id);
			
			os << "\t</transaction_definition>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportMetricDefinitions (std::ostream &os) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
	std::string sql = "SELECT metric_id,metric_name,md5_digest,appl_id,metric_units,metric_format,metric_usage"
		" FROM metric";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			arm_id_t metric_id = *((arm_id_t *) sqlite3_column_blob (statement_ptr, 0));
			os << "\t<metric_definition>\n";
			os << "\t\t<metric_id>";
				Arm4dbCommon::printId (os, metric_id);
			os << "</metric_id>\n";
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t<digest>";
				Arm4dbCommon::printDigest (os, (const char *) sqlite3_column_blob (statement_ptr, 2));
			os << "</digest>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 3));
			os << "</appl_id>\n";
			os << "\t\t<units>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 4), ARM_NAME_MAX_CHARS);
			os << "</units>\n";
			os << "\t\t<format>" << sqlite3_column_int (statement_ptr, 5) << "</format>\n";
			os << "\t\t<usage>" << sqlite3_column_int (statement_ptr, 6) << "</usage>\n";
			os << "\t</metric_definition>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplicationGroup (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT app_group"
		" FROM application_group"
		" WHERE app_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<group>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_NAME_MAX_CHARS);
			os << "</group>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplicationInstanceNames (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT instance_name"
		" FROM application_instance_id"
		" WHERE app_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplicationContextValues (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT slot,property_name"
		" FROM application_context_values"
		" WHERE app_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<context_value>\n";
			os << "\t\t\t<slot>" << sqlite3_column_int (statement_ptr, 0) << "</slot>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</context_value>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplicationAddress (std::ostream &os, arm_app_start_handle_t app_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT address_format, address_length, id_valid, id, address"
		" FROM application_address"
		" WHERE app_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, app_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<address>\n";
			os << "\t\t\t<address_format>" << sqlite3_column_int (statement_ptr, 0) << "</address_format>\n";
			int address_length = sqlite3_column_int (statement_ptr, 1);
			os << "\t\t\t<address_length>" << address_length << "</address_length>\n";
			int id_valid = sqlite3_column_int (statement_ptr, 2);
			os << "\t\t\t<id_valid>" << id_valid << "</id_valid>\n";
			os << "\t\t\t<id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 3));
			os << "</id>\n";
			os << "\t\t\t<address>";
				Arm4dbCommon::printAddress (os, (arm_uint8_t *) sqlite3_column_blob (statement_ptr, 4), address_length);
			os << "</address>\n";
			os << "\t\t</address>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportApplications (std::ostream &os) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT app_handle,appl_id,active"
		" FROM application_instance";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			arm_app_start_handle_t app_handle = sqlite3_column_int64 (statement_ptr, 0);

			os << "\t<application_instance>\n";
			os << "\t\t<handle>" << app_handle << "</handle>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 1));
			os << "</appl_id>\n";
			os << "\t\t<active>" << sqlite3_column_int (statement_ptr, 2) << "</active>\n";

			exportApplicationGroup (os, app_handle);
			exportApplicationInstanceNames (os, app_handle);
			exportApplicationContextValues (os, app_handle);
			exportApplicationAddress (os, app_handle);
			
			os << "\t</application_instance>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionContextValues (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT slot,value"
		" FROM transaction_context_values"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<context_value>\n";
			os << "\t\t\t<slot>" << sqlite3_column_int (statement_ptr, 0) << "</slot>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 1), ARM_NAME_MAX_CHARS);
			os << "</value>\n";
			os << "\t\t</context_value>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionContextUri (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT uri"
		" FROM transaction_context_uri"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<context_uri>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_PROPERTY_URI_MAX_CHARS);
			os << "\t\t</context_uri>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionContextMetricValue (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT slot,format,usage,valid,metric_value,string_value"
		" FROM transaction_metric_value"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			arm_metric_t metric;

			metric.slot = sqlite3_column_int (statement_ptr, 0);
			metric.format = sqlite3_column_int (statement_ptr, 1);
			metric.usage = sqlite3_column_int (statement_ptr, 2);
			metric.valid = sqlite3_column_int (statement_ptr, 3);
			memcpy (&metric.metric_u, sqlite3_column_blob (statement_ptr, 4), sizeof (metric.metric_u));

			os << "\t\t<metric_value>\n";
			os << "\t\t\t<slot>" << (int) metric.slot << "</slot>\n";
			os << "\t\t\t<format>" << (int) metric.format << "</format>\n";
			os << "\t\t\t<usage>" << metric.usage << "</usage>\n";
			os << "\t\t\t<valid>" << metric.valid << "</valid>\n";
			os << "\t\t\t<value>";
				Arm4dbCommon::printMetricValue (os, metric, (char *) sqlite3_column_text (statement_ptr, 5));
			os << "</value>\n";
			os << "\t\t</metric_value>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionContextUser (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT name,valid,id"
		" FROM transaction_user"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<user>\n";
			os << "\t\t\t<name>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_NAME_MAX_CHARS);
			os << "</name>\n";
			os << "\t\t\t<id_valid>" << sqlite3_column_int (statement_ptr, 1) << "</id_valid>\n";
			os << "\t\t\t<id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 2));
			os << "</id>\n";
			os << "\t\t</user>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactionContextDiag (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT detail"
		" FROM transaction_diag"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<diag>";
				Arm4dbCommon::escapeForXml(os, sqlite3_column_text (statement_ptr, 0), ARM_DIAG_DETAIL_MAX_CHARS);
			os << "\t\t</diag>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportCorrelators (std::ostream &os, arm_tran_start_handle_t tran_handle) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);

	// First the parent correlator
	std::string sql = "SELECT parent_correlator"
		" FROM parent_correlator"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<parent_correlator>";
			Arm4dbCommon::printCorrelator(os, (arm_correlator_t *) sqlite3_column_blob (statement_ptr, 0));
			os << "\t\t</parent_correlator>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	// Then the current correlator
	sql = "SELECT current_correlator"
		" FROM current_correlator"
		" WHERE tran_handle = ?";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}

	status = sqlite3_bind_int64 (statement_ptr, 1, tran_handle);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t\t<correlator>";
			Arm4dbCommon::printCorrelator(os, (arm_correlator_t *) sqlite3_column_blob (statement_ptr, 0));
			os << "\t\t</correlator>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTransactions (std::ostream &os) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT tran_handle,appl_id,tran_id,app_handle,"
		"arrival_time,arrival_time_usec,stop_time,stop_time_usec,response_time,response_time_usec,blocked_time,blocked_time_usec,"
		"tran_status,active"
		" FROM transaction_instance";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			arm_tran_start_handle_t tran_handle = sqlite3_column_int64 (statement_ptr, 0);
			
			os << "\t<transaction_instance>\n";
			os << "\t\t<handle>" << tran_handle << "</handle>\n";
			os << "\t\t<appl_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 1));
			os << "</appl_id>\n";
			os << "\t\t<tran_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 2));
			os << "</tran_id>\n";
			os << "\t\t<app_handle>" << sqlite3_column_int64 (statement_ptr, 3) << "</app_handle>\n";
			os << "\t\t<arrival_time>" << sqlite3_column_int64 (statement_ptr, 4) << "."
					<< std::setfill ('0') << std::setw (6) << sqlite3_column_int (statement_ptr, 5) << "</arrival_time>\n";
			os << "\t\t<stop_time>" << sqlite3_column_int64 (statement_ptr, 6) << "."
					<< std::setfill ('0') << std::setw (6) << sqlite3_column_int (statement_ptr, 7) << "</stop_time>\n";
			os << "\t\t<response_time>" << sqlite3_column_int64 (statement_ptr, 8) << "."
					<< std::setfill ('0') << std::setw (6) << sqlite3_column_int (statement_ptr, 9) << "</response_time>\n";
			os << "\t\t<blocked_time>" << sqlite3_column_int64 (statement_ptr, 10) << "."
					<< std::setfill ('0') << std::setw (6) << sqlite3_column_int (statement_ptr, 11) << "</blocked_time>\n";
			
			os << "\t\t<status>" << sqlite3_column_int (statement_ptr, 12) << "</status>\n";
			os << "\t\t<active>" << sqlite3_column_int (statement_ptr, 13) << "</active>\n";
			
			exportTransactionContextValues (os, tran_handle);
			exportTransactionContextUri (os, tran_handle);
			exportTransactionContextMetricValue (os, tran_handle);
			exportTransactionContextUser (os, tran_handle);
			exportTransactionContextDiag (os, tran_handle);
			exportCorrelators (os, tran_handle);
			
			os << "\t</transaction_instance>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}

std::ostream&
Sqlite::exportTraceStats (std::ostream &os) const
{
	int status;
	sqlite3_stmt *statement_ptr;
	const char *unused_ptr;

    /* Ensure we have database access */
    DB_ASSERT (m_db != NULL);
    
	std::string sql = "SELECT time_seconds,tran_id,counter"
		" FROM trace_stats";
	status = prepare(
			m_db,
			sql.c_str (), sql.length (),
			&statement_ptr,
			&unused_ptr);
	if (status != SQLITE_OK)
	{
		logError ();
		return os;
	}
	
	do
	{
		status = sqlite3_step(statement_ptr);
		if (status == SQLITE_ROW)
		{
			os << "\t<transaction_trace>\n";
			os << "\t\t<transaction_trace_sample_time>" << sqlite3_column_int64 (statement_ptr, 0) << "</transaction_trace_sample_time>\n";
			os << "\t\t<transaction_trace_tran_id>";
				Arm4dbCommon::printId (os, (arm_id_t *) sqlite3_column_blob (statement_ptr, 1));
			os << "</transaction_trace_tran_id>\n";
			os << "\t\t<transaction_trace_counter>" << sqlite3_column_int64 (statement_ptr, 2) << "</transaction_trace_counter>\n";
			os << "\t</transaction_trace>\n";
		}
	}
	while (status == SQLITE_ROW);
	sqlite3_finalize(statement_ptr);

	return os;
}
